.TH ODL.pm 3 2004.07.07
.SH NAME
ODL.pm \- a perl module for working with ODL files
.SH CONVENTIONS
Use of the word "object" can mean one of two things in this documentation
depending on whether or not it is in all upper case. "OBJECT" refers to a type
of ODL statement, formatted in ODL file as follows:
.nf

OBJECT = <object_identifier>
  <substatement 1>
  <substatement 2>
  ...
  <substatement n>
END_OBJECT = <object_identifier>
.fi

"object" refers to a perl object, created by one of many ODL functions which
construct ODL objects. When refered to as a variable (in sample code, for
example) the object will be represented as $odlObj.
.P
The word statement refers to a logical unit in ODL files. A statement can be
a simple identifier/value pair, OBJECT or GROUP statements, or the complete
ODL file in general. The word statement refers to either the actual ODL
statement in the ODL file, or to the perl object which represents the ODL
statement, depending on how the word "statement" is used. If "statement" is
refering to an element in perl code, that statement is also an object. Not all
objects, however, are statements, since objects are also used to represent the
values associated with identifiers in attribute and pointer statements.
.SH SYNOPSIS
This ODL module can be included in any perl program by including the following
line.

.B "    require ODL;"

There are two ways to call functions included in the module, depending on
whether or not the function requires an ODL object. To call a function which
doesn't require an ODL object, use the following style:

.BI "    ODL::function_name ([" $param1 ", [...]]);"

To call a function which does require an ODL object, use the following style,
(assuming $odlObj is a reference to the ODL object.)

.BI "    $odlObj->function_name ([" $param1 ", [...]]);"
.SH DESCRIPTION
The ODL module lets you create, read, write, and modify ODL labels. Since the
PDS label format is based on ODL, this module can be used to read PDS files as
well. The following sections describe the various types of functions provided
by this module. The sections are split up into functions for creating ODL
labels, getting information from ODL labels, modifying ODL labels, and writing
ODL labels to files.
.P
ODL objects come in one of two types: statements and values. Statements cover
simple identifier/value pairs as well as OBJECT and GROUP statements which in
turn may contain more statements. Every identifier/value statement contains one
value object. Value objects can be scalar values (numbers, strings, etc.), or
sets, one dimensional lists, or two dimensional lists of scalar values.
.P
There is one special type of statement, not formally defined in ODL but used
for the sake of clean code, which is the root statement. This statement can be
thought of as the entire label, and acts like an OBJECT or GROUP statement in
that it can (and most likely will) contain sub statements.
.SH CREATING ODL OBJECTS
.SS
.BI "ODL::create ([" $filename "])"
The
.B create
function creates and returns a root ODL statement. If
.I $filename
is provided, the statement created will be filled with the data from the file
by that name.
.SS
.IB $odlObj "->new_object_statement (" $identifier ")"
This function creates an OBJECT statement whose identifier is given by
.IR $identifier .
The
.I $odlObj
object must be either a root, OBJECT, or GROUP statement, and the new OBJECT
statement is inserted after the last statement within
.IR $odlObj .
.SS
.IB $odlObj "->new_group_statement (" $identifier ")"
This function works just like the
.B new_object_statement
except that it creates a GROUP statement rather than an OBJECT statement.
.SS
.IB $odlObj "->new_attribute_statement (" $identifier ", " $value ")"
This function creates a attribute statement with the given identifier and
value.
.I $value
must be a properly formatted ODL value, set, 1 dimensional list, or 2
dimensional list. See the
.B DATA TYPES
section for more information on formatting.
.SS
.IB $odlObj "->new_pointer_statement (" $identifier ", " $value ")"
This function creates a new pointer statement with the given identifier and
value. If
.I $identifier
doesn't begin with a caret (^) character, one will be added.
.SH READING ODL OBJECTS
.SS
.IB $odlObj "->get_attribute_identifiers ()"
This function returns an array of all attribute identifiers present in the
root, OBJECT, or GROUP statement. Nested OBJECTs and GROUPs are not searched.
If $odlObj is not a root, OBJECT, or GROUP statement, or if $odlObj doesn't
contain any attributes, undef is returned.
.SS
.IB $odlObj "->get_pointer_identifiers ()"
This function acts just like
.B get_attribute_identifiers
except that all pointer identifiers are returned instead of all attribute
identifiers.
.SS
.IB $odlObj "->get_object_identifiers ()"
This function acts just like
.B get_attribute_identifiers
except that all OBJECT identifiers are returned instead of all attribute
identifiers.
.SS
.IB $odlObj "->get_group_identifiers ()"
This function acts just like
.B get_attribute_identifiers
except that all GROUP identifers are returned instead of all attribute
identifiers.
.SS
.IB $odlObj "->get_collection_identifiers ()"
This function acts just like
.B get_attribute_identifiers
except that all OBJECT and GROUP identifiers are returned instead of all
attribute identifiers.
.SS
.IB $odlObj "->get_all_identifiers ()"
This function acts just like
.B get_attribute_identifers
except that all (attribute, pointer, OBJECT, and GROUP) identifers are returned
instead of all attribute identifers.
.SS
.IB $odlObj "->get_statement (" $identifer ", [" $number "])"
This function searches
.I $odlObj
for a statement with the
.I $identifier
identifier. If
.I $number
is given, and multiple statements exist within
.I $odlObl
with the given identifier,
.I $number
represents which to return (1 for 1st, 2 for 2nd, etc.) If a matching statement
is found, it is returned, otherwise, undef is returned.
.SS
.IB $odlObj "->get_object (" $object_identifier ", [" $number "])"
This function acts just like
.B get_statement
except that only OBJECT statements are searched.
.SS
.IB $odlObj "->get_group (" $group_identifier ", [" $number "])"
This function acts like
.B get_object
except that GROUP statements are searched instead of OBJECT statements.
.SS
.IB $odlObj "->get_collection (" $identifier ", [" $number "])"
This function acts like
.B get_object
except that both OBJECT and GROUP statements are searched instead of OBJECT
statements.
.SS
.IB $odlObj "->get_pointer (" $pointer_identifier ", [" $number "])"
This function acts like
.B get_object
except that pointer statements are searched instead of OBJECT statements.
.SS
.IB $odlObj "->get_attribute (" $attribute_identifier ", [" $number "])"
This function acts like
.B get_object
except that attribute statements are searched instead of OBJECT statements.
.SS
.IB $odlObj "->get_value ([" $index1 " [, " $index2 "]])"
This function returns the value stored in $odlObj. If
.IR $index1 ", or " $index1 " and " $index2 ", "
are given, and the value is a set or sequence, the indices are used to extract
the correct element.
.I $index2
should only be used if the value is a two dimensional list. If too few indices
are given (only
.I $index1
on a two dimensional list, or none on any list or set, then a set or list
will be returned. In this case,
.B get_value
can be called again with more indices to get the correct scalar. For example,
the three pieces of code all return the same value (the second value in the
first one-dimensional list) assuming
.B $odlObj
is a statement whose value is a two dimensional list:
.nf

.B $odlObj->get_value (1, 2);
.B $odlObj->get_value (1)->get_value (2);
.B $odlObj->get_value ()->get_value (1)->get_value (2);
.fi
.P
If the statement reads as follows in the ODL file, the return value will be
"green":
.nf

statement = ((red, green, blue)(1, 2, 3))
.fi
.SH DETERMINING THE TYPES ODL OBJECTS
The following functions are provided to help determine the type of an ODL
object.
.SS
.IB $odlObj "->is_statement ()"
This function returns true if
.I $odlObj
is a statement object, and false if it is a value object.
.SS
.IB $odlObj "->is_value ()"
This function returns true if
.I $odlObj
is a value object, and false if it is a statement object.
.SS
.IB $odlObj "->statement_is_root_node ()"
This function returns true if
.I $odlObj
is a root node (entire ODL file), and false otherwise.
.SS
.IB $odlObj "->statement_is_object ()"
This function returns true if
.I $odlObj
is an OBJECT statement, and false otherwise.
.SS
.IB $odlObj "->statement_is_group ()"
This function returns true if
.I $odlObj
is a GROUP statement, and false otherwise.
.SS
.IB $odlObj "->statement_is_attribute ()"
This function returns true if
.I $odlObj
is an attribute statement, and false otherwise.
.SS
.IB $odlObj "->statement_is_pointer ()"
This function returns true if
.I $odlObj
is a pointer statement, and false otherwise.
.SS
.IB $odlObj "->statement_has_value ()"
This function returns true if 
.I $odlObj
contains a value (i.e. it is an attribute or pointer statement), and false
otherwise.
.SS
.IB $odlObj "->statement_is_collection ()"
This function returns true if
.I $odlObj
is a collection of statements (i.e. it is a root node, or a OBJECT or GROUP
statement), and false otherwise.
.SS
.IB $odlObj "->value_is_number ()"
This function returns true if the value represented in
.I $odlObj
is of a numeric type, and false otherwise. If
.I $odlObj
is a statement with a value, that value is checked instead.
.SS
.IB $odlObj "->value_is_integer ()"
This function returns true if the value represented in
.I $odlObj
is of an integer type, and false otherwise. If
.I $odlObj
is a statement with a value, that value is checked instead.
.SS
.IB $odlObj "->value_is_date_time ()"
This function returns true if the value represented in
.I $odlObj
is a date, time, or date/time value, and false otherwise. If
.I $odlObj
is a statement with a value, that value is checked instead.
.SS
.IB $odlObj "->value_is_string ()"
This function returns true if the value represented in
.I $odlObj
is a string, and false otherwise. If
.I $odlObj
is a statement with a value, that value is checked instead.
.SS
.IB $odlObj "->value_is_identifier ()"
This function returns true if the value represented in
.I $odlObj
is an identifier, and false otherwise. If
.I $odlObj
is a statement with a value, that value is checked instead.
.SS
.IB $odlObj "->value_is_scalar ()"
This function returns true if the value represented in
.I $odlObj
is a number, date/time, string, or identifier, and false otherwise. If
.I $odlobj
is a statement with a value, that value is checked instead.
.SS
.IB $odlObj "->value_is_set ()"
This function returns true if the value represented in
.I $odlObj
is a set of scalars, and false otherwise. If
.I $odlObj
is a statement with a value, that value is checked instead.
.SS
.IB $odlObj "->value_is_sequence ()"
This function returns true if the value represented in
.I $odlObj
is a one or two dimensional list of scalars, and false otherwise. If
.I $odlObj
is a statement with a value, that value is checked instead.
.SS
.IB $odlObj "->value_is_1D_sequence ()"
This function returns true if the value represented in
.I $odlObj
is a one dimensional list of scalars, and false otherwise. If
.I $odlObj
is a statement with a value, that value is checked instead.
.SS
.IB $odlobj "->value_is_2D_sequence ()"
This function returns true if the value represented in
.I $odlObj
is a two dimensional list of scalars, and false otherwise. If
.I $odlObj
is a statement with a value, that value is checked instead.
.SS
.IB
.SH DATA TYPES
Please refer to Chapter 12, Object Description Language Specification and
Usage, from the PDS Standards Reference available at
http://pds.jpl.nasa.gov/documents/sr/, for data formatting rules.