#!/usr/bin/perl

# This package provides a collection of functions usefull for reading,
# creating, and modifying ODL and PDS files. Below are a list of functions,
# what they do, and how to use them.
#
# Each function may be called in one of two ways, depending on whether or not
# the function requires an ODL object to work on. Functions defined by
# ODL::function_name are to be called without an ODL object, and functions
# defined as $odlObj->function_name require an ODL object. To call these
# second type of function, replace $odlObj with the name of the variable
# holding the ODL object.
#
# ODL::create ($filename)
# This function creates and returns an ODL object. If $filename is provided
# that file will be opened and read into the ODL object. The file must be a
# valid ODL file.
#
# example:
# $empty = ODL::create ();
# $file  = ODL::create ("lspn0001.pds");
#
# $odlObj->get_pointer_identifiers ()
# This function returns an array of all pointer identifiers contained within
# the $odlObj ODL object. This function does not, however, traverse OBJECT
# or GROUP statements contained within the ODL object; it only returns pointers
# contained directly by the object.
#
# example:
# @pointers = $odlObj->get_pointer_identifiers ();
#

# declare package name
package ODL;
	     
# declare some constants
use constant FALSE => 0;
use constant TRUE  => not FALSE;

# declare some output modes.
# COMPRESSED outputs ODL statements with as little whitespace as required.
#            Strings are cut off at the end of a line and continued on the
#            next line. Indentations are kept for sets and lists to make the
#            output more readable.
# SPACED outputs ODL statements with one space between each element of a
#        set or 1d or 2d list.
# LINED outputs ODL statements where each element of a 1d list or set is
#       started on a new line. The inner lists of 2d lists are kept on one
#       line as SPACED outputs
# EXPANDED is like LINED, only the inner lists of 2d lists are also outputed
#          with one on each line.
my $output_modes = {COMPRESSED => 1,
                    SPACED     => 2,
                    LINED      => 3,
                    EXPANDED   => 4};
# declare the default output mode
my $output_modes_default = $output_modes->{SPACED};

# declare object types.
my $types = {ERROR            =>  0,

	     ROOT_NODE        =>  1,
	     
	     # statement types:
	     ATTRIBUTE        =>  2,
	     POINTER          =>  3,
	     OBJECT           =>  4,
	     GROUP            =>  5,
	     END_OBJECT       =>  6,
	     END_GROUP        =>  7,
	     END_LABEL        =>  8,
	     STATEMENT_ERROR  =>  9,

	     # value types
	     UNKNOWN_TYPE     => 10,
	     BLANK_TYPE       => 11,
	     # strings
	     TEXT_STRING      => 12,
	     SYMBOL_STRING    => 13,
	     # identifiers
	     IDENTIFIER       => 14,
	     # numbers
	     INTEGER          => 15,
	     UNSIGNED_INTEGER => 16,
	     BASED_INTEGER    => 17,
	     UNSCALED_REAL    => 18,
	     SCALED_REAL      => 19,
	     # dates and times
	     DATE             => 20,
	     TIME             => 21,
	     DATE_TIME        => 22,
	     # sets and sequences
	     SET              => 23,
	     SEQUENCE1D       => 24,
	     SEQUENCE2D       => 25};

# define some regular expressions that are used a lot
my $re_identifier       = "(?:(?i)(?:[[:alpha:]](?:(?:_|[[:alnum:]])*[[:alnum:]])?))";
my $re_units_factor     = "$re_identifier(?:\\*\\*(?:\\+|-)?\\d+)?";
my $re_units_expression = "(?:\\<$re_units_factor(?:(?:\\*|\\/)$re_units_factor)*\\>)";
my $re_hour_min_sec     = "(\\d+)\\:(\\d+)(?:\\:((?:\\d+(?:\\.\\d*)?)|(?:\\.\\d+)))?";

# global variables, for debuging
my $last_line_read;

=head1

ODL::create ( [<filename>] ) creates an ODL object, and loads the contents of
the file given by <filename>, if present.

 $data = ODL::create ($filename);
results in the same behavior as:
 $data = ODL::create ();
 $data->load_from_file ($filename);

=cut

sub create
{
    # get parameters
    my ($filename) = @_;
    # create ODL object, root node type
    my $self = bless({TYPE => $types->{ROOT_NODE}});
    # load from file if filename given
    $self = $self->load_from_file ($filename) if ($filename);
    # return the ODL object
    return $self;
}

sub new_object_statement ($)
{
    # get self and parameters
    my ($self, $objectIdentifier) = @_;
    # make sure new_object_statement was called on a collection statement
    # (these are OBJECT and GROUP statements as well as the root node
    #  which contains the entire ODL file.)
    if ($self->statement_is_collection ())
    {
	# make sure object identifier given is a valid identifier
	if ($objectIdentifier =~ /^$re_identifier$/)
	{
	    # create ODL OBJECT statement with given keyword
	    $object = bless({TYPE       => $types->{OBJECT},
			     IDENTIFIER => uc($objectIdentifier),
			     VALUE      => []});
	    # add to collection object new_object_statement was called on
	    push (@{$self->{STATEMENTS}}, $object);
	    # return the new object
	    return $object;
	}
	else
	{
	    # invalid identifier, display warning
	    warn ("WARNING - object identifier $objectIdentifier invalid");
	}
    }
    else
    {
	# function called on incompatable object, display warning
	warn ("WARNING - new_object_statement on a non collection statement");
    }
    # if we got here, there was a problem. return undef.
    return undef;
}

sub new_group_statement ($)
{
    # get self and parameters
    my ($self, $groupIdentifier) = @_;
    # declare local variables
    my ($group);
    # make sure new_group_statement was called on a collection statement
    # (these are OBJECT and GROUP statements as well as the root node
    #  which contains the entire ODL file.)
    if ($self->statement_is_collection ())
    {
	# make sure group identifier given is a valid identifier
	if ($groupIdentifier =~ /^$re_identifier$/)
	{
	    # create ODL GROUP statement with given keyword
	    $group = bless({TYPE       => $types->{GROUP},
			    IDENTIFIER => uc($groupIdentifier),
			    VALUE      => []});
	    # add to collection object new_group_statement was called on
	    push (@{$self->{STATEMENTS}}, $group);
	    # return the new object
	    return $group;
	}
	else
	{
	    # invalid identifier, display warning
	    warn ("WARNING - group identifier $groupIdentifier invalid");
	}
    }
    else
    {
	# function called on incompatable object, display warning
	warn ("WARNING - new_group_statement  on a non collection statement");
    }
    # if we got here, there was a problem. return undef
    return undef;
}

sub new_attribute_statement ($$)
{
    # get self and parameters
    my ($self, $identifier, $value) = @_;
    # declare local variables
    my ($valueObj);
    # make sure new_addribute_statement was called on a collection statement
    # (these are OBJECT and GROUP statements as well as the root node
    #  which contains the entire ODL file
    if ($self->statement_is_collection ())
    {
	# make sure attribute identifier given is a valid identifier
	if ($identifier =~ /^($re_identifier\:)?$re_identifier$/)
	{
	    # create value object from given value
	    $valueObj = finish_top_level_value ($value);
	    # make sure value was parsable
	    if ($valueObj->{TYPE} != $types->{UNKNOWN_TYPE})
	    {
		# create ODL attribute statement with given keyword and value
		$attribute = bless({TYPE       => $types->{ATTRIBUTE},
				    IDENTIFIER => uc($identifier),
				    VALUE      => $valueObj});
		# add to collection object new_attribute_statement was called
		# on
		push (@{$self->{STATEMENTS}}, $attribute);
		# return the new object
		return $attribute;
	    }
	    else
	    {
		# value not parsable, display warning
		warn ("WARNING - attribute created with invalid value $value");
	    }
	}
	else
	{
	    # invalid identifier, display warning
	    warn ("WARNING - attribute identifier $identifier invalid");
	}
    }
    else
    {
	# function called on incompatable object, display warning
	warn ("WARNING - new_attribute_statement on a non collection statement");
    }
    # if we got here, there was a problem. return undef
    return undef;
}

sub new_pointer_statement ($$)
{
    # get self and parameters
    my ($self, $identifier, $value) = @_;
    # make sure new_pointer_statement was called on a collection statement
    # (these are OBJECT and GROUP statements as well as the root node
    #  which contains the entire ODL file.)
    if ($self->statement_is_collection ())
    {
	# add a leading caret (^) to the pointer if non exists
	$identifier =~ s/^([^\^])/\^$1/;
	# make sure the identifier given is a valid identifier
	if ($identifier =~ /^\^$re_identifier$/)
	{
	    # create value object from given value
	    $value = finish_top_level_value ($value);
	    # make sure value was parsable
	    if ($valueObj->{TYPE} != $types->{UNKNOWN_TYPE})
	    {
		# create ODL pointer statement with given keyword and value
		$pointer = bless({TYPE       => $types->{POINTER},
				  IDENTIFIER => uc($identifier),
				  VALUE      => $value});
		# add to collection object new_pointer_statement was called on
		push (@{$self->{STATEMENTS}}, $pointer);
		# return new object
		return $pointer;
	    }
	    else
	    {
		# value not parsable, display warning
		warn ("WARNING - pointer created with invalid value $value");
	    }
	}
	else
	{
	    # invalid identifier, display warning
	    warn ("WARNING - pointer identifier $identifier invalid");
	}
    }
    else
    {
	# function called on incompatable object, display warning
	warn ("WARNING - new_pointer_statement on a non collection statement");
    }
    # if we got here, there was a problem, return undef.
    return undef;
}

=head1

ODL::get_pointer_identifiers () returns a list of all pointer identifiers that
exist within the ODL structure statement list, or undef if the ODL object isn't
the root-node, or an object or group statement.

=cut

sub get_pointer_identifiers ()
{
    # get self
    my ($self) = @_;
    # declare local variables
    my @identifiers;
    # return undef if statement is not a collection of substatements
    return undef unless ($self->statement_is_collection ());
    # extract identifiers from all pointer statements
    push (@identifiers, $_->{IDENTIFIER})
	foreach grep {$_->statement_is_pointer()} @{$self->{STATEMENTS}};
    # return array, or undef if there are no pointers
    return ($#identifiers >= 0)? @identifiers : undef;
}

sub get_attribute_identifiers ()
{
    # get self
    my ($self) = @_;
    # declare local variables
    my (@identifiers, $s, $identifier);
    # return undef if statement is not a collection of substatements
    return undef unless ($self->statement_is_collection ());
    # foreach statement that is an attribute
    foreach $s (grep {$_->statement_is_attribute()} (@{$self->{STATEMENTS}}))
    {
	# get the identifier
	$identifier = $s->{IDENTIFIER};
	# prepend the namespace if present
	$identifier = "$s->{NAMESPACE}:".$identifier if ($s->{NAMESPACE});
	# add the identifier to the list
	push (@identifiers, $identifier);
    }
    # return the identifiers
    return @identifiers;
}

sub get_group_identifiers ()
{
    # get self
    my ($self) = @_;
    # declare local variables
    my @identifiers;
    # return undef if statement is not a collection of substatements
    return undef unless ($self->statement_is_collection ());
    # extract identifiers from all group statements
    push (@identifiers, $_->{IDENTIFIER})
	foreach grep {$_->statement_is_group()} @{$self->{STATEMENTS}};
    # return identifier array
    return @identifiers;
}

sub get_object_identifiers ()
{
    # get self
    my ($self) = @_;
    # declare local variables
    my @identifiers;
    # return undef if statement is not a collection of substatements
    return undef unless ($self->statement_is_collection ());
    # extract identifiers from all object statements
    push (@identifiers, $_->{IDENTIFIER})
	foreach grep {$_->statement_is_object()} @{$self->{STATEMENTS}};
    # return identifier array
    return @identifiers;
}

sub get_collection_identifiers ()
{
    # get self
    my ($self) = @_;
    # declare local variables
    my @identifiers;
    # return undef if statement is not a collection of substatements
    return undef unless ($self->statement_is_collection ());
    # extract identifiers from all collection statements
    push (@identifiers, $_->{IDENTIFIER})
	foreach grep {$_->statement_is_collection()} @{$self->{STATEMENTS}};
    # return identifier array
    return @identifiers;
}

sub get_all_identifiers ()
{
    # get self
    my ($self) = @_;
    # declare local variables
    my @identifiers;
    # return undef if statement is not a collection of substatements
    return undef unless ($self->statement_is_collection ());
    # for each substatement
    foreach $stmnt (@{$self->{STATEMENTS}})
    {
	if ($stmnt->{TYPE} == $types->{ATTRIBUTE})
	{
	    # substatement is an attribute
	    # get the identifier
	    $identifier = $stmnt->{IDENTIFIER};
	    # add the namespace if given
	    $identifier = $stmnt->{NAMESPACE}.":".$identifier
		if ($stmnt->{NAMESPACE});
	    # add the identifier to the list
	    push (@identifiers, $identifier);
	}
	else
	{
	    # substatement is not an attribute
	    # just add the identifier
	    push (@identifiers, $stmnt->{IDENTIFIER});
	}
    }
    # return the list of identifiers
    return @identifiers;
}

sub get_statement ($;$)
{
    # get parameters
    my ($self, $identifier, $index) = @_;
    # declare local variables
    my ($object_name, $namespace, $id);
    # return undef if statement not collection or invalid index given
    return undef unless ($self->statement_is_collection () and
			 (not defined ($index) or $index > 0));
    # default index to one if not given
    $index = 1 unless (defined ($index));
    # set identifier to upper case (since case is irrelevant, upper case is
    # always used.)
    $identifier = uc ($identifier);
    # If identifier has OBJECT or GROUP identifiers, traverse through
    # ODL tree to get final statement.
    # If $identifier is IMAGE.CREATOR.NAME is given, the CREATOR
    # statement will be stored in $self when this loop is finished,
    # and $identifier will contain NAME
    while ($identifier =~ /^($re_identifier)\.(.+)$/)
    {
	# split object name and identifier from regex
	($object_name, $identifier) = ($1, $2);
	# get substatement (must be collection)
	$self = $self->get_collection ($object_name);
	# return failure if no such substatement exists
	return undef unless ($self);
    }
    # extract namespace and identifier if namespace exists
    ($namespace, $identifier) = (uc($1), uc($2))
	if ($identifier =~ /^($re_identifier)\:($re_identifier)$/);
    # foreach statement in the final collection statement...
  search_value:
    foreach $statement (@{$self->{STATEMENTS}})
    {
	# if the namespace and identifier match
	if ($statement->{NAMESPACE}  eq $namespace and
	    $statement->{IDENTIFIER} eq $identifier)
	{
	    # return statement if it's the one specified by $index
	    return $statement if ($index == 1);
	    # decrement $index, so that if $index is given as 2, the second
	    # time a matching statement is found it is returned, etc.
	    --$index;
	}
    }
    # if we got here, no matching statement was found, return failure
    return undef;
}

sub get_units ()
{
    # get self
    my ($self) = @_;
    # get value object from statement if statement has value
    $self = $self->{VALUE} if ($self->statement_has_value ());
    # return failure if not a value or value isn't a number
    return undef unless ($self->is_value () and
			 $self->value_is_number ());
    # extract units from value, if present, and return
    return $1 if ($self->{VALUE} =~ /^.*\<(.*)\>/);
    # no units, return failure.
    return undef;
}

sub get_number ()
{
    # get self
    my ($self) = @_;
    # get value object from statement if statement has value
    $self = $self->{VALUE} if ($self->statement_has_value ());
    # return failure if not a value or value isn't a number
    return undef unless ($self->is_value() and
			 $self->value_is_number ());
    # return numeric portion (converted to a perl scalar number by adding zero)
    # if the value has a numeric field
    return $1 + 0 if ($self->{VALUE} =~ /^([^\<]*)(\<.*\>)?$/);
    # no number defined, return failure.
    return undef;
}

sub get ($;$$)
{
    # get self and parameters
    my ($self, $identifier, $index1, $index2) = @_;
    # get statement identified by identifier
    $statement = $self->get_statement ($identifier);
    # return failure if no such statement exists
    return undef unless $statement;
    # return value indexed by index1 and index2 unless the statement is just
    # a scalar
    return $statement->get_value ($index1, $index2)
	unless $statement->value_is_scalar();
    # return value unless indexes given (not allowed at this point, since the
    # statement is just a scalar.)
    return $statement->get_value_object ()->to_string ()
	unless (defined ($index1) or defined ($index2));
    # indexes given on scalar value, display warning and return failure
    warn ("Indexes used retrieving scalar not contained in a list or set.\n");
    return undef;
}

# TODO: handle case when identifier and indexes don't retrieve a scalar value.
sub get_noq ($;$$)
{
    # get self and parameters
    my ($self, $identifier, $index1, $index2) = @_;
    # declare local variables
    my ($value);
    # get value as defined by identifier and indexes
    $value = $self->get ($identifier, $index1, $index2);
    # return failure if value doesn't exist
    return undef unless $value;
    # extract quotes
    $value = $1 if ($value =~ /^\"(.*)\"$/);
    # return value
    return $value;
}

sub get_value (;$$)
{
    # get self and parameters
    my ($self, $index1, $index2) = @_;
    # get the value object associated with the given (optional) indices
    $obj = $self->get_value_object ($index1, $index2);
    # return the value object is the resulting object is a sequence or set
    return $obj if ($obj and ($obj->value_is_sequence () or
			      $obj->value_is_set ()));
    # return the value scalar if the resulting object exists (and is a scalar)
    return $obj->{VALUE} if ($obj);
    # no object, return failure
    return undef;
}

sub get_size (;$)
{
    # get self and parameters
    my ($self, $index) = @_;
    # get the value object associated with the statement
    $obj = $self->get_value_object ();
    # make sure the value is the kind that can have a size, and that
    # if an index is given, the value is a 2d sequence
    return undef if (not ($obj->value_is_sequence () or
			  $obj->value_is_set ())
		     or
		     (defined ($index) and
		      not $obj->value_is_2D_sequence ()));
    # get the list of subvalues
    @arr = @{$obj->{VALUE}};
    # get the sublist if $index was given
    @arr = @{$arr[$index - 1]->{VALUE}} if (defined ($index));
    # return the size of the array
    return $#arr;
}

sub get_type (;$$)
{
    # get self and parameters
    my ($self, $index1, $index2) = @_;
    # return the type string associated with the value at the given index
    return get_type_string ($self->get_type_code ($index1, $index2));
}

sub get_type_code (;$$)
{
    # get self and parameters
    my ($self, $index1, $index2) = @_;
    # get the value object at the given index
    $self = $self->get_value_object ($index1, $index2)
	unless (not $self or $self->statement_is_collection ());
    # return the type code if a value exists at the given index
    return $self->{TYPE} if ($self);
    # no value at given index, return failure
    return undef;
}

sub get_type_string ($)
{
    # get parametrs
    my ($type_code) = @_;
    # get key associated with given type code
    my @results = grep {$types->{$_} == $type_code} (keys (%{$types}));
    # return key if one exists
    return $results[0] if ($#results == 0);
    # no such type code, return failure
    return undef;
}

sub get_value_object (;$$)
{
    # get self and parameters
    my ($self, $index1, $index2) = @_;
    # make sure statement isn't a collection (GROUP, OBJECT, or root node)
    return undef if ($self->statement_is_collection ());
    # get value object if given a statement
    $self = $self->{VALUE} if ($self->is_statement ());
    # if the first index is given:
    if (defined ($index1))
    {
	# shift one (perl starts at zero, ODL starts at one)
	$index1 = $index1 - 1;
	# make sure we're working with a sequence or set
	return undef unless ($self->value_is_sequence () or
			     $self->value_is_set ());
	# if we have a second index:
	if (defined ($index2))
	{
	    # shift one (perl starts at zero, ODL starts at one)
	    $index2 = $index2 - 1;
	    # make sure we have a two dimensional list for the two indices
	    return undef unless ($self->value_is_2D_sequence ());
	    # return the value at the given index
	    $value = $self->{VALUE}->[$index1]->{VALUE}->[$index2];
	}
	else
	{
	    # return the value at the given index
	    $value = $self->{VALUE}->[$index1];
	}
    }
    else
    {
	# make sure we have a value. (no indices)
	return undef unless ($self->is_value ());
	# save the value
	$value = $self;
	#return undef if ($self->statement_is_collection ());
	#return $self->{VALUE}->{VALUE};
    }
    # return the value if it exists
    return $value if ($value);
    # no such value, return failure
    return undef;
}

sub get_object ($;$)
{
    # get self and parameters
    my ($self, $object_identifier, $index) = @_;
    # make sure we have a collection and the index given is valid
    return () unless ($self->statement_is_collection () and
		      (not defined ($index) or $index > 0));
    # default index to one
    $index = 1 unless (defined ($index));
    # make the object identifier uppercase
    $object_identifier = uc ($object_identifier);
    # foreach matching OBJECT
    foreach $statement grep {($_->statement_is_object() and
			      $_->{IDENTIFIER} eq $object_identifier)}
                       (@{$self->{STATEMENTS}})
    {
#	if ($statement->statement_is_object () and
#	    $statement->{IDENTIFIER} eq $object_identifier)
#	{
	# return the statement if we've counted down to one using $index
	return $statement if ($index == 1);
	# decrement $index
	--$index;
#	}
    }
    return undef;
}

sub get_group ($;$)
{
    # get parameters
    my ($self, $object_identifier, $index) = @_;
    # make sure statement is a collection and the index given is valid
    return () unless ($self->statement_is_collection () and
		      (not defined ($index) or $index > 0));
    # default index to one
    $index = 1 unless (defined ($index));
    # set the object identifier to upper case
    $object_identifier = uc ($object_identifier);
    # foreach matching OBJECT statement
    foreach $statement grep {($_->statement_is_object() and
			      $_->{IDENTIFIER} eq $object_identifier)}
                       (@{$self->{STATEMENTS}})
    {
#	if ($statement->statement_is_group () and
#	    $statement->{IDENTIFIER} eq $object_identifier)
#	{
	# return the statement if we've counted down to one using $index
	return $statement if ($index == 1);
	# decrement index
	--$index;
#	}
    }
    # no such statement found, return undef
    return undef;
}

sub get_collection ($;$)
{
    # self and get parameters
    my ($self, $object_identifier, $index) = @_;
    # make sure statement is collection and index is valid
    return () unless ($self->statement_is_collection () and
		      (not defined ($index) or $index > 0));
    # default index to one
    $index = 1 unless (defined ($index));
    # make object identifier upper case
    $object_identifier = uc ($object_identifier);
    # foreach matching OBJECT statement
    foreach $statement grep {($_->statement_is_object() and
			      $_->{IDENTIFIER} eq $object_identifier)}
                       (@{$self->{STATEMENTS}})
    {
#	if ($statement->statement_is_collection() and
#	    $statement->{IDENTIFIER} eq $object_identifier)
#	{
	# return the statement if we've counted down to one using $index
	return $statement if ($index == 1);
	# decrement index
	--$index;
#	}
    }
    # no matching collection, return undef
    return undef;
}

sub get_pointer ($;$)
{
    # get self and parameters
    my ($self, $pointer_identifier, $index) = @_;
    # make sure statement is a collection and index is valid
    return () unless ($self->statement_is_collection () and
		      (not defined ($index) or $index > 0));
    # default index to one
    $index = 1 unless (defined ($index));
    # make pointer identifier upper case
    $pointer_identifier = uc ($pointer_identifier);
    # for each matching pointer...
    foreach $statement grep {($_->statement_is_pointer() and
			      $_->{IDENTIFIER} eq $pointer_identifier)}
                       (@{$self->{STATEMENTS}})
    {
#	if ($statement->statement_is_pointer () and
#	    $statement->{IDENTIFIER} eq $pointer_identifier)
#	{
	# return the statement if we've counted down to one using $index
	return $statement if ($index == 1);
	# decrement index
	--$index;
#	}
    }
    # no such pointer, return undef
    return undef;
}

sub get_attribute ($;$)
{
    # get self and parameters
    my ($self, $attribute_identifier, $index) = @_;
    # make sure statement is collection and index is valid
    return () unless ($self->statement_is_collection () and
		      (not defined ($index) or $index > 0));
    # default index to one
    $index = 1 unless (defined ($index));
    # make attribute identifier upper case
    $attribute_identifier = uc ($attribute_identifier);
    # foreach matching attribute...
    foreach $statement grep {($_->statement_is_attribute() and
			      $_->{IDENTIFIER} eq $attribute_identifier)}
                            (@{$self->{STATEMENTS}})
    {
#	if ($statement->statement_is_attribute () and
#	    $statement->{IDENTIFIER} eq $attribute_identifier)
#	{
	# return the statement if we've counted down to one using $index
	return $statement if ($index == 1);
	# decrement index
	--$index;
#	}
    }
    # no matching attribute, return undef
    return undef;
}

=head1 

Loading content

=cut

sub load_from_file ($)
{
    # get parameters
    my ($self, $filename) = @_;
    # open the file
    if (not open (ODL_STREAM, "<", $filename))
    {
	# open failed, print warning and return undef
	warn ("WARNING - Could not open ODL file at $filename");
	return undef;
    }
    # save filename
    $self->{FILENAME} = $filename;
    # load from file
    $self = $self->load_from_stream (ODL_STREAM);
    # close file
    close (ODL_STREAM);
    # return self
    return $self;
}

sub load_from_stream ($)
{
    # get self, parameters
    my ($self, $stream) = @_;
    # make sure statement is the root node
    return undef unless ($self->{TYPE} == $types->{ROOT_NODE});
    # create empty statement array
    $statements = [];
    # read statements from the stream
  read_odl_stream:
    while ($statement = read_odl_statement ($stream))
    {
	# skip errors
	next read_odl_stream
	    if ($statement->{TYPE} == $types->{STATEMENT_ERROR});
	# stop on the END label
	last read_odl_stream
	    if ($statement->{TYPE} == $types->{END_LABEL});
	# set statements parent to self
	$statement->{PARENT} = $self;
	# END_OBJECT and END_GROUP lines should be found and discarded
	# by read_odl_statement. If it returns any, it means there was an
	# error in the stream
	if ($statement->{TYPE} == $types->{END_OBJECT} or
	    $statement->{TYPE} == $types->{END_GROUP})
	{
	    # print warning
	    warn ("WARNING - error in odl stream".
		  $last_line_read);
	    # stop reading
	    last read_odl_stream;
	}
	# add statement
	push (@{$statements}, $statement);
    }
    # set statement array
    $self->{STATEMENTS} = $statements;
    # return self
    return $self;
}

=head1

ODL::set_output_mode($) takes either an integer from 1 to 4 or one of the
following strings: "COMPRESSED", "SPACED", "LINED", or "EXPANDED" and sets
the output mode for the ODL object to the parameter (1 = compressed,
2 = spaced, etc.). The strings are not case sensitive.

=cut

sub set_output_mode ($)
{
    # get self, parameters
    my ($self, $output_mode) = @_;
    # set output mode if string given
    $output_mode = 1 if ($output_mode =~ /^COMPRESSED$/i);
    $output_mode = 2 if ($output_mode =~ /^SPACED$/i);
    $output_mode = 3 if ($output_mode =~ /^LINED$/i);
    $output_mode = 4 if ($output_mode =~ /^EXPANDED$/i);
    # don't set unless output mode is valid
    return undef unless ($output_mode == 1 or
			 $output_mode == 2 or
			 $output_mode == 3 or
			 $output_mode == 4);
    # set output mode
    $self->{OUTPUTMODE} = $output_mode;
    # return success, new output moden
    return $self->{OUTPUTMODE}
}

=head1

ODL::get_output_mode() returns the string representation of the output mode set
for the object. Possible values are "COMPRESSED", "SPACED", "LINED", and
"EXPANDED". If none is specified for the given object, undef is returned.

=cut

sub get_output_mode ()
{
    # get self
    my ($self) = @_;
    # declare local variables
    my (@results);
    # get key for current output mode
    @results = grep {$self->{OUTPUTMODE} == $output_modes->{$_}}
                    (keys (%{$output_modes}));
    # return key if known
    return $results[0] if ($#results == 0);
    # key not known, return undef
    return undef;
}

=head1

ODL::write_to_file ($;$) takes a file name and an optional output mode, opens
the file and calls write_to_stream on the file.

=cut

sub write_to_file ($;$)
{
    # get self, parameters
    my ($self, $filename, $output_mode) = @_;
    # open file
    if (not open (ODL_STREAM, ">", $filename))
    {
	# open failed, return undef
	warn ("WARNING - Couldn't open output file for writing: $filename");
	return FALSE;
    }
    # write object to stream
    $self->write_to_stream (ODL_STREAM, $output_mode);
    # close file
    close (ODL_STREAM);
    # return success
    return TRUE;
}

=head1

ODL::write_to_stream (;$$$) takes optional stream, output mode, and prefix 
parameters and writes to the stream (or STDOUT if none given) the ODL statement
formatted in the appropriate output mode. Note for object or group statements:
all internal statements will be written to the stream with an additional "  "
prefix to the given prefix. This will insure propper indentation.


=cut

sub write_to_stream (;$$$)
{
    # get self, parameters
    my ($self, $odl_stream, $output_mode, $prefix) = @_;
    # default stream to STDOUT if none given
    $odl_stream = STDOUT unless ($odl_stream);
    # default output mode to objects output mode if none given
    $output_mode = $self->{OUTPUTMODE} if (defined ($self->{OUTPUTMODE}) and
					   $output_mode == undef);
    # set default output mode if none found sofar
    $output_mode = $output_modes_default if ($output_mode == undef);
    # declare local variables
    my ($line);
    # check type of statement and handle appropriately
    if ($self->statement_has_value ())
    {
	# statement is attribute or pointer, convert to string and print
	print {$odl_stream} ($self->to_string ($prefix, $output_mode));
    }
    elsif ($self->statement_is_root_node ())
    {
	# statement is entire ODL file.
	# write each statement
	$_->write_to_stream ($odl_stream,
			     $_->{OUTPUTMODE} || $output_mode,
			     $prefix)
	    foreach (@{$self->{STATEMENTS}});
	# add the end statement
	print {$odl_stream} ("END\n");
    }
    elsif ($self->statement_is_object ())
    {
	# statement is an OBJECT statement
	# print the OBJECT header
	print {$odl_stream} $prefix."OBJECT = ".$self->{IDENTIFIER}."\n";
	# print each substatement with a prefix of two additional spaces
	$_->write_to_stream ($odl_stream,
			     $_->{OUTPUTMODE} || $output_mode,
			     $prefix."  ")
	    foreach (@{$self->{STATEMENTS}});
	# print the OBJECT footer
	print {$odl_stream} $prefix."END_OBJECT = ".$self->{IDENTIFIER}."\n";
    }
    elsif ($self->statement_is_group ())
    {
	# statement is a GROUP statement
	# print the GROUP header
	print {$odl_stream} $prefix."GROUP = ".$self->{IDENTIFIER}."\n";
	# print each substatement with a prefix of two additional spaces
	$_->write_to_stream ($odl_stream,
			     $_->{OUTPUTMODE} || $output_mode,
			     $prefix."  ")
	    foreach (@{$self->{STATEMENTS}});
	# print the GROUP footer
	print {$odl_stream} $prefix."END_GROUP = ".$self->{IDENTIFIER}."\n";
    }
    else
    {
	# unknown type, display warning and return failure
	warn ("write_to_stream called on incompatable type: ".
	      $self->get_type()."\n");
	return undef;
    }
    return TRUE;
}

=head1

ODL::to_string($$) takes a prefix string and output mode and returns a string
representation of the ODL statement or value which can be appended to the
prefix and keep the 80-character line limit (including the trailing \r\n)

Note that the prefix argument is not returned as part of the string.

=cut

sub to_string ($$)
{
    # get self and parameters
    my ($self, $prefix, $output_mode) = @_;
    # declare local variables
    my ($line, $all, $pre, $indent, $index, $element, $start, $end);
    # save prefix
    $pre = $prefix;
    # check to see if $self is a statement or value
    if ($self->is_statement ())
    {
	# see if $self contains a value
	if (not $self->statement_has_value ())
	{
	    # no value (collection), print warning and return undef
	    warn ("WARNING - to_string called on a root node, group, ".
		  "or object statement.\n");
	    return undef;
	}
	if (not $pre =~ /^\s*$/)
	{
	    # statement with value, but prefix is not entirely whitespace
	    # non-whitespace prefixes are used for to_string on values
	    # print warning and return undef
	    warn ("WARNING - to_string called on a statement with a prefix ".
		  "containing non whitespace. Prefix:\"$prefix\"");
	    return undef;
	}
	# begin with prefix
	$line = $pre;
	#$line .= "^" if $self->statement_is_pointer ();
	# add namespace and identifier
	$line .= $self->{NAMESPACE}.":" if ($self->{NAMESPACE} ne "");
	$line .= $self->{IDENTIFIER}." = ";
	# add value, properly indented
	$line .= $self->{VALUE}->to_string ($line, $output_mode)."\r\n";
    }
    else
    {
	# $self is a value
	if ($self->value_is_scalar ())
	{
	    # $self is a scalar value
	    if ($self->value_is_string ())
	    {
		# $self is a string
		# just return the string if it's short enough
		return $self->{VALUE}
		    if (length ($pre.$self->{VALUE}) <= 78);
		# populate buffer with the value
		$buffer = $self->{VALUE};
		# check the output mode and setup variables appropriately
		if ($output_mode == $output_modes->{COMPRESSED})
		{
		    # conserve space
		    $all = "";
		    $line = $pre;
		    $pre = "";
		}
		elsif ($output_mode == $output_modes->{SPACED})
		{
		    # conserve a little space
		    $all = "";
		    $line = $pre;
		    $pre = " " x (length ($pre) + 1);
		}
		else
		{
		    # conserve very little space
		    $all = $pre . "\"\r\n";
		    substr ($buffer, 0, 1) = "";
		    $pre =~ s/^(\s*).*/$1  /;
		    $line = $pre;
		}
	      write_string:
		while ($buffer ne "")
		{
		    if (length ($line.$buffer) <= 78)
		    {
			$all .= $line.$buffer."\r\n";
			last write_string;
		    }
		    $index = rindex ($buffer, " ", 78 - length ($line));
		    if ($index == -1)
		    {
			$all .= $line.
			    substr ($buffer, 0, 77 - length ($line))."-\r\n";
			$buffer = substr ($buffer, 78 - length ($line));
			$line = $pre;
		    }
		    else
		    {
			$all .= $line.substr ($buffer, 0, $index)."\r\n";
			$buffer = substr ($buffer, $index + 1);
			$line = $pre;
		    }
		}
		substr ($all, 0, length ($prefix)) = "";
		substr ($all, length ($all) - 2) = "";
		return $all;
	    }
	    else
	    {
		# return the value with a new line, if it can fit
		return $self->{VALUE}
		    if (length ($pre.$self->{VALUE}) <= 78);
		# if it can't even fit on one line, just return the value on
		# it's own line
		return "\r\n".$self->{VALUE}
		    if (length ($self->{VALUE}) > 78);
		# it has to go on it's own line, so we should right justify
		# ex.
 		# |<-- 80 chars, (hiding \r\n) --->|
		# /--------------------------------\
		# |IDENTIFIER =                    |
                # |       VERY_VERY_VERY_LONG_VALUE|
		# \--------------------------------/
		# get the amount we should index
		$indent = 78 - length ($self->{VALUE});
		return "\r\n".(" " x $indent).$self->{VALUE};
	    }
	}
	elsif ($self->value_is_set () or
	       $self->value_is_1D_sequence ())
	{
	    if ($self->value_is_set ())
	    {
		# sets use braces
		$start = "{";
		$end = "}";
	    }
	    else
	    {
		# sequences use parenthesis
		$start = "(";
		$end = ")";
	    }
	    if ($output_mode == $output_modes->{COMPRESSED} or
		$output_mode == $output_modes->{SPACED})
	    {
		$all = "";
		$line = $prefix.$start;
		$delimeter = ",";
		$delimeter .= " " if ($output_mode == $output_modes->{SPACED});
		foreach $element (@{$self->{VALUE}})
		{
		    if (length ($line.$element->{VALUE}.$delimeter) > 78)
		    {
			$all .= $line."\r\n";
			if ($output_mode == $output_modes->{SPACED} and
			    length ($prefix.$element->{VALUE}.$delimeter) < 78)
			{
			    $line = " " x (length ($prefix) + 1).
				$element->{VALUE}.$delimeter;
			}
			elsif ($output_mode == $output_modes->{COMPRESSED})
			{
			    $line = $element->{VALUE}.$delimeter;
			}
			else
			{
			    $line = $element->{VALUE}.",";
			    $line = (" " x (78 - length ($line))).$line;
			}
		    }
		    else
		    {
			$line .= $element->{VALUE}.$delimeter;
		    }
		}
		$all .= $line;
		substr ($all, 0, length ($prefix)) = "";
		$all =~ s/\,\s*$//;
		$all .= $end;
		return $all;
	    }
	    else
	    {
		$all = "";
		$line = $prefix.$start;
		foreach $element (@{$self->{VALUE}})
		{
		    if (length ($line.$element->{VALUE}) <= 77)
		    {
			$all .= $line.$element->{VALUE}.",\r\n";
			$line = " " x (length ($prefix) + 1);
		    }
		    elsif ($line =~ /^\s*$/)
		    {
			$line = $element->{VALUE}.",";
			$all .= (" " x (78 - length ($line))).$line."\r\n";
			$line = " " x (length ($prefix) + 1);
		    }
		    else
		    {
			$all .= $line."\r\n";
			$line = $element->{VALUE}.",";
			$all .= (" " x (78 - length ($line))).$line."\r\n";
			$line = " " x (length ($prefix) + 1);
		    }
		}
		$all .= $line;
		substr ($all, 0, length ($prefix)) = "";
		$all =~ s/\,\s*$//;
		$all .= $end;
		return $all;
	    }
	}
	elsif ($self->value_is_2D_sequence ())
	{
	    if ($output_mode == $output_modes->{COMPRESSED} or
		$output_mode == $output_modes->{SPACED})
	    {
		$d1 = ",";
		$d2 = "";
		if ($output_mode == $output_modes->{SPACED})
		{
		    $d1 .= " ";
		    $d2 .= " ";
		}
		$all = "";
		$line = $prefix."(";
		foreach $sequence (@{$self->{VALUE}})
		{
		    if (length ($line) > 77)
		    {
			$all .= $line."\r\n";
			if ($output_mode == $output_modes->{SPACED})
			{
			    $line = " " x (length ($prefix) + 1)."(";
			}
			else
			{
			    $line = "(";
			}
		    }
		    else
		    {
			$line .= "(";
		    }
		    foreach $element (@{$sequence->{VALUE}})
		    {
			if (length ($line.$element->{VALUE}.$d1) > 78)
			{
			    $all .= $line."\r\n";
			    if ($output_mode == $output_modes->{SPACED} and
				length ($prefix.$element->{VALUE}.$d1) < 76)
			    {
				$line = " " x (length ($prefix) + 2).
				    $element->{VALUE}.$d1;
			    }
			    elsif ($output_mode == $output_modes->{COMPRESSED})
			    {
				$line = $element->{VALUE}.$d1;
			    }
			    else
			    {
				$line = $element->{VALUE}.",";
				$line = (" " x 78 - length ($line)).$line;
			    }
			}
			else
			{
			    $line .= $element->{VALUE}.$d1;
			}
		    }
		    $line =~ s/\,\s*$//;
		    if (length ($line) > 77 - length ($d2))
		    {
			$all .= $line."\r\n";
			if ($output_mode == $output_modes->{SPACED})
			{
			    $line = " " x (length ($prefix) + 1);
			}
			else
			{
			    $line = "";
			}
			$line .= ")".$d2;
		    }
		    else
		    {
			$line .= ")".$d2;
		    }
		}
		$line =~ s/\s*$//;
		if (length ($line) < 78)
		{
		    $line .= ")";
		}
		else
		{
		    $all .= $line;
		    if ($output_mode == $output_modes->{SPACED})
		    {
			$line = " " x (length ($prefix) + 1).")";
		    }
		    else
		    {
			$line = ")";
		    }
		}
		$all .= $line;
		substr ($all, 0, length ($prefix)) = "";
		return $all;
	    }
	    elsif ($output_mode == $output_modes->{LINED})
	    {
		$all = "";
		$line = $prefix."(";
		foreach $sequence (@{$self->{VALUE}})
		{
		    $line .= "(";
		    foreach $element (@{$sequence->{VALUE}})
		    {
			if (length ($line.$element->{VALUE}) <= 76)
			{
			    $line .= $element->{VALUE}.", ";
			}
			else
			{
			    if ($line =~ /^\s*$re_identifier\s*\=/)
			    {
				$all .= $line."\r\n";
				if (length ($prefix.$element->{VALUE}) <= 74)
				{
				    $line = " " x (length ($prefix) + 2).
					$element->{VALUE}.", ";
				}
				else
				{
				    $line = $element->{VALUE}.", ";
				    $line = " " x (78 - length ($line)).$line;
				}
			    }
			    else
			    {
				$line =~ /^(\s*)(.*)$/;
				($length, $rest) = (length($1), $2);
				if (length ($line.$element->{VALUE}) <= 76 and
				    $length < length ($prefix) + 2)
				{
				    $line = $rest;
				    $line .= $element->{VALUE}.", ";
				    $line = " " x (78 - length ($line)).$line;
				}
				else
				{
				    $all .= $line."\r\n";
				    if (length ($prefix.$element->{VALUE}) <= 74)
				    {
					$line = " " x (length ($prefix) + 2).
					    $element->{VALUE}.", ";
				    }
				    else
				    {
					$line = $element->{VALUE}.", ";
					$line = " " x (78 - length ($line)).$line;
				    }
				}
			    }
			}
		    }
		    $line =~ s/\,\s*$//;
		    if (length ($line) < 78)
		    {
			$all .= $line.")\r\n";
		    }
		    else
		    {
			$all .= $line."\r\n";
			$all .= " " x (length ($prefix) + 2).")\r\n";
		    }
		    $line = " " x (length ($prefix) + 1);
		}
		substr ($all, 0, length ($prefix)) = "";
		return $all;
	    }
	    else # ($output_mode == $output_modes->{EXPANDED})
	    {
		$all = "";
		$line = $prefix."(";
		foreach $sequence (@{$self->{VALUE}})
		{
		    $line .= "(";
		    foreach $element (@{$sequence->{VALUE}})
		    {
			if (length ($line.$element->{VALUE}) < 78)
			{
			    $line .= $element->{VALUE}.",";
			    $all .= $line."\r\n";
			    $line = " " x (length ($prefix) + 2);
			}
			else
			{
			    $all .= $line."\r\n"
				if ($line =~ /^\s*$re_identifier\s*\=/);
			    $line = $element->{VALUE}.",";
			    $all .= " " x (78 - length ($line)).$line."\r\n";
			    $line = " " x (length ($prefix) + 2);
			}
		    }
		    if (substr ($line, length ($line) - 1) eq "(")
		    {
			$line .= ")";
			$all .= $line."\r\n";
		    }
		    else
		    {
			$all =~ s/\,\s*$//;
			$all .= ")\r\n";
		    }
		    $line = " " x (length ($prefix) + 1);
		}
		$all =~ s/\s*$//;
		$last = rindex ($all, "\n");
		$line = substr ($all, $last + 1);
		if (length ($line) < 78)
		{
		    $all .= ")";
		}
		else
		{
		    $line =~ /^(\s*)(.*)$/;
		    ($length, $rest) = (length ($1), $2);
		    if ($length > length ($prefix))
		    {
			$all =~ s/\n[^\n]*$//;
			$all .= $rest.")";
		    }
		    else
		    {
			$all .= "\r\n";
			$all .= " " x length ($prefix).")";
		    }
		}
		substr ($all, 0, length ($prefix)) = "";
		return $all;
	    }
	}
	else
	{
	    return $self->{VALUE};
	}
    }
}

=head1

read_odl_statement ($) takes an odl stream and returns the next statement
from the stream. compound statements (like the object or group statements)
will result in additional calls to read_odl_statement, but will return a single
object or group statement object, which will contain all the substatements.

=cut

sub read_odl_statement ($)
{
    # get parameters
    my ($odl_stream) = @_;
    # declare local variables
    my ($statement, $line, $identifier, $value);
    # read statement, skipping blank lines
    readStatement: while ($line = read_odl_line ($odl_stream, TRUE))
    {
	# strip comments
	$line = strip_comments ($line);
	# stop loop when $line isn't blank after stripping comments
	last readStatement unless ($line eq "");
    }
    # if $line is still blank, there are no more statements
    return undef unless ($line);
    # check type of line and handle appropriately
    if ($line =~ /^(\^$re_identifier)\s*=\s*(.*\S)\s*$/)
    {
	# line is pointer statement
	# get upper cased identifier, and unmodifiied value
	($identifier, $value) = (uc ($1), $2);
	# finish reading the value if it spans multiple lines
	$value = finish_top_level_value ($value, $odl_stream);
	# construct the statement
	$statement = bless({TYPE       => $types->{POINTER},
			    IDENTIFIER => $identifier,
			    VALUE      => $value});
    }
    elsif ($line =~ /^(?:BEGIN_)?OBJECT\s*=\s*($re_identifier)\s*$/i)
    {
	# line is OBJECT statement
	# get upper cased identifier
	$identifier = uc ($1);
	# construct statement
	$statement = bless({TYPE       => $types->{OBJECT},
			    IDENTIFIER => $identifier,
			    STATEMENTS => []});
	# while a substatement can be read.
      read_object:
	while ($sub_statement = read_odl_statement ($odl_stream))
	{
	    # skip erroneous statements
	    next read_object
		if ($sub_statement->{TYPE} == $types->{STATEMENT_ERROR});
	    # set substatements parent
	    $sub_statement->{PARENT} = $statement;
	    # handle statements
	    if ($sub_statement->{TYPE} == $types->{END_OBJECT})
	    {
		# make sure identifier used in END_OBJECT matches that
		# used in BEGIN_OBJECT
		if ($sub_statement->{IDENTIFIER} ne "" and
		    $sub_statement->{IDENTIFIER} ne $statement->{IDENTIFIER})
		{
		    # identifiers don't match, print warning
		    warn ("WARNING - END_OBJECT statement for ".
			  $sub_statement->{IDENTIFIER}.
			  "found while not in such an OBJECT\n".
			  "Statement ignored.\n".
			  $last_line_read);
		}
		else
		{
		    last read_object;
		}
	    }
	    elsif ($sub_statement->{TYPE} == $types->{END_GROUP})
	    {
		# END_GROUP found illegally, print warning
		warn ("WARNING - END_GROUP statement found while "
		      ."not in a GROUP\n".
		      "Statement ignored.\n".
		      $last_line_read);
	    }
	    else
	    {
		# add the statement to the list
		push (@{$statement->{STATEMENTS}}, $sub_statement);
	    }
	}
	if (not $sub_statement)
	{
	    # $sub_statement should be the END_OBJECT statement now.
	    # if not (and blank) an error occured.
	    warn ("WARNING - End of ODL stream found while reading ".
		  "within an OBJECT\n".
		  $last_line_read);
	}
    }
    elsif ($line =~ /^(?:BEGIN_)?GROUP\s*=\s*($re_identifier)\s*$/i)
    {
	# line is GROUP statement
	# get upper-cased identifier
	$identifier = uc ($1);
	# construct object
	$statement = bless({TYPE       => $types->{GROUP},
			    IDENTIFIER => $identifier,
			    STATEMENTS => []});
	# read substatements
      read_group:
	while ($sub_statement = read_odl_statement ($odl_stream))
	{
	    # skip erroneous statements
	    next read_object
		if ($sub_statement->{TYPE} == $types->{STATEMENT_ERROR});
	    # set parent
	    $sub_statement->{PARENT} = $statement;
	    if ($sub_statement->{TYPE} == $types->{END_GROUP})
	    {
		# handle statements
		if ($sub_statement->{IDENTIFIER} ne "" and
		    $sub_statement->{IDENTIFIER} ne $statement->{IDENTIFIER})
		{
		    # END_OBJECT identifier found with illegal keyword
		    # display warning
		    warn ("WARNING - END_GROUP statement for ".
			  $sub_statement->{IDENTIFIER}.
			  "found while not in such a GROUP\n".
			  "Statement ignored.\n".
			  $last_line_read);
		}
		else
		{
		    # END_GROUP found, stop reading group
		    last read_group;
		}
	    }
	    elsif ($sub_statement->{TYPE} == $types->{END_OBJECT})
	    {
		# END_OBJECT found illegally, display warning
		warn ("WARNING - END_OBJECT statement found while not ".
		      "in an OBJECT\n".
		      "Statement ignored.\n".
		      $last_line_read);
	    }
	    else
	    {
		# add statement
		push (@{$statement->{STATEMENTS}}, $sub_statement);
	    }
	}
	if (not $sub_statement)
	{
	    # sub statement should be the END_GROUP statement. If missing,
	    # an error occured
	    warn ("WARNING - End of ODL stream found while reading ".
		  "within a GROUP\n".
		  $last_line_read);
	}
    }
    elsif ($line =~ /^END_OBJECT(?:\s*=\s*($re_identifier))?\s*$/i)
    {
	# END_OBJECT keyword found, get upper-case identifier
	$identifier = uc ($1);
	# create statement
	$statement = bless({TYPE       => $types->{END_OBJECT},
			    IDENTIFIER => $identifier});
    }
    elsif ($line =~ /^END_GROUP(?:\s*=\s*($re_identifier))?\s*$/i)
    {
	# END_GROUP keyword found, get upper-case identifier
	$identifier = uc ($1);
	# create statement
	$statement = bless({TYPE       => $types->{END_GROUP},
			    IDENTIFIER => $identifier});
    }
    elsif ($line =~ /^(?:($re_identifier)\:)?($re_identifier)\s*=(?:\s*(.*\S))?\s*$/)
    {
	# ATTRIBUTE statement found, get namespace, identfiier, and value
	($namespace, $identifier, $value) = (uc ($1), uc ($2), $3);
	if ($value eq "")
	{
	    # handle no-value attributes
	    warn ("WARNING - Attribute statement found with no value.\n".
		  $last_line_read);
	    # construct value
	    $value = bless({TYPE  => $types->{BLANK_TYPE},
			    VALUE => undef});
	    # construct statement
	    $statement = bless({TYPE       => $types->{ATTRIBUTE},
				NAMESPACE  => $namespace,
				IDENTIFIER => $identifier,
				VALUE      => $value});
	}
	else
	{
	    # finish reading value, if it spans multiple lines
	    $value = finish_top_level_value ($value, $odl_stream);
	    # make sure value is valid
	    if ($value->{TYPE} == $types->{UNKNOWN_TYPE})
	    {
		# invalid value, display warning
		warn ("WARNING - Couldn't determine type ".
		      "of value in attribute.\n".
		      $last_line_read);
	    }
	    # construct statement
	    $statement = bless({TYPE       => $types->{ATTRIBUTE},
				NAMESPACE  => $namespace,
				IDENTIFIER => $identifier,
				VALUE      => $value});
	}
    }
    elsif ($line =~ /^END\s*$/)
    {
	# END keyword found, construct statement
	$statement = bless({TYPE => $types->{END_LABEL}});
    }
    else
    {
	# unknown statement type found, display warning
	warn ("WARNING - Malformed line in ODL stream: ignored\n".
	      $last_line_read);
	# construct error-statement
	$statement = bless({TYPE => $types->{STATEMENT_ERROR}});
    }
    return $statement;
}

=head1

read_odl_line($$) takes a stream and a comment flag and reads the next non-
blank line from the odl stream, stripping comments if the comment flag is
true.

=cut

sub read_odl_line ($$)
{
    # get parameters
    my ($stream, $comments) = @_;
    # declare local variables
    my ($line);
    # try to read a line
  readLine:
    while ($last_line_read = $line = <$stream>)
    {
	# strip trailing format effectors
	$line =~ s/[\r\n\f\013]*$//;
	# strip leading spacing characters
	$line =~ s/^[\t\040]*//;
	$line = strip_comments ($line) if ($comments);
	last readLine if ($line ne "");
    }
    # return the line read
    return $line;
}

=head1

finish_top_level_value ($;$) takes in initial text and an optional stream from
which more text can be read, and reads an ODL value which may be either a
scalar, a set, or a one or two dimensional sequence. A ODL value object
is returned.

=cut

sub finish_top_level_value ($;$)
{
    # get parameters
    my ($value, $odl_stream) = @_;
    # define local variables
    my ($error_return, $need_comma, $tlv);
    $error_return = bless({TYPE     => $types->{UNKNOWN_TYPE},
			   VALUE    => $value});
    # determine if the value is a scalar, ordered list, or set
    if (substr ($value, 0, 1) eq "(")
    {
	# get rid of the opening parenthesis and any spaces
	$value =~ s/\(\s*//;
	# get the next line, if there isn't anything more on this one
	$value = read_odl_line ($odl_stream, TRUE) if ($value =~ /^\s*$/);
	# check to see if the first character is another
	# opening parenthesis
	if (substr ($value, 0, 1) ne "(")
	{
	    # the inner-sequence doesn't begin with a "(",
	    # so just read a scalar list
	    ($sequence, $rest) = read_scalar_list ($value, $odl_stream, ")");
	    # check what we got back from read_scalar_list
	    if ($sequence != undef and
		$rest =~ /^\s*;?\s*$/)
	    {
		# we sucessfully read the list.
		# save the value
		$tlv = bless({TYPE     => $types->{SEQUENCE1D},
			      VALUE    => $sequence});
	    }
	    else
	    {
		# there was an error reading the list
	        $tlv = $error_return;
	    }
	}
	else
	{
	    # the inner-sequence begins with a "(",
	    # so it's a list of lists
	    $sequence2 = [];
	    # keep going until the loop is exited from within
	  read_2d_sequence:
	    while (TRUE)
	    {
		# re-read value if it is blank
		$value = read_odl_line ($odl_stream, TRUE) if ($value eq "");
		# stop on closing parenthesis
		last read_2d_sequence if ($value =~ /^\)/);
		# make sure we have an opening parenthesis for the next itme
		return $error_return unless ($value =~ /^\(/);
		# remove the opening parenthesis
		$value =~ s/^\(\s*//;
		# read a scalar list, stoping on a closing parenthesis
		# replace $value with the remainder (2nd return value)
		($sequence, $value) =
		    read_scalar_list ($value, $odl_stream, ")");
		# trim closing spaces
		$value =~ s/\s*$//;
		# check if we recieved a valid sequence
		if ($sequence)
		{
		    # valid sequence, construct object
		    $s1 = bless({TYPE  => $types->{SEQUENCE1D},
				 VALUE => $sequence});
		    # add the 1-dimensional sequence to the
		    # 2-dimensional sequence
		    push (@{$sequence2}, $s1);
		}
		else
		{
		    # no valid sequence, return error
		    $tlv = $error_return;
		    # abort
		    last read_2d_sequence;
		}
	    }
	    # stop reading on closing parenthesis (and optional semicolon)
	    if ($value =~ /^\)\s*;?\s*$/)
	    {
		# create value
		$tlv = bless({TYPE     => $types->{SEQUENCE2D},
			      VALUE    => $sequence2});
	    }
	    else
	    {
		# return error
		$tlv = $error_return;
	    }
	}
    }
    elsif (substr ($value, 0, 1) eq "{")
    {
	# value is a set
	# get rid of the opening winged brace and any spaces
	$value =~ s/\{\s*//;
	# read the list, ending on a closing brace
	($set, $rest) = read_scalar_list ($value, $odl_stream, "}");
	# check the return values from read_scalar_list
	if ($set and $rest =~ /^\s*;?\s*$/)
	{
	    # no errors, create value
	    $tlv = bless({TYPE     => $types->{SET},
			  VALUE    => $set});
	}
	else
	{
	    # error, display warning and return error
	    warn ("WARNING - error after termination of set\n".
		  $last_line_read);
	    $tlv = $error_return;
	}
    }
    else
    {
	# value is simple scalar
	# read scalar
	$retValue = finish_scalar ($value, $odl_stream);
	# check return values from finish_scalar
	if ($retValue->{TYPE} == $types->{UNKNOWN_TYPE} or
	    not $retValue->{LEFTOVER} =~ /^\s*;?\s*$/)
	{
	    # no errors, create value
	    $tlv = bless({TYPE     => $types->{UNKNOWN_TYPE},
			  VALUE    => $value,
			  LEFTOVER => $retValue->{LEFTOVER}});
	}
	else
	{
	    # return error
	    $tlv = $retValue;
	}
    }
    # return value
    return ($tlv);
}

=head1

read_scalar_list($$$) takes in a initial value (which should contain the
beginning of the scalar list, minus the initial parenthesis or brace, the
stream from which more text can be read if needed, and a terminating character
to look for: ")" for sequences and "}" for sets.

=cut

sub read_scalar_list ($$$)
{
    # get parameters
    my ($value, $odl_stream, $terminator) = @_;
    # initialize the set
    $set = [];
    $need_comma = FALSE;
    $term = "\\$terminator";
  read_scalar_list:
    while (TRUE)
    {
	while ($value eq "")
	{
	    $value = read_odl_line ($odl_stream, TRUE);
	    if ($value eq "")
	    {
		warn ("WARNING - end of stream found within ODL list.\n".
		      $last_line_read);
		return (undef, $value);
	    }
	    elsif ($value =~ /^\s*,/ && $need_comma)
	    {
		$need_comma = FALSE;
		$value =~ s/^\s*,\s*//;
	    }
	    elsif (not $need_comma)
	    {
		$value =~ s/^\s*//;
	    }
	    else
	    {
		$value =~ s/^\s*//;
		if (substr ($value, 0, 1) eq $terminator)
		{
		    $value =~ s/^$term\s*//;
		    return ($set, $value);
		}
		else
		{
		    warn ("WARNING - comma absent in ODL list.\n".
			  $last_line_read);
		    return (undef, $value);
		}
	    }
	}
	if ($#{@{$set}} == -1 && substr ($value, 0, 1) eq $terminator)
	{
	    $value =~ s/^$term\s*//;
	    return ($set, $value);
	}
	$retValue = finish_scalar ($value, $odl_stream);
	if ($retValue == undef or
	    $retValue->{TYPE} == $types->{UNKNOWN_TYPE})
	{
	    warn ("WARNING - error reading scalar within ODL list.\n".
		  $last_line_read);
	    return (undef, $value);
	}
	push (@{$set}, $retValue);
	$value = delete ($retValue->{LEFTOVER});
	$value =~ s/^\s*//;
	if (substr ($value, 0, 1) eq ",")
	{
	    $value =~ s/,\s*//;
	    next read_scalar_list;
	}
	elsif (substr ($value, 0, 1) eq $terminator)
	{
	    $value =~ s/^$term\s*//;
	    last read_scalar_list;
	}
	elsif ($value ne "")
	{
	    warn ("WARNING - error reading scalar within ODL list: ".
		  "comma absent.\n".
		  $last_line_read);
	    return (undef, $value);
	}
	else
	{
	    $need_comma = TRUE;
	}
    }
    return ($set, $value);
}

=head1

finish_scalar($$) takes in the beginning of a value and a stream from which
the rest of the value (if needed) will be read. The value type is determined
and read into an ODL object of the appropriate type. The return value is
an array containing first the ODL value object, and second a string containing
any leftover text. This is usefull when reading sets and sequences since
each call on finish_scalar will remove one value from the string and return
the remaining values.

=cut

sub finish_scalar ($$)
{
    my ($value, $odl_stream) = @_;
    my ($finished, $line_end, $error_return, $scalar);
    $value = strip_comments ($value);
  read_scalar:
    {
	if ($value =~ /^\"/)
	{
	    # text string
	    if ($value =~ /^(\"[^\"]*\")(.*)$/s)
	    {
		$finished = $1;
		$value = $2;
	    }
	    else
	    {
		$finished = $value;
	      read_text_string:
		while ($value = read_odl_line ($odl_stream, FALSE))
		{
		    # check to see how we add the next line: a hyphen
		    # supresses a space
		    if ($finished =~ /-$/)
		    {
			# remoce the hyphen
			chop ($finished);
			# add the next line
			$finished .= $value;
		    }
		    else
		    {
			# trim any trailing spacing characters,
			# replace with a space
			$finished =~ s/[\t\040]*$/\040/;
			# add the next line
			$finished .= $value;
		    }
		    # strip comments, which will only work if
		    # the quotes are closed
		    $finished = strip_comments ($finished);
		    # check to see if we're done
		    if ($finished =~ /^(\"[^\"]*\")(.*)$/s)
		    {
			$finished = $1;
			$value = $2;
			last read_text_string;
		    }
		}
		if (not $finished =~ /^\".*\"$/s)
		{
		    warn ("WARNING - String value not terminated.".
			  $last_line_read);
		}
	    }
	    $scalar = {TYPE     => $types->{TEXT_STRING},
		       VALUE    => $finished,
		       LEFTOVER => $value};
	}
	elsif ($value =~ /^\'/)
	{
	    # symbol string
	    if ($value =~ /^(\'[^\'[:cntrl:]]+\')(.*)$/)
	    {
		($value, $leftover) = ($1, $2);
		$scalar = {TYPE     => $types->{SYMBOL_STRING},
			   VALUE    => $value,
			   LEFTOVER => $leftover};
	    }
	}
	elsif ($value =~ /^($re_identifier(?:\:$re_identifier)?)(.*)$/)
	{
	    ($value, $leftover) = ($1, $2);
	    $scalar = {TYPE     => $types->{IDENTIFIER},
		       VALUE    => uc ($value),
		       LEFTOVER => $leftover};
	}
	elsif ($value =~ /^(((?:2|3|4|5|6|7|8|9)|(?:1(?:0|1|2|3|4|5|6)))\#(?:\+|-)?([[:alnum:]]+)\#(?:\s*$re_units_expression)?)(.*)$/)
	{
	    ($value, $radix, $extended_digit, $leftover) = ($1, $2, $3, $4);
	    for ($i = 0; $i < $radix and $i < 10; ++$i)
	    {
		$digits .= $i;
	    }
	    for ($i = 10; $i < $radix; ++$i)
	    {
		$digits .= chr (ord ("A") + $i - 10);
	    }
	    $scalar = {TYPE     => $types->{BASED_INTEGER},
		       VALUE    => $value,
		       LEFTOVER => $leftover}
	    if ($extended_digit =~ /^[$digits]+$/);
	}
	elsif ($value =~ /^((?:\+|-)?(?:(?:[[:digit:]]+\.[[:digit:]]*)|(?:\.[[:digit:]]+)|(?:[[:digit:]]+))(?:e|E)(?:\+|-)?[[:digit:]]+(?:\s*$re_units_expression)?)(.*)$/)
	{
	    ($value, $leftover) = ($1, $2);
	    $scalar = {TYPE     => $types->{SCALED_REAL},
		       VALUE    => $value,
		       LEFTOVER => $leftover};
	}
	elsif ($value =~ /^((?:\+|-)?(?:(?:[[:digit:]]+\.[[:digit:]]*)|(?:\.[[:digit:]]+))(?:\s*$re_units_expression)?)(.*)$/)
	{
	    ($value, $leftover) = ($1, $2);
	    $scalar = {TYPE     => $types->{UNSCALED_REAL},
		       VALUE    => $value,
		       LEFTOVER => $leftover};
	}
	elsif ($value =~ /^(([[:digit:]]{4})-([[:digit:]]+)T$re_hour_min_sec(?:Z|(?:(?:\+|-)(\d+)(?:\:(\d+))?))?)(.*)$/)
	{
	    ($value, $year, $doy, $hour, $minute, $second, $h, $m, $leftover) =
		($1, $2, $3, $4, $5, $6, $7, $8, $9);
	    if ( $year %   4 == 0 and
		 ($year % 100 != 0 or
		  $year % 400 == 0))
	    {
		last read_scalar if ($doy > 366);
	    }
	    elsif ($doy > 365)
	    {
		last read_scalar;
	    }
	    last read_scalar unless ((0 <= $hour   and $hour   <= 23) and
				     (0 <= $h      and $h      <= 12) and
				     (0 <= $minute and $minute <= 59) and
				     (0 <= $m      and $m      <= 59) and
				     (0 <= $second and $second <  60));
	    $scalar = {TYPE     => $types->{DATE_TIME},
		       VALUE    => $value,
		       LEFTOVER => $leftover};
	}
	elsif ($value =~ /^(([[:digit:]]{4})-([[:digit:]]+)-([[:digit:]]+)T$re_hour_min_sec(?:Z|(?:(?:\+|-)(\d+)(?:\:(\d+))?))?)(.*)$/)
	{
	    ($value,  $year,   $month, $day, $hour,
	     $minute, $second, $h,     $m,   $leftover) =
		 ($1, $2, $3, $4, $5,
		  $6, $7, $8, $9, $10);
	    last read_scalar unless (1 <= $month and $month <= 12);
	    if    ($month ==  1) { $max_day = 31; }
	    elsif ($month ==  2)
	    {
		if ($year % 4 == 0 and
		    ($year % 100 != 0 or
		     $year % 400 == 0))
		{
		    $max_day = 29;
		}
		else
		{
		    $max_day = 28;
		}
	    }
	    elsif ($month ==  3) { $max_day = 31; }
	    elsif ($month ==  4) { $max_day = 30; }
	    elsif ($month ==  5) { $max_day = 31; }
	    elsif ($month ==  6) { $max_day = 30; }
	    elsif ($month ==  7) { $max_day = 31; }
	    elsif ($month ==  8) { $max_day = 31; }
	    elsif ($month ==  9) { $max_day = 30; }
	    elsif ($month == 10) { $max_day = 31; }
	    elsif ($month == 11) { $max_day = 30; }
	    elsif ($month == 12) { $max_day = 31; }
	    last read_scalar unless ((1 <= $day and $day <= $max_day) or
				     ((0 <= $hour   and $hour   <= 23) and
				      (0 <= $h      and $h      <= 12) and
				      (0 <= $minute and $minute <= 59) and
				      (0 <= $m      and $m      <= 59) and
				      (0 <= $second and $second <  60)));
	    $scalar = {TYPE     => $types->{DATE_TIME},
		       VALUE    => $value,
		       LEFTOVER => $leftover};
	}
	elsif ($value =~ /^(([[:digit:]]{4})-([[:digit:]]+)-([[:digit:]]+))(.*)$/)
	{
	    ($value, $year, $month, $day, $leftover) = ($1, $2, $3, $4, $5);
	    last read_scalar unless (1 <= $month and $month <= 12);
	    if    ($month ==  1) { $max_day = 31; }
	    elsif ($month ==  2)
	    {
		if ( $year %   4 == 0 and
		     ($year % 100 != 0 or
		      $year % 400 == 0))
		{
		    $max_day = 29;
		}
		else
		{
		    $max_day = 28;
		}
	    }
	    elsif ($month ==  3) { $max_day = 31; }
	    elsif ($month ==  4) { $max_day = 30; }
	    elsif ($month ==  5) { $max_day = 31; }
	    elsif ($month ==  6) { $max_day = 30; }
	    elsif ($month ==  7) { $max_day = 31; }
	    elsif ($month ==  8) { $max_day = 31; }
	    elsif ($month ==  9) { $max_day = 30; }
	    elsif ($month == 10) { $max_day = 31; }
	    elsif ($month == 11) { $max_day = 30; }
	    elsif ($month == 12) { $max_day = 31; }
	    last read_scalar unless (1 <= $day and $day <= $max_day);
	    $scalar = {TYPE     => $types->{DATE},
		       VALUE    => $value,
		       LEFTOVER => $leftover};
	}
	elsif ($value =~ /^(([[:digit:]]{4})-([[:digit:]]+))(.*)$/)
	{
	    ($value, $year, $doy, $leftover) = ($1, $2, $3, $4);
	    # check for leap year
	    if ( $year %   4 == 0 and
		 ($year % 100 != 0 or
		  $year % 400 == 0))
	    {
		last read_scalar if ($doy > 366);
	    }
	    elsif ($doy > 365)
	    {
		last read_scalar;
	    }
	    $scalar = {TYPE     => $types->{DATE},
		       VALUE    => $value,
		       LEFTOVER => $leftover}
	}
	elsif ($value =~ /^($re_hour_min_sec(?:Z|(?:(?:\+|-)(\d+)(?:\:(\d+))?))?)(.*)$/)
	{
	    ($value, $hour, $minute, $second, $h, $m, $leftover) =
		($1, $2, $3, $4, $5, $6, $7);
	    last read_scalar unless ((0 <= $hour   and $hour   <= 23) and
				     (0 <= $h      and $h      <= 12) and
				     (0 <= $minute and $minute <= 59) and
				     (0 <= $m      and $m      <= 59) and
				     (0 <= $second and $second <  60));
	    $scalar = {TYPE     => $types->{TIME},
		       VALUE    => $value,
		       LEFTOVER => $leftover};
	}
	elsif ($value =~ /^((?:\+|-)?[[:digit:]]+(?:\s*$re_units_expression)?)(.*)$/)
	{
	    ($value, $leftover) = ($1, $2);
	    $scalar = {TYPE     => $types->{INTEGER},
		       VALUE    => $value,
		       LEFTOVER => $leftover};
	}
	else
	{
	    $scalar = {TYPE     => $types->{UNKNOWN_TYPE},
		       VALUE    => $value};
	}
    }
    bless ($scalar) if ($scalar);
    return $scalar;
}

=head1

strip_comments($) takes a string parameter and removes any ODL comments from
it. an ODL comment is defined as beginning with the "/*" character sequence
and ending with the "*/" character sequence. Any text on a line after the
comment is ignored. Comments within quotes are kept.

strip_comments ("one \"/* fake comment */\" and /* one real comment */ end")
would return
"one \"/* fake comment */\" and "

=cut

sub strip_comments ($)
{
    # get parameters
    my ($line) = @_;
    # strip out any /*[...]*/..., excluding comments found in quotes
    $line =~ s/^((?:[^\"\']*)|(?:(?:[^\"\']*((\"[^\"]*\")|(\'[^\']*\'))[^\"\']*)*)\S?)?\s*\/\*.*\*\/.*$/$1/;
    # strip any leading whitespace
    $line =~ s/^\s*//;
    # return the modified line
    return $line;
}

=head1

Object Identifier Functions

These functions make determining the type of an ODL object easier. They provide
the means to determine if the object is a statement or a value within a
statement, and what kind of statement or value it is. These types can be
directly observed by looking at the TYPE key within the hash that is each
object. (A table of type-codes is present in the $types hash)

ODL objects can be either statements or values. Most statements contain values,
and some statements may contain multiple or nested values. The following are
examples of ODL statements:

a pointer statement:
^DATA        = ("file.dat", 3)

a attribute statement:
RECORD_BYTES = 2880

an object statement (with two nested attribute statements):
OBJECT = IMAGE
  WIDTH  = 320
  HEIGHT = 200
END_OBJECT = IMAGE

Values occur as values to the right of an equals sign in pointer and attribute
statements, and may either be scalar values, sets, 1 dimensional sequences, or
2 dimensional sequences. The following are examples of ODL values, following
the statements which contain them:

^DATA = ("file.dat", 3)
value is a 1 dimensional list containing the scalars "file.dat" and 3

RECORD_BYTES = 2880
value is the scalar 2880

COLORS = {RED, GREEN, BLUE}
value is a set containing the scalars RED, GREEN and BLUE

=cut

=head1

ODL::is_statement() returns true if the ODL object is a statement, and false if
the ODL object is a value.

=cut

sub is_statement ()
{
    # get self
    my ($self) = @_;
    # check type of object
    return ($self->{TYPE} < $types->{UNKNOWN_TYPE});
}

=head1

ODL::is_value() returns true if the ODL object is a value, and false if the
ODL object is a statement.

=cut

sub is_value ()
{
    # get self
    my ($self) = @_;
    # check type of object
    return ($self->{TYPE} >= $types->{UNKNOWN_TYPE});
}

=head1

ODL::statement_is_root_node() returns true if the ODL object is a root node
statement, and false otherwise. A root node statement is analogous to a PDS
label file, and what is returned by the ODL::create(;$) method.

=cut

sub statement_is_root_node ()
{
    # get self
    my ($self) = @_;
    # check type of object
    return ($self->{TYPE} == $types->{ROOT_NODE});
}

=head1

ODL::statement_is_object() returns true if the ODL statement is an object
statement, false otherwise. Object statements in ODL are defined as follows:
BEGIN_OBJECT = <object_name>
  <statement1>
  <statement2>
  ...
  <statementN>
END_OBJECT [ = <object_name> ]

=cut

sub statement_is_object ()
{
    # get self
    my ($self) = @_;
    # check type of object
    return ($self->{TYPE} == $types->{OBJECT});
}

=head1

ODL::statement_is_group() returns true if the ODL statement is a group
statement, false otherwise. Group statements in ODL are defined as follows:
BEGIN_GROUP = <group_name>
  <statement1>
  <statement2>
  ...
  <statementN>
END_GROUP [ = <group_name> ]

=cut

sub statement_is_group ()
{
    # get self
    my ($self) = @_;
    # check type of object
    return ($self->{TYPE} == $types->{GROUP});
}

=head1

ODL::statement_is_attribute() returns true if the ODL object is an attribute
statement (or conversely, not a value, nor a pointer, object, group, or
root-level statement).

attribute statements take the form
[<namespace>:]<label> = <value>
=cut

sub statement_is_attribute ()
{
    # get self
    my ($self) = @_;
    # check type of object
    return ($self->{TYPE} == $types->{ATTRIBUTE});
}

=head1

ODL::attribute_is_pointer returns true if the ODL object is a pointer
statement.

=cut

sub statement_is_pointer ()
{
    # get self
    my ($self) = @_;
    # check type of object
    return ($self->{TYPE} == $types->{POINTER});
}

=head1

ODL::statement_has_value returns true if the ODL object is either an attribute
statement or a pointer statement, which are the only statements which hold
values (scalar, array, etc.) All other statements contain a collection of
statements.

=cut

sub statement_has_value ()
{
    # get self
    my ($self) = @_;
    # check type of object
    return ($self->{TYPE} == $types->{ATTRIBUTE} or
	    $self->{TYPE} == $types->{POINTER});
}

=head1

ODL::statement_is_collection returns true if the ODL object is either a root
node statement, an object statement, or a group statement.

=cut

sub statement_is_collection ()
{
    # get self
    my ($self) = @_;
    # check type of object
    return ($self->{TYPE} == $types->{ROOT_NODE} or
	    $self->{TYPE} == $types->{OBJECT} or
	    $self->{TYPE} == $types->{GROUP});
}

#=== DETERMINING TYPE ===

=head1

ODL::value_is_number() returns true if the ODl value is a numeric value, and
false otherwise.

=cut

sub value_is_number ()
{
    # get self
    my ($self) = @_;
    # replace object with its value object if given a statement object
    $self = $self->{VALUE} if ($self->statement_has_value ());
    # check type of value object
    return ($self->{TYPE} == $types->{INTEGER} or
	    $self->{TYPE} == $types->{UNSIGNED_INTEGER} or
	    $self->{TYPE} == $types->{BASED_INTEGER} or
	    $self->{TYPE} == $types->{UNSCALED_REAL} or
	    $self->{TYPE} == $types->{SCALED_REAL});
}

sub value_is_integer ()
{
    # get self
    my ($self) = @_;
    # replace object with its value object if given a statement object
    $self = $self->{VALUE} if ($self->statement_has_value ());
    # check type of value object
    return ($self->{TYPE} == $types->{INTEGER} or
	    $self->{TYPE} == $types->{UNSIGNED_INTEGER});
}

=head1

ODL::value_is_date_time() returns true if the ODL value is a date, time, or
date/time value, and false otherwise.

=cut

sub value_is_date_time ()
{
    # get self
    my ($self) = @_;
    # replace object with its value object if given a statement object
    $self = $self->{VALUE} if ($self->statement_has_value ());
    # check type of value object
    return ($self->{TYPE} == $types->{DATE} or
	    $self->{TYPE} == $types->{TIME} or
	    $self->{TYPE} == $types->{DATE_TIME});
}

=head1

ODL::value_is_string() returns true if the ODl value is a text or symbol
string, and false otherwise.

=cut

sub value_is_string ()
{
    # get self
    my ($self) = @_;
    # replace object with its value object if given a statement object
    $self = $self->{VALUE} if ($self->statement_has_value ());
    # check type of value object
    return ($self->{TYPE} == $types->{TEXT_STRING} or
	    $self->{TYPE} == $types->{SYMBOL_STRING});
}

=head1

ODL::value_is_identifier() returns true if the ODL value is an identifier, and
false otherwise.

=cut

sub value_is_identifier ()
{
    # get self
    my ($self) = @_;
    # replace object with its value object if given a statement object
    $self = $self->{VALUE} if ($self->statement_has_value ());
    # check type of value object
    return ($self->{TYPE} == $types->{IDENTIFIER});
}

=head1

ODL::value_is_scalar() returns true if the ODL value is a scalar, and false
if it is a set, one dimensional sequence, or two dimensional sequence.

=cut

sub value_is_scalar ()
{
    # get self
    my ($self) = @_;
    # check type of value
    return ($self->value_is_number () or
	    $self->value_is_date_time () or
	    $self->value_is_string () or
	    $self->value_is_identifier ());
}

=head1

ODL::value_is_set() returns true if the ODL value is a set, and false
otherwise. Sets are a sequence of scalars separated by commas, with
braces. example: {RED, GREEN, BLUE}

=cut

sub value_is_set ()
{
    # get self
    my ($self) = @_;
    # replace object with its value object if given a statement object
    $self = $self->{VALUE} if ($self->statement_has_value ());
    # check type of value object
    return ($self->{TYPE} == $types->{SET});
}

=head1

ODL::value_is_sequence returns true if the ODL value is a sequence, either one
or two dimensional, and false otherwise. One dimensional sequences are defined
like sets, only with parenthesis instead of braces. example: (ONE, TWO, THREE).
Two dimensional sequences are a list of one dimensional sequences, with
an additional set of parenthesis around the whole. Example:
((ONE, TWO, THREE)(RED, GREEN, BLUE)(1.0, 2.0, 3.0)).
Notice the absence of commas between each one dimensional sequence within
the entire two dimensional sequence.

=cut

sub value_is_sequence ()
{
    # get self
    my ($self) = @_;
    # replace object with its value object if given a statement object
    $self = $self->{VALUE} if ($self->statement_has_value ());
    # check type of value object
    return ($self->{TYPE} == $types->{SEQUENCE1D} or
	    $self->{TYPE} == $types->{SEQUENCE2D});
}

=head1

ODL::value_is_1D_sequence() returns true if the ODL value is a one dimensional
sequence, and false otherwise. See the documentation for ODL::value_is_sequence
for notes on one dimensional sequences.

=cut

sub value_is_1D_sequence ()
{
    # get self
    my ($self) = @_;
    # replace object with its value object if given a statement object
    $self = $self->{VALUE} if ($self->statement_has_value ());
    # check type of value object
    return ($self->{TYPE} == $types->{SEQUENCE1D});
}

=head1

ODL::value_is_2D_sequence() returns true if the ODL value is a two dimensional
sequence, and false otherwise. See the documentation for ODL::value_is_sequence
for notes on two dimensional sequences.

=cut

sub value_is_2D_sequence ()
{
    # get self
    my ($self) = @_;
    # replace object with its value object if given a statement object
    $self = $self->{VALUE} if ($self->statement_has_value ());
    # check type of value object
    return ($self->{TYPE} == $types->{SEQUENCE2D});
}

# package must return true
1;
