#!/usr/bin/perl

package ODLlib;

use lib '/n/sbnops01/lcltools/lib/perl5';
use TRY;

use constant false => 0;
use constant true  => not false;

my $objectTypes = {ERROR =>             0,
		   # Implicit file object
		   FILE =>              1,
		   # Statement types:
		   ATTRIBUTE        =>  2,
		   POINTER          =>  3,
		   OBJECT           =>  4,
		   GROUP            =>  5,
		   # Termination statements:
		   END_FILE         =>  6,
		   END_OBJECT       =>  7,
		   END_GROUP        =>  8,
		   # Value types:
		   TEXT_STRING      =>  9,
		   SYMBOL_STRING    => 10,
		   IDENTIFIER       => 11,
		   INTEGER          => 12,
		   UNSIGNED_INTEGER => 13,
		   BASED_INTEGER    => 14,
		   UNSCALED_REAL    => 15,
		   SCALED_REAL      => 16,
		   DATE             => 17,
		   TIME             => 18,
                   DATE_TIME        => 19,
		   SET              => 20,
		   SEQUENCE_1D      => 21,
		   SEQUENCE_2D      => 22
		  };

my $tokenTypes = {EOF               =>  1,
		  WHITESPACE        =>  2,
		  COMMENT           =>  3,
		  IDENTIFIER        =>  4,
		  SPECIAL_CHARACTER =>  5,
		  TEXT_STRING       =>  6,
		  SYMBOL_STRING     =>  7,
		  INTEGER           =>  8,
		  UNSIGNED_INTEGER  =>  9,
		  BASED_INTEGER     => 10,
		  SCALED_REAL       => 11,
		  UNSCALED_REAL     => 12,
		  DATE              => 13,
		  TIME              => 14,
		  DATE_TIME         => 15
		 };

# Sanity checking profiles. $odlSanity is always used
my $odlSanity = {SEMICOLON     => [],
		 BEGIN_GROUP   => [],
		 BEGIN_OBJECT  => [],
		 IDENTIFIER    => [],
		 KEYWORD       => [],
		 BASED_INTEGER => [\&checkBasedInteger],
		 DATE          => [\&checkDate],
		 TIME          => [\&checkTime],
		 DATE_TIME     => [],
		};
# Default sanity checks that can be overwritten. Any sanity check in this
# structure will only be used if there are no sanity checks for the same
# item given by either the user, or the custom base.
my $odlSanityOverride = {SEMICOLON     => [\&odlDenySemicolon],
			 BEGIN_GROUP   => [\&odlDenyBeginGroup],
			 BEGIN_OBJECT  => [\&odlDenyBeginObject]};

# Base sanities for variants of ODL.
my $baseSanities = {PDS => {SEMICOLON     => [],# [\&pdsAllowSemicolon],
			    BEGIN_GROUP   => [\&pdsAllowBeginGroup],
			    BEGIN_OBJECT  => [\&pdsAllowBeginObject],
			    KEYWORD       => [\&pdsCheckKeyword],
			    BASED_INTEGER => [\&pdsCheckBasedInteger],
			    DATE          => [\&pdsCheckDate],
			    TIME          => [\&pdsCheckTime]
			   }
		   };

sub create (;$$)
{
    my (
	$class,
	$filename,
	$customSanity
	) = @_;
    my $this = {type => $objectTypes->{FILE}};
    bless ($this, $class);
    $this->loadFromFile ($filename, $customSanity) if ($filename);
    return $this;
}

sub sanityMerge ($$;$)
{
    my (
	$sanity1,
	$sanity2,
	$noOverwrite
	) = @_;
    $mergedSanity = {};
    foreach $key (keys (%$sanity1))
    {
	$mergedSanity->{$key} = [];
	push (@{$mergedSanity->{$key}}, $_) foreach (@{$sanity1->{$key}});
    }
    foreach $key (keys (%$sanity2))
    {
	$mergedSanity->{$key} = [] unless exists $mergedSanity->{$key};
	unless ($noOverwrite &&	scalar (@{$mergedSanity->{$key}}) > 0)
	{
	    push (@{$mergedSanity->{$key}}, $_) foreach (@{$sanity2->{$key}});
	}
    }
    return $mergedSanity;
}

# Attempt some block of code, rolling back in the ODL stream if an error
# was encountered.
sub attempt (&)
{
    my (
	$block
	) = @_;
    my (
	$accumulator
	);
    # Start collecting tokens read from the ODL stream
    $accumulator = [];
    $tokenAccumulator->registerAccumulator ($accumulator);
    # Attempt to execute the block
    try
    {
	&$block ();
    }
    catch
    {
	# There were errors.
	# Restore the tokens read from the stream
	$tokenAccumulator->restoreTokens ($accumulator);
	# Re-throw the error
	die ($_);
    }
    finally
    {
	# Whether or not the block succeeded, make sure we stop
	# accumulating tokens.
	$tokenAccumulator->removeAccumulator ($accumulator);
    };
}

sub loadFromFile ($;$)
{
    my (
	$this,
	$filename,
	$customSanity
	) = @_;
    $this->{FILENAME} = $filename;
    die ("Could not open ODL file $filename")
	unless (open (odlStream, "<", $filename));
    try
    {
	$this->loadFromStream (odlStream, $customSanity);
    }
    finally
    {
	close (ODL_STREAM);
    };
}

sub loadFromStream ($;$)
{
    my (
	$this,
	$stream,
	$customSanity
	) = @_;
    my (
	$reading,
	$statement,
	$lineNumber,
	$errorMessage,
	$lastLineRead
	);
    local $sanity;
    local $tokenAccumulator = $this;
    $base = $customSanity->{base};
    if (exists ($baseSanities->{$base}))
    {
	$customSanity = sanityMerge ($customSanity, $baseSanities->{$base});
    }
    $customSanity = sanityMerge ($customSanity, $odlSanityOverride, true);
    $sanity = sanityMerge ($odlSanity, $customSanity);
    die ("Can't read ODL file into a non implicit-file object.")
	unless ($this && $this->{type} == $objectTypes->{FILE});
    $this->{input} = {stream       => $stream,
		      inputLines   => [],
		      lineNumber   => -1,
		      accumulators => [],
		      remainder    => "",
		      tokenStack   => []};
    $this->{statements} = [];
    $reading = true;
    try
    {
	while ($reading)
	{
	    $statement = $this->readODLStatement ();
	    die ("END_OBJECT without matching OBJECT")
		if ($statement->{type} == $objectTypes->{END_OBJECT});
	    die ("END_GROUP without matching GROUP")
		if ($statement->{type} == $objectTypes->{END_GROUP});
	    push (@{$this->{statements}}, $statement);
	    $reading = ($statement->{type} != $objectTypes->{END_FILE});
	}
    }
    catch
    {
	$errorMessage = $_;
	$lineNumber = 0 + $.;
	$lastLineRead = $this->getLastInputLine ();
	if (exists ($this->{FILENAME}))
	{
	    $lineNumber .= " of ".$this->{FILENAME};
	}
	die ("Error reading ODL stream at or before line ". $lineNumber .":\n".
	     $lastLineRead.
	     $errorMessage);
    };
}

sub createEndFileStatement ()
{
    my (
	$statement
	);
    $statement = {type => $objectTypes->{END_FILE}};
    return bless ($statement);
}

sub readObjectStatement ()
{
    my (
	$this
	) = @_;
    my (
	$statement,
	$value,
	$subStatement
	);
    $statement = {type       => $objectTypes->{OBJECT},
		  statements => []};
    attempt
    {
	$this->ensureEqualSymbol ();
	$value = $this->readValue ();
	$this->ensureEndOfLine ();
	die ("Syntax must be \"OBJECT = <identifier>\".")
	    if ($value->{type} != $objectTypes->{IDENTIFIER});
	$statement->{object} = $value->{value};
    };
    do
    {
	$subStatement = $this->readODLStatement ();
	die ("END_GROUP without matching group.")
	    if ($subStatement->{type} == $objectTypes->{END_GROUP});
	die ("END while still nested in objects or groups.")
	    if ($subStatement->{type} == $objectTypes->{END_FILE});
	push (@{$statement->{statements}}, $subStatement)
	    unless ($subStatement->{type} == $objectTypes->{END_OBJECT});
    } until ($subStatement->{type} == $objectTypes->{END_OBJECT});
    die ("OBJECT and END_OBJECT statements have mismatched identifiers.")
	if (exists ($subStatement->{object}) &&
	    $subStatement->{object} ne $statement->{object});
    return bless ($statement);
}

sub readEndObjectStatement ()
{
    my (
	$this
	) = @_;
    my (
	$statement,
	$accumulator,
	$value
	);
    $statement = {type => $objectTypes->{END_OBJECT}};
    $accumulator = [];
    $this->registerAccumulator ($accumulator);
    try
    {
	try
	{
	    $this->ensureEqualSymbol ();
	    $value = $this->readValue ();
	    $this->ensureEndOfLine ();
	}
	catch
	{
	    try
	    {
		$this->ensureEndOfLine ();
	    }
	    catch
	    {
		$this->restoreTokens ($accumulator);
		die ("Syntax must be \"END_OBJECT [= <identifier>]\"");
	    };
	};
	if (defined ($value) &&
	    $value->{type} != $objectTypes->{IDENTIFIER})
	{
	    $this->restoreTokens ($accumulator);
	    die ("Syntax must be \"END_OBJECT [= <identifier>]\n".
		 "\"".$value->{value}."\" is not an identifier.");
	}
	elsif (defined ($value))
	{
	    $statement->{object} = $value->{value};
	}
    }
    finally
    {
	$this->removeAccumulator ($accumulator);
    };
    return bless ($statement);
}

sub readGroupStatement ()
{
    my (
	$this
	) = @_;
    my (
	$statement,
	$value,
	$subStatement
	);
    $statement = {type       => $objectTypes->{GROUP},
		  statements => []};
    attempt
    {
	$this->ensureEqualSymbol ();
	$value = $this->readValue ();
	$this->ensureEndOfLine ();
	die ("Syntax must be \"GROUP = <identifier>\".")
	    if ($value->{type} != $objectTypes->{IDENTIFIER});
	$statement->{group} = $value->{value};
    };
    do
    {
	$subStatement = $this->readODLStatement ();
	die ("END_OBJECT without matching object.")
	    if ($subStatement->{type} == $objectTypes->{END_OBJECT});
	die ("END while still nested in objects or groups.")
	    if ($subStatement->{type} == $objectTypes->{END_FILE});
	push (@{$statement->{statements}}, $subStatement)
	    unless ($subStatement->{type} == $objectTypes->{END_GROUP});
    } until ($subStatement->{type} == $objectTypes->{END_GROUP});
    die ("GROUP and END_GROUP statements have mismatched identifiers.")
	if (exists ($subStatement->{group}) &&
	    $subStatement->{group} ne $statement->{group});
    return bless ($statement);
}

sub readEndGroupStatement ()
{
    my (
	$this
	) = @_;
    my (
	$statement,
	$accumulator,
	$value
	);
    $statement = {type => $objectTypes->{END_GROUP}};
    $accumulator = [];
    $this->registerAccumulator ($accumulator);
    try
    {
	try
	{
	    $this->ensureEqualSymbol ();
	    $value = $this->readValue ();
	    $this->ensureEndOfLine ();
	}
	catch
	{
	    try
	    {
		$this->ensureEndOfLine ();
	    }
	    catch
	    {
		$this->restoreTokens ($accumulator);
		die ("Syntax must be \"END_GROUP [= <identifier>]\"");
	    };
	};
	if (defined ($value) &&
	    $value->{type} != $objectTypes->{IDENTIFIER})
	{
	    $this->restoreTokens ($accumulator);
	    die ("Syntax must be \"END_GROUP [= <identifier>]\n".
		 "\"".$value->{value}."\" is not an identifier.");
	}
	elsif (defined ($value))
	{
	    $statement->{group} = $value->{value};
	}
    }
    finally
    {
	$this->removeAccumulator ($accumulator);
    };
    return bless ($statement);
}

sub readAttributeOrPointerStatement ($)
{
    my (
	$this,
	$keyword
	) = @_;
    my (
	$value,
	$statement
	);
    $this->ensureEqualSymbol ();
    $value = $this->readValue ();
    $this->ensureEndOfLine ();
    $statement = {type    => $keyword->{pointer}?$objectTypes->{POINTER}:
		                                 $objectTypes->{ATTRIBUTE},
		  keyword => $keyword,
		  value   => $value};
    return bless ($statement);
}

sub readODLStatement ()
{
    my (
	$this,
	) = @_;
    my (
	$keyword,
	$value,
	$token,
	$accumulator,
	$statement,
	$check,
	$subStatement,
	$identifier
	);
    $keyword = $this->readKeyword ();
    if ($keyword->{special} &&
	$keyword->{END_FILE})
    {
	$statement = createEndFileStatement ();
    }
    elsif ($keyword->{special} &&
	   $keyword->{OBJECT})
    {
	if ($keyword->{element} eq "BEGIN_OBJECT")
	{
	    foreach $check (@{$sanity->{BEGIN_OBJECT}})
	    {
		&$check ();
	    }
	}
	$statement = $this->readObjectStatement ();
    }
    elsif ($keyword->{special} &&
	   $keyword->{GROUP})
    {
	if ($keyword->{element} eq "BEGIN_GROUP")
	{
	    foreach $check (@{$sanity->{BEGIN_GROUP}})
	    {
		&$check ();
	    }
	}
	$statement = $this->readGroupStatement ();
    }
    elsif ($keyword->{special} &&
	   $keyword->{END_OBJECT})
    {
	$statement = $this->readEndObjectStatement ();
    }
    elsif ($keyword->{special} &&
	   $keyword->{END_GROUP})
    {
	$statement = $this->readEndGroupStatement ();
    }
    else
    {
	$statement = $this->readAttributeOrPointerStatement ($keyword);
    }
    return $statement;
}

sub readKeyword ()
{
    my (
	$this
	) = @_;
    my (
	$token,
	$pointer,
	$namespace,
	$identifier,
	$accumulator,
	$keyword
	);
    # Start collecting tokens read from the ODL stream.
    $accumulator = [];
    $this->registerAccumulator ($accumulator);
    # Attempt to read a keyword.
    try
    {
	# Get the first token
	$token = $this->getInputToken (true);
	# Throw an error.
	# If it's an EOF when we're expecting more statements.
	die ("premature end of stream")
	    if ($token->{type} == $tokenTypes->{EOF});
	# If the next token is the '^' character, we're dealing with a pointer
	if ($token->{type} == $tokenTypes->{SPECIAL_CHARACTER} &&
	    $token->{value} eq "^")
	{
	    # The statement is a pointer.
	    $pointer = true;
	    # Get the next token, which should now be the identifier.
	    $token = $this->getInputToken ();
	}
	else
	{
	    # The statement isn't a pointer.
	    $pointer = false;
	}
	# Make sure the last token we got was an identifier. This could
	# happen two ways:
	# 1) It's a normal statement of the form
	#      KEYWORD ...
	#      NAMESPACE:KEYWORD ...
	# 2) It's a pointer of the form
	#      ^KEYWORD ...
	#    and we already processed the '^' token and got the next one
	die ("Statements must begin with an identifier")
	    if ($token->{type} != $tokenTypes->{IDENTIFIER});
	# Remember the value of the identifier
	$identifier = $token->{value};
	# Get the next token, so we can find out if we're actually working
	# with a namespace
	$token = $this->getInputToken ();
	# Check if the token is the ':' token.
	if ($token->{type} == $tokenTypes->{SPECIAL_CHARACTER} &&
	    $token->{value} eq ":")
	{
	    # We're dealing with a namespace. Get the next token, which
	    # will be the real identifier. The identifier we got previously
	    # is actually the namespace.
	    $token = $this->getInputToken ();
	    # Make sure the next token is an identifier.
	    die ("Statements cannot have a namespace identifier\n".
		 "without an element identifier")
		if ($token->{type} != $tokenTypes->{IDENTIFIER});
	    # Remember the namespace as the first identifier we got.
	    $namespace = $identifier;
	    # Remember the new identifier.
	    $identifier = $token->{value};
	}
	else
	{
	    # There isn't a namespace, so we've read too far.
	    # Replace the last token into the stream.
	    $this->restoreTokens($token);
	}
	# Perform any sanity checks on the keyword we just parsed.
	foreach $check (@{$sanity->{KEYWORD}})
	{
	    &$check (($pointer?"^":"").
		     ($namespace?$namespace.":":"").
		     $identifier,
		     $pointer,
		     $namespace,
		     $identifier);
	    
	}
    }
    catch
    {
	# There was some error in the handling of the keyword.
	# Restore any tokens we got out of the stream.
	$this->restoreTokens ($accumulator);
	# Re-throw the error.
	die ($_);
    }
    finally
    {
	# Now that we're done, whether successful or not, stop accumulating
	# tokens.
	$this->removeAccumulator ($accumulator);
    };
    # Return the keyword, as we've parsed it.
    if ($pointer)
    {
	die ("Namespaces cannot be used with pointers.")
	    if ($namespace ne "");
	$keyword = {special => false,
		    pointer => true,
		    object  => $identifier};
    }
    elsif ($namespace eq "" &&
	   ($identifier eq "OBJECT" ||
	    $identifier eq "BEGIN_OBJECT"))
    {
	$keyword = {special => true,
		    OBJECT  => true,
		    element => $identifier};
    }
    elsif ($namespace  eq "" &&
	   $identifier eq "END_OBJECT")
    {
	$keyword = {special    => true,
		    END_OBJECT => true};
    }
    elsif ($namespace eq "" &&
	   ($identifier eq "GROUP" ||
	    $identifier eq "BEGIN_GROUP"))
    {
	$keyword = {special => true,
		    GROUP   => true,
		    element => $identifier};
    }
    elsif ($namespace  eq "" &&
	   $identifier eq "END_GROUP")
    {
	$keyword = {special   => true,
		    END_GROUP => true};
    }
    elsif ($namespace  eq "" &&
	   $identifier eq "END")
    {
	$keyword = {special  => true,
		    END_FILE => true}
    }
    else
    {
	$keyword = {special   => false,
		    pointer   => false,
		    element   => $identifier};
	$keyword->{namespace} = $namespace
	    if (defined ($namespace));
    }
    return $keyword;
}

sub ensureEqualSymbol ()
{
    my (
	$this
	) = @_;
    my (
	$token,
	$accumulator
	);
    # Attempt to read the equals symbol
    attempt
    {
	# Get the next non-whitespace and non-comment token
	$token = $this->getInputToken (true);
	# Make sure it's the equals symbol
	if ($token->{type} != $tokenTypes->{SPECIAL_CHARACTER} ||
	    $token->{value} ne "=")
	{
	    # It wasn't... perhaps it's missing? Throw an error.
	    die ("missing '=' in statement (found \"".$token->{value}."\" instead.)");
	}
    };
}

sub createTextValue ($)
{
    my (
	$text
	) = @_;
    my (
	$value
	);
    $value = {type  => $objectTypes->{TEXT_STRING},
	      value => $text};
    return bless ($value);
}

sub createSymbolValue ($)
{
    my (
	$symbol
	) = @_;
    my (
	$value
	);
    $value = {type  => $objectTypes->{SYMBOL_STRING},
	      value => $symbol};
    return bless ($value);
}

sub createIdentifierValue ($)
{
    my (
	$identifier
	) = @_;
    my (
	$value
	);
    $value = {type  => $objectTypes->{IDENTIFIER},
	      value => $identifier};
    return bless ($value);
}

sub createDateValue ($)
{
    my (
	$date
	) = @_;
    my (
	$value
	);
    $value = {type  => $objectTypes->{DATE},
	      value => $date};
    return bless ($value);
}

sub createTimeValue ($)
{
    my (
	$time
	) = @_;
    my (
	$value
	);
    $value = {type  => $objectTypes->{TIME},
	      value => $time};
    return bless ($value);
}

sub createDateTimeValue ($)
{
    my (
	$dateTime
	) = @_;
    my (
	$value
	);
    $value = {type  => $objectTypes->{DATE_TIME},
	      value => $dateTime};
    return bless ($value);
}

sub createNumericValue ($$$)
{
    my (
	$number,
	$units,
	$specificType
	) = @_;
    my (
	$type,
	$value
	);
    ($type) = grep {$tokenTypes->{$_} == $specificType} keys (%$tokenTypes);
    $value = {type  => $objectTypes->{$type},
	      value => $number,
	      units => $units};
    return bless ($value);
}

sub readSetValue ()
{
    my (
	$this
	) = @_;
    my (
	$value,
	$token,
	$subValue
	);
    $value = {type  => $objectTypes->{SET},
	      value => []};
    $token = $this->lookAhead (true);
    if ($token->{type} == $tokenTypes->{SPECIAL_CHARACTER} &&
	$token->{value} eq "}")
    {
	$this->getInputToken (true);
    }
    else
    {
	while ($token->{type} != $tokenTypes->{SPECIAL_CHARACTER} ||
	       $token->{value} ne "}")
	{
	    $subValue = $this->readValue (1);
	    push (@{$value->{value}}, $subValue);
	    $token = $this->getInputToken (true);
	    die ("Illegal text \"".$token->{value}."\" found where ".
		 "\"}\" or \",\" is expected.")
		if ($token->{type} != $tokenTypes->{SPECIAL_CHARACTER} ||
		    ($token->{value} ne "," &&
		     $token->{value} ne "}"));
	}
    }
    return bless ($value);
}

sub readSequence1dValue ()
{
    my (
	$this
	) = @_;
    my (
	$value,
	$token,
	$subValue
	);
    $value = {type  => $objectTypes->{SEQUENCE_1D},
	      value => []};
    $token = $this->lookAhead (true);
    if ($token->{type} == $tokenTypes->{SPECIAL_CHARACTER} &&
	$token->{value} eq ")")
    {
	$this->getInputToken (true);
    }
    else
    {
	while ($token->{type} != $tokenTypes->{SPECIAL_CHARACTER} ||
	       $token->{value} ne ")")
	{
	    $subValue = $this->readValue (1);
	    push (@{$value->{value}}, $subValue);
	    $token = $this->getInputToken (true);
	    die ("Illegal text \"".$token->{value}."\" found where ".
		 "\")\" or \",\" is expected.")
		if ($token->{type} != $tokenTypes->{SPECIAL_CHARACTER} ||
		    ($token->{value} ne "," &&
		     $token->{value} ne ")"));
	}
    }
    return bless ($value);
}

sub readSequence2dValue ()
{
    my (
	$this
	) = @_;
    my (
	$value,
	$subValue,
	$token
	);
    $value = {type  => $objectTypes->{SEQUENCE_2D},
	      value => []};
    do
    {
	$subValue = $this->readValue (2);
	push (@{$value->{value}}, $subValue);
	$token = $this->lookAhead (true);
    } until ($token->{type} == $tokenTypes->{SPECIAL_CHARACTER} &&
	     $token->{value} eq ")");
    $token = $this->getInputToken (true);
    return bless ($value);
}

sub readValue (;$)
{
    my (
	$this,
	$mode
	) = @_;
    my (
	$token,
	$value,
	$type,
	$units
	);
    # Modes: 0 or undef: Accept all types of values.
    #        1: Accept only scalars
    #        2: Accept only singly-nested lists
    attempt
    {
	$token = $this->getInputToken (true);
	if ($token->{type} == $tokenTypes->{TEXT_STRING})
	{
	    die ("Scalar found where list expected")
		if ($mode == 2);
	    $value = createTextValue ($token->{value});
	}
	elsif ($token->{type} == $tokenTypes->{SYMBOL_STRING})
	{
	    die ("Scalar found where list expected")
		if ($mode == 2);
	    $value = createSymbolValue ($token->{value});
	}
	elsif ($token->{type} == $tokenTypes->{IDENTIFIER})
	{
	    die ("Scalar found where list expected")
		if ($mode == 2);
	    $value = createIdentifierValue ($token->{value});
	}
	elsif ($token->{type} == $tokenTypes->{DATE})
	{
	    die ("Scalar found where list expected")
		if ($mode == 2);
	    $value = createDateValue ($token->{value});
	}
	elsif ($token->{type} == $tokenTypes->{TIME})
	{
	    die ("Scalar found where list expected")
		if ($mode == 2);
	    $value = createTimeValue ($token->{value});
	}
	elsif ($token->{type} == $tokenTypes->{DATE_TIME})
	{
	    die ("Scalar found where list expected")
		if ($mode == 2);
	    $value = createDateTimeValue ($token->{value});
	}
	elsif ($token->{type} == $tokenTypes->{INTEGER} ||
	       $token->{type} == $tokenTypes->{UNSIGNED_INTEGER} ||
	       $token->{type} == $tokenTypes->{BASED_INTEGER} ||
	       $token->{type} == $tokenTypes->{SCALED_REAL} ||
	       $token->{type} == $tokenTypes->{UNSCALED_REAL})
	{
	    die ("Scalar found where list expected")
		if ($mode == 2);
	    $value = createNumericValue ($token->{value},
					 $this->getAnyUnits (),
					 $token->{type});
	}
	elsif ($token->{type} == $tokenTypes->{SPECIAL_CHARACTER} &&
	       $token->{value} eq "{")
	{
	    die ("Set found where list expected")
		if ($mode == 2);
	    die ("Set found where scalar expected")
		if ($mode == 1);
	    $value = $this->readSetValue ();
	}
	elsif ($token->{type} == $tokenTypes->{SPECIAL_CHARACTER} &&
	       $token->{value} eq "(")
	{
	    # We're reading some sort of list
	    die ("List found where scalar expected")
		if ($mode == 1);
	    # Read the next token, and then put it back.
	    $token = $this->lookAhead (true);
	    # See what type of values the list contains
	    if ($token->{type} == $tokenTypes->{SPECIAL_CHARACTER} &&
		$token->{value} eq "(")
	    {
		# The list contains sublists
		# If mode is 2, then we can only read singly nested lists
		# e.g. (1, 2, 3), not lists with sublists ((1,2)(3,4)(5,6))
		die ("List found where scalar is expected")
		    if ($mode == 2);
		$value = $this->readSequence2dValue ();
	    }
	    else
	    {
		$value = $this->readSequence1dValue ();
	    }
	}
	else
	{
	    die ("Unexpected token where value is expected");
	}
    };
    return bless ($value);
}

sub lookAhead (;$)
{
    my (
	$this,
	$noSemanticVoids
	) = @_;
    my (
	$accumulator,
	$token
	);
    $accumulator = [];
    $this->registerAccumulator ($accumulator);
    try
    {
	$token = $this->getInputToken ($noSemanticVoids);
    }
    finally
    {
	$this->restoreTokens ($accumulator);
	$this->removeAccumulator ($accumulator);
    };
    return $token;
}

sub getAnyUnits ()
{
    my (
	$this
	) = @_;
    my (
	$accumulator,
	$token,
	$unitsExpression
	);
    $accumulator = [];
    $this->registerAccumulator ($accumulator);
    try
    {
	$token = $this->getInputToken (true);
	if ($token->{type} == $tokenTypes->{SPECIAL_CHARACTER} &&
	    $token->{value} eq "<")
	{
	    $token = $this->getInputToken ();
	    die ("Illegal content in units expression")
		if ($token->{type} != $tokenTypes->{IDENTIFIER});
	    $token = $this->getInputToken ();
	    if ($token->{type} == $tokenTypes->{SPECIAL_CHARACTER} &&
		$token->{value} eq "**")
	    {
		$token = $this->getInputToken ();
		die ("Illegal exponent in units expression")
		    if ($token->{type} != $tokenTypes->{INTEGER} &&
			$token->{type} != $tokenTypes->{UNSIGNED_INTEGER});
		$token = $this->getInputToken ();
	    }
	    while ($token->{type} != $tokenTypes->{SPECIAL_CHARACTER} ||
		   $token->{value} ne ">")
	    {
		die ("Illegal operator in units expression")
		    if ($token->{type} != $tokenTypes->{SPECIAL_CHARACTER} ||
			($token->{value} ne "*" &&
			 $token->{value} ne "/"));
		$token = $this->getInputToken ();
		die ("Illegal content in units expression")
		    if ($token->{type} != $tokenTypes->{IDENTIFIER});
		$token = $this->getInputToken ();
		if ($token->{type} == $tokenTypes->{SPECIAL_CHARACTER} &&
		    $token->{value} eq "**")
		{
		    $token = $this->getInputToken ();
		    die ("Illegal exponent in units expression")
			if ($token->{type} != $tokenTypes->{INTEGER} &&
			    $token->{type} != $tokenTypes->{UNSIGNED_INTEGER});
		    $token = $this->getInputToken ();
		}
	    }
	    $unitsExpression = "";
	    $unitsExpression .= $_->{value} foreach @$accumulator;
	    $unitsExpression =~ s/^\s+//;
	}
	else
	{
	    $unitsExpression = undef;
	    $this->restoreTokens ($accumulator);
	}
    }
    catch
    {
	$this->restoreTokens ($accumulator);
	die ($_);
    }
    finally
    {
	$this->removeAccumulator ($accumulator);
    };
    return $unitsExpression;
}

sub ensureEndOfLine ()
{
    my (
	$this
	) = @_;
    my (
	$token,
	$check,
	$accumulator
	);
    # Attempt to read until the end of the line.
    # If we fail, we will backtrack in the stream to where reading
    # started.
    attempt
    {
	# retrieve tokens until one of two things happen:
	# 1) we find a whitespace token representing a new line
	# 2) we find a non whitespace and non comment token
	do
	{
	    $token = $this->getInputToken ();
	} until (($token->{type} == $tokenTypes->{WHITESPACE} &&
		  $token->{value} eq "\n") ||
		 ($token->{type} != $tokenTypes->{COMMENT} &&
		  $token->{type} != $tokenTypes->{WHITESPACE}));
	# If the last token found wasn't whitespace, then it has to also be a
	# non-comment token.
	if ($token->{type} != $tokenTypes->{WHITESPACE})
	{
	    # If it was a semicolon (a valid, but unadvised part of the
	    # PDS implementation of PVL-Specific extensions), then we still
	    # need to look for the end of the line FOLLOWING the semicolon
	    if ($token->{type} == $tokenTypes->{SPECIAL_CHARACTER} &&
		$token->{value} eq ";")
	    {
		# First, check to see if there are any sanity checks regarding
		# semicolons, and if so, execute them.
		foreach $check (@{$sanity->{SEMICOLON}})
		{
		    &$check ();
		}
		# Keep looking, as before, for the end of the line.
		do
		{
		    $token = $this->getInputToken ();
		} until (($token->{type} == $tokenTypes->{WHITESPACE} &&
			  $token->{value} eq "\n") ||
			 ($token->{type} != $tokenTypes->{COMMENT} &&
			  $token->{type} != $tokenTypes->{WHITESPACE}));
	    }
	}
	# Now, we've reached either the end of the line, or something that
	# doesn't belong. Check to see if it's whitespace (which it will only
	# be if it's the end of the line)
	if ($token->{type} != $tokenTypes->{WHITESPACE})
	{
	    # It's not whitespace, so whatever is there isn't allowed.
	    die ("Extra content (\"".$token->{value}."\") past end of statement");
	}
    };
}

sub restoreTokens ($)
{
    my (
	$this,
	$value
	) = @_;
    if (ref ($value) eq "ARRAY")
    {
	@accumulator = @$value;
	while ($value = pop (@accumulator))
	{
	    $this->restoreTokens ($value);
	}
    }
    else
    {
	push (@{$this->{input}->{tokenStack}}, $value);
	if ($value->{type} == $tokenTypes->{WHITESPACE} &&
	    $value->{value} eq "\n")
	{
	    -- $this->{input}->{lineNumber};
	}
	foreach $accumulator (@{$this->{input}->{accumulators}})
	{
	    pop (@$accumulator);
	}
    }
}

sub registerAccumulator ($)
{
    my (
	$this,
	$accumulator
	) = @_;
    push (@{$this->{input}->{accumulators}}, $accumulator);
}

sub removeAccumulator ($)
{
    my (
	$this,
	$accumulator
	) = @_;
    for ($i = 0; $i < scalar (@{$this->{input}->{accumulators}}); ++$i)
    {
	if ($this->{input}->{accumulators}->[$i] == $accumulator)
	{
	    splice (@{$this->{input}->{accumulators}}, $i, 1);
	    last;
	}
    }
}

sub getInputToken (;$)
{
    my (
	$this,
	$noSemanticVoids
	) = @_;
    my (
	$rest,
	$token,
	$right,
	$nextLine,
	$date,
	$time,
	$remainder,
	$token
	);
    if ($noSemanticVoids)
    {
	do
	{
	    $token = $this->getInputToken ();
	} until ($token->{type} != $tokenTypes->{WHITESPACE} &&
		 $token->{type} != $tokenTypes->{COMMENT});
	return $token;
    }
  FIND_TOKEN:
    {
	if (scalar (@{$this->{input}->{tokenStack}}) > 0)
	{
	    $token = pop (@{$this->{input}->{tokenStack}});
	    if ($token->{type} == $tokenTypes->{WHITESPACE} &&
		$token->{value} eq "\n")
	    {
		++ $this->{input}->{lineNumber};
	    }
	    last;
	}
	# Get more content if there is none remaining in the current line and
	# we haven't reached the end of the stream yet.
	if (defined ($this->{input}->{remainder}) &&
	    $this->{input}->{remainder} eq "")
	{
	    $this->{input}->{remainder} = $this->getInputLine();
	    # Return a new-line whitespace token to signal that a new line has
	    # been read, if we're not on the first line.
	    if ($. > 1)
	    {
		$token = {type  => $tokenTypes->{WHITESPACE},
			  value => "\n"};
		last;
	    }
	}
	# If the remaining content is (still?) undefined, we've reached the end
	# of the stream. Return the EOF token.
	if (!defined ($this->{input}->{remainder}))
	{
	    $token = {type => $tokenTypes->{EOF},
		      value => undef};
	    last;
	}
	$rest = $this->{input}->{remainder};
	# Handle white-space tokens
	if ($rest =~ /^(\s+)(\S.*)?$/)
	{
	    $token = $1;
	    $this->{input}->{remainder} = defined ($2)?$2:"";
	    $token = {type => $tokenTypes->{WHITESPACE},
		      value => $token};
	    last;
	}
	# Special case for new-lines
	elsif ($rest eq "\n")
	{
	    # Clear the remaining content
	    $this->{input}->{remainder} = "";
	    # Get the next token. This will now trigger a new-line token.
	    return $this->getInputToken();
	}
	# Handle comments
	elsif ($rest =~ /^\/\*/)
	{
	    if ($rest =~ /^(
			    \/\*                    # Opening slash-asterisk
			    (?:[^\*]|(?:\*(?!\/)))* # Content
			    \*\/                    # closing asterisk-slash 
			   )(.*)$/x)
	    {
		$token = $1;
		$right = $2;
		die ("Content (\"$right\") follows comment")
		    if ($right !~ /^\s*$/);
		$this->{input}->{remainder} = "";
		$token = {type => $tokenTypes->{COMMENT},
			  value => $token};
		last;
	    }
	    else
	    {
		die ("Comment not terminated");
	    }
	}
	# Handle identifiers:
	# There is a tricky bit of BNF in the PDS Standards Reference regarding
	# to underscores in identifiers. While in most regular languages,
	# things of this nature are usually: a letter or underscore followed by
	# any number of letters, digits, and underscores. The PDS states that 
	# identifiers must start with a letter, and any underscore in the
	# identifier must be followed by a letter or digit. So, while some
	# regular langues allow:
	#  _IDENTIFIER           (opening with an underscore)
	#  ANOTHER__IDENTIFIER   (two underscores following eachother)
	#  A_THIRD_IDENTIFIER_   (closing with an underscore)
	# The PDS does not allow any of these
	elsif ($rest =~ /^(
			   [A-Za-z]          # Must begin with a letter
			   (?:
			    [A-Za-z0-9]      # Can contain any letters or digits
			    |                # and
			    (?:_[A-Za-z0-9]) # Underscores followed by
			                     # letters or digits
			   )*
			  )
	                  (.*)$/x)
	{
	    $token = $1;
	    $this->{input}->{remainder} = $2;
	    foreach $check (@{$sanity->{IDENTIFIER}})
	    {
		&$check ($token);
	    }
	    $token = {type => $tokenTypes->{IDENTIFIER},
		      value => $token};
	    last;
	}
	# Handle special characters
	elsif ($rest =~ /^(\*\*|[\=\,\*\/\^\<\>\(\)\{\}\:\;])(.*)$/)
	{
	    $token = $1;
	    $this->{input}->{remainder} = $2;
	    $token = {type => $tokenTypes->{SPECIAL_CHARACTER},
		      value => $token};
	    last;
	}
	# Handle text strings
	elsif ($rest =~ /^\"/)
	{
	    if ($rest =~ /^(\"[^\"]+\")(.*)$/)
	    {
		$token = $1;
		$this->{input}->{remainder} = $2;
		$token = {type => $tokenTypes->{TEXT_STRING},
			  value => $token};
		last;
	    }
	    else
	    {
		$token = $rest."\n";
		for ($nextLine = $this->getInputLine();
		     defined ($nextLine) && $nextLine !~ /\"/;
		     $nextLine = $this->getInputLine())
		{
		    $token .= $nextLine;
		}
		die ("Reached EOF before string terminated")
		    if (!defined ($nextLine));
		$nextLine =~ /^([^\"]*\")(.*)/;
		$token .= $1;
		$this->{input}->{remainder} = $2;
		$token = {type => $tokenTypes->{TEXT_STRING},
			  value => $token};
		last;
	    }
	}
	# Handle symbol strings
	elsif ($rest =~ /^(\'[^\']*\')(.*)$/)
	{
	    $token = $1;
	    $this->{input}->{remainder} = $2;
	    $token = {type => $tokenTypes->{SYMBOL_STRING},
		      value => $token};
	    last;
	}
	# Handle date-times
	elsif ($rest =~ /^(
			   (                      # Date portion:
			    (?:\d+-\d+-\d+)       # Year-Month-Day
			    |                     # or
			    (?:\d+-\d+)           # Year-DOY
			   )
			   T                      # 'T' combines date and time
			   (                      # Time portion:
			    \d+\:\d+              # Hour:Minute
			    (?:\:                 # :, for (optional) secconds
			     (?:                  # Types of 'second' values:
			      (?:\d+(?:\.\d*)?)   #   1, 1., 1.2
			      |                   # or
			      (?:\.\d+)           #   .2
			     )
			    )?
			    (?:                   # Optional time zone portion:
			     Z                    # Z, for UTC time
			     |                    # or
			     (?:                  # Zoned time:
			      [\+\-]              # Sign
			      \d+                 # Hour
			      (?:\:\d+)?          # Optional minute.
			     )
			    )?
			   )
			  )
	                  (.*)$/x)
	{
	    ($token, $date, $time, $remainder) = ($1, $2, $3, $4);
	    foreach $check (@{$sanity->{DATE}})
	    {
		&$check ($date);
	    }
	    foreach $check (@{$sanity->{TIME}})
	    {
		&$check ($time);
	    }
	    foreach $check (@{$sanity->{DATE_TIME}})
	    {
		&$check ($token);
	    }
	    $this->{input}->{remainder} = $remainder;
	    $token = {type => $tokenTypes->{DATE_TIME},
		      value => $token};
	    last;
	}
	# Handle dates
	elsif ($rest =~ /^(
			   (?:\d+-\d+-\d+) # Year-Month-Day
			   |               # or
			   (?:\d+-\d+)     # Year-DOY
			  )
	                  (.*)$/x)
	{
	    $token = $1;
	    $this->{input}->{remainder} = $2;
	    foreach $check (@{$sanity->{DATE}})
	    {
		&$check ($token);
	    }
	    $token = {type => $tokenTypes->{DATE_TIME},
		      value => $token};
	    last;
	}
	# Handle times
	elsif ($rest =~ /^(
			   \d+\:\d+              # Hour:Minute
			   (?:\:                 # :, for (optional) secconds
			    (?:                  # Types of 'second' values:
			     (?:\d+(?:\.\d*)?)   #   1, 1., 1.2
			     |                   # or
			     (?:\.\d+)           #   .2
			    )
			   )?
			   (?:                   # Optional time zone portion:
			    Z                    # Z, for UTC time
			    |                    # or
			    (?:                  # Zoned time:
			     [\+\-]              # Sign
			     \d+                 # Hour
			     (?:\:\d+)?          # Optional minute.
			    )
			   )?
			  )
	                  (.*)$/x)
	{
	    $token = $1;
	    $this->{input}->{remainder} = $2;
	    foreach $check (@{$sanity->{TIME}})
	    {
		&$check ($token);
	    }
	    $token = {type => $tokenTypes->{DATE_TIME},
		      value => $token};
	    last;
	}
	# Handle based integers
	elsif ($rest =~ /^(\d+\#[\+\-]?[0-9A-Za-z]+\#)(.*)$/)
	{
	    $token = $1;
	    $this->{input}->{remainder} = $2;
	    foreach $check (@{$sanity->{BASED_INTEGER}})
	    {
		&$check ($token);
	    }
	    $token = {type => $tokenTypes->{BASED_INTEGER},
		      value => $token};
	    last;
	}
	# Handle reals
	elsif ($rest =~ /^(
			   [\+\-]?             # Optional sign
			   (?:                 # Types of unscaled reals:
			    (?:\d+(?:\.\d*)?)  #   1, 1., 1.1
			    |                  # or
			    (?:\.\d+)          #   .1
			   )
			   (?:[Ee][\+\-]?\d+)? # Optional exponent
			  )
	                  (.*)?$/x)
	{
	    $token = $1;
	    $this->{input}->{remainder} = $2;
	    if ($token =~ /[\.Ee]/)
	    {
		$token = {type => $token =~ /[eE]/?$tokenTypes->{SCALED_REAL}:
			                           $tokenTypes->{UNSCALED_REAL},
			  value => $token};
		last;
	    }
	    else
	    {
		$token = {type => $token =~ /^[\+\-]/?$tokenTypes->{INTEGER}:
			                              $tokenTypes->{UNSIGNED_INTEGER},
			  value => $token};
		last;
	    }
	}
	else
	{
	    die ("Unrecognized token beginning at \"".$rest."\"");
	}
    }
    foreach $accumulator (@{$this->{input}->{accumulators}})
    {
	push (@{$accumulator}, $token);
    }
    return $token;
}

sub getInputLine ($)
{
    my (
	$this,
	) = @_;
    my (
	$stream,
	$line
	);
    $stream = $this->{input}->{stream};
    return undef if (eof ($stream));
    $line = <$stream>;
    ++ $this->{input}->{lineNumber};
    push (@{$this->{input}->{inputLines}}, $line);
    return $line;
}

sub getLastInputLine ()
{
    my (
	$this
	) = @_;
    return $this->{input}->{inputLines}->[-1];
}

##############################################################################
###   Identity functions ###
############################

sub getSize ()
{
    my (
	$this
	) = @_;
    die ("Only sets and sequences have sizes.")
	unless ($this->isaSequence() || $this->isaValue("SET"));
    return scalar(@{$this->{value}});
}

sub getType ()
{
    my (
	$this
	) = @_;
    my (
	$type
	);
    ($type) = grep {$objectTypes->{$_} == $this->{type}} keys (%{$objectTypes});
    return $type;
}

sub isaValue (;$)
{
    my (
	$this,
	$type
	) = @_;
    return $this->isType ($type, qw{TEXT_STRING
				    SYMBOL_STRING
				    IDENTIFIER
				    INTEGER
				    UNSIGNED_INTEGER
				    BASED_INTEGER
				    UNSCALED_REAL
				    SCALED_REAL
				    DATE
				    TIME
				    DATE_TIME
				    SET
				    SEQUENCE_1D
				    SEQUENCE_2D});
}

sub isaSequence (;$)
{
    my (
	$this,
	$type
	) = @_;
    return $this->isType ($type, qw{SEQUENCE_1D
				    SEQUENCE_2D});
}

sub isaScalar (;$)
{
    my (
	$this,
	$type
	) = @_;
    return $this->isType ($type, qw{TEXT_STRING
				    SYMBOL_STRING
				    IDENTIFIER
				    INTEGER
				    UNSIGNED_INTEGER
				    BASED_INTEGER
				    UNSCALED_REAL
				    SCALED_REAL
				    DATE
				    TIME
				    DATE_TIME});
}

sub isaNumber (;$)
{
    my (
	$this,
	$type
	) = @_;
    return $this->isType ($type, qw{INTEGER
				    UNSIGNED_INTEGER
				    BASED_INTEGER
				    UNSCALED_REAL
				    SCALED_REAL});
}

sub isaDateOrTime (;$)
{
    my (
	$this,
	$type
	) = @_;
    return $this->isType ($type, qw{DATE
				    TIME
				    DATE_TIME});
}

sub isaStatement (;$)
{
    my (
	$this,
	$type
	) = @_;
    $type = END_FILE if ($type eq END);
    return $this->isType ($type, qw{FILE
				    ATTRIBUTE
				    POINTER
				    OBJECT
				    GROUP
				    END_FILE});
}

sub isaValueStatement (;$)
{
    my (
	$this,
	$type
	) = @_;
    return $this->isType ($type, qw{ATTRIBUTE
				    POINTER});
}

sub isaNestingStatement (;$)
{
    my (
	$this,
	$type
	) = @_;
    return $this->isType ($type, qw{FILE
				    OBJECT
				    GROUP});
}

sub isType (;$@)
{
    my (
	$this,
	$type,
	) = @_;
    my (
	@types
	);
    @types = splice (@_, 2);
    @types = keys (%$objectTypes) if (scalar(@types) == 0);
    if (grep {$type eq $_} @types)
    {
	return $this->{type} == $objectTypes->{$type};
    }
    elsif (!defined ($type))
    {
	return grep {$this->{type} == $objectTypes->{$_}} @types;
    }
    else
    {
	return false;
    }
}

##############################################################################
###   Value retrieval   ###
###########################

sub getValueObject (;$)
{
    my (
	$this,
	$keyword
	) = @_;
    my (
	$objectIdentifier,
	$rest,
	$index,
	@statements,
	$statement
	);
    if ($keyword =~ /^([A-Za-z](?:\w|\:)*)(.*)$/)
    {
	die ("Only OBJECT, GROUP, or FILE objects contain substatements.")
	    unless ($this->isaNestingStatement ());
	($identifier, $rest) = ($1, $2);
	if ($rest =~ /^\[(\d+)\](.*)$/)
	{
	    ($statementIndex, $rest) = ($1, $2);
	}
	else
	{
	    $statementIndex = 1;
	}
	@statements = $this->getStatements ($identifier);
	die ("No ".$identifier." exists at index ".$statementIndex."\n")
	    if ($statementIndex > scalar(@statements));
	$statement = $statements[$statementIndex - 1];
	if ($rest =~ /^\(/ ||
	    $rest =~ /^\./ ||
	    $rest eq "")
	{
	    $rest =~ s/^\.//;
	    return $statement->getValueObject ($rest);
	}
	else
	{
	    die ("Invalid statement locator: ".$rest);
	}
    }
    elsif ($keyword =~ /^\(/)
    {
	die ("Only sets and sequences have indexed values")
	    unless ($this->isaValueStatement() &&
		    ($this->{value}->isaSequence() ||
		     $this->{value}->isaValue("SET")));
	if ($keyword =~ /^\((\d+)(?:,(\d+))?\)$/)
	{
	    ($index1, $index2) = ($1, $2);
	    die ("Only 2D sequences have dual-indexed values")
		unless ($this->{value}->isaSequence("SEQUENCE_2D") ||
			$index2 eq "");
	    die ("No value at index ".$index1)
		unless (0 < $index1 && $index1 <= scalar(@{$this->{value}->{value}}));
	    $value = $this->{value}->{value}->[$index1 - 1];
	    if ($index2 ne "")
	    {
		die ("No value at index ".$index1.",".$index2)
		    unless (0 < $index2 &&
			    $index2 <= scalar(@{$value->{value}}));
		$value = $value->{value}->[$index2 - 1];
	    }
	    return $value;
	}
	else
	{
	    die ("Invalid indexing: ".$keyword);
	}
    }
    elsif ($keyword eq "")
    {
	die ("Only attributes and pointers have values")
	    unless ($this->isaValueStatement());
	return $this->{value};
    }
}

sub getStatements ($)
{
    my (
	$this,
	$keyword
	) = @_;
    my (
	@list
	);
    foreach $statement (@{$this->{statements}})
    {
	if (($statement->{type} == $objectTypes->{OBJECT} &&
	     $statement->{object} eq $keyword) ||
	    ($statement->{type} == $objectTypes->{GROUP} &&
	     $statement->{group} eq $keyword) ||
	    ($statement->isaValueStatement() &&
	     $statement->getWholeKeyword() eq $keyword))
	{
	    push (@list, $statement);
	}
    }
    return @list;
}

sub getPointers ()
{
    my (
	$this
	) = @_;
    my (
	@pointers
	);
    foreach $statement (@{$this->{statements}})
    {
	push (@pointers, $statement)
	    if ($statement->isaValueStatement("POINTER"));
    }
    return @pointers;
}

sub getObjects ()
{
    my (
	$this
	) = @_;
    my (
	@objects
	);
    foreach $statement (@{$this->{statements}})
    {
	push (@objects, $statement)
	    if ($statement->isaNestingStatement("OBJECT"));
    }
    return @objects;
}

sub getRawValueNoUnits ()
{
    my (
	$self
	) = @_;
    die ("Only scalars can be converted into raw values")
	unless $self->isaScalar();
    if ($self->isaScalar("TEXT_STRING"))
    {
	$value = $self->{value};
	$value =~ s/^\"(.*)\"$/\1/;
    }
    elsif ($self->isaScalar("SYMBOL_STRING"))
    {
	$value = $self->{value};
	$value =~ s/^\'(.*)\'$/\1/;
    }
    elsif ($self->isaScalar("IDENTIFIER") ||
	   $self->isaDateOrTime())
    {
	$value = $self->{value};
    }
    elsif ($self->isaNumber())
    {
	$value = $self->{value};
    }
    return $value;
}

sub getRawValue ()
{
    my (
	$self
	) = @_;
    die ("Only scalars can be converted into raw values")
	unless $self->isaScalar();
    if ($self->isaScalar("TEXT_STRING"))
    {
	$value = $self->{value};
	$value =~ s/^\"(.*)\"$/\1/;
    }
    elsif ($self->isaScalar("SYMBOL_STRING"))
    {
	$value = $self->{value};
	$value =~ s/^\'(.*)\'$/\1/;
    }
    elsif ($self->isaScalar("IDENTIFIER") ||
	   $self->isaDateOrTime())
    {
	$value = $self->{value};
    }
    elsif ($self->isaNumber())
    {
	$value = $self->{value};
	$value .= " ".$self->{units} if (defined ($self->{units}));
    }
    return $value;
}

sub getWholeKeyword ()
{
    my (
	$this
	) = @_;
    my (
	$keyword
	);
    die ("Only Attributes and Pointers have keywords")
	unless ($this->isaValueStatement ());
    if ($this->{type} == $objectTypes->{ATTRIBUTE})
    {
	if (exists ($this->{keyword}->{namespace}))
	{
	    $keyword = $this->{keyword}->{namespace}.":";
	}
	$keyword .= $this->{keyword}->{element};
    }
    elsif ($this->{type} == $objectTypes->{POINTER})
    {
	$keyword = "^".$this->{keyword}->{object};
    }
    else
    {
	die ("Logical flaw in ODL library.");
    }
}

sub getPerlNumber ()
{
    my (
	$this
	) = @_;
    die ("Can't get Perl Number from a non-numeric value")
	unless ($this->isaNumber());
    if ($this->{type} == $objectTypes->{BASED_INTEGER})
    {
	if ($this->{value} =~ /^(\d+)\#([\+\-]?[0-9A-Fa-f]+)\#$/)
	{
	    ($radix, $digits) = ($1, $2);
	    $digits = uc($digits);
	    $negative = ($digits =~ /^-/);
	    $digits =~ s/^[\+\-]//;
	    $value = 0;
	    while (length ($digits) > 0)
	    {
		$digit = substr ($digits, 0, 1, "");
		if ($digit !~ /^[0-9]$/)
		{
		    $digit = ord($digit) - ord("A") + 10;
		}
		$value *= $radix;
		$value += $digit;
	    }
	    if ($negative)
	    {
		$value = -$value;
	    }
	    return $value;
	}
	else
	{
	    die ("Malformed BASED_INTEGER: ".$this->{value}."\n");
	}
    }
    else
    {
	return (0 + $this->{value});
    }
}

##############################################################################
###   Sanity Checks ###
#######################

sub odlDenyBeginGroup ()
{
    die ("BEGIN_GROUP is not allowed in ODL.\n".
	 "Use GROUP instead.");
}

sub odlDenyBeginObject ()
{
    die ("BEGIN_OBJECT is not allowed in ODL.\n".
	 "Use OBJECT instead.");
}

sub odlDenySemicolon ()
{
    die ("Semicolons are not allowed in ODL.");
}

sub pdsAllowBeginGroup ()
{
    #warn ("PDS allows BEGIN_GROUP");
}

sub pdsAllowBeginObject ()
{
    #warn ("PDS allows BEGIN_OBJECT");
}

sub pdsAllowSemicolon ()
{
    #warn ("PDS allows semicolons.");
}

sub pdsCheckKeyword ($$$$)
{
    my (
	$whole,
	$pointer,
	$namespace,
	$element
	) = @_;
    die ("Namespace in attribute \"".$whole."\" is too long.")
	if (length ($namespace) > 30);
    die ("Element identifier in attribute \"".$whole."\" is too long.")
	if (length ($item) > 30);
}

sub pdsCheckTime ($)
{
    my (
	$time
	) = @_;
    die ("PDS does not allow time zones in times")
	if ($time =~ /[\+\-]/);
}

sub pdsCheckDate ($)
{
    my (
	$date
	) = @_;
    my (
	$year,
	$month,
	$day,
	$doy
	);
    if ($date =~ /^(\d+)-(\d+)-(\d+)$/)
    {
	($year, $month, $day) = ($1, $2, $3);
	die ("PDS requires a 4-digit year be used.")
	    if (length ($year) != 4);
	die ("Values in integral parts of dates must be padded on the\n".
	     "left with zeros as necessary to fill the field.")
	    if (length ($month) != 2 ||
		length ($day)   != 2);
    }
    elsif ($date =~ /^(\d+)-(\d+)$/)
    {
	($year, $doy) = ($1, $2);
	die ("PDS requires a 4-digit year be used.")
	    if (length ($year) != 4);
	die ("Values in integral parts of dates must be padded on the\n".
	     "left with zeros as necessary to fill the field.")
	    if (length ($doy) != 3);
    }
}

sub checkDate ($)
{
    my (
	$date
	) = @_;
    my (
	$year,
	$month,
	$day,
	$doy,
	$days
	);
    if ($date =~ /^(\d+)-(\d+)-(\d+)$/)
    {
	($year, $month, $day) = ($1, $2, $3);
	die ("Month must be an integer between 1 and 12.")
	    if ($month < 1 || 12 < $month);
	# January, March, May, July, August, October, and December:
	# 31 days.
	if (grep {$month == $_} (1, 3, 5, 7, 8, 10, 12))
	{
	    $days = 31;
	}
	# April, June, September, and November: 30 days
	elsif (grep {$month == $_} (4, 6, 9, 11))
	{
	    $days = 30;
	}
	# February on a leap year: 29 days
	elsif (($year % 4) == 0 &&
	       (($year % 100) != 0 ||
		($year % 400) == 0))
	{
	    $days = 29;
	}
	# February on a non-leap year: 28 days
	else
	{
	    $days = 28;
	}
	die ("Day of month must be an integer in the range 1 to 31,\n".
	     "as appropriate for the particular month and year.")
	    if ($day < 1 || $days < $day);
    }
    elsif ($date =~ /^(\d+)-(\d+)$/)
    {
	($year, $doy) = ($1, $2);
	# Leap years have 366 days
	if (($year % 4) == 0 &&
	    (($year % 100) != 0 ||
	     ($year % 400) == 0))
	{
	    $days = 366;
	}
	# Non leap years have 355 days
	else
	{
	    $days = 365;
	}
	die ("Day or year must be in the range 1 to 365,\n".
	     "or 366 in a leap year.")
	    if ($doy < 1 || $days < $doy);
    }
    else
    {
	die ("Malformed date: \"".$date."\".");
    }
}

sub checkTime ($)
{
    my (
	$time
	) = @_;
    my (
	$hour,
	$minute,
	$second,
	$zHour,
	$zMinute
	);
    if ($time =~ /^(\d+)\:(\d+)          # Hour:Minute
	           (?:\:                 # :, for (optional) secconds
		    (                    # Types of 'second' values:
		     (?:\d+(?:\.\d*)?)   #   1, 1., 1.2
		     |                   # or
		     (?:\.\d+)           #   .2
		    )
		   )?
	           (?:                   # Optional time zone portion:
		    Z                    # Z, for UTC time
		    |                    # or
		    (?:                  # Zoned time:
		     [\+\-]              # Sign
		     (\d+)               # Hour
		     (?:\:(\d+))?        # Optional minute.
		    )
		   )?$/x)
    {
	($hour, $minute, $second, $zHour, $zMinute) = ($1, $2, $3, $4, $5, $6);
	#print ("[$hour][$minute][$second][$zHour][$zMinute]\n");
	die ("Hours must be in the range 0 to 23")
	    if ($hour < 0 || 23 < $hour);
	die ("Minutes must be in the range 0 to 59")
	    if ($minute < 0 || 59 < $minute);
	die ("Seconds, if specified, must be greater than or\n".
	     "equal to 0 and less than 60.")
	    if ($second ne "" &&
		$second < 0 || 60 <= $second &&
		$second !~ /59.9*$/);
	die ("Within zoned time values:\n".
	     "Hours must be in the range -12 to +12.")
	    if ($zHour ne "" &&
		$zHour < 0 || 12 < $zHour);
	die ("Within zomed time values:\n".
	     "Minutes, if specified, must be in the range 0 to 59.")
	    if ($zMinute ne "" &&
		$zMinute < 0 || 59 < $zMinute);
    }
    else
    {
	die ("Malformed time");
    }
}

sub pdsCheckBasedInteger ($)
{
    my (
	$integer
	) = @_;
    my (
	$base,
	$extended,
	$digits
	);
    die ("Signs may not be used in non-decimal numbers.")
	if ($integer =~ /\-/);
    if ($integer =~ /^(\d+)\#[\+\-]?([0-9A-Za-z]+)\#$/)
    {
	($base, $extended) = ($1, $2);
	die ("Only the bases 2, 8, and 16 may be used for ".
	     "non-decimal numbers")
	    if ($base !=  2 &&
		$base !=  8 &&
		$base != 16);
    }
}

sub checkBasedInteger ($)
{
    my (
	$integer
	) = @_;
    my (
	$base,
	$extended,
	$digits
	);
    if ($integer =~ /^(\d+)\#[\+\-]?([0-9A-Za-z]+)\#$/)
    {
	($base, $extended) = ($1, $2);
	die ("Based integers must have a radix between 2 and 16")
	    if ($base < 2 || 16 < $base);
	if ($base < 11)
	{
	    $digits = "[0-".($base - 1)."]";
	}
	else
	{
	    $digits = ("[0-9".
		       "A-".chr(ord("A") + $base - 11).
		       "a-".chr(ord("a") + $base - 11)."]");
	}
	die ("Illegal extended digit in based integer")
	    if ($extended !~ /^$digits+$/);
    }
    else
    {
	die ("Malformed based integer");
    }
}

1;
