#!/usr/bin/perl

# ADDCOLHDR
#===========
#
# Routine to add the numeric column headings and a corresponding HEADER file
# to an ASCII TABLE file and label.
#
# Format:  % addcolhdr [-n] [-f] [-o] [ input_label [ output_label ] ]
#  
#    where    -n      indicates the COLUMN_NUMBERs should be assigned,
#                     replacing any existing COLUMN_NUMBERs
#             -f      indicates the lines should be blank padded out to 79
#                     bytes and carriage returns should be added
#             -o      overwrite existing label and data files
#
# Standard I/O is used if no label names are supplied.  The name of the data
# file is taken from the input label, with the suffix ".NEW" added if the 
# original file is not being overwritten.
#
# 09 Dec 1998, a.c.raugh
# 14 Dec 1998, acr: fixed minor bug causing abutting columns to show up as
#                   overlapping
# 26 Sep 2006, acr: Modified to use Getopt::Std
#
#==============================================================================

use Getopt::Std;

# First, collect the options:

getopts('nfo');

# Now, check for filenames and open them accordingly:

$inlbl  = (@ARGV > 0)? $ARGV[0] : "-";

# If we're overwriting the input files, we'll need a temporary file to hold
# the output label:

if ($opt_o)
  { $outlbl = "%ADDCOLHDRtmplabel"; }
else
  { $outlbl = (@ARGV > 1)? $ARGV[1] : "-"; }

open(INLBL, $inlbl) || die "Unable to open input label, $inlbl, ";

if ($opt_f)
  { open (OUTLBL, "| fixlen -r 79 | addcr > $outlbl") ||
      die "Unable to open output pipe to $outlbl"; 
  }
else
  { open (OUTLBL, ">$outlbl") || die "Unable to open output label, $outlbl, "; }

#_______________________________________________________________________________
#
# OK, we begin by stepping through the input label.  We need to make sure there
# is no existing header, get the name of the data file, and check that this is
# an ASCII table.  We should have all this information before we get to the 
# first COLUMN definition.
#

$looking_for_COLUMN = 1;
$lcount = 0;

# Check the first line to see if the input file is blank-padded and CRLF
# delimited.  If so, then the output file must be regardless of whether 
# the "-c" option was included:

$line = <INLBL>;
if (length($line) == 80  &&  $line =~ /\r$/)
  { $opt_f = 1; }

while ($looking_for_COLUMN)
  { $line = <INLBL>;
    $line =~ s/[\n\r]//g;
    $lcount++;

    # Check for pointer.  It better be a TABLE pointer and it better not have
    # a current offset.  In either case we stop processing; otherwise we save
    # the name of the data file:

    if ($line =~ /^\s*\^/)
      { if ($line !~ /^\s*\^[A-Z0-9_]*TABLE\s*=\s/)
          { printf "Encountered non-TABLE pointer, line $lcount:\n> $line\n\n";
            die;
          }

        $line =~ s/^.*=\s*//;
        $line =~ s/\s*$//;
        if ($line =~ /^\(/)
          { die "TABLE pointer already has offset, line $lcount.\n"; }
        else
          { $line =~ s/["']//g;
            $datafile = $line;
          }

        # Now that we have the name of the datafile, make sure it exists.  We
        # try both upper and lower case versions:

        if (!(-e $datafile))
          { $datafile =~ tr/a-z/A-Z/;
            if (!(-e $datafile))
              { $datafile =~ tr/A-Z/a-z/; }
          }

        if (!(-e $datafile))
          { die "Could not find datafile $line";
            undef $datafile;
          }

        $newdatafile = "$datafile.NEW";
      }

    elsif ($line =~ /INTERCHANGE_FORMAT/)
      { if ($line =~ /=.*BINARY/)
          { die "Not as ASCII file."; }
      }

    elsif ($line =~ /ROW_BYTES/)
      { $line =~ s/^.*=\s*//;
        $line =~ s/\s*$//;
        $row_bytes = $line;
      }

    elsif ($line =~ /^\s*OBJECT\s*=\s*COLUMN/)
      { $looking_for_COLUMN = 0; }
  }

#
# If we made it this far we have a file we can work with.  We continue scanning
# through the COLUMN information, collecting size and checking for overlapping
# columns:
#

$not_done = 1;
$colnum   = 1;     # Note, we're currently in the first COLUMN definition.
$overlap[1] = 0;

while ($not_done)
  { $line = <INLBL>;
    $line =~ s/[\n\r]//g;

    if ($line =~ /^\s*OBJECT\s*=\s*COLUMN/)
      { $colnum++; }

    elsif ($line =~ /^\s*START_BYTE/)
      { $line =~ s/^.*=\s*//;
        $line =~ s/\s*$//;
        $start_byte[$colnum] = $line - 1;
      }

    elsif ($line =~ /^\s*BYTES/)
      { $line =~ s/^.*=\s*//;
        $line =~ s/\s*$//;
        $bytes[$colnum] = $line;
      }

    elsif ($line =~ /^\s*END_OBJECT\s*=\s*COLUMN/  &&  $colnum > 1)
      { $overlap[$colnum] = 0;

        # At the end of the column definition we check back through the already
        # existing definitions, if any, to see if the current one overlaps an
        # existing one.  If so, we set a flag.

        for ($i=1; $i<$colnum; $i++)
          { $sb = $start_byte[$colnum];
            if ($start_byte[$i] <= $sb  &&  $sb < ($start_byte[$i]+$bytes[$i]))
              { $overlap = 1;
                $overlap[$colnum] = 1;
              }

            # Check for double overlap:

            if ($overlap[$colnum]  &&  $overlap[$i])
              { die "Double overlap in columns $i and $colnum."; }

          }
      }

    elsif ($line =~ /^\s*END$/  ||  $line =~ /^\s*END\s+/)
      { $not_done = 0; }
  }


# We're done with the first pass through the label, so we can close the file
# in preparation for re-opening it later:

close(INLBL);

#______________________________________________________________________________
#
# Now we'll re-open the label and make a second pass, this time modifying the
# label lines as needed to accommodate the header.  Recall that the output 
# label file is already opened.

open(INLBL,$inlbl) || die "Could not re-open $infile";

$looking_for_COLUMN = 1;
while ($looking_for_COLUMN)
  { $line=<INLBL>;
    $line =~ s/\s*[\r\n]+//g;

    # Check for lines that need special processing:

    if ($line =~ /FILE_RECORDS/)
      { $records = $line;
        $records =~ s/^.*=\s*//;
        $records = ($overlap)? $records+4 : $records+2;
        $line =~ s/=\s*[0-9]+/= $records/;
      }

    elsif ($line =~ /^\s*\^[A-Z0-9_]*TABLE\s*=/)
      {
        # We must add a pointer to the HEADER object and an offset for the
        # table pointer:

        $i = index($line,"=");
        $newline = sprintf ("%-${i}.${i}s= (\"$datafile\",1)","^HEADER");
        &printline($newline);

        $offset = ($overlap)? 5 : 3;
        $line = sprintf ("%-${i}.${i}s= (\"$datafile\",$offset)","^TABLE");
      }

    elsif ($line =~ /^\s*OBJECT\s*=.*TABLE/)
      {
        # We'll insert the HEADER object definition before the TABLE object
        # definition:

        $hdrrecords = ($overlap)? 4 : 2;
        $hdrbytes = $row_bytes * $hdrrecords;

        &printline("OBJECT     = HEADER");
        &printline("  HEADER_TYPE = COLUMN_LABEL");
        &printline("  BYTES       = $hdrbytes");
        &printline("  RECORDS     = $hdrrecords");
        &printline("END_OBJECT = HEADER");
        &printline(" ");
      }

    elsif ($line =~ /^\s*OBJECT\s*=\s*COLUMN/)
      { $looking_for_COLUMN = 0; 
        next;
      }

    &printline($line);
  }

# If we're adding COLUMN_NUMBERs, we have to check each COLUMN definitions
# for a COLUMN_NUMBER line and change it or add it as needed.  This means
# we can't print out the definition until AFTER we've read the whole thing 
# in:

if ($opt_n)
  { $colnum = 1;
    undef $cbuff;
    $cbuff[0] = $line;
    $cb = 1;
    $number_found = 0;
    $in_column = 1;
    $not_done  = 1;

    while ($not_done)
      { $line = <INLBL>;
        $line =~ s/\s*[\r\n]+$//;

        if ($line =~ /COLUMN_NUMBER/)
          { $line =~ s/=\s*[0-9]+/= $colnum/;
            $number_found = 1;
            $colnumline = $line;
          }

        elsif ($line =~ /END_OBJECT\s*=\s*COLUMN/)
          { if ($number_found)
              { &printline($cbuff[0]);
                &printline($colnumline);
                for ($i=1; $i<@cbuff; $i++)
                  { &printline($cbuff[$i]); }
              }
            else
              { $eq = index($cbuff[1],"=");
                $buff = $cbuff[1];
                $buff =~ s/^(\s*).*$/$1/;
                $buff .= "COLUMN_NUMBER ";

                # We have to make sure there is enough room before the current
                # "=" to print the entire COLUMN_NUMBER keyword.  If so, we're 
                # OK.  If not, we need to move ALL the "=" in the other 
                # keywords:

                if ($eq >= length($buff))
                  { $colnumline = sprintf("%-${eq}.${eq}s= $colnum",$buff);

                    &printline($cbuff[0]);
                    &printline($colnumline);
                    for ($i=1; $i<@cbuff; $i++)
                      { &printline($cbuff[$i]); }
                  }

                else
                  { # Find the new position for the "=" and the amount of blank
                    # padding needed, saving it as a string to be spliced into 
                    # the output lines via a substitution:

                    $sp = length($buff) - $eq;
                    $spacer = sprintf("%${sp}.${sp}s"," ");
                    $colnumline = "$buff= $colnum";
    
                    &printline($cbuff[0]);
                    &printline($colnumline);
                    for ($i=1; $i<(@cbuff-1); $i++)
                      { $cbuff[$i] =~ s/=/$spacer=/;
                        &printline($cbuff[$i]);
                      }
                  }

              }
            &printline($line);

            undef @cbuff;
            $cb = 0;
            $number_found = 0;
            $in_column = 0;
          }

        elsif ($line =~ /^\s*OBJECT\s*=\s*COLUMN/)
          { $in_column = 1; 
            $cbuff[$cb] = $line;
            $cb++;
            $colnum++;
          }

        elsif ($in_column)
          { $cbuff[$cb] = $line;
            $cb++;
          }

        elsif ($line =~ /^\s*END_OBJECT\s*=\s*TABLE/)
          { $not_done = 0; }

        else
          { &printline($line); }
      }
  }

# Finish transferring the rest of the label lines:

&printline($line);
while ($line=<INLBL>)
  { $line =~ s/\s*[\r\n]+$//;
    &printline($line);
  }

# Close up the label files and we're done with them for now:

close(INLBL);
close(OUTLBL);

#______________________________________________________________________________
#
# OK, now we're ready to set up the banner for the table itself.  We allow one
# level of overlapping columns.  If we have overlap, then we'll need 2 
# additional lines of header.  We label the overlapped columns at the higher
# level.
#
# NOTE that it is possible for the column number to exceed the width of the
# fields, although this is unlikely.  In these cases we write as much of the
# number as possible.  The trimming must be done outside of the print statement,
# however, to avoid messing up the byte count in the output header line.
# Open the datafile and read the first record so that we can determine the
# type of carriage control:

open (INDATA,$datafile) || die "Unable to open datafile $datafile";
open (TMP,">$newdatafile")  || die "Unable to open temp file $newdatafile";

$line = <INDATA>;
$CRLF = ($line =~ /\r$/)? 1 : 0;

# Carriage control will be added in printing, so we should not leave room for 
# it here:

$rb = ($CRLF)? $row_bytes - 2 : $row_bytes - 1;
for ($i=0; $i<$rb; $i++)
  { $barline[$i] = '='; }
$barline = join('',@barline);
$buffer[0] = sprintf "%${rb}.${rb}s", " ";
$buffer[1] = $buffer[0];
$buffer[2] = $buffer[0];
$buffer[3] = $buffer[0];

for ($i=1; $i<=$colnum; $i++)
  { $num = $i;
    if (length($num) > $bytes[$i])
      { $num = substr($i,0,$bytes[$i]); }

    if ($overlap[$i])
      { substr($buffer[2],$start_byte[$i],$bytes[$i]) = 
          sprintf("%-${bytes[$i]}d",$num); 
        substr($buffer[3],$start_byte[$i],$bytes[$i]) =
          sprintf ("%-${bytes[$i]}.${bytes[$i]}s",$barline);
      }
    else
      { substr($buffer[0],$start_byte[$i],$bytes[$i]) = 
          sprintf("%-${bytes[$i]}d",$num); 
        substr($buffer[1],$start_byte[$i],$bytes[$i]) =
          sprintf ("%-${bytes[$i]}.${bytes[$i]}s",$barline);
      }
  }

# Now we print the banner lines to the temporary data file, followed by the 
# first line of data, which we've already read in.  Then we just copy lines
# to the output file:

$cc = ($CRLF)? "\r\n" : "\n";

if ($overlap)
  { printf TMP "$buffer[2]$cc";
    printf TMP "$buffer[3]$cc";
  }
printf TMP "$buffer[0]$cc";
printf TMP "$buffer[1]$cc";
printf TMP "$line";

# Copy the remaining lines:

while ($line=<INDATA>)
  { printf TMP "$line"; }
close(INDATA);
close(TMP);

#______________________________________________________________________________
#
# OK, we've successfully written the label and data temporary files.  If we're
# overwriting the input files, we just rename the temporary versions:

if ($opt_o)
  { rename $outlbl, $inlbl;
    rename $newdatafile, $datafile;
  }

#______________________________________________________________________________
#______________________________________________________________________________
#
sub printline
  { my ($line) = $_[0];     # line to be printed 

    if ($optc)
      { printf OUTLBL "%-78.78s\r\n", $line; }
    else
      { printf OUTLBL "$line\n"; }

    return 1;
  }
