#!/usr/bin/perl

use Getopt::Std;

# addcolnum
#
#  This routine adds a COLUMN_NUMBER keyword to each COLUMN object found in the
#  input label file.  It searches for the "OBJECT = COLUMN" line, then adds the
#  new keywords as the next line.  In the new line the '=' is aligned with the
#  '=' in "OBJECT = COLUMN", and blank padding and record delimiters are the
#  same as in that previous record.
#
# 27 January 1998, acr.
#
#  Format:   % addcolnum [-f] [ input_label [ output_label ]]
#
#    where   -f   indicates that the output file should be forced to 80-byte
#                 fixed-length records with CRLF carriage-control
#
# 11 Dec 1998, acr: Upgraded to replace/update existing COLUMN_NUMBER keywords
#                   as well; added "-f" option.
# 12 Mar 1999, acr: Fixed dumb bug that was missing END_OBJECT lines without
#                   specific object type values; fixed adjustment of '=' in 
#                   last line in object definition.
# 23 May 2005, acr: Fixed bug that looped forever in labels without COLUMN
#                   objects.
# 26 Sep 2006, acr: Modified to use Getopt::Std
#
#==============================================================================
# Start by checking for file arguments:

getopts('f');

if (@ARGV > 2)
  { die "Usage: % addCOLNUM [ input_label [ output_label ] ].\n"; }

$output_file = (@ARGV>1)? $ARGV[1] : "-";
$input_file  = (@ARGV>0)? $ARGV[0] : "-";

# Open files:

open (IN,$input_file) || die "Unable to open $input_file for input.";
open (OUT,">$output_file") || die "Unable to open $output_file for output.";

# Read the first line of the file to determine carriage control and whether
# this file is already blank-padded to 80 bytes:

$line = <IN>;
$FIXED = (length($line) > 78  ||  $opt_f)?      1 : 0;
$CRLF  = ($line =~ /\r\n$/    ||  $opt_f)? "\r\n" : "\n";


# Now we loop through input lines until we find the beginning of column
# definitions:

while ($line !~ /^\s*OBJECT\s*=\s*COLUMN/)
  { $line =~ s/\s*[\r\n]+$//;
    &printline($line);
    $line = <IN>;

    # Check for unexpected EOF reading from <IN>:

    if (eof)
      { close(IN);
        close(OUT);
        die "No COLUMN objects found.\n"; 
      }
  }

# Now we'll loop through the column definitions.  Each definition is read in
# in its entirety so we can check for existing COLUMN_NUMBER lines before
# printing it out:

$line =~ s/\s*[\r\n]+$//;
undef $cbuff;
$cbuff[0] = $line;
$cb       = 1;

$colnum       = 1;
$number_found = 0;
$in_column    = 1;
$not_done     = 1;

while ($not_done)
  { $line = <IN>;
    $line =~ s/\s*[\r\n]+$//;

    if ($line =~ /COLUMN_NUMBER/)
      { $line =~ s/=\s*[0-9]+/= $colnum/;
        $number_found = 1;
        $colnumline = $line;
      }

    elsif ($line =~ /END_OBJECT\s*={0,1}/  &&  $in_column)
      { if ($number_found)
          { &printline($cbuff[0]);
            &printline($colnumline);
            for ($i=1; $i<@cbuff; $i++)
              { &printline($cbuff[$i]); }
          }
        else
          { $eq = index($cbuff[1],"=");
            $buff = $cbuff[1];
            $buff =~ s/^(\s*).*$/$1/;
            $buff .= "COLUMN_NUMBER ";

            # We have to make sure there is enough room before the current
            # "=" to print the entire COLUMN_NUMBER keyword.  If so, we're 
            # OK.  If not, we need to move ALL the "=" in the other keywords:

            if ($eq >= length($buff))
              { $colnumline = sprintf("%-${eq}.${eq}s= $colnum",$buff);

                &printline($cbuff[0]);
                &printline($colnumline);
                for ($i=1; $i<@cbuff; $i++)
                  { &printline($cbuff[$i]); }
              }

            else
              { # Find the new position for the "=" and the amount of blank
                # padding needed, saving it as a string to be spliced into 
                # the output lines via a substitution:

                $sp = length($buff) - $eq;
                $spacer = sprintf("%${sp}.${sp}s"," ");
                $colnumline = "$buff= $colnum";

                &printline($cbuff[0]);
                &printline($colnumline);
                for ($i=1; $i<(@cbuff); $i++)
                  { $cbuff[$i] =~ s/=/$spacer=/;
                    &printline($cbuff[$i]);
                  }
              }
          }
        &printline($line);

        undef @cbuff;
        $cb = 0;
        $number_found = 0;
        $in_column = 0;
        $colnum++;
      }

    elsif ($line =~ /^\s*OBJECT\s*=\s*COLUMN/)
      { $in_column = 1; 
        $cbuff[$cb] = $line;
        $cb++;
      }

    elsif ($in_column)
      { $cbuff[$cb] = $line;
        $cb++;
      }

    elsif ($line =~ /^\s*END_OBJECT\s*={0,1}/)
      { $not_done = 0; }

    else
      { &printline($line); }

  }

# Finish transferring the rest of the label lines:

&printline($line);
while ($line=<IN>)
  { $line =~ s/\s*[\r\n]+$//;
    &printline($line); 
  }



# Done:

close(IN);
close(OUT);

#______________________________________________________________________________
#______________________________________________________________________________
#
sub printline
  { my ($line) = $_[0];     # line to be printed 

    if ($FIXED)
      { printf OUT "%-78.78s$CRLF", $line; }
    else
      { printf OUT "$line$CRLF"; }

    return 1;
  }
