#!/usr/bin/perl

use Getopt::Std;
use POSIX;

# Routine to change the DATA_SET_ID values in a directory tree.
#
# Format: % change_dsid [-a] [-n] [-r root_dir] [old_DSID] new_DSID
#
#  where -a           means check all files for attached labels
#        -n           suppresses recursion (default is to recurse through
#                      subdirectories)
#        -r root_dir  gives the root directory in which to work (default
#                      is ".")
#        old_DSID     is the specific DSID to be changed (default is to
#                      change all DATA_SET_ID keyword values to new_DSID).
#                      If this value is in mixed or lowercase, it is forced
#                      to uppercase.
#        new_DSID     The new DATA_SET_ID value.  If this values is in mixed
#                      or lowercase, it is forced to uppercase.
#
# In its simplest mode, change_dsid searches all label files (ending in 
# ".lbl" or ".LBL") and catalog files (ending in ".cat" or ".cat") and
# looks for "DATA_SET_ID =" lines.  It then replaces the existing value 
# with the value passed in by the user.  It overwrites the existing file 
# with the new file, preserving line lengths, and recurses down through all
# subdirectories looking for more label or catalog files.
#
# When the "-a" option is specified, change_dsid checks every file it 
# encounters to see if it contains an attached PDS label (the first line
# in the file will start with "PDS_VERSION_ID").  If so, it looks for the
# DATA_SET_ID and replaces it with the new value, padding with or removing
# trailing blanks as needed to preserve the original byte count and spacing.
#
# Alternately, the user can give a specific DATA_SET_ID value that is to be
# changed.  If the value is found, a new label overwrites the old one.  
# Otherwise, no changes are made.
#
# NOTES
#
#  o All warnings and errors are printed to STDOUT.
#
#  o change_dsid cannot handle multi-valued DATA_SET_ID keywords.  If a 
#    DATA_SET_ID keyword has a value that begins with either "{" or "(",
#    a warning will be printed and the file will not be changed.
#
#  o change_dsid will not look in text descriptions for DATA_SET_ID values,
#    even if an old_DSID is given on the command line.  Only the values of
#    the DATA_SET_ID keyword are checked and changed as needed.
#
#  o change_dsid will attempt to preserve the original line length when 
#    replacing values.  If this is not possible and the label is detached,
#    the label will be changed but a warning message will be printed.  If
#    it is not possible and the label is attached, change_dsid will print 
#    a failure message and the original file will not be changed.
#
#  o This program only works on directories, not files.
#
# MODIFICATION HISTORY
# ====================
#
# 2008-11-24, A.C.Raugh   Fixed a couple of idiot mistakes that the
#                         anonymous author never noticed were preventing
#                         recursion from working.  Test your freakin'
#                         code, people!
#
#=============================================================================

# grabing options and argument(s) #######################################

%options=();
getopts("anr", \%options);
$len = @ARGV; 

# finding the root directory, old dsid (if any) and new dsid ############

$new_dsid = $ARGV[$len-1];
$old_dsid = "";
$root_dir = ".";

# if there are more than one arguement after the option flags ###########

if ($len > 1) {
    if (defined $options{r}) {
	$root_dir = $ARGV[0];
	if ($len > 2) {
	    $old_dsid = $ARGV[$len-2];
	}
    }
    else {
	$old_dsid = $ARGV[$len-2];
    }
}

# NOTE 1: Do not proceed if the user uses the -r option, but doesn't give a 
# directory and gives us an old dsid to change.
# NOTE 2: If the new DATA_SET_ID has '{' or '(' included, then stop

if ($new_dsid !~ /\{|\(/) {
    dodir ($root_dir, defined $options{n}, defined $options{a});
}
else {
    # there are multiple values in the new dataset id
    print "change_dsid cannot handle multi-valued DATA_SET_ID keywords\n";
}

sub dodir
{ 
    local ($dir, $nore, $all) = @_;   #directory name and recursion flag
    my (@dlist, $dl); 
    my (@flist, $fl); 
    my ($file, $subdir);

    opendir (DIR, $dir) || die "Could not open directory '$dir'\n";
    $dl = 0;    $fl = 0;
#    if ($dir eq ".") {	$dir = "";    }

    foreach $file (sort readdir (DIR)) {
	if ($file =~ /^\./) {}
	#storing the sub-directories in the current directory
        elsif (-d "$dir/$file") {
            $dlist[$dl] = "$dir/$file";
            $dl++;
        } 

	#checking the files within the current directory
        else { 
	    # if -a flag selected, then check all files
	    if ($all) {		#print "$file\n";
		$flist[$fl] = "$file";
		$fl++;
	    }

	    # if -a flag is not selected, then check only .cat and .lbl files
	    else {
		if (check_ext ($file) eq "yes") {
		    $flist[$fl] = "$file";
		    $fl++; 
		} 
	    }
	}
    }

    closedir (DIR); 

    foreach $file (@flist) {
        &dofile ($file, $dir, $all);
    } 
    if ($nore eq "" && $dl > 0) {
        foreach $file (@dlist) {
            &dodir($file, $nore, $all);
        }
    }
}

sub check_ext
{
    local ($filename) = @_;
    
    if ($filename =~ /\.lbl$/ || $filename =~ /\.LBL$/ || 
	$filename =~ /\.cat$/ || $filename =~ /\.CAT$/) {
	return "yes";
    }
    else {
	return "";
    }
}

sub dofile
{
    #variables ############################################################

    local ($file, $dir, $all) = @_;
    my ($line, $length, $new, $failed);
    my (@dsid);

    #opening the appropriate files  ######################################

    open (LABEL, "$dir/$file") || die "Cannot open $dir/$file \n";
    open (FILE, ">st_$file") || die "Cannot open st_$file \n";

    #making the blasted changes    #######################################

    $new = "";    $failed = "";
    if ($all) {
	
	#print "got the -a flag\n";

	$line = <LABEL>;   chomp $line;
	if ($line !~ /PDS_VERSION_ID/)
	{  $failed = "yes";  }
	else 
	{  print FILE "$line\n";   }
    }

    #checking if the file has a PDS label
    while ($line = <LABEL>) {

	chomp $line;	#print "$line\n";
	
	#if the line has the DATA_SET_ID tag
	if ($line =~ /DATA_SET_ID/) {
	    
	    # handling the case where there is multi-valued DATA_SET_ID.
	    if ($line =~ /\{|\(/ ) {
		print "$dir/$file: change_dsid cannot handle multi-valued ";
		print "DATA_SET_ID keywords\n";
		$failed = "yes";
		last;
	    }
	    
	    $length = length($line);
	    @dsid = split/=/,$line;
	    
	    # if an old dsid is provided, then compare it with the current
	    # value of the DSID. change only if the value is the same.
	    if ($old_dsid) {

		#print "$dsid[1] =~ $old_dsid\n";
		#print "matches\n" if $dsid[1] =~ m/$old_dsid/i;

		if ($dsid[1] =~ m/$old_dsid/i)
		{  $new = "$dsid[0]"."="."\"$new_dsid\"";  }
		else 
		{  $new = $line;  }
	    }
	    
	    # if no specific dsid is provided, change all DATA_SET_ID
	    # values to the new DSID provided
	    else {
		$new = "$dsid[0]"."="."\"$new_dsid\"";
	    }
	    
	    # padding white space to the end of the DSID line, if needed.
	    while ($length > length($new)) {
		$new = $new." ";
	    }
	    
	    # It is possible the line length of the new DATA_SET_ID line  
	    # is longer than the original length. If it is the label is   
	    # detached, then proceed. If it is an attatched label, stop. 
	    if ($length < length($new)) {
		if (check_ext($file)) {
		    print "WARNING: new DATA_SET_ID keyword longer than ";
		    print "original line length.\n";
		    print FILE "$new\n"; 
		}
		else {
		    $failed = "yes";
		    print "STOP: cannot write to attached label.\n";
		}
	    }
	    else {
		print FILE "$new\n"; }
	}
	else {
	    print FILE "$line\n";
	}

    }

    # mv the temporary file to the original file if there had been no 
    # failures. If there were any, remove the temporary file.
    if ($failed) {
	close (FILE);
	unlink("st_$file") || die "Can't remove temorary file st_$file\n";
    }
    else {
	system ("mv st_$file $dir/$file");
    }
}
