#!/usr/bin/perl

use lib "/n/sbnops01/lcltools/lib/perl5";
use Getopt::Std;

require "nextPDSODL.pl";  # Also contains 'getval'

# checkids
#
# Routine to read through all PDS labels and catalog files in the given
# directory tree and make sure there are object files available for all
# essential ID fields contained in those files.
#
# Format: % checkids [-a] [-o <output>] [-c <cat_dir_list>] [-v]
#                     [<directory_root>]
#
#  where:   -a                check data files for attached labels
#           -c <cat_dir_list> additional catalog directories, relative to
#                             the current working directory, separated by
#                             commas
#           -o <output>       names an output file (stdout by default)
#           -v                verbose mode - sends progress messages to stderr
#           <directory_root>  is the directory tree to search ('.' by default)
#
# checkids assumes that all catalog files are located in the subdirectory
# 'catalog' (and/or 'CATALOG') under the given root directory (there may be
# subdirectories under the main catalog directory), and also requires that
# all catalog files have the suffix '.cat' or '.CAT'.  Additional catalog 
# directories may be specified using the '-c' option. 
#
# After collecting the information from the catalog directories, checkids
# reads through all directories under the given root directory looking for 
# ODL files and collecting lists of _ID and _NAME (i.e., catalog key) values
# encountered.  These are compared against the defined IDs and NAMEs from the
# catalog files, and a list of undefined keys is generated.  
#
# Label files are identified by having the '.lbl' or '.LBL' suffix.  The user
# may specify that all files should be checked for attached labels by using 
# the "-a" option.
#
# Once all label files have been scanned, checkids generates a second list 
# of archive objects that were never referenced in any ODL file.
#
# The objects included in this search are:
#
#    DATA_SET
#    DATA_SET_COLLECTION
#    MISSION
#    INSTRUMENT
#    INSTRUMENT_HOST
#    TARGET
#    REFERENCE
#
# NOTE: This routine assumed properly formatted ODL in all catalog and label
#       files!
#
#  25 Jan 1999, acr.
#  26 Sep 2006, acr: Modified to use Getopt::Std and "use lib"
#
#==============================================================================
#
# Associative arrays are used to collect the information wanted.  This is 
# complicated by the fact that sometimes NAME fields are used or required
# where ID fields would be expected.  All fields are forced to upper case 
# for purposes of matching.
#
# Global associative arrays:
#
#   %refcount{<refid>:<id>} = number of references to this object
#   %deffile{<refid>:<id>}  = source file for the object definition
#   %aka{<altid>:<name>}    = "<refid>:<id>" (i.e., name->ID translator)
#
#  These arrays are keyed on the type of the key ids (TARGET_NAME, DATA_SET_ID, 
#  etc.) and the value found in the catalog file. The "aka" array is used to
#  decode references to alternate IDs (INSTRUMENT_NAME, for example) which 
#  might be used as references.  It returns the definitive ID tag for use with
#  %refcount.
#
#  These arrays are maintained globally.  The division into subroutines is to
#  simplify and clarify the logic of the procedure.  Explicit parameters are
#  passed only when necessary.
#
#  The recognized archive keys are:

@KEYS = ( "DATA_SET_ID",
          "DATA_SET_COLLECTION_ID",
          "INSTRUMENT_ID",
          "INSTRUMENT_HOST_ID",
          "MISSION_NAME",
          "TARGET_NAME",
          "REFERENCE_KEY_ID"
        );
$KEYLIST = ":" . join(":",@KEYS) . ":";

%ALTKEYSMAP = ( "INSTRUMENT_NAME",      "INSTRUMENT_ID",
                "INSTRUMENT_HOST_NAME", "INSTRUMENT_HOST_NAME"
              );
$ALTLIST = ":" . join(":",(keys %ALTKEYS)) . ":";

#
#  This array maps the archive keys to their defining context(s):

%DEF   = ( "DATA_SET_ID",            ":DATA_SET:",
           "DATA_SET_COLLECTION_ID", ":DATA_SET_COLLECTION:",
           "INSTRUMENT_ID",          ":INSTRUMENT:",
           "INSTRUMENT_NAME",        ":INSTRUMENT:",
           "INSTRUMENT_HOST_ID",     ":INSTRUMENT_HOST:",
           "INSTRUMENT_HOST_NAME",   ":INSTRUMENT_HOST:",
           "MISSION_NAME",           ":MISSION:",
           "TARGET_NAME",            ":TARGET:",
           "REFERENCE_KEY_ID",       ":REFERENCE:"
          );

%DEFID = ( "DATA_SET",            "DATA_SET_ID",           
           "DATA_SET_COLLECTION", "DATA_SET_COLLECTION_ID",
           "INSTRUMENT",          "INSTRUMENT_ID",         
           "INSTRUMENT_HOST",     "INSTRUMENT_HOST_ID",
           "MISSION",             "MISSION_NAME",          
           "TARGET",              "TARGET_NAME",           
           "REFERENCE",           "REFERENCE_KEY_ID",      
          );

#
#==============================================================================
#
# Begin by getting the command line options and opening the root directory:

getopts('o:ac:v');

if (@ARGV > 1)
  { die "Usage: checkids [-o <output>] [-a] [-c <catdir>] [-v] [<directory>]\n";
  }

if (@ARGV == 1)
  { $rootdir = $ARGV[0];
    if (!(-d $rootdir))
      { die "checkids: \"$ARGV[0]\" is not a directory.\n"; }
  }
else
  { $rootdir = "."; }

# Open the output file:

if ($opt_o)
  { open (OUT, "> $opt_o") || die "Could not open output file \"$opt_o\"."; }
else
  { open (OUT, ">-"); }

# If auxilliary catalog directories were supplied, add them to the list:

@catdirlist = ("$rootdir/catalog","$rootdir/CATALOG");
if ($opt_c)
  { push @catdirlist, split(/,/,$opt_c); }

#---------------------------------------------
# We begin by dealing with the catalog files, looping through the directory
# list:

$catcount = 0;
foreach $dir (@catdirlist)
  { if (!(-d $dir))
      { if ($dir ne "$rootdir/catalog"  &&  $dir ne "$rootdir/CATALOG") 
          { printf STDERR "'$dir' is not a directory. Ignored. \n"; }
        next;
      }
    $catcount++;
    &runcatdir($dir);
  }

if ($catcount == 0)
  { die "checkids: No catalog directories found.  Stopping"; }

#---------------------------------------------
# Now we call the subroutine to collect information from all the files in the
# directory tree.  Note that the command line option "-a" value is assumed to 
# be in the global variable "$opt_a" in all subroutines.

# ...banner printing...

$barline  = "====================================================";
printf OUT "References to undefined objects:\n\n";
printf OUT "%-22.22s  %-25.25s  File\n","Name/ID","Value";
printf OUT "%-22.22s  %-25.25s  %-25.25s\n",$barline,$barline,$barline;

$ERRCOUNT = 0;        # Global error counter;
&rundir($rootdir);


## DEBUG PRINTING::::
##
##open (DMP,">refcount");
##foreach $id (sort keys(%refcount))
##  { printf DMP "'%-50.50s' '%s'\n",$id,$refcount{$id};}
##close(DMP);
##open (DMP,">deffile");
##foreach $id (sort keys(%deffile))
##  { printf DMP "'%-50.50s' '%s'\n",$id,$deffile{$id};}
##close(DMP);
##open (DMP,">aka");
##foreach $id (sort keys(%aka))
##  { printf DMP "'%-50.50s' '%s'\n",$id,$aka{$id};}
##close(DMP);
##
##


if ($ERRCOUNT == 0)
  { printf OUT "  [None]\n\n"; }
elsif ($ERRCOUNT == 1)
  { printf OUT "\n  ...1 unresolved reference.\n\n"; }
else
  { printf OUT "\n  ...$ERRCOUNT unresolved references.\n\n"; }

# Now we check the %objref array to make sure that all catalog objects were
# referenced at least once.  We display a list of those that weren't:

$ERRCOUNT = 0;
printf OUT "\nUnreferenced catalog objects:\n\n";
printf OUT "%-22.22s  %-25.25s  Catalog File\n","Object Type","Name/ID";
printf OUT "%-22.22s  %-25.25s  %-25.25s\n",$barline,$barline,$barline;

foreach $id (sort (keys %deffile))
  { if (!$refcount{$id})
      { ($obj,$name) = split(/:/,$id);
        $obj =~ s/_ID$//;
        $obj =~ s/_NAME//;
        printf OUT "%-22.22s  %-25.25s  %s\n",
                    $obj, $name, $deffile{$id};
        $ERRCOUNT++;
      }
  }

if ($ERRCOUNT == 0)
  { printf OUT "  [None]\n\n"; }
elsif ($ERRCOUNT == 1)
  { printf OUT "\n  ...1 unreferenced object.\n\n"; }
else
  { printf OUT "\n  ...$ERRCOUNT unreferenced objects.\n\n"; }

# Close the output file, and we're done:

close(OUT);

#______________________________________________________________________________
#______________________________________________________________________________
#______________________________________________________________________________
#

sub rundir

# Subroutine to loop through all files in the input directory looking for ODL
# files to process and subdirectories to check.

{ local ($dir=$_[0]);    # Input (unopened) directory
  
  my (@odl);             # list of ODL files
  my (@subdir);          # list of subdirectories
  my ($of,$sd);          # subscripts
  my ($file,$i,$line);
  my ($filename);        # full file name

  # Open the directory passed in:

  opendir (DIR,$dir) || die "dsids: Could not open \"$dir\".";

  # If we're running in verbose mode, print the status message:

  if ($opt_v)
    { printf STDERR "\nDirectory: $dir\n";
      printf STDERR   "----------\n";
    }

  # We'll now loop through the list of files in this directory, looking for
  # those of interest.  Only if $opt_a is defined do we check each file for
  # an ODL header.

  $of = $sd = 0;
  foreach $file (sort readdir(DIR))
    { next if ($file =~ /^\./);           # Ignore hidden files

      # Add all files with .lbl and .cat extensions to the ODL list:

      $filename = "$dir/$file";
      if ($filename =~ /\.lbl$/  ||  $filename =~ /\.LBL$/  ||
          $filename =~ /\.cat$/  ||  $filename =~ /\.CAT$/)
        { $odl[$of] = $filename;
          $of++;
        }

      # Add all directories to the subdirectory list:

      elsif (-d $filename)
        { $subdir[$sd] = $filename;
          $sd++;
        }

      # If the "-a" option was given, check remaining files for ODL headers,
      # indicated by the presence of PDS_VERSION_ID as the first real line
      # in the file:

      elsif ($opt_a)
        { open(ODL,"$filename");
          $line = &nextPDSODL(*ODL);
          if ($line =~ /PDS_VERSION_ID/)
            { $odl[$of] = $filename;
              $of++;
            }
          close(ODL);
        }
    }
  closedir(DIR);

  # Now call the subroutine to process each of the ODL files:

  foreach $file (@odl)
    { &process_odl($file); }

  # Add call this routine recursively on each subdirectory encountered:

  foreach $file (@subdir)
    { &rundir($file); }

  # And we're done:

}
#______________________________________________________________________________
#

sub process_odl

# Subroutine to read through an ODL file, determine its type, look for specific
# identification keywords and add the values found to the global associative
# arrays.  We assume properly formatted ODL.
#
# NOTE: We also assume that all archive keys of interest end with either
#       "_ID" or "_NAME"!

{ local ($odlname= $_[0]);   # Name of the file to be examined

  my ($line);                # Input line buffer
  my ($key,$val,$id,$kv);    # Keyword, value and index strings
  my ($notdone);             # loop flag
  my ($objtype);
  my ($v);

  # Open the input file.  This should always work:

  open (ODL,$odlname) || die "checkids: Unable to open ODL file ($odlname)";

  # Once the file is opened, we can remove the "./" string that might appear
  # at the beginning of the file name.  This is for readability on output:

  $odlname =~ s/^\.\///;

  # In verbose mode, print the status message:

  printf STDERR "   => Reading $odlname..." if ($opt_v);

  # Loop through input lines until the "END" statement (or EOF) is found:

  undef $objtype;
  $notdone = 1;
  while ($notdone)
    { $line = &nextPDSODL(*ODL);

      # Check for the "END" statement, and also for lines with no value
      # parts, which are of no interest to us:

      if (!$line)
        { printf STDERR "No END statement found in '$odlname'.\n";
          $notdone = 0;
          next;
        }
      elsif ($line eq "END")
        { $notdone = 0;
          next;
        }
      elsif ($line !~ /=/)
        { next; }

      # Separate the key and value parts:

      $key = $line;
      $key =~ s/=.*$//;
      $val = &getval($line);

      # If the keyword found isn't an _ID, _NAME or OBJECT, we don't care:

      next if ($key !~ /_ID$/ && $key !~ /_NAME$/ && $key !~ /OBJECT$/);

      # Force value to uppercase and remove quotes:

      $val =~ tr/a-z/A-Z/;
      $val =~ s/"//g;
      $val =~ s/^\{//;
      $val =~ s/\}$//;
      $id = "$key:$val";

      # If we're left with a value of "N/A" or "UNK", skip it:

      next if ($val eq "N/A"  ||  $val eq "UNK");

      # We'll need the object type for later reference, so if we haven't got it,
      # collect it when it first shows up:

      if ($key eq "OBJECT" && !$objtype)
        { $objtype = $val;
          next;
        }

      # If we encounter an END_OBJECT matching the $objtype, we reset for
      # the next object:

      if ($key eq "END_OBJECT"  &&  $val eq $objtype)
        { undef $objtype;
          next;
        }

      # We know we have a name or ID field, but it may not be a key of 
      # interest, so we look for it in the $KEYLIST string:

      next if ($KEYLIST !~ /:$key:/);

      # Now we know we have an archive key, but we need to double-check to
      # make sure we're not in a defining context (INSTRUMENT_NAME found in 
      # an INSTRUMENT object, e.g.).  We use the %DEF array for this, keying
      # on the keyword name and checking against the file type:

      next if ($objtype  &&  $DEF{$key} =~ /:$objtype:/);

      # At this point we've recognized an archive key in a referencing context.
      # We now parse the value string (there may be a set of values) and check
      # each individually:

      foreach $v (split(/,/,$val))
        { $kv = "$key:$v";
          $id = ($aka{$kv})? $aka{$kv} : $kv;

          # We checked for a defined source file.  If there is none, we 
          # signal an error:

          if ($deffile{$id})
            { $refcount{$id}++; }
          else
            { printf OUT "%-22.22s  %-25.25s  %s\n", $key, $v, $odlname;
              $ERRCOUNT++;
            }
        }
    }


  # We're done with the ODL, so we close the file and return:

  close(ODL);
  printf STDERR "done\n" if ($opt_v);
}
#______________________________________________________________________________
#

sub runcatdir

# Subroutine to look for catalog files in the given catalog directory and all
# its subdirectories. Catalog files must end with ".cat" or ".CAT".

{ local ($dir) = $_[0];       # Directory to search
  my    (@dirlist,@filelist);
  my    ($dl, $fl);
  my    ($file,$tmpfile);

  # Open the catalog directory.  If we're running in verbose mode, print
  # a status message to STDERR:

  opendir (DIR, $dir) || die "Unable to open directory '$dir'.";
  if ($opt_v)
    { $tmpfile = $dir;
      $tmpfile =~ s/^\.\///;
      printf STDERR "\nCatalog Directory: $tmpfile\n";
      printf STDERR   "------------------\n";
    }
  
  # Loop through files in this directory:

  foreach $file (sort readdir(DIR))
    { next if ($file =~ /^\./);         # Skip hidden files

      $tmpfile = "$dir/$file";
      if ($file =~ /\.cat$/  ||  $file =~ /\.CAT$/)
        { $filelist[$fl] = $tmpfile;
          $fl++;
        }

      elsif (-d $tmpfile)
        { $dirlist[$dl] = $tmpfile;
          $dl++
        }
      }
    closedir(DIR);

  # Now read each catalog file found:

  foreach $file (sort @filelist)
    { &runcatfile($file); }

  # And recurse through any subdirectories:

  foreach $file (sort @dirlist)
    { &runcatdir($file); }

  # And we're done:
}

#______________________________________________________________________________
#

sub runcatfile

# Subroutine to read through a catalog file and extract just key field 
# definitions for initializing the reference tracking arrays:

{ local ($catfile) = $_[0];
  my    ($line,$key,$val,$id,$primid);
  my    ($notdone, $objtype);

  # Open the file.  This should always work:

  open (CAT,$catfile) || 
    die "checkids: Unable to open catalog file ($catfile)";

  # Do a little minor cleanup on the filename for output purposes:

  $catfile =~ s/^\.\///;

  # If we're in verbose mode, print a message:

  printf STDERR "   => Reading $catfile..." if ($opt_v);

  # Loop through input lines until the END statement is found:

  undef $objtype;
  $notdone = 1;
  while ($notdone)
    { $line = nextPDSODL(*CAT);

      # Check for terminating conditions:

      if (!$line)   # EOF without "END" statment
        { printf STDERR "No END statement found in '$catfile'.\n";
          $notdone = 0;
          next;
        }
      elsif ($line eq "END")
        { $notdone = 0;
          next;
        }
      elsif ($line !~ /=/)
        { next; }

      # Separate key and value parts:

      $key = $line;
      $key =~ s/=.*$//;
      $val = &getval($line);

      # If the keyword found isn't an _ID, _NAME or OBJECT, we don't care:

      next if ($key !~ /_ID$/ && $key !~ /_NAME/ && $key !~ /OBJECT$/);

      # Force value to uppercase and remove quotes:

      $val =~ tr/a-z/A-Z/;
      $val =~ s/"//g;
      $val =~ s/^\{//;
      $val =~ s/\}$//;
      $id = "$key:$val";

      # If we're left with a value of "N/A" or "UNK", skip it:

      next if ($val eq "N/A"  ||  $val eq "UNK");

      # The first OBJECT encountered determines the object type:

      if ($key eq "OBJECT"  &&  !$objtype)
        { $objtype = $val;
          next;
        }

      # If we encounter an END_OBJECT matching the $objtype, we reset for
      # the next object:

      if ($key eq "END_OBJECT"  &&  $val eq $objtype)
        { undef $objtype;
          undef $primid;
          next;
        }

      # We check to see if we've encountered a keyword of interest in a
      # defining context. If so, we set the %deffile array value accordingly:

      if ($KEYLIST =~ /:$key:/  &&  $DEF{$key} =~ /:$objtype:/)
        { $deffile{$id} = $catfile; 
          $primid = $id;
        }

      # On the other hand, if we've got an alternate ID, we set the %aka array.
      # Note that we are taking advantage of the implicit requirement in the 
      # object structures that the secondary reference key ALWAYS occur after
      # the primary IDs in the catalog objects.

      elsif ($ALTLIST =~ /:$key:/  &&  $DEF{$key} =~ /:$objtype:/)
        { $aka{$id} = $primid; }

      # Next keyword.

    }

  # Close the file and we're done:

  close(CAT);
  printf STDERR "done\n" if ($opt_v);
}


