#!/usr/bin/perl

# chkisonames
#
# Routine to check a directory tree to make sure all directory and file
# names are ISO 9660/Level 2-compliant.  Specifically, all names must
# be 27.3 format, with only letters, numbers and the underscore in them.
# Directory names may be 31 characters long, with the same character set
# (no '.').
#
# Format: chkisonames [-u | -l] [-s] [ root_dir ]
#
#  where  -u        make sure all filenames are upper case
#         -l        make sure all filenames are lower case
#         -s        shutup about things that are PDS-specific, like
#                   non-null file names; extensions of more than 
#                   3 characters; and names longer than 27 characters
#         root_dir  is the root of the tree to be checked
#
# Default is to allow upper or lower case but to flag mixed-case names
# as a problem.  The -u and -l options are mutually exclusive.
#
# 29 Jan 2007, A.C.Raugh
# 16 Feb 2007, ACR: Added code to check: total path length does not exceed
#                   255 character; directory hierarchy does not exceed 8.
#                   Tightened up character case checks on file names.
#
#==========================================================================

use Getopt::Std;

$TRUE  = 1;
$FALSE = 0;

getopts('uls');

if ($opt_u and $opt_l)
  { die "Usage: chkisonames [-u | -l] { root_dir ]\n"; }

$root = (@ARGV == 0)? "." : $ARGV[0];
if ($root eq ".")
  { @junk = split(/\//,$ENV{"PWD"});
    $PWD_NAME = pop(@junk);
  }
else
  { $PWD_NAME = $root; }

$PDS_rules = ($opt_s)? $FALSE : $TRUE;

# If the input directory isn't actually a directory, check the name for
# ISO compliance and end:

if (! -d $root)
  { check_iso_file($root); 
    exit;
  }

# Otherwise, we'll start the directory loop:

do_dir($root);

# And we're done.

exit;

#----------------------------------------------------------------------------

sub do_dir

#  Loop through files and subdirectories of the directory passed in.

{ local ($indir=$_[0]);      # Input directory
  local (@flist,$fl,@dlist,$dl);
  local ($file,$i);
  local ($pds, $iso);
  local ($dir,$tfile,$tdir);
  local (@temp);

  # Open the directory for reading:

  opendir(DIR,$indir) || die "Could not open directory $indir for reading ($!)";

  # Check to see if this directory name is valid:

  $pds = $TRUE;
  $iso = $TRUE;

  $dir = $indir;

  # Start by making sure there aren't more than 8 directory levels:

  @temp = split(/\//,$dir);
  if (@temp > 8)
    { printf "%60.60s","Directory levels exceeds eight:";
      printf " '$indir'\n";
    }

  # Get the last element of the path:

  if ($dir =~ /\//)
    { $dir =~ /.*\/([^\/]+)$/;
      $file = $1;
    }
  else
    { $file = $dir; }

  # If we're in the special case of the "." directory, we need to check the
  # actual name, so we'll use a temporary file name holder for the directory
  # name check:

  if ($file eq ".")
    { $tfile = $PWD_NAME;
      $tdir  = $PWD_NAME;
    }
  else
    { $tfile = $file; 
      $tdir  = $dir;
    }

  # Now for the checks on the actual directory name:

  if ($opt_u)       # Require uppercase
    { if ($tfile =~ /[a-z]/)
        { printf "%60.60s","Lowercase characters in directory identifier:";
          printf " '$tdir'\n"; 
        }
    }
  elsif ($opt_l)     # Require lowercase
    { if ($tfile =~ /[A-Z]/)
        { printf "%60.60s","Uppercase characters in directory identifier:";
          printf " '$tdir'\n"; 
        }
    }
  else
    { if ($tfile =~ /[a-z]/  &&  $file =~ /[A-Z]/)
        { printf "%60.60s","Mixed case in directory identifier:";
          printf " '$tdir'\n";
        }
      if ($tfile !~ /^[A-Za-z0-9_]+$/)
        { printf "%60.60s","Invalid characters in directory identifier:";
          printf " '$tdir'\n";
        }
    }

  # Loop through contents:

  $fl = 0;
  $dl = 0;

  foreach $file (sort readdir(DIR))
    { next if ($file eq "."  ||  $file eq "..");

      if (-d "$dir/$file")
        { $dlist[$dl] = $file;
          $dl++;
        }

      else
        { $flist[$fl] = $file;
          $fl++;
        }
    }
  closedir(DIR);

  # First, we recurse through subdirectories:

  for ($i=0; $i<$dl; $i++)
    { $subdir = "$dir/$dlist[$i]";
      do_dir($subdir);
    }

  # Then we check the file names in this directory:

  for ($i=0; $i<$fl; $i++)
    { check_iso_file($dir,$flist[$i]); }

  # Done.

  return;
}

#----------------------------------------------------------------------------

sub check_iso_file

# Routine to see if a particular file name meets ISO (and PDS) requirements.

{ local ($dir = $_[0]);
  local ($file = $_[1]);

  my ($name,$ext);
  my (@parts);
  my ($nlen, $xlen);
  my ($fullname);
  my (@parts);


  # Check on total length constraint (255 characters for the fully qualified
  # file name).  If the first directory level is the current working directory
  # (i.e., "."), we replace it with its name (in $PWD_NAME):

  $fullname = "$dir/$file";
  $fullname =~ s/^\./$PWD_NAME/;

  if (length($fullname) > 255)
    { printf "%60.60s","Total identifier length exceeds 255 characters:";
      printf " '$fullname'\n";
    }

  # Character set constraints:

  if ($opt_u  &&  $file =~ /[a-z]/)     # Require uppercase
    { printf "%60.60s","Lowercase characters in file identifier:";
      printf " '$dir/$file'\n";
    }
  elsif ($opt_l  &&  $file =~ /[A-Z]/)  # Require lowercase
    { printf "%60.60s", "Uppercase characters in file identifier:";
      printf " '$dir/$file'\n"; 
    }
  elsif ($file !~ /^[A-Za-z0-9_.]+$/)
    { printf "%60.60s", "Invalid characters in file identifier:";
      printf " '$dir/$file'\n"; 
    }

  # Check for mixed case as well:

  if ($file =~ /[a-z]/  &&  $file =~ /[A-Z]/)
    { printf "%60.60s","Mixed case in file identifier:";
      printf " '$dir/$file'\n"; 
    }

  # Now check for a separator.  If there is none and we're not 
  # ignoring PDS rules, that's a potential problem:

  if ($file !~ /\./  &&  $PDS_rules)
    { printf "%60.60s","Potential PDS problem - file name without extension:";
      printf " '$dir/$file'\n";
    }

  # Split the identifier into name and extension:

  @parts = split(/\./,$file);
  if (@parts > 2)
    { printf "Invalid file identifier - too many '.': '$dir/$file'\n"; }

  if (@parts == 1  &&  $file =~ /\./  &&  $PDS_rules)
    { printf "%60.60s",
             "Potential PDS problem - separator not followed by extension:";
      printf " '$dir/$file'\n";
    }
  elsif (@parts == 1)
    { $name = $parts[0];
      undef $ext;
    }
  else
    { $name = $parts[0];
      $ext  = $parts[1];
    }

  # Checking length constraints, starting with the looser ISO lengths:

  $nlen = length($name);
  $xlen = ($ext)? length($ext) : 0;

  if (($nlen + $xlen) > 30)
    { printf "%60.60s","ISO 9660 violation - total identifier length > 30:";
      printf " '$dir/$file'\n";
    }

  if ($nlen == 0  &&  $xlen == 0)
    { printf "%60.60s","Invalid null file identifier: ";
      printf "'$dir/$file'\n";
    }

  # Now the PDS constraints, if we're checking:

  if ($PDS_rules)
    { if ($nlen > 27)
        { printf "%60.60s","Invalid PDS file name  - length > 27:";
          printf " '$dir/$file'\n"; 
        }
      elsif ($nlen == 0)
        { printf "%60.60s","Invalid PDS file identifier - no name:";
          printf " '$dir/$file'\n"; 
        }

      if ($xlen > 3)
        { printf "%60.60s","Invalid PDS file extension - length > 3:";
          printf " '$dir/$file'\n";
        }
    }

  # And we're done.  I think.

  return;
}
