#!/usr/bin/perl

use Getopt::Std;
#require "getopts.pl";

# chunkdir
#
#
#
#  Format:  chunkdir [OPTIONS] <size> [<directory>]
#
#  Options:
#     -b           Will print BAD (exluded) files/dirs in report
#     -o <file>    Will output report to specifed <file>
#     -p           PDS, will ignore ALL CAPS files/dir and the dataset.html
#                  file, inside of the root dir.
#     -s <scaler>  Will designate the scaler (K|M|G|T) for <size>
#
#  Known Issues:
#     *  Will ignore hidden files at the root level, but doesn't in sub dirs.
#     *  <size> must be an integer or have no more than two decimal places.
#        Also it cannot start with a decimal point.
#     *  Does not work with links to dirs.


# Default values

$blockSize = 512;    # default block size for ISO media

# Function Declarations

sub Print($); # Replaces the "print" command so it will print to two outputs.
sub roundup($;$); # Rounds number to next highest block size.
sub rlist; # file/dir list in conjunction with File::Find
sub dirSize($);  # calculates directory size to next higher ISO block
sub bestUnit($); # finds best scalar to display for a given size in bytes
sub cleandirpath($); # removes junk from directory path


#===================================================================

# Gather all arguments and options from utility declaration.

getopts('o:s:bp');

if (@ARGV > 2 || ! @ARGV)
{
    die "Usage: % chunkdir [ OPTIONS ] size [ directory ].\n";
}
elsif (@ARGV == 2)
{
    $chunkSize = $ARGV[0];
    $rootdir = $ARGV[1];
    if (! (-d $rootdir))
    {
	die "ERROR: Could not open source directory \"$rootdir\".\n";
    }
    elsif ($rootdir =~ /^(.*)\/$/) { $rootdir = $1; }
}
else
{
    $chunkSize = $ARGV[0];
    $rootdir = ".";
}

if ($opt_o)
{
    open(OUTPUT, "> $opt_o") or die "ERROR: Cannot open \"$opt_o\": $!\n";
}


if ($opt_s)
{
    if ($opt_s =~ /^[KMGT]$/)
    {
	if ($chunkSize =~ /^[0-9]+$/)
	{
	    $scaler = $opt_s;
	}
	else
	{
	    die "ERROR: \"$chunkSize\" must be an integer when used in ".
		"conjunction with -o option.\n";
	}
    }
    else { die "ERROR: \"$opt_s\" is an invalid -s argument.\n"; }
}
elsif ($chunkSize =~ /^([0-9]+(\.[0-9][0-9]?)?)([KMGT])$/)
{
    $chunkSize = $1;
    $scaler = $3;
}
elsif ($chunkSize =~ /^[0-9]+(\.[0-9][0-9]?)?$/)
{
    $scaler = "M";  # Default Scaler
}
else
{
    die "ERROR: \"$chunkSize\" is not a valid chunk size.\n";
}

# Calculate Exact Chunk Size
if    ($scaler eq "K") { $chunkSize *= 2 ** 10; }
elsif ($scaler eq "M") { $chunkSize *= 2 ** 20; }
elsif ($scaler eq "G") { $chunkSize *= 2 ** 30; }
elsif ($scaler eq "T") { $chunkSize *= 2 ** 40; }
else { die "Logic Error: -o option and chunkSize calc.\n"; }
$chunkSize = &roundup($chunkSize,1);

#===================================================================

# Get list of files and directories in $rootdir

use File::Basename;
use Cwd;
use File::Find;

$pwd = cwd()."/";
$rootdir = &cleandirpath($rootdir);
if (! -d $rootdir) {die "ERROR: cleandirpath() function failed.\n";}

@list = glob("$rootdir/*");  # this will not get hidden files/dirs

foreach (@list)
{
    ($tst_n,$tst_p,$tst_e) = fileparse($_);
    if ($opt_p && ("$tst_n$tst_e" eq "dataset.html" ||
	("$tst_n$tst_e" eq "\U$tst_n$tst_e\E" && "$tst_n$tst_e" =~ /[A-Z]+/)))
    {
	if (-d $_) { push(@listBad,"$_/"); }
	else       { push(@listBad,$_); }
    }
    elsif (-d $_) { push(@listDirs,"$_/"); }
    else          { push(@listFiles,$_); }
}


#==================================================================

# Print Report

use File::stat;

Print "# Present Working Directory: \"$pwd\"\n";
Print "# Standard Chunk Size: $chunkSize bytes ".&bestUnit($chunkSize)."\n";
Print "# Root Directory: \"$rootdir\/\"\n";

if ($opt_b && @listBad)
{
    foreach (@listBad)
    {
	if (! -d $_)
	{
	    $stats = stat($_);
	    $totSize += &roundup($stats->size);
	}
	else { $totSize += &dirSize($_); }
    }
    Print "\n# Files/Dirs not included; SIZE = $totSize bytes ".
	&bestUnit($totSize)."\n\n";
    foreach my $value (@listBad)
    {
	$value =~ s/^$rootdir\///;
	Print "0 $value\n";
    }
    Print "\n";
}

$ccount = 1; # Chunk Count Number
$index = 0; # Current index number in list**** arrays
@listFD = (@listFiles,@listDirs);
$totSize = 0;
$totSizeTmp = 0;
@listChunk = ();
@warn = ();


while ($index <= $#listFD)
{
    if (! -d $listFD[$index])
    {
	$stats = stat($listFD[$index]);
	$totSizeTmp += &roundup($stats->size);
    }
    else { $totSizeTmp += &dirSize($listFD[$index]); }

    my $abrvpath = $listFD[$index];
    $abrvpath =~ s/^$rootdir\///;

    if ($totSizeTmp == $chunkSize)
    {
	push (@listChunk,$abrvpath);
	Print "\n# Chunk $ccount; SIZE = $totSizeTmp bytes ".
	    &bestUnit($totSizeTmp)."; WARNING: matches chunk size.\n\n";
	Print "$ccount ".join("\n$ccount ",@listChunk)."\n";
	push (@warn,$ccount);

	@listChunk = ();
	$totSizeTmp = 0;
	$totSize = 0;
	$index++;
	$ccount++;
    }
    elsif ($totSizeTmp > $chunkSize && @listChunk == ())
    {
	Print "\n# Chunk $ccount; SIZE = $totSizeTmp bytes ".
	    &bestUnit($totSizeTmp)."; ".
	    "WARNING: Greater than standard chunk size.\n\n";
	Print $ccount." ".$abrvpath."\n";
	push (@warn,$ccount);

	$totSizeTmp = 0;
	$index++;
	$ccount++;
    }
    elsif ($totSizeTmp > $chunkSize)
    {
	Print "\n# Chunk $ccount; SIZE = $totSize bytes ".
	&bestUnit($totSize)."\n\n$ccount ";
	Print join("\n$ccount ",@listChunk)."\n";

	@listChunk = ();
	$totSizeTmp = 0;
	$totSize = 0;
	$ccount++;
    }
    else
    {
	$totSize = $totSizeTmp;
	push(@listChunk,$abrvpath);
	$index++;
    }
}

if ($totSize>0)
{
    Print "\n# Chunk $ccount; SIZE = $totSize bytes ".
	&bestUnit($totSize)."\n\n$ccount ";
    Print join("\n$ccount ",@listChunk)."\n";
}

# End Comments/Reports

Print "\n\n";

if (@warn) { Print "# WARNINGS given for chunk(s): ".join(", ",@warn)."\n"; }
else { Print "# NO WARNINGS given\n"; }

Print "\n\n# EOF\n";


#===================================================================

# Close program

if ($opt_o)
{
    close(OUTPUT);
    print "\nReport succesfully outputted to \"$opt_o\".\n";
}


#===================================================================
#===================================================================

# Functions

sub Print($)
# Function to replace 'print'.  Will print to both STDOUT and another
# location if another location was provided.
{
    $text = $_[0];
    if ($opt_o) { print OUTPUT $text; }
    print $text;
    return 1;
} # end Print()


sub roundup($;$)
# Rounds up the value given to the nearest ISO block size (or another
# size if provided).
{
    my ($num,$near) = ($_[0],$_[1]);
    if (! $near) { $near = $blockSize; }
    return int( $num / $near + .9999) * $near;
} # end roundup()


sub rlist
# Used exclusively by dirSize() function to create a list of files and sub
# directories of a given directory.
{
    push(@rlisttmp,$File::Find::name);
} # end rlist


sub dirSize($)
# Calculates the total size of a directory and all of its contents rounded
# up to the nearest ISO block size.
{
    my $dir = $_[0];
    my $rv = 0;
    @rlisttmp = ();

    find(\&rlist,$dir);

    foreach (@rlisttmp)
    {
	$dstats = stat($_);
	$rv += &roundup($dstats->size);
    }

    return $rv;
} # end dirSize()


sub bestUnit($)
# Displays best abreviated unit of byte size to one decimal place.
# EX: "10000" bytes = "(9.8 KB)"
{
    my $num = $_[0];

    if ($num < 2 ** 10) { return; }
    elsif ($num < 2 ** 20)
    {
	return "(".&roundup($num,(2 ** 10)/10)/(2 ** 10)." KB)";
    }
    elsif ($num < 2 ** 30)
    {
	return "(".&roundup($num,(2 ** 20)/10)/(2 ** 20)." MB)";
    }
    elsif ($num < 2 ** 40)
    {
	return "(".&roundup($num,(2 ** 30)/10)/(2 ** 30)." GB)";
    }
    else { return "(".&roundup($num,(2 ** 40)/10)/(2 ** 40)." TB)"; }

} # end bestUnit()


sub cleandirpath($)
# Returns a cleaned up Directory path.  NO "." or ".." or "~"
{
    my $dir = $_[0];
    my $odir = $dir;
    my $home = $ENV{"HOME"};

    # add trailing "/"
    if ($dir !~ /\/$/) {$dir = $dir."/";}

    # replace home directory ("~" value)
    if ($dir =~ /^\~\// || $dir eq "~/")
    {
	$dir =~ /^(\~)(.*)$/;
	$dir = $home.$2;
    }

    # replace/remove current directory ("." value)
    if ($dir eq "./") {$dir = $pwd;}
    elsif ($dir =~ /^(\.\/)(.+)$/) {$dir = $pwd.$2;}
    $dir =~ s/\/\.\//\//g;

    # remove parent directories (".." value)
    if ($dir =~ /^\.\./) {$dir = $pwd.$dir;}
    while ($dir =~ /\/\.\./)
    {
	if ($dir !~ s/\/[^\/]+\/\.\.\//\//)
	{
	    die "ERROR: Location does not exist.  Too many parent directory ".
		"values.\n\t\Dir: \"$odir\"\n";
	}
    }

    # remove trailing "/"
    if ($dir =~ /^(.+)\/$/) {$dir = $1;}


    return $dir;
}




1;
