#!/usr/bin/perl

use Getopt::Std;

# Routine to split a PDS attached header off from the associated data file.
#
# Format:
#
#   % detachpds [-s] [-d | -l] <PDS file list>
#
#   where  -s   does a simple lable/data split (without pointer updates)
#          -d   indicates that data segments should be discarded
#          -l   indicates that labels should be discarded
#
#   Each input file is split into label and data, and written to output
#   files with the same name as the input file. Label files will have the
#   extension ".lll", data files the extension ".ddd". 
#
#
#  09 Oct 2001, a.c.raugh
#  24 Oct 2001, acr: Upgraded to alter output labels appropriately.
#  09 Nov 2001, acr: Added simple split option
#  26 Sep 2006, acr: Modified to use Getopt::Std
#
#==============================================================================

# Initializing:

$no_labels = 0;
$no_data   = 0;

# Get options and check for a file list:

getopts('dls');

if ((@ARGV == 0)  ||  ($opt_d  &&  $opt_l))
  { die "Usage: detachpds [-s] [-d | -l] <file list>\n"; }

# Loop through each file:

foreach $file (@ARGV) 
  { &detach($file); }

# Done.

#-----------------------------------------------------------------------------
#-----------------------------------------------------------------------------

sub detach

# Routine to detach the PDS label from the data, update the pointers and
# write label and data into separate files.

{ local ($infile) = @_[0];     # file name to be checked.

  my ($file,$line);
  my ($not_done,$not_pds);
  my ($record_bytes);
  my ($abel_size);
  my ($offset,$byte_offset);
  my ($buffer);

  # Check that the file exists and open it:

  if (-s $infile)
    { open (IN,$infile) || die "Could not open $infile for reading."; }
  else
    { printf stderr "\"$infile\" does not exist (or might be empty)\n";
      return;
    }

  # If this is a PDS file, the first keyword should be "PDS_VERSION_ID":

  $not_done = 1;
  while ($not_done)
    { $line = <IN>;
      if ($line =~ /^\s*PDS_VERSION_ID/)
        { $not_pds = 0;
          $not_done = 0;
        }
      elsif ($line !~ /^\s*$/  &&  $line !~ /^\s*\/\*/)
        { printf stderr "\"$infile\" is not a PDS file.\n";
          $not_pds  = 1;
          $not_done = 0;
        }
    }

  # If it's not a PDS file, we're done now:

  if ($not_pds)
    { close(IN);
      return;
    }

  # If we're going to be modifying pointers, we need to collect some parameters
  # in advance. We'll read through the label until we find them:

  if (!$opt_s)
    { chop $line;
      $label_records = 0;
      $record_bytes  = 0;

      while ($line !~ /^\s*END\s/  &&  $line !~ /^\s*END$/)
        { if ($line =~ /^\s*RECORD_BYTES/)
            { $record_bytes = $line;
              $record_bytes =~ s/^[^=]*=\s*//;
              $record_bytes =~ s/([0-9]+)\s.*$/$1/;
            }
          $line = <IN>;
          chop $line;
        }
    }

  # OK, we've got all the info we need. Now we'll rewind, step all the
  # way through the file and split into separate files aas we go:

  seek(IN,0,0);

  $file = $infile;
  $file =~ s/\.[^.]*$//;   # Remove extension

  # To avoid making a third logical check on each line, we'll split here
  # depending on whether we're saving the label:

  if ($opt_l)     # No label
    { $line = <IN>;
      while ($line !~ /^\s*END\s*$/  &&  $line !~ /^\s*END\s*\/\*/)
        { $line = <IN>; }
    }

  else
    { open (LBL,">$file.lll") || die "Could not open $file.lll for writing.";
 
      $line = <IN>;
      while ($line !~ /^\s*END\s*$/  &&  $line !~ /^\s*END\s*\/\*/)

        { # If we're detaching rather than just splitting, we need to locate
          # the data object pointers and update them:

          if (!$opt_s &&
              $line =~ /^\s*\^/        &&  $line !~ /TEXT/          &&
              $line !~ /STRUCTURE/    &&  $line !~ /CATALOG/       &&
              $line !~ /INDEX_TABLE/  &&  $line !~ /DESCRIPTION /  &&
              $line !~ /DATA_SET_MAP_PROJECTION/)
            { chop $line;
              $offset = $line;
              $offset =~ s/^[^=]+=\s+//;
              $offset =~ s/([0-9]+).*$/$1/;
              $byte_offset = ($line =~ /<BYTES>/);

              # If we have not yet set the label size, set it now in the
              # offset units:

              $label_size = $offset - 1 if ($label_size == 0);

              # Write a new pointer line:

              if ($byte_offset)
                { $buffer = sprintf "(\"$file.ddd\", %d <BYTES>)\n",
                                    $offset-$label_size; }
              else
                { $buffer = sprintf "(\"$file.ddd\", %d)\n", 
                                    $offset-$label_size; }
              $line =~ s/$offset\s*(<BYTES>|)/$buffer/;
            }

          # Now we can write the line:

          printf LBL $line; 
          $line = <IN>;
        }

      close(LBL);
    }

  # We'll do the same for the data part. Note, however, that the data might
  # be binary. So we switch to direct read and write operations:

  if ($opt_d)      # No data
    { close(IN); }

  elsif (eof(IN))  # No attached data
    { printf stderr "$infile is not attached to data."; }

  else
    { open (DAT,">$file.ddd") || die "Could not open $file.ddd for writing."; 
      while (read(IN,$line,512))
        { print DAT $line; }
      close(IN);
      close(DAT);
    }

  # And we're done.

  return 1;
}

#-----------------------------------------------------------------------------
