#!/usr/bin/perl

use lib "/n/sbnops01/lcltools/lib/perl5";
use Getopt::Std;

require "nextPDSODL.pl";

# dsids
#
#    Routine to search through a directory tree containing PDS .lbl files
#    and find all the different DATA_SET_ID values for listing in in the
#    VOLDESC.CAT file.
#
#  NOTE:  This routine assumes all PDS labels are detached and named "*.lbl".
#
#
#  Format: % dsids [-o <output_file>] [<directory>]
#
#  27 Jul 1995, acr.
#  20 May 1997, acr: fixed handling of DATA_SET_IDs appearing in set notation,
#                    added call to nextPDSODL subroutine.
#  01 Sep 1998, acr: Tweaked output format so that DATA_SET_ID fields won't be
#                    truncated if they run long.
#  07 Oct 1998, acr: Changed 'require' statement to use default system libraries
#  26 Sep 2006, acr: Modified to use Getopt::Std and "use lib"
#
#=============================================================================
#  First, get the input arguments if any and open the output file:

getopts('o:');

if (@ARGV > 1)
  { die "Usage: % dsids [ -o output ] [ directory ].\n"; }
elsif (@ARGV == 1)
  { $rootdir = $ARGV[0]; 
    if (!(-d $rootdir))
      { die "dsids: \"$ARGV[0]\" is not a directory.\n"; }
  }
else
  { $rootdir = "."; }

if ($opt_o)
  { open (OUTPUT,"> $opt_o")  ||  
       die "Could not open output file \"$opt_o\".";
  }
else
  { open (OUTPUT,">-"); }

#=============================================================================
# Process the directory tree and print the results:

&runlabels($rootdir);

$count = (keys %idlist);              # The total number of ids found
if ($count < 1) 
  { die "dsids: No DATA_SET_IDs found in \"$rootdir\".\n"; }

$blanks = "                         ";
$i = 0;

foreach $id (sort (keys %idlist))
  { $i++;

    if ($count == 1  && $idlist{$id} == 1)
      { printf OUTPUT "DATA_SET_ID = $id   /* 1 label */\n"; }
    elsif ($count == 1  && $idlist{$id} > 1)
      { printf OUTPUT "DATA_SET_ID = %s   /* %d labels */\n", 
              $id, $idlist{$id}; }

    elsif ($i == 1  &&  $idlist{$id} == 1)
      { printf OUTPUT "DATA_SET_ID = { %-42s,  /* %3d label  */\n", 
              $id, $idlist{$id}; }
    elsif ($i == 1  &&  $idlist{$id} > 1)
      { printf OUTPUT "DATA_SET_ID = { %-42s,  /* %3d labels */\n", 
              $id, $idlist{$id}; }

    elsif ($i < $count  &&  $idlist{$id} == 1)
      { printf OUTPUT "%16.16s%-42s,  /* %3d label  */\n", 
              $blanks, $id, $idlist{$id}; }
    elsif ($i < $count  &&  $idlist{$id} > 1)
      { printf OUTPUT "%16.16s%-42s,  /* %3d labels */\n", 
              $blanks, $id, $idlist{$id}; }

    elsif ($i == $count  &&  $idlist{$id} == 1)
      { printf OUTPUT "%16.16s%-42s   /* %3d label  */\n", 
              $blanks, $id, $idlist{$id}; }
    elsif ($i == $count  &&  $idlist{$id} > 1)
      { printf OUTPUT "%16.16s%-42s   /* %3d labels */\n", 
              $blanks, $id, $idlist{$id}; }

  }
if ($count > 1)
  { printf OUTPUT "%14.14s}\n",$blanks; }


#-----------------------------------------------------------------------------
#-----------------------------------------------------------------------------

sub runlabels

#
# Subroutine to run the loop through label files and subdirectories as needed
#
#   Arguments:  $_[0] = directory (unopened)
#
  { local ($dir=$_[0]);
    local (@label_file);                 # file names ending with ".lbl"
    local (@subdir);                     # subdirectory names
    local ($lf,$sd);                     # element counters for filename lists
    local ($file,$i);                    # loop/subscript variables
    local ($full_name);                  # full file name

    # Open the directory:

#print "runlabels: $dir\n";
    opendir(DIR,$dir) || die "dsids: Could not open \"$dir\".\n";

    # Get a sorted list of files in the directory, and parse it into lists
    # of label file names and subdirectories:

    $lf = $sd = 0;
    foreach $file (sort readdir(DIR))
      { if ($file =~ /^\./)
          { next; }
        else
          { $file = $dir . '/' . $file; }
        $tfile = $file;
        $tfile =~ tr/a-z/A-Z/;
        if ($tfile =~ /\.LBL$/)
          { $label_file[$lf] = $file;
            $lf++;
          }
        elsif (-d $file)
          { $subdir[$sd] = $file;
            $sd++;
          }
      }

    #  Now process the label files:

    foreach $file (@label_file)
      { open (LABEL,$file) || die "dsidx: Could not open \"$file\".\n";
        &process_label($file);
      }

    # Close the current directory and descend through subdirectories:

    close(DIR);

    foreach $file (@subdir)
      { &runlabels($file,$rflag); }
  }

#-----------------------------------------------------------------------------

sub process_label

  #  Routine to read through the label file to find the DATA_SET_ID keyword.
  #  This is added to the list of DATA_SET_IDs found so far.  The associative
  #  array keyed on DATA_SET_ID contains a count of how many references to 
  #  that ID were encountered.

  { local ($file = $_[0]);     # Name of currently open file (LABEL)

    #  Loop through label lines until the DATA_SET_ID keyword is found:
 
    $found = 0;
    $line = &nextPDSODL(*LABEL);
    while (!($found)  &&  $line)
      { if ($line =~ /^\s*DATA_SET_ID/)
          { $str = $line; 
            $str =~ s/^.*=\s*//;
            if ($str =~ /^[\(\{]/)
              { $str =~ s/^[({]\s*//;
                $str =~ s/\s*[)}].*$//;
                $str =~ s/\s*,\s*/,/g;
              }
            foreach $id (split(/\s*,\s*/,$str))
              { $idlist{$id}++; }
            $found = 1;
          }
        $line = &nextPDSODL(*LABEL);
      }

    # Make sure we actually found a DATA_SET_ID:

    if (!$found)
      { printf STDERR "dsids: \"$file\" did not contain a DATA_SET_ID.\n"; }

  }


#=============================================================================
