#!/usr/bin/perl

use Getopt::Std;
use lib "/n/sbnops01/lcltools/lib/perl5";
require "nextPDSODL.pl";

# dstree
#
#    Routine to search through a directory tree containing PDS files
#    and construct the DIRECTORY and FILE objects describing the tree
#    structure, for use in the VOLUME object of VOLDESC.CAT file.  The
#    program will optionally also produce a command file with the appropriate
#    tar commands to ensure that the files are actually stored in the tar
#    file in the order indicated by the tree structure.
#
#    If the "a" option is specified, dstree will determine the RECORD_TYPE and
#    FILE_RECORDS for all ASCII files, and will search for (.lbl) files
#    to retrieve those values for binary files.  For binary data, the label
#    files must have the same filename as the data file and be in the same
#    directory.
#
#    For the "c" option, dstree create an output file of the given name 
#    which will contain the sequence of tar commands required to create and
#    fill an archive file, called "volume.tar", with the listed files in the
#    order in which they appear in the object listing.
#
#  NOTE:  This routine assumes all PDS labels are detached and named "*.lbl".
#
#
#  Format: % dstree [-o <output_file>] [-c <command_file>] [-a] [<directory>]
#
#  31 Jul 1995, acr.
#
#  Modification History:
#
#    17 Dec 1996, acr: Changed record type of "FIXED" to "FIXED_LENGTH"
#    28 May 1997, acr: Corrected error in substitution which prevented finding
#                      labels for binary files; fixed handling for binary
#                      file labels.
#    07 Oct 1998, acr: Changed 'require' to use default system libraries
#    26 Sep 2006, acr: Modified to use Getopt::Std and and "use lib"
#
#=============================================================================
#  First, get the input arguments if any and open the output file(s):

getopts('ao:c:');

if (@ARGV > 1)
  { die "Usage: % dstree [-o output] [-c command_file] [-a] [directory]\n"; }
elsif (@ARGV == 1)
  { $rootdir = $ARGV[0]; 
    if (!(-d $rootdir))
      { die "dstree: \"$ARGV[0]\" is not a directory.\n"; }
  }
else
  { $rootdir = "."; }

if ($opt_o)
  { open (OUTPUT,"> $opt_o")  ||  
       die "Could not open output file \"$opt_o\".";
  }
else
  { open (OUTPUT,">-"); }

if ($opt_c)
  { open (COMMAND,"> $opt_c") ||
       die "Could not open command file \"$opt_c\" for output.";
  }

#=============================================================================
# Process the directory tree and print the results:

$indent = 0;
$indent_increment = 2;
$sequence = 0;
$blanks = "                                                             ";
$tar_flags = "cvf";

&runlabels($rootdir);



#-----------------------------------------------------------------------------
#-----------------------------------------------------------------------------

sub runlabels

#
# Subroutine to run the loop through label files and subdirectories as needed
#
#   Arguments:  $_[0] = directory (unopened)
#
  { local ($dir=$_[0]);
    local (@filename);                   # file names 
    local (@subdir);                     # subdirectory names
    local ($fn,$sd);                     # element counters for filename lists
    local ($file,$i);                    # loop/subscript variables
    local ($full_name);                  # full file name
    local ($dirname);                    # lowest level directory name

    # Open the directory:

#print "runlabels: $dir\n";
    opendir(DIR,$dir) || die "dstree: Could not open \"$dir\".\n";

    # Print the output lines:

    $dirname = $dir;
    if ($dirname eq ".")
      { $dirname = "\".\""; }
    else
      { $dirname =~ tr/a-z/A-Z/;
        $dirname =~ s/^.*\/([-A-Z0-9_]+)$/"$1"/;
      }

    printf OUTPUT "%${indent}.${indent}sOBJECT     = DIRECTORY\n",$blanks;
    $indent += $indent_increment;
    printf OUTPUT "%${indent}.${indent}sNAME     = $dirname\n",$blanks;

    # Get a sorted list of files in the directory, and parse it into lists
    # of label file names and subdirectories:

    $fn = $sd = 0;
    foreach $file (sort readdir(DIR))
      { if ($file =~ /^\./)
          { next; }
        else
          { $file = $dir . '/' . $file; }

        if (-d $file)
          { $subdir[$sd] = $file;
            $sd++;
          }
        else
          { $filename[$fn] = $file;
            $fn++;
          }
      }

    #  Now process the regular files:

    foreach $file (@filename)
      { if ($opt_a)
          { open (DATA,$file) || die "dstree: Could not open \"$file\".\n"; }
        printf OUTPUT "\n";
        &process_file($file);
      }

    # Close the current directory and descend through subdirectories:

    close(DIR);

    foreach $file (@subdir)
      { printf OUTPUT "\n";
        &runlabels($file); 
      }

    # Finish the DIRECTORY object:

    printf OUTPUT "\n";
    $indent -= $indent_increment;
    printf OUTPUT "%${indent}.${indent}sEND_OBJECT = DIRECTORY\n",$blanks;

  }

#-----------------------------------------------------------------------------

sub process_file

  #  Routine to process each regular file in the directory.  If no file
  #  typing was requested, this consists of merely writing out the file 
  #  name and a couple of blank keywords for filling in later.  However,
  #  if typing was requested, the file type is determined and the program
  #  attempts to determine the number and type of records in the file.
  #  If a tar command file is being generated, the command line for this
  #  file is written to it.

  { local ($file = $_[0]);     # Name of currently open file (DATA)
    local ($bare);             # bare file name, stripped of directory
    local ($max,$min,$rcount); # file record attributes
    local ($line,$len);        # input buffer
    local ($found);            # logical flag
    local ($rt,$fr,$rb);       # file attributes from label

    # Format the file name for printing by stripping off the directory,
    # forcing to uppercase and surrounding with double quotes:

    $bare = $file;
    $bare =~ s/^.*\///;
    $bare =~ tr/a-z/A-Z/;
    $bare = "\"$bare\"";

    #  If we're making a tar command file, write the appropriate line:

    if ($opt_c)
      { printf COMMAND "tar $tar_flags volume.tar $file\n";
        $tar_flags = "rvf";
      }

    #  Begin the output FILE object:

    $sequence++;
    printf OUTPUT "%${indent}.${indent}sOBJECT     = FILE\n",$blanks;
    $indent += $indent_increment;
    printf OUTPUT "%${indent}.${indent}sFILE_NAME       = %s\n",$blanks,$bare;
    printf OUTPUT "%${indent}.${indent}sSEQUENCE_NUMBER = $sequence\n", $blanks;

    # If no file typing, write the simple lines and exit:

    if (!$opt_a)
      { printf OUTPUT "%${indent}.${indent}sRECORD_TYPE     = \n", $blanks;
        printf OUTPUT "%${indent}.${indent}sFILE_RECORDS    = \n", $blanks;
        printf OUTPUT "%${indent}.${indent}sRECORD_BYTES    = \n", $blanks;
        $indent -= $indent_increment;
        printf OUTPUT "%${indent}.${indent}sEND_OBJECT = FILE\n", $blanks;
        return;
      }

    # File typing requested.  Determine the file type:

    if (-T DATA)
      { # It's a text file.  Determine the number and type of records:

        $rcount = 0;
        $max = -1;
        $min = 99999999;
        while ($line=<DATA>)
          { $rcount++;
            $len = length($line);
            $max = ($len > $max) ? $len : $max;
            $min = ($len < $min) ? $len : $min;
          }

        # Finish the FILE object:

        if ($max == $min)
        { printf OUTPUT "%${indent}.${indent}sRECORD_TYPE     = FIXED_LENGTH\n",
                           $blanks;
          }
        else
          { printf OUTPUT "%${indent}.${indent}sRECORD_TYPE     = STREAM\n",
                           $blanks;
          }
        printf OUTPUT "%${indent}.${indent}sFILE_RECORDS    = $rcount\n", 
                           $blanks;
        printf OUTPUT "%${indent}.${indent}sRECORD_BYTES    = $max\n", $blanks;
        $indent -= $indent_increment;
        printf OUTPUT "%${indent}.${indent}sEND_OBJECT = FILE\n", $blanks;
      }

    elsif (-B DATA)
      { # It's a binary file.  Find and open the label:

        $label = $file;
        $label =~ s/^(.*)\.[A-Za-z0-9]+$/$1.lbl/;

        $found = -e $label;
        if (!$found)
          { $label =~ s/\.lbl/.LBL/;
            $found = -e $label;
          }

        # If we couldn't find a label, write blank lines to the output
        # file and notify the user:

        if (!$found)
          { printf STDERR "dstree: Could not find label file for ";
            printf STDERR "binary file \"$file\".\n";

            printf OUTPUT "%${indent}.${indent}sRECORD_TYPE     = \n", $blanks;
            printf OUTPUT "%${indent}.${indent}sFILE_RECORDS    = \n", $blanks;
            printf OUTPUT "%${indent}.${indent}sRECORD_BYTES    = \n", $blanks;
            $indent -= $indent_increment;
            printf OUTPUT "%${indent}.${indent}sEND_OBJECT = FILE\n", $blanks;
            return;
          }

        # Otherwise, open the label and look for the appropriate keywords:

        open (LABEL,$label)  ||
           die "dstree: Could not open label file \"$label\".";

        $rt = $rb = $fr = "";
        while ($line = &nextPDSODL(*LABEL))
          { if ($line =~ /RECORD_TYPE/)
              { $rt = getval($line); 
                $rt =~ s/["']//g;
                $rt =~ s/ /_/g;
              }
            elsif ($line =~ /FILE_RECORDS/)
              { $fr = &getval($line); }

            elsif ($line =~ /RECORD_BYTES/)
              { $rb = &getval($line); }

            # Quit once all the information has been gathered:

            last if ($rt ne ""  &&  $fr ne ""  &&  $rb ne "");
          }

        close(LABEL);

        # Write the values:

        printf OUTPUT "%${indent}.${indent}sRECORD_TYPE     = $rt\n", $blanks;
        printf OUTPUT "%${indent}.${indent}sFILE_RECORDS    = $fr\n", $blanks;
        printf OUTPUT "%${indent}.${indent}sRECORD_BYTES    = $rb\n", $blanks;
        $indent -= $indent_increment;
        printf OUTPUT "%${indent}.${indent}sEND_OBJECT = FILE\n", $blanks;
      }

    # If the file is neither text nor binary, complain:

    else
      { printf STDERR "dstree: Unrecognized file typs for \"$file\".\n";

        printf OUTPUT "%${indent}.${indent}sRECORD_TYPE     = \n", $blanks;
        printf OUTPUT "%${indent}.${indent}sFILE_RECORDS    = \n", $blanks;
        printf OUTPUT "%${indent}.${indent}sRECORD_BYTES    = \n", $blanks;
        $indent -= $indent_increment;
        printf OUTPUT "%${indent}.${indent}sEND_OBJECT = FILE\n", $blanks;
      }
  }

#=============================================================================
