#!/usr/bin/perl

# Routine to create a references.cat file by extracting the full references
# from "allrefs.cat" corresponding to the list of REFERENCE_KEY_ID values
# passed in.
#
# Format:
#
#    % gather_ref [-a master_file] [ input_file [ output_file ] ]
#
# Input reference ID list format is fairly lenient: it may include the 
# "REFERENCE_KEY_ID =" tag or not; references may be in or outside of 
# quotes; there may be more than one on a line (separated by commas);
# grep output file name tags will be ignored.  Anything extracted from 
# a syntactically valid PDS label or catalog file should work.
#
# The input references are tabulated and sorted, with duplicates being
# removed.  Then the allrefs.cat master file is consulted and the full
# reference object copied from it to the output file.  Unmatched 
# references are reported to stderr.
#
# Standard I/O is used in the absence of specific file names.  The
# -a option allows the user to specify an alternate master reference
# file (include path relative to the current working directory as
# needed).
#
# 13 Feb 2004, acr.
# 21 Apr 2004, acr: Added grep file tag handling
#
#=============================================================================

use Getopt::Std;

$DEFAULT_MASTER = "/n/sbnops01/lcltools/doc/allrefs.cat";

# Sort out the command line specs:

getopts('a:');
if (@ARGV > 2)
  { die "Usage: gather_ref [-a master_file] [ input_file [ output_file ] ]\n"; }

$MASTER_REF = ($opt_a)? $opt_a : $DEFAULT_MASTER;

$infile  = (@ARGV > 0)? $ARGV[0] : "-";
$outfile = (@ARGV > 1)? $ARGV[1] : "-";


# Open the input file and read the REFERENCE_KEY_ID values into a hash:

open (IDLIST,$infile) || die "Could not open $infile for reading ($!)";

while ($line=<IDLIST>)
  { chop $line;
    $line =~ s/^\s*//;         # Trim blanks
    $line =~ s/\s*$//;

    # Remove a grep file name tag, if any:

    $line =~ s/^[^:]*: *//;    

    # Remove the REFERENCE_KEY_ID tag, if any:

    $line =~ s/^\s*REFERENCE_KEY_ID\s*=\s*//;

    # Remove opening or closing braces or parentheses and trailing commas:

    $line =~ s/^\s*[{(]\s*//;
    $line =~ s/\s*[})]\s*$//;
    $line =~ s/\s*,\s*$//;

    # Now peel off the IDs one at a time.  They may be delimited by 
    # quotes or commas:

    while (length($line))
      { if ($line =~ /^"/)             # Quotes delimit next ID
          { $line =~ s/^"([^"]+)"//;
            $id = $1;
            $line =~ s/^(,|)\s*//;     # Remove separator stuff
          }
        elsif ($line =~ /^[^",]+,/)    # Comma delimits next ID
          { $line =~ s/^([^,]+)\s*,\s*//;
            $id = $1;
          }
        else                           # There's only one ID on the line
          { $line =~ s/^\s*(\S+)\s*//;
            $id = $1;
          }

        # We'll ignore "N/A" values:

        $idlist{$id}++ unless ($id eq "N/A");
      }

    # Next line...
  }

close(IDLIST);

# Now we open the master list file, which we're going to reaf through
# sequentially.  This is a brute-force, rather inelegant method, but it
# gets the job done.

open (MASTER,$MASTER_REF) || 
  die "Could not open $MASTER_REF file for reading ($!)";

# We also open the output pipe and start the file:

if (@ARGV > 1)
  { open (OUT, "| sortref > $outfile") ||
      die "Could not open output pipe to $outfile ($!)";
  }
else
  { open (OUT, "| sortref"); }

printf OUT "PDS_VERSION_ID      = PDS3\n";
printf OUT "LABEL_REVISION_NOTE = \"\n";
printf OUT "  Generated by 'gather_ref' from '$MASTER_REF'\n";
printf OUT "  \"\n\n";

# The plan is to loop through the allref.cat file looking for the 
# REFERENCE_KEY_ID lines, then checking their values for matches in the
# %idlist hash.  Matching references are printed to the output and
# removed from the hash.  Any remaining unmatched IDs are output as
# blank REFERENCE templates and list on stderr.

$idcount = 0;

while ($line=<MASTER>)
  { if ($line =~ /REFERENCE_KEY_ID\s*=\s*"([^"]+)"/)
      { $refid = $1;
        if ($idlist{$refid})
          { printf OUT "OBJECT     = REFERENCE\n";
            while ($line !~ /END_OBJECT/)
              { printf OUT $line;
                $line = <MASTER>;
              }
            printf OUT $line;
            undef $idlist{$refid};
            $idcount++;
          }
      }
  }
close(MASTER);

# Now let's see what's left:

$first = 1;

foreach $id (sort keys(%idlist))
  { if ($idlist{$id})
      { printf OUT "OBJECT     = REFERENCE\n";
        printf OUT "  REFERENCE_KEY_ID = \"$id\"\n";
        printf OUT "  REFERENCE        = \n";
        printf OUT "END_OBJECT = REFERENCE\n";
        $idcount++;

        if ($first)
          { printf STDERR "\nUnmatched REFERENCE_KEY_ID values: \n";
            $first = 0;
          }

        printf STDERR "  $id\n";
      }
  }

printf OUT "END\n";

# All done:

close(OUT);
printf STDERR "\n$idcount unique REFERENCE objects written.\n";
