#!/usr/bin/perl

#
#  Routine to generate a file index of a directory tree.  The index list
#  produced includes path name, filename and file size in bytes, with 
#  a summary at the bottom of total number of files, directories, and
#  bytes (note that directory files are counted as files).
#
#  Format:    % index [ top_of_path ]
#
#             If no path name is supplied, current working directory 
#             is used.
#
#  'index' works with a directory listing produced by the "ls -lAR" command
#  on the path supplied.
#
#  02 Mar 1995, acr
#  19 Jun 1995, acr: Fixed a bug in collecting the path name which looked
#                    suspiciously like the result of an 'ls' output format
#                    change.
#  19 Sep 1995, acr: modified to run under Solaris
#

#
#  Open the input stream.  If a path was given, set the value:
#

if ($#ARGV >= 0)
  { $path = $ARGV[0]; }
open (LISTING, "ls -lAR $path |") || die "Unable to open input pipe.\n";

#
#  Print some titles to the output stream:
#

printf ("%-83.83s %-9.9s\r\n","File Path and Name","  Bytes  ");
printf ("%-83.83s %-9.9s\r\n","------------------","  -----  ");
printf ("\r\n");

while ($line = <LISTING>)
  { chop $line;

    #
    # Ignore the "total" lines and any blank lines:
    #

    if (($line =~ /^total /) || ($line =~ /^\s*$/))
      { next; }

    #
    # Save a new path line:
    #

    if ($line =~ /.*:$/)
      { $line =~ s/:.*$//;
        $path = $line;
        next;
      }

    #
    # Process a file line.
    #

    # First, break the line down into fields:

    ($perm, $link, $o, $g, $size, $m, $d, $ty, $file) = split(/\s+/,$line);

    # ... check for soft links, hard link, directories and zero-length 
    #     files:

    ($perm =~ /^l/) && $total_sl++;
    ($link > 1)     && $total_hl++;
    ($perm =~ /^d/) && $total_dir++;
    ($size == 0)    && $total_zero++;

    # ... update the total byte and file counts:

    $total_size += $size;
    $total_files++;

    # ... output the index line:

    $filename = $path . "/" . $file;
    printf ("%-83.83s %9d\r\n", $filename, $size);
  
  }  # next $line

#
#  Print the Totals and exit:
#

print ("\r\n\r\n");
for ($i=0; $i<93; $i++)
  { print "="; }
print ("\r\n\r\n");
print ("Totals:\r\n-------\r\n\r\n");
printf ("Files: %10d   Directories: %10d\r\n",$total_files, $total_dir);
printf ("Bytes: %10d     Megabytes: %10.3f\r\n\r\n",
               $total_size, $total_size/1000000);

#
#  Finally, report unusual files which may screw up the counts:
#

$total_sl   && printf ("%10d soft links were found.\r\n",$total_sl);
$total_hl   && printf ("%10d hard links were found.\r\n",$total_hl);
$total_zero && printf ("%10d zero-length files were found.\r\n",$total_zero);

close (LISTING);

