#!/usr/bin/perl

use Getopt::Std;

# Routine to seek out text files in a PDS archive tree and add default TEXT
# object labels to those which don't already have some sort of label (attached
# or detached). The labels are just one cut above minimal. This is very much
# a "Quick & Dirty" solution. Editing of the subsequent labels is highly 
# recommended.
#  
#  Format:
#
#   % labeltext [-r] [-d] <file list>
#
#  where    -r   indicates that directories should be searched recursively
#           -d   indicates detached labels should be created
#
# LABELTEXT may be run on a single local file, a list of files, or recursively
# down a directory tree (in which case the directory should be the level at
# which to start looking for unlabeled text files). Only files with extensions
# of ".TXT" or ".txt" are checked. Each file will be checked for both attached
# and detached labels, but detached labels must reside in the same directory as
# the file itself.
#
# The default action is to create attached labels, in which case the new file
# will overwrite the original file. The PDS labels created will always have
# the required TEXT object fields (PUBLICATION_DATE and DEFAULT_NOTE); default
# values are used if none are supplied by the user. The PDS labels written by
# LABELTEXT will always have 80-byte, CR/LF delimited lines, regardless of the
# state of the referenced text file. No attempt is made to evaluate or validate
# the text file itself (beyond checking it for an attached label).
#
# Files with attached labels are recognized by the "PDS_VERSION_ID" keyword
# line, which must be the first non-blank, non-comment line in the file.
#
#  04 May 2000, acr: Creation.
#  23 Aug 2000, acr: Fixed a couple of stupid bugs.
#
#==============================================================================
#
# Retrieve command options, if any:

getopts('rd');

# Initialize default values and global variables for label writing:

($a, $b, $c, $day, $month, $year, @junk) = localtime(time);
$TODAY = sprintf("%4.4d-%2.2d-%2.2d",$year+1900,$month+1,$day);
$DEFAULT_NOTE = "";

$NOTE[0]   = $DEFAULT_NOTE;
$NOTELINES = 1;
$PUBLICATION_DATE = $TODAY;

$TMPFILENAME = "__QQxTMP__";

# Now we'll read through the file list from the command line, separating
# directories and text files:

undef @textfile;
undef @directory;
$t = 0;
$d = 0;

# If no arguments were included on the command line, we assume the current
# default directory:

if (@ARGV == 0)
  { $ARGV[0] = "."; }

while ($file = shift(@ARGV))
  { if (-d $file)
      { $directory[$d] = $file;
        $d++;
      }

    # If it's not a directory name, we check for an unlabeled text file:

    elsif (&unlabeled_textfile($file))
      { $textfile[$t] = $file;
        $t++;
      }
  }

# Loop through the file list, calling the label generator for each:

foreach $file (@textfile)
  { &mklbl($file,$opt_d,); }

# Loop through the directory list:

foreach $dir (@directory)
  { &dodir($dir,$opt_r,$opt_d); }
      
#
#==============================================================================

sub unlabeled_textfile

  # Returns TRUE (1) if the file passed in is an unlabelled text file.

{ my ($file) = $_[0];
  my ($TRUE,$FALSE) = (1,0);
  my ($lbl);
  my ($line,$not_found);

  # Begin by checking for the required extension:

  return $FALSE if ($file !~ /\.txt$/  &&  $file !~ /\.TXT$/);

  # The file must have non-zero size and be a text file:

  return $FALSE if (!(-s $file)  ||  !(-T $file));

  # If there is already a file in the same directory with a ".LBL" or ".lbl"
  # extension, we won't write another label:

  $lbl = $file;
  $lbl =~ s/\.TXT$/\.LBL/;
  $lbl =~ s/\.txt$/.lbl/;
  return 0 if (-e $lbl);

  # Finally, we check inside the candidate file itself. If the first non-blank,
  # non-comment line id a PDS_VERSION_ID line, we assume the file already has
  # a label and leave it alone:

  open (TXT,$file) || die "Unable to open '$file' to check for attached label.";

  # Find first non-blank, non-comment line:

  $not_found = 1;
  while ($not_found)
    { $line=<TXT>;
      chomp $line;
      next if ($line =~ /^\s*$/  ||  $line =~ /^\s*#/); 
      $line =~ s/^\s+//;
      $not_found = 0;
    }

  if ($line =~ /^PDS_VERSION_ID\s*=\s*/)
    { return $FALSE; }
  else
    { return $TRUE; }
}
      
#------------------------------------------------------------------------------

sub dodir

  # Routine to process a directory looking for unlabeled text files. May be
  # called recursively.

{ my ($dir,$recurse,$detach) = @_;

  my ($file, @textlist, @dirlist);
  my ($tl,$dl);

  # Re-initialize:

  undef @textlist;
  undef @dirlist;
  $tl = 0;
  $dl = 0;

  # Open the directory and loop through its contents:

  opendir(DIR,$dir) || die "Unable to open directory '$dir'.";

  foreach $file (readdir(DIR))
    { next if $file =~ /^\./;
      $file = "$dir/$file";

      if (-d $file && $recurse)
        { $dirlist[$dl] = $file; 
          $dl++;
        }
      
      elsif (&unlabeled_textfile($file))
        { $textlist[$tl] = $file;
          $tl++;
        }
    }

  # Call the file handler for each unlabeled text file: 

  foreach $file (@textlist)
    { &mklbl($file,$detach); }

  # Call this routine recursively for each subdirectory (the list is empty if
  # we're not recursing):

  foreach $file (@dirlist)
    { &dodir($file,$recurse,$detach); }

  # And we're done:

  return;
}
#
#==============================================================================

sub mklbl

  # This routine opens the given file and solicits input for the TEXT label.
  # Default values are maintained in global variables. 

  # This is not particularly elegant, but it gets the job done.

{ my ($file,$detach) = @_;
  my ($line);
  my ($invalid_input);
  my ($y,$m,$d,$leap);
  my (@daysin) = (0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);
  my ($tempfile);
  my ($filename);                # Input file without directory information
  my ($not_done);                # Loop flag
  my ($null_line, $blank_line);  # Logical flags

  # Open the temp file to hold the output. The temp file is written in the 
  # same directory as the file to be labelled:

  if ($file =~ /\//)
    { $tempfile = $file; 
      $tempfile =~ s/[^\/]*$/$TMPFILENAME/;
      $filename = $file;
      $filename =~ s/^.*\///;
    }
  else
    { $tempfile = $TMPFILENAME; 
      $filename = $file;
    }
 
  open (TMP,">$tempfile") || die "Unable to open temp file for output.";
  open (IN,$file) || die "Unable to open source file '$file'.";

  # Print a banner with the file name and the first 5 lines of the file:

  printf "\n";
  printf "==================================================================\n";
  printf "File: $file\n";
  printf "==================================================================\n";
  printf "\n";
  for ($i=0; $i<5; $i++)
    { $line = <IN>;
      chomp $line;
      printf ": %-77.77s\n",$line;
    }
  printf "\n";
  close(IN);
  open (IN,$file);   # Resets file position

  # Get the PUBLICATION_DATE and NOTE values:


  $invalid_input = 1;
  while ($invalid_input)
    { printf "PUBLICATION_DATE [$PUBLICATION_DATE]: ";
      $line=<STDIN>;
      chomp $line;
      $invalid_input = 0;

      # We allow a special flag value here. If the user decides NOT to 
      # label this text file, he may enter "SKIP" or "skip" (or actually,
      # just an the first letter) and we'll abort this subroutine, but 
      # continue with the main processing:

      if ($line =~ /^[Ss]/)
        { close(IN);
          close(TMP);
          system "rm $tempfile";
          return;
        }

      if (length($line) != 0)
        { ($y,$m,$d) = split(/-/,$line);
          $leap = (($y % 4)==0  &&  (($y % 100) != 0 || ($y % 400) == 0))?
                  1 : 0;
          $daysin[2] = ($leap)? 29 : 28;

          # We'll run a few checks on the input to look for obviously 
          # invalid entries:

          if ($line !~ /^[0-9]{4}-[0-9]{2}-[0-9]{2}$/)
            { printf "  ...Invalid date format. Use 'YYYY-MM-DD'.\n";
              $invalid_input = 1;
            }
          elsif ($m < 1  ||  $m > 12)
            { printf "  ...Invalid month ($m).\n";
              $invalid_input = 1;
            }
          elsif ($d < 1  ||  $d > $daysin[$m])
            { printf "  ...Invalid day number ($d).\n";
              $invalid_input = 1;
            }
        }
    }

  $PUBLICATION_DATE = $line if (length ($line) > 0);

  # Notes may be several lines long, and we're assuming we have to add
  # the quotation marks. The note lines are formatted here, wrapped on 
  # characters if necessary, and saved in an array which is written out
  # later.

  $not_done = 1;
  while ($not_done)
    { if ($NOTELINES == 1  ||  length($NOTE[0]) == 0)
        { printf "NOTE [%s]:\n> ", substr($NOTE[0],0,60); }
      else 
        { printf "NOTE [%s...]:\n> ", substr($NOTE[0],21,60); }
      $line=<STDIN>;
      chomp $line;                # Remove delimiters
      $line =~ s/[^ ]+\s+$//;     # Trim trailing blanks
      $line =~ s/"/''/g;          # Replace double quote with two apostrophes

      $not_done = 0;

      if (length($line) > 0)
        { $NOTELINES = 0;
          undef @NOTES;

          # Handling special cases for the first line:

          if ($line =~ /^\.$/)
            { $NOTE[0] = ""; }

          elsif (length($line) == 0)
            { if (length($DEFAULT_NOTE) > 0)
                { $NOTE[0] = "  NOTE             = $DEFAULT_NOTE";
                  $NOTELINES = 1;
                }
              else
                { $NOTE[0] = ""; }
            }

          else
            { $NOTELINES = 0;
              $NOTE[0] = "  NOTE             = \"$line";
              while (length($NOTE[$NOTELINES]) > 78)
                { $NOTE[$NOTELINES+1] = "    " . substr($NOTE[$NOTELINES],74);
                  $NOTE[$NOTELINES]   = substr($NOTE[$NOTELINES],0,74);
                  $NOTELINES++;
                }
              $NOTELINES++;

              printf "> ";
              $line = <STDIN>;
              chomp $line;
              $line =~ s/\s+$//;
              $line =~ s/"/''/g;

              while ($line !~ /^\.$/)
                { $NOTE[$NOTELINES] = "    " . $line;
                  while (length($NOTE[$NOTELINES]) > 78)
                    { $NOTE[$NOTELINES+1] = "    " . 
                                            substr($NOTE[$NOTELINES],74);
                      $NOTE[$NOTELINES]   = substr($NOTE[$NOTELINES],0,74);
                      $NOTELINES++;
                    }
                  $NOTELINES++;
                  printf "> ";
                  $line = <STDIN>;
                  chomp $line;
                  $line =~ s/\s+$//;
                  $line =~ s/"/''/g;
                }
              if (length($NOTE[$NOTELINES-1]) < 78)
                { $NOTE[$NOTELINES-1] .= "\""; }
              else
                { $NOTE[$NOTELINES] = "\""; 
                  $NOTELINES++;
                }
            }
        }

      # Do not allow a null NOTE field:

      if (length($NOTE[0]) == 0)
        { printf "ERROR: The NOTE field is required. ";
          printf "You MUST supply a value.\n\n";
          $not_done = 1;
        }

    }  # end while ($not_done)

  # Write the label:

  printf TMP "%-78.78s\r\n","PDS_VERSION_ID      = PDS3";
  printf TMP "%-78.78s\r\n","RECORD_TYPE         = STREAM";

  # Add a pointer if we're writing a detached label: 

  if ($detach)
    { printf TMP "\n%-78.78s\r\n","\^TEXT               = \"$filename\""; }

  printf TMP "%-78.78s\r\n"," ";
  printf TMP "%-78.78s\r\n","OBJECT     = TEXT";
  printf TMP "%-78.78s\r\n","  PUBLICATION_DATE = $PUBLICATION_DATE";
  for ($i=0; $i<$NOTELINES; $i++)
    { printf TMP "%-78.78s\r\n",$NOTE[$i]; }
  printf TMP "%-78.78s\r\n","END_OBJECT = TEXT";
  printf TMP "%-78.78s\r\n"," ";
  printf TMP "%-78.78s\r\n","END";

  # If we've been asked for a detached label, we stop here and modify the final
  # file name used for the 'rename' operation to create a new file:

  if ($detach)
    { $file =~ s/\.txt/.lbl/;     # We use the same case as the text file
      $file =~ s/\.TXT/.LBL/;
    }

  # Otherwise, we transfer contents from the original file to the temp file:

  else
    { while ($line=<IN>)
        { printf TMP "%s",$line; }
    }

  # Close everything up, rename the temp file, and we're done:

  close IN;
  close TMP;
  rename $tempfile, $file;

  return 1;
}
