#!/usr/bin/perl

use "/n/sbnops01/lcltools/lib/perl5";

require "nextPDSODL.pl";

#
# This routine reads the given label and prints a summary of its contents, by
# object.  The parameters listed depend on the object type encountered.
#
#  Format:  lblsumm   <label file>
#
# The label may be attached or detached.  This routine should be able to handle 
# ^STRUCTURE include files in-line.  The parameters printed are:
#
#  All Labels
#
#    PDS_VERSION_ID
#    DATA_SET_ID
#    Data file name        (may be the same as the input file if attached)
#    FILE_RECORDS
#    RECORD_BYTES
#    RECORD_TYPE
#
#  Supported OBJECTs:
#
#    DOCUMENT                     QUBE 
#    FILE (with subobjects)       SERIES
#    HEADER                       SPECTRUM
#    HISTOGRAM                    SPICE_KERNEL
#    IMAGE                        TABLE
#    INDEX_TABLE                  TEXT
#  
#
#  05 Jun 1997, acr: creation
#
#==============================================================================
# Begin by checking the input arguments:

if (@ARGV > 2 )
  { die "Usage: lblsumm [ label_file  [ output_file ] ]\n"; }


$outfile = (@ARGV == 2)? $ARGV[1] : "-";
$infile  = (@ARGV > 0) ? $ARGV[0] : "-";

open (LBL,$infile)     || die "lblsumm: Unable to open $infile for input";
open (OUT,">$outfile") || die "lblsumm: Unable to open $outfile for output";

$infile  = "[ STDIN ]"  if ($infile eq "-");
$outfile = "[ STDOUT ]" if ($outfile eq "-");

# Defaults:

$pds_version_id = "PDS2";

$not_done = 1;

while ($not_done)
  { $line = &nextPDSODL(*LBL);

    if ($line =~ /PDS_VERSION_ID/)
      { $pds_version_id = &getval_symbol($line); }

    elsif ($line =~ /DATA_SET_ID/)
      { $dsid = &getval_symbol($line); }

    elsif ($line =~ /FILE_RECORDS/)
      { $file_records = &getval($line); }

    elsif ($line =~ /RECORD_BYTES/)
      { $record_bytes = &getval($line); }

    elsif ($line =~ /RECORD_TYPE/)
      { $record_type = &getval_symbol($line); }

    elsif ($line =~ /INSTRUMENT_NAME/)
      { $instrument_name = &getval_symbol($line); }

    elsif ($line =~ /TARGET_NAME/)
      { $target_name = &getval_symbol($line); }

    elsif ($line =~ /^\^/)
      { $val = &getval($line);

        # OK, we've got an object pointer here.  First thing to do is see
        # if it consists entirely of an offset, in which case the data file
        # name is the same as the current input file name:

        if ($val =~ /^[0-9]+\s*(<BYTES>|)/)
          { $val = $ARGV[0]; }

        # Alternately, it might be a series of file names.  There might also
        # be a series of pointers.  We need to collect the file names into a
        # list, forced to lower case and separated by commas:

        elsif ($val =~ /^\{/)     # A sequence of file names
          { $val =~ s/[{}"']//g;
            $val =~ s/\s*,\s*/, /g;
            $val =~ tr/A-Z/a-z/;
          }

        else                      # A single file name with or without offset
          { $val =~ s/\s*,.*$//;  # drop offset
            $val =~ s/[()"']//g;
            $val =~ tr/A-Z/a-z/;
          }

        # If the current file name string is already in the list, do not add
        # it again.  Otherwise it is concatenated to the end of the current
        # list:

        if ($datafile !~ /$val/)
          { $datafile .= ", $val"; 
            $datafile =~ s/^, //;   # Just in case this is the first value
          }

      }
    elsif ($line =~ /^OBJECT\s*=/)
      { $not_done = 0; }

  }

# Before proceeding with the label objects, print out whatever we've found so 
# far:

printf OUT "\n            SUMMARY OF \"$infile\", $pds_version_id LABEL\n\n";
printf OUT " Data Set ID: $dsid\n" if ($dsid);
if ($datafile)
  { if ($datafile =~ /,/)
      { printf OUT "  Data Files: $datafile\n"; }
    else
      { printf OUT "   Data File: $datafile\n"; }
  }
printf OUT "     RECORDS: $file_records\n" if ($file_records);
printf OUT " RECORD TYPE: $record_type\n" if ($record_type);
printf OUT "RECORD BYTES: $record_bytes\n" if ($record_bytes);
printf OUT "  INSTRUMENT: $instrument_name\n" if ($instrument_name  && 
                                                     $instrument_name ne "N/A");
printf OUT "      TARGET: $target_name\n" if ($target_name  &&  
                                                 $target_name ne "N/A");


# Print a small banner for the OBJECT definitions:

printf OUT "\n.............OBJECT Definitions.............\n\n";

# At this point $line still contains the first line of an object reference.  We 
# start a loop to process all the OBJECT definitions in the label:

$indent = "";

while ($line  &&  $line ne "END")
  { if ($line =~ /OBJECT\s*=/)
      { $object = &getval_symbol($line);

        # Call the appropriate object subroutine:

        &call_object(*LBL,$indent,$object);
      }

    $line = &nextPDSODL(*LBL);

  }

# Done:

close(LBL);
close(OUT);

#==============================================================================

sub getval_symbol

  # Routine to retrieve the value part of the line as a single symbol.
  # This is basically a step-saver for values expected to be simple symbols.
  # Quotes are removed, all letters are forced to upper case and blanks are 
  # replaced by underscores.

  { local ($line) = $_[0];   # The input line
    local ($val);

    $val = &getval($line);
    $val =~ s/["']//g;
    $val =~ s/\s+/_/g;
    $val =~ tr/a-z/A-Z/;

    return $val;
  }

#==============================================================================
#==============================================================================
#==============================================================================
# This section contains Perl subroutine for processing individual objects.  
#
# 05 Jun 1997, acr.
#
#==============================================================================

sub call_object

  # Routine to call the appropriate object subroutine for the input object.

  { local (*LBL,$indent,$object) = @_;  # File handle, indentation, object type

    # Now branch on the object type.  Note that object types may be modified,
    # so we check for a match at the end of the string only:

    if ($object =~ /COLUMN$/)
      { &do_column(*LBL,$indent); }
    elsif ($object =~ /DOCUMENT$/)
      { &do_document(*LBL,$indent); }
    elsif ($object =~ /FILE$/)
      { &do_file(*LBL,$indent); }
    elsif ($object =~ /HEADER$/)
      { &do_header(*LBL,$indent); }
    elsif ($object =~ /HISTOGRAM$/)
      { &do_histogram(*LBL,$indent); }
    elsif ($object =~ /IMAGE$/)
      { &do_image(*LBL,$indent); }
    elsif ($object =~ /QUBE$/)
      { &do_qube(*LBL,$indent); }
    elsif ($object =~ /SPICE_KERNEL$/)
      { &do_spice_kernel(*LBL,$indent); }
    elsif ($object =~ /TABLE$/    ||      # All TABLE variants are handled in 
           $object =~ /SERIES$/   ||      # the same subroutine.
           $object =~ /SPECTRUM$/ )
      { &do_table(*LBL,$indent,$object); }
    elsif ($object =~ /TEXT$/)
      { &do_text(*LBL,$indent); }
    else
      { &do_unknown_object(*LBL,$indent,$object); }

    # That's it:

    return 1;
  }

#==============================================================================

sub read_structure

  # Routine to run processing through a ^STRUCTURE incuded file.  Only OBJECT
  # definition lines are processed.  Nested ^STRUCTUREs are not recognized.

  { local ($structure,$indent) = @_;  # File name and indentation
    local ($line);
    local ($newobj);

    # First, open the file:

    open (STR,$structure) ||
               die "lblsumm: Unable to open structure file $structure";

    # Now read through the OBJECT definitions in the structure file:

    $line = &nextPDSODL(*STR);
    while ($line  &&  $line ne "END")
      { if ($line =~ /^OBJECT\s*=/)
          { $newobj = &getval_symbol($line);  # Get new object

            # Call the object subroutine:

            &call_object(*STR,$indent,$newobj);
          }

        $line = &nextPDSODL(*STR); 
      }

    # Done with the structure file:

    close(STR);

    # And we're outta here:

    return 1;
  }


#==============================================================================

sub do_unknown_object

  # Routine to run through an unknown object.  It attempts to collect a NAME
  # field, but otherwise just loops through looking for sub-objects it 
  # recognizes and STRUCTURE Pointers to expand.

  { local (*LBL,$indent,$object) = @_;   # Input handle, indentation, obj type
    local ($line);
    local ($newind, $newobj);
    local ($name);
    local ($interchange_format);

    local ($structure);    # Structure file name

    printf OUT "$indent";
    printf OUT "Unknown Object: $object\n";

    $line = &nextPDSODL(*LBL);
    while ($line !~ /^END_OBJECT/  &&  $line ne "END")
      { if ($line =~ /^NAME/)
          { $name = &getval_symbol($line);
            printf OUT "$indent   Name: $name\n";
          }
        elsif ($line =~ /^INTERCHANGE_FORMAT/)
          { $interchange_format= &getval_symbol($line); 
            printf OUT "$indent Format: $interchange_format\n";
          }

        # For sub-objects, call the appropriate routine:

        elsif ($line =~ /^OBJECT\s*=/)
          { $newind = "$indent     ";         # Increase indentation
            $newobj = &getval_symbol($line);  # Get new object

            # Call the object subroutine:

            &call_object(*LBL,$newind,$newobj);
          }

        # If we encounter a structure pointer, we must open the file and
        # process the OBJECT definitions within:

        elsif ($line =~ /^\^STRUCTURE/)
          { # First, retrieve the file name from the pointer and open:

            $structure = &getval_symbol($line);
            $structure =~ tr/A-Z/a-z/;

            # Now call the structure subroutine:

            &read_structure($structure,"$indent    ");

          }
 
        $line = &nextPDSODL(*LBL);
      }

    # And we're done:

    return 1;

  }            

#==============================================================================

sub do_column

  # Subroutine to handle COLUMN objects.  STRUCTURE pointers and sub-objects
  # are not procesed.
  
  { local (*LBL,$indent) = @_;  # Input handle, indentation string
    local ($line);
    local ($name,$data_type,$start_byte,$bytes);
    local ($items, $item_bytes);
    local ($format, $unit);
    local ($sampling_parameter_name);
    local ($maximum,$minimum);
    local ($sequence_number);

    $line = &nextPDSODL(*LBL);
    while ($line !~ /^END_OBJECT/  &&  $line ne "END")
      { if ($line =~ /^NAME/)
          { $name = &getval_symbol($line); }
        elsif ($line =~ /^DATA_TYPE/)
          { $data_type = &getval_symbol($line); }
        elsif ($line =~ /^START_BYTE/)
          { $start_byte = &getval($line); }
        elsif ($line =~ /^BYTES/)
          { $bytes = &getval($line); }
        elsif ($line =~ /^ITEMS/)
          { $items = &getval($line); }
        elsif ($line =~ /^ITEM_BYTES/)
          { $item_bytes = &getval($line); }
        elsif ($line =~ /^FORMAT/)
          { $format = &getval_symbol($line); }
        elsif ($line =~ /^UNIT/)
          { $unit = &getval_symbol($line); }
        elsif ($line =~ /^SAMPLING_PARAMETER_NAME/)
          { $sampling_parameter_name = &getval_symbol($line); }
        elsif ($line =~ /MAXIMUM$/)
          { $maximum = &getval($line); }
        elsif ($line =~ /MINIMUM$/)
          { $minimum = &getval($line); }
        elsif ($line =~ /^SEQUENCE_NUMBER/)
          { $sequence_number = &getval($line); }

        $line = &nextPDSODL(*LBL);
      }

    # Print out the collected values:

    printf OUT "\n$indent";
    printf OUT "OBJECT: COLUMN";
    printf OUT " $sequence_number:" if ($sequence_number);
    printf OUT " \"$name\"\n";
    printf OUT "$indent Start byte: $start_byte\n";
    printf OUT "$indent  Data type: $data_type";
    if ($items)
      { printf OUT "*$item_bytes X $items\n"; }
    else
      { printf OUT "*$bytes\n"; }
    printf OUT "$indent     Format: $format\n"  if ($format);
    printf OUT "$indent      Units: $unit\n"    if ($unit);
    printf OUT "$indent    Maximum: $maximum\n" if ($maximum);
    printf OUT "$indent    Minimum: $minimum\n" if ($minimum);
    printf OUT "$indent Sampled on: $sampling_parameter_name\n" 
                    if ($sampling_parameter_name);
    printf "\n";

    # And we're done:

    return 1;
  }

#==============================================================================

sub do_document

  # Subroutine to handle DOCUMENT objects.  No STRUCTURE pointers are 
  # recognized here.  The last line read was the OBJECT definition line.

  { local (*LBL,$in) = @_;          # Input file, indentation string
    local ($interchange_format);
    local ($document_name);
    local ($document_topic_type);
    local ($document_format);
    local ($publication_date);
    local ($encoding_type);
    local ($files);
    local ($line);

    $line = &nextPDSODL(*LBL);
    while ($line !~ /END_OBJECT/  &&  $line ne "END")
      { if ($line =~ /^DOCUMENT_NAME/)
          { $document_name = &getval($line); }
        elsif ($line =~ /^DOCUMENT_TOPIC_TYPE/)
          { $document_topic_type = &getval_symbol($line);
            undef $document_topic_type if ($document_topic_type eq "N/A");
          }
        elsif ($line =~ /^DOCUMENT_FORMAT/)
          { $document_format = &getval_symbol($line); }
        elsif ($line =~ /^PUBLICATION_DATE/)
          { $publication_date = &getval($line); }
        elsif ($line =~ /^ENCODING_TYPE/)
          { $encoding_type = &getval_symbol($line); }
        elsif ($line =~ /^FILES/)
          { $files = &getval($ine); }

        # Next line:

        $line = &nextPDSODL(*LBL);
      }

    # Now print the collected values:

    printf OUT "\n$indent";
    printf OUT "OBJECT: $document_format DOCUMENT, $document_name\n";
    printf OUT "$indent        ($document_topic_type)\n"
                                                      if ($document_topic_type);
    printf OUT "$indent     Date: $publication_date\n" if ($publication_date);
    printf OUT "$indent    Files: $files\n"            if ($files);
    printf OUT "$indent Encoding: $encoding_type\n"    if ($encoding_type);
    printf OUT "\n";

    # Done:

    return 1;
  }
    

#==============================================================================

sub do_file

  # Subroutine to handle FILE objects.  STRUCTURE pointers and sub-objects are
  # handled.

  { local (*LBL,$indent) = @_;   # Input handle, indentation string
    local ($record_type);
    local ($file_name);
    local ($record_bytes);
    local ($file_records);
    local ($object);
    local ($structure);     # STRUCTURE file name

    $line = &nextPDSODL(*LBL);
    while ($line !~ /END_OBJECT/  &&  $line ne "END")
      { if ($line =~ /^RECORD_TYPE/)
          { $record_type = &getval_symbol($line); }
        elsif ($line =~ /^FILE_NAME/)
          { $file_name = &getval($line);
            $file_name =~ s/["']//g;
            $file_name =~ tr/A-Z/a-z/;
          }
        elsif ($line =~ /^FILE_RECORDS/)
          { $file_records = &getval($line); }
        elsif ($line =~ /^RECORD_BYTES/)
          { $record_bytes = &getval($line); }

        # Handle sub-objects:

        elsif ($line=~ /^OBJECT\s*=/)
          { $object = &getval_symbol($line); 
            &call_object(*LBL,"$indent    ",$object);
          }

        # Handle ^STRUCTURE pointers:

        elsif ($line =~ /^\^STRUCTURE/)
          { $structure = &getval_symbol($line);
            $structure =~ tr/A-Z/a-z/;
            &read_structure($structure,"$indent    ");
          }

        $line = &nextPDSODL(*LBL);
      }

    # Now print the collected values:

    printf OUT "\n$indent";
    printf OUT "OBJECT: FILE";
    printf OUT " \"$file_name\"" if ($file_name);
    printf OUT "\n";
    printf OUT "$indent     Records: $file_records\n"        if ($file_records);
    printf OUT "$indent Record Size: $record_bytes bytes\n"  if ($record_bytes);
    printf OUT "$indent              $record_type records\n" if ($record_type);
    printf OUT "\n";

    # And we're done:

    return 1;
  }


#==============================================================================

sub do_header

  # Subroutine to handle HEADER objects.  STRUCTURE pointers are not processed;
  # no allowed sub-objects.

  { local (*LBL,$indent) = @_;   # Input handle, indentation string
    local ($bytes);
    local ($header_type);
    local ($records);
    local ($interchange_format);
    local ($line);

    $line = &nextPDSODL(*LBL);
    while ($line !~ /END_OBJECT/  &&  $line ne "END")
      { if ($line =~ /^BYTES/)
          { $bytes = &getval($line); }
        elsif ($line =~ /^HEADER_TYPE/)
          { $header_type = &getval_symbol($line); }
        elsif ($line =~ /^RECORDS/)
          { $records = &getval($line); }
        elsif ($line =~ /^INTERCHANGE_FORMAT/)
          { $interchange_format = &getval_symbol($line); }

        $line = &nextPDSODL(*LBL);
      }

    # Print the collected values:

    printf OUT "\n$indent";
    printf OUT "OBJECT: $header_type HEADER\n";
    printf OUT "$indent Size: $bytes bytes";
    printf OUT ", $records" if ($records);
    printf OUT " ($interchange_format file)" if ($interchange_format);
    printf OUT "\n\n";

    # And we're done:

    return 1;
  }

#==============================================================================

sub do_histogram

  # Subroutine to handle HISTOGRAM objects.  STRUCTURE pointers and sub-objects
  # are not allowed.
  
  { local (*LBL,$indent) = @_;  # Input handle, indentation string
    local ($line);
    local ($items);
    local ($data_type);
    local ($item_bytes);
    local ($offset);
    local ($scaling_factor);
    local ($interchange_format);

    $line = &nextPDSODL(*LBL);
    while ($line !~ /^END_OBJECT/  &&  $line ne "END")
      { if ($line =~ /^ITEMS/)
          { $items = &getval($line); }
        elsif ($line =~ /^DATA_TYPE/)
          { $data_type = &getval_symbol($line); }
        elsif ($line =~ /^ITEM_BYTES/)
          { $item_bytes = &getval($line); }
        elsif ($line =~ /^OFFSET/)
          { $offset = &getval($line); }
        elsif ($line =~ /^SCALING_FACTOR/)
          { $scaling_factor = &getval($line); }
        elsif ($line =~ /^INTERCHANGE_FORMAT/)
          { $interchange_format = &getval_symbol($line); }

        $line = &nextPDSODL(*LBL);
      }

    # Print out the collected values:

    printf OUT "\n$indent";
    printf OUT "OBJECT: HISTOGRAM";
    printf OUT " ($interchange_format File)" if ($interchange_format);
    printf OUT "\n";
    printf OUT "$indent           Bins: $items  ($data_type*$item_bytes)\n";
    printf OUT "$indent         Offset: $offset\n" if ($offset);
    printf OUT "$indent Scaling Factor: $scaling_factor\n" if ($scaling_factor);
    printf "\n";

    # And we're done:

    return 1;
  }

#==============================================================================

sub do_image

  # Subroutine to handle IMAGE objects.  STRUCTURE pointers and sub-objects are
  # ignored.
  
  { local (*LBL,$indent) = @_;  # Input handle, indentation string
    local ($line);
    local ($lines, $samples, $sample_type, $sample_bits);
    local ($bands, $band_storage_type);
    local ($encoding_type);

    $line = &nextPDSODL(*LBL);
    while ($line !~ /^END_OBJECT/  &&  $line ne "END")
      { if ($line =~ /^LINES/)
          { $lines = &getval($line); }
        elsif ($line =~ /^LINE_SAMPLES/)
          { $samples = &getval($line); }
        elsif ($line =~ /^SAMPLE_TYPE/)
          { $sample_type = &getval_symbol($line); }
        elsif ($line =~ /^SAMPLE_BITS/)
          { $sample_bits = &getval($line); }
        elsif ($line =~ /^BANDS/)
          { $bands = &getval($line); }
        elsif ($line =~ /^BAND_STORAGE_TYPE/)
          { $band_storage_type = &getval_symbol($line); }
        elsif ($line =~ /^ENCODING_TYPE/)
          { $encoding_type = &getval_symbol($line); }

        $line = &nextPDSODL(*LBL);
      }

    # Print out the collected values:

    printf OUT "\n$indent";
    printf OUT "OBJECT: ";
    printf OUT "$encoding_type encoded " if ($encoding_type  &&
                                             $encoding_type ne "N/A");
    printf OUT "IMAGE\n";
    printf OUT "$indent Size: $lines lines X $samples samples\n";
    printf OUT "$indent       $sample_type*%d pixels\n", $sample_bits/8;
    printf OUT "$indent       $bands bands" if ($bands);
    printf OUT ", stored $band_storage_type\n" if ($band_storage_type);
    printf "\n";

    # And we're done:

    return 1;
  }

#==============================================================================

sub do_qube

  # Subroutine to handle QUBE objects.  STRUCTURE pointers and sub-objects
  # are not allowed.
  
  { local (*LBL,$indent) = @_;  # Input handle, indentation string
    local ($line);
    local ($axes, $axis_name);
    local ($core_items, $core_item_bytes, $core_item_type);
    local ($core_name);
    local ($val);

    $line = &nextPDSODL(*LBL);
    while ($line !~ /^END_OBJECT/  &&  $line ne "END")
      { if ($line =~ /^AXES/)
          { $axes = &getval($line); }
        elsif ($line =~ /^AXIS_NAME/)
          { $axis_name = &getval($line);
            $axis_name =~ s/[()"']//g;
            $axis_name =~ s/\s*,\s*/, /g;
          }
        elsif ($line =~ /^CORE_ITEMS/)
          { $core_items = &getval($line); 
            $core_items =~ s/[()]//g;
            $core_items =~ s/\s*,\s*/, /g;
          }
        elsif ($line =~ /^CORE_ITEM_BYTES/)
          { $core_item_bytes = &getval($line); }
        elsif ($line =~ /^CORE_ITEM_TYPE/)
          { $core_item_type = &getval_symbol($line); }
        elsif ($line =~ /^CORE_NAME/)
          { $core_name = &getval_symbol($line); }

        $line = &nextPDSODL(*LBL);
      }

    # Print out the collected values:

    printf OUT "\n$indent";
    printf OUT "QUBE\n";
    printf OUT "$indent  Core Data: $core_name\n" if ($core_name);
    printf OUT "$indent     $axes Axes: $core_items\n";
    printf OUT "$indent             $axis_names\n";
    printf OUT "$indent  Data Type: $core_item_type*$core_item_bytes\n";
    printf "\n";

    # And we're done:

    return 1;
  }

#==============================================================================

sub do_spice_kernel

  # Subroutine to handle SPICE_KERNEL objects.  STRUCTURE pointers and 
  # sub-objects are not allowed.
  
  { local (*LBL,$indent) = @_;  # Input handle, indentation string
    local ($line);
    local ($kernel_type);
    local ($interchange_format);

    $line = &nextPDSODL(*LBL);
    while ($line !~ /^END_OBJECT/  &&  $line ne "END")
      { if ($line =~ /^KERNEL_TYPE/)
          { $kernel_type = &getval_symbol($line); }
        elsif ($line =~ /^INTERCHANGE_FORMAT/)
          { $interchange_format =~ &getval_symbol($line); }

        $line = &nextPDSODL(*LBL);
      }

    # Print out the collected values:

    printf OUT "\n$indent";
    printf OUT "SPICE Kernel\n";
    printf OUT "$indent   Type: $kernel_type\n";
    printf OUT "$indent Format: $interchange_format\n";
    printf "\n";

    # And we're done:

    return 1;
  }

#==============================================================================

sub do_table

  # Subroutine to handle TABLE objects.  STRUCTURE pointers and sub-objects
  # are handled.
  
  { local (*LBL,$indent,$table) = @_;  # Input handle, indentation, table type
    local ($line);
    local ($interchange_format);
    local ($rows,$columns,$row_bytes);
    local ($name,$table_storage_type);
    local ($index_type);
    local ($sp_name,$sp_unit,$sp_interval);
    local ($no_banner_yet);            # Flag for output production
    local ($object);
    local ($structure);                # ^STRUCTURE file name

    $no_banner_yet = 1;

    $line = &nextPDSODL(*LBL);
    while ($line !~ /^END_OBJECT/  &&  $line ne "END")
      { # We need to print the TABLE information before processing COLUMNS
        # (or ^STRUCTURE pointers), so check for this first:

        if ($no_banner_yet  &&  
            ($line =~ /^OBJECT\s*=/  ||  $line =~ /^\^STRUCTURE/))
          { printf OUT "$indent";
            printf OUT "OBJECT: $interchange_format $table";
            printf OUT " \"$name\"" if ($name);
            printf OUT ", $index_type Index" if ($index_type);
            printf OUT "\n";
            printf OUT "$indent    Size: $rows Rows X $row_bytes bytes\n";
            printf OUT "$indent Columns: $columns\n";
            printf OUT "$indent In $table_storage_type order\n"
                                                       if ($table_storage_type);
            printf OUT "$indent Sampled on $sp_name"   if ($sp_name);
            if ($sp_unit  &&  $sp_interval)
              { printf OUT " at $sp_interval $sp_unit intervals\n"; }
            else
              { printf OUT "\n" if ($sp_name);}
            printf OUT "\n";

            $no_banner_yet = 0;
          }

        # Now check for keywords to match:

        if ($line =~ /^INTERCHANGE_FORMAT/)
          { $interchange_format = &getval_symbol($line); }
        elsif ($line =~ /^ROWS/)
          { $rows = &getval($line); }
        elsif ($line =~ /^COLUMNS/)
          { $columns = &getval($line); }
        elsif ($line =~ /^ROW_BYTES/)
          { $row_bytes = &getval($line); }
        elsif ($line =~ /^NAME/)
          { $name = &getval_symbol($line); }
        elsif ($line =~ /^TABLE_STORAGE_TYPE/)
          { $table_storage_type = &getval_symbol($line);
            $table_storage_type =~ s/_/ /g;
          }
        elsif ($line =~ /^INDEX_TYPE/)
          { $index_type = &getval_symbol($line); }

        # Handling sub-objects:

        elsif ($line =~ /^OBJECT\s*=/)
          { $object = &getval_symbol($line);
            &call_object(*LBL,"$indent    ",$object);
          }

        # Handling ^STRUCTUREs:

        elsif ($line =~ /^\^STRUCTURE/)
          { $structure = &getval_symbol($line);
            $structure =~ tr/A-Z/a-z/;
            &read_structure($structure,"$indent    ");
          }

        $line = &nextPDSODL(*LBL);
      }

    printf OUT "\n";

    # And we're done:

    return 1;
  }

#==============================================================================

sub do_text

  # Subroutine to handle TEXT objects.  STRUCTURE pointers and sub-objects
  # are not allowed.
  
  { local (*LBL,$indent) = @_;  # Input handle, indentation string
    local ($line);
    local ($publication_date);
    local ($interchange_format);

    $line = &nextPDSODL(*LBL);
    while ($line !~ /^END_OBJECT/  &&  $line ne "END")
      { if ($line =~ /^PUBLICATION_DATE/)
          { $publication_date = &getval($line); }
        elsif ($line =~ /^INTERCHANGE_FORMAT/)
          { $interchange_format = &getval_symbol($line); }

        $line = &nextPDSODL(*LBL);
      }

    # Print out the collected values:

    printf OUT "\n$indent";
    printf OUT "$interchange_format " if ($interchange_format);
    printf OUT "TEXT\n";
    printf OUT "$indent PublicationDate: $publication_date\n";
    printf "\n";

    # And we're done:

    return 1;
  }

#==============================================================================
