#!/usr/bin/perl

# mkcumidx
#
#  Purpose:  To combine index files created by the mkvolidx tool, to create a
#            cumulative index file.
#
#  Format:   mkcumidx [OPTIONS] <volume_1> <volume_2> {<volume_i>}
#
#    <volume_#>   Is the volume path for the volume.  There must be at least 
#                 two volume paths provided.
#
#  Options:
#     -h          Help - shows brief description of options
#     -o <path>   Will output cumindex table and label file to a specified path
#                 instead of in each [VOLUME]/index directory.
#     -v          Will show what actions are being taken by the utility.
#

# Created by Tilden F Barnes IV
# Completed: 2/28/2007
#
# Updates:
#   20070301 - updated calculations for cumulative START_TIME and STOP_TIME 
#              to take into account "UNK", "N/A", "NULL" and empty values.
#   20070306 - added DOY - YMD converting functions.  Also updated how to 
#              calculate START_TIME and STOP_TIME if their some times are in 
#              both DOY and YMD format.  If there is a mix, then they will be 
#              converted into YMD format.


#==============================================================================

# Utilities/packages used

use File::Copy;
use Getopt::Std;
use Cwd 'abs_path';
use ODL; # created by Andrew Decore

# ==============================
# functions internal to this program

sub printv($);         # prints status report message if option -v used
sub convert_to_doy($); # converts YYYY-MM-DD to YYYY-DDD
sub convert_to_ymd($); # converts YYYY-DDD to YYYY-MM-DD
sub is_leap_year($);   # detects if a year is a leap year

# ==============================
# Constants

$delimiter = ",";              # Column delimiter for tables
$dlength = length($delimiter); # Length of the delimiter (# of chars)
$tmpdir = $ENV{"TMPDIR"};      # gets the location to put temporary files


# ==============================
# program values

@volume_list = (); # will contain of list of the locations of each volume

# index label KEYWORD values.  
#    (Array index corresponds with index in @volume_list)
@file_records = (undef);
@record_bytes = (undef);
@start_time = (undef);
@stop_time = (undef);

# index.lbl OBJECT = INDEX_TABLE values.
@rows      = (undef);
@row_bytes = (undef);
@columns   = (undef);

# index.lbl OBJECT = COLUMN values.
#   format: $array[volume#,column#]
%i_start_byte  = undef; # START_BYTE keyword
%i_bytes       = undef; # BYTES keyword

# cumindex.lbl KEYWORD values
$file_records = undef;
$record_bytes = undef;
$start_time   = undef;
$stop_time    = undef;

# cumindex.lbl OBJECT = INDEX_TABLE values.
$rows      = undef;
$row_bytes = undef;
$columns   = undef;

# cumindex.lbl OBJECT = COLUMN (level 2) values.  Array index is column number.
@name        = (0); # NAME keyword
@start_byte  = (0); # START_BYTE keyword
@bytes       = (0); # BYTES keyword
@data_type   = (0); # DATA_TYPE keyword
@description = (0); # DESCRIPTION keyword


#==============================================================================
# Gather all arguments and options from utility declaration

$options = "ho:v";
getopts($options);

if ($opt_h) # Help option evoked
{
    print "usage: % mkcumidx [ OPTIONS ] <volume_1> <volume_2> ".
	            "{ <volume_i> }\n".
	  "OPTIONS:\n".
	  "  -h         Help - shows brief description of options\n".
	  "  -o <path>  Will output cumindex table and label file to a spec".
                    "ified path\n".
	  "             instead of in each '[VOLUME]/index/' directory.\n".
	  "  -v         Will show what actions are being taken by the utility".
	            ".\n";
    exit;
}

$die = 0; # to count how many errors occurred before finally dieing
if (@ARGV < 2 || ! @ARGV)
{
    # There should be at least two volumes to combine otherwise utility die.
    die "usage: % mkcumidx [ -$options ] volume_1 volume_2 { volume_i }\n";
}
else
{
    # Check to see if each provided volume is a real directory, and if 
    # the index files therein are accessible and the folders are writable 
    # otherwise the utility dies.  If everything is okay then the @vol_lst
    # array gets populated with the volume name.
    foreach my $volume (@ARGV)
    {
	$volume = abs_path($volume);

	if (-d $volume)
	{
	    if (! -e $volume."/index/index.tab") 
	    { 
		print "ERROR: \"$volume/index/index.tab\" does not exist.\n";
		$die++;
	    }
	    elsif (! -r $volume."/index/index.tab") 
	    { 
		print "ERROR: \"$volume/index/index.tab\" cannot be read.\n";
		$die++;
	    }
	    if (! -e $volume."/index/index.lbl")
	    {
		print "ERROR: \"$volume/index/index.lbl\" does not exist.\n";
		$die++;
	    }
	    elsif (! -r $volume."/index/index.lbl")
	    {
		print "ERROR: \"$volume/index/index.lbl\" cannot be read.\n";
		$die++;
	    }
	    if (! -w $volume."/index/" && ! $opt_o)
	    {
		print "ERROR: \"$volume/index/\" cannot be written to.\n";
		$die++;
	    }
	}
	else { print "ERROR: \"$volume\" is not a directory.\n"; $die++; }
	push(@volume_list,abs_path($volume));
    }
} # end review @ARGV

# Verifies if the -o option's directory is valid and can be written to.
if ($opt_o)
{
    if (! -d $opt_o)
    {
	print "ERROR: Option -o \"$opt_o\" must be a directory.\n";
	$die++;
    }
    elsif (! -w $opt_o)
    {
	print "ERROR: Option -o \"$opt_o\" has no write permissions.\n";
	$die++;
    }
    else { $opt_o = abs_path($opt_o); }
} # end verify -o option

# If ERRORS Then utility dies
if ($die) { die "mkcumidx: Program terminated with $die ERROR(S).\n"; }

#==============================================================================
# Find best values from all index labels for all index tables

for (my $vol_num = 0;$vol_num<=$#volume_list;$vol_num++)
{
    my $filename = $volume_list[$vol_num]."/index/index.lbl";
    printv "Reading \"$filename\"\n";

    # create links to file
    my $file = ODL::create($filename);
    my $idx_tbl = $file->get_statement("INDEX_TABLE");

    # Store FILE_RECORDS
    printv "Reading in FILE_RECORDS value\n";
    $file_records[$vol_num] = $file->get("FILE_RECORDS");
    if ($file_records[$vol_num] <= 0)
    {
	die "ERROR: Invalid or missing FILE_RECORDS value.\n";
    }

    # Store RECORD_BYTES
    printv "Reading in RECORD_BYTES value\n";
    $record_bytes[$vol_num] = $file->get("RECORD_BYTES");
    if ($record_bytes[$vol_num] <= 0)
    {
	die "ERROR: Invalid or missing RECORD_BYTES value.\n";
    }

    # Store START_TIME value
    printv "Reading in START_TIME value\n";
    my $time = $file->get_statement("START_TIME");
    $start_time[$vol_num] = $file->get("START_TIME");
    if (! $start_time[$vol_num])
    {
	printv "No START_TIME value found.\n";
    }
    elsif ($start_time[$vol_num] =~ /^\"(N\/A|UNK|NULL)\"$/) {;}
    elsif (length($start_time[$vol_num]) < 8)
    {
	die "ERROR: Invalid START_TIME value.\n";
    }

    # Store STOP_TIME value
    printv "Reading in STOP_TIME value\n";
    $stop_time[$vol_num] = $file->get("STOP_TIME");
    if (! $stop_time[$vol_num])
    {
	printv "No STOP_TIME value found.\n";
    }
    elsif ($stop_time[$vol_num] =~ /^\"(N\/A|UNK|NULL)\"$/) {;}
    elsif (length($stop_time[$vol_num]) < 8)
    {
	die "ERROR: Invalid or missing STOP_TIME value.\n";
    }

    # Store and Calculate INDEX_TABLE->ROWS value
    printv "Reading and calculating INDEX_TABLE->ROWS value\n";
    $rows[$vol_num] = $idx_tbl->get("ROWS");
    if ($rows[$vol_num] <= 0)
    {
	die "ERROR: Invalid or missing INDEX_TABLE->ROWS value.\n";
    }
    else { $rows += $rows[$vol_num]; }

    # Store INDEX_TABLE->ROW_BYTES value
    printv "Reading INDEX_TABLE->ROW_BYTES value\n";
    $row_bytes[$vol_num] = $idx_tbl->get("ROW_BYTES");
    if ($row_bytes[$vol_num] <= 0)
    {
        print "ERROR: Invalid or missing INDEX_TABLE->ROW_BYTES value.\n";
    }

    # Check total number of COLUMNS
    printv "Reading in and checking COLUMNS value\n";
    $columns[$vol_num] = $idx_tbl->get("COLUMNS");
    if ($columns[$vol_num] <= 0)
    {
	die "ERROR: Invalid or missing COLUMNS value.\n";
    }
    elsif ($columns && $columns != $columns[$vol_num])
    {
	die "ERROR: Each index table must have the same number of COLUMNS.\n";
    }
    elsif (! $columns) 
    { 
	$columns = $columns[$vol_num]; 
	printv "Set COLUMNS = $columns\n";
    }

    # Start taking in values from INDEX_TABLE->OBJECT=COLUMN
    for (my $col_num = 1;$col_num<=$columns;$col_num++)
    {
	printv "Reading Column $col_num\n";
	# create link to a COLUMN $col_num object
	my $column = $idx_tbl->get_statement("COLUMN", $col_num);

	# Read and compare NAME value
	printv "Reading in and checking NAME value\n";
	my $name = $column->get_statement("NAME")->get_value_object();
	if (! $name->value_is_string())
	{
	    die "ERROR: Invalid or missing NAME value.\n";
	}
	elsif ($name[$col_num] && $name[$col_num] ne $name->get_value_noq())
	{
	    die "ERROR: Each column NAME must be the same for each ".
		"index label.\n";
	}
	elsif (! $name[$col_num]) 
	{ 
	    $name[$col_num] = $name->get_value_noq();
	    printv "Set cumulative NAME = \"$name[$col_num]\"\n";
	}
	# end read/calculate NAME

	# Read in START_BYTE value
	printv "Reading in START_BYTE value\n";
	$i_start_byte{"$vol_num,$col_num"} = $column->get("START_BYTE");
	if ($i_start_byte{"$vol_num,$col_num"} <= 0)
	{
	    die "ERROR: Invalid or missing START_BYTE value.\n";
	}
	# end read START_BYTE

	# Read in and calculate BYTES value
	printv "Reading in and calculating BYTES value\n";
	$i_bytes{"$vol_num,$col_num"} = $column->get("BYTES");
	if ($i_bytes{"$vol_num,$col_num"} <= 0)
	{
	    die "ERROR: Invalid or missing BYTES value.\n";
	}
	elsif ($bytes[$col_num] < $i_bytes{"$vol_num,$col_num"})
	{
	    $bytes[$col_num] = $i_bytes{"$vol_num,$col_num"};
	    printv "Set cumulative BYTES = $bytes[$col_num]\n";
	} 
	# end read/calculate BYTES

	# Read in and calculate DATA_TYPE value
	printv "Reading in and calculating DATA_TYPE value\n";
	my $data_type = 
	    $column->get_statement("DATA_TYPE")->get_value_object();
	if (! $data_type->value_is_string())
	{
	    die "ERROR: Invalid or missing NAME value.\n";
	}
	elsif ($data_type[$col_num] =~ /^(TIME|DATE)$/ && 
	       $data_type->get_value_noq() =~ /^(TIME|DATE)$/)
	{
	    if ($data_type[$col_num] eq "DATE" &&
		$data_type->get_value_noq() eq "TIME")
	    {
		$data_type[$col_num] = $data_type->get_value_noq();
		printv "Set cumulative DATA_TYPE = \"$data_type[$col_num]\"\n";
	    }
	}
	elsif ($data_type[$col_num] && 
	       $data_type[$col_num] ne $data_type->get_value_noq())
	{
	    die "ERROR: Each column DATA_TYPE must be the same for each ".
		"index label.\n";
	}
	elsif (! $data_type[$col_num]) 
	{ 
	    $data_type[$col_num] = $data_type->get_value_noq();
	    printv "Set cumulative DATA_TYPE = \"$data_type[$col_num]\"\n";
	}
	# end read/calculate DATA_TYPE

	# Read in DESCRIPTION
	if (! $description[$col_num])
	{
	    printv "Reading in DESCRIPTION value\n";
	    my $desc=$column->get_statement("DESCRIPTION")->get_value_object();
	    $description[$col_num] = 
		$column->get_statement("DESCRIPTION")->get_value_lined_noq();
	    if ( ! $description[$col_num] || $description[$col_num] eq "")
	    {
		printv "No DESCRIPTION value provided.\n";
		$description[$col_num] = undef;
	    }
	    elsif (! $desc->value_is_string())
	    {
		die "ERROR: Invalid DESCRIPTION value.\n";
	    }
	    else
	    {
		printv "Set DESCRIPTION to: \n".
		    "            \"$description[$col_num]\"\n";
	    }
	} # end reading in DESCRIPTION

	printv "Finished reading Column $col_num\n";
    } # end taking in values from INDEX_TABLE->OBJECT=COLUMN for loop

    printv "Finished reading \"$filename\"\n";
} # end taking in values from index labels for loop

#====================
# Calculate remaining cumulative values

# Calculate FILE_RECORDS from ROWS
$file_records = $rows;
printv "Calculated cumulative FILE_RECORDS = ROWS = $rows\n";

# Calculate START_BYTE
if ($data_type[1] eq "CHARACTER") { $start_byte[1] = 2; }
else { $start_byte[1] = 1; }
printv "Calculated Column 1 cumulative START_BYTE = $start_byte[1]\n";
for (my $col_num = 2;$col_num<=$columns;$col_num++)
{
    $start_byte[$col_num] = $start_byte[$col_num-1] + $bytes[$col_num-1] + 
	                    $dlength;
    if ($data_type[$col_num-1] eq "CHARACTER") { $start_byte[$col_num]++; }
    if ($data_type[$col_num] eq "CHARACTER") { $start_byte[$col_num]++; }
    printv "Calculated Column $col_num cumulative START_BYTE = ".
	   "$start_byte[$col_num]\n";

} # end calculate START_BYTE for loop

# Calculate ROW_BYTES and RECORD_BYTES
$row_bytes = $start_byte[$columns] + $bytes[$columns] + 1; # +1 for "\r"
if ($data_type[$columns] eq "CHARACTER") { $row_bytes++; }
$record_bytes = $row_bytes;
printv "Calculated cumulative RECORD_BYTES = ROW_BYTES = $row_bytes\n";

# calculate START_TIME and STOP_TIME
{
    # calculate cumulative START_TIME
    my @tmp_start_time = sort(@start_time);
    if ($tmp_start_time[$#tmp_start_time] !~ /^\"(N\/A|UNK|NULL)\"$/)
    {
	@tmp_start_time = grep(! /^\"(N\/A|UNK|NULL)\"$/, @tmp_start_time);
    }
    elsif ($tmp_start_time[$#tmp_start_time] =~ /^\"(N\/A|UNK|NULL)\"$/)
    {
	@tmp_start_time = grep(!/^\"$1\"$/,@tmp_start_time);
	if ($tmp_start_time[$#tmp_start_time] =~ /^.+$/)
	{
	    @tmp_start_time = ("");
	}
	else { @tmp_start_time = ("\"$1\""); }
    }

    if (grep(/^\d{4}-\d{3}/,@tmp_start_time) && 
	grep(/^\d{4}-\d\d-\d\d/,@tmp_start_time))
    {
	my @fixed_start_time;
	foreach my $temp_time (@tmp_start_time)
	{
	    if ($temp_time =~ /^\d{4}-\d{3}/) 
	    { 
		$temp_time = &convert_to_ymd($temp_time); 
	    }
	    push(@fixed_start_time,($temp_time));
	}
	@tmp_start_time = sort(@fixed_start_time);
    }

    $start_time = shift(@tmp_start_time);
    while (! $start_time)
   { 
	$start_time = shift(@tmp_start_time);
	if ($#tmp_start_time == -1) { last; }
    }
    if (length($start_time) < 1) { $start_time = "\"NULL\""; }
    printv "Calculated cumulative START_TIME = $start_time\n";
    # end calculating cumulative START_TIME

    # calculate cumulative STOP_TIME
   my @tmp_stop_time  = sort(@stop_time);
    if ($tmp_stop_time[$#tmp_stop_time] =~ /^\"(N\/A|UNK|NULL)\"$/)
    {
	@tmp_stop_time = grep(!/^\"$1\"$/,@tmp_stop_time);
	if ($tmp_stop_time[$#tmp_stop_time] =~ /^.+$/)
	{
	    @tmp_stop_time = ("");
	}
	else { @tmp_stop_time = ("\"$1\""); }
    }

    if (grep(/^\d{4}-\d{3}/,@tmp_stop_time) && 
	grep(/^\d{4}-\d\d-\d\d/,@tmp_stop_time))
    {
	my @fixed_stop_time;
	foreach my $temp_time (@tmp_stop_time)
	{
	    if ($temp_time =~ /^\d{4}-\d{3}/) 
	    { 
		$temp_time = &convert_to_ymd($temp_time); 
	    }
	    push(@fixed_stop_time,($temp_time));
	}
	@tmp_stop_time = sort(@fixed_stop_time);
    }

    $stop_time  = pop(@tmp_stop_time);
    if (length($stop_time) == 0) { $stop_time = "\"NULL\""; }
    printv "Calculated cumulative STOP_TIME  = $stop_time\n";
    # end calculating cumulative STOP_TIME
}


#=============================================================================
# Create cumulative index table

printv "Starting to create temporary cumulative index table\n";

#====================
# Open temp cumulative index table 'cumidex.tab.mkcumidx.tmp'

open(CUMTAB, "> ".$tmpdir."/cumindex.tab.mkcumidx.tmp") or
    die "ERROR: Could not create temp table file to write in \"$tmpdir/\".\n";
printv "Created \"$tmpdir/cumindex.tab.mkcumidx.tmp\" for writing\n";

#====================
# populate temporary cumulative index table

for (my $vol_num = 0;$vol_num<=$#volume_list;$vol_num++)
{
    # open volume's index table
    open(TABLE, $volume_list[$vol_num]."/index/index.tab") or 
	die "ERROR: Could not open \"$volume_list[$vol_num]\".\n";
    printv "Opened \"$volume_list[$vol_num]/index/index.tab\" for reading\n";

    # read in each line from TABLE and spit it out to CUMTAB
    for (my $linenum = 1;$linenum<=$rows[$vol_num];$linenum++)
    {
	# read in line
	printv "Reading in Line $linenum of $rows[$vol_num]\n";
	my $line = <TABLE> or die "ERROR: EOF reached before expected.\n";

	# output each field to CUMTAB
	for (my $col_num = 1;$col_num<=$columns;$col_num++)
	{
	    # parse out individual field data
	    my $leading_chars = $i_start_byte{"$vol_num,$col_num"} - 1;
	    $line =~ /^(.{$leading_chars})(.{$i_bytes{"$vol_num,$col_num"}})/ 
		or die "ERROR: EOR reached before expected.\n";
	    my $field = $2;

	    # print delimiter
	    if ($col_num > 1) { print CUMTAB $delimiter; }

	    # print field to CUMTAB
	    if ($data_type[$col_num] =~ /^(TIME|DATE)$/)
	    {
		printf CUMTAB "%-$bytes[$col_num]s",$field;
	    }
	    elsif ($data_type[$col_num] eq "CHARACTER")
	    {
		printf CUMTAB "\"%-$bytes[$col_num]s\"",$field;
	    }
	    elsif ($data_type[$col_num] eq "INTEGER")
	    {
		printf CUMTAB "%$bytes[$col_num]s",$field;
	    }
	    else { die "ERROR: \"$data_type\" unknown.\n"; }
	} # end output each field to CUMTAB for loop
	print CUMTAB "\r\n";
	printv "Printed line successfully\n";
    } # end reading each line from TABLE and spiting out to CUMTAB

    # close volume's index table
    close(TABLE);
    printv "Closed \"$volume_list[$vol_num]/index/index.tab\" for reading\n";

} # end populate temp cumulative index table for loop

#====================
# Close temporary cumulative index label and fix the record length to 80

close(CUMTAB);
printv "Closed \"$tmpdir/cumindex.tab.mkcumidx.tmp\" for writing\n";

printv "Finished creating temporary cumulative index table\n";


#==============================================================================
# Create cumulative index label 

printv "Starting to create temporary cumulative index label\n";

#====================
# Open temp cumulative index label 'cumindex.lbl.mkcumidx.tmp'

open(CUMLBL, "> ".$tmpdir."/cumindex.lbl.mkcumidx.tmp") or
    die "ERROR: Could not create temp label file to write in \"$tmpdir/\".\n";
printv "Created \"$tmpdir/cumindex.lbl.mkcumidx.tmp\" for writing\n";

#====================
# print pre COLUMN lines

printv "Printing pre OBJECT = INDEX_TABLE statements\n";

print CUMLBL 
    "PDS_VERSION_ID = PDS3\n".
    "\n".
    "RECORD_TYPE    = \"FIXED_LENGTH\"\n".
    "FILE_RECORDS   = $file_records\n".
    "RECORD_BYTES   = $record_bytes\n".
    "\n".
    "START_TIME     = $start_time\n".
    "STOP_TIME      = $stop_time\n".
    "\n".
    "^INDEX_TABLE   = \"cumindex.tab\"\n".
    "\n";

printv "Printing pre OBJECT = COLUMN statements\n";

print CUMLBL
    "OBJECT     = INDEX_TABLE\n".
    "  INDEX_TYPE         = \"CUMULATIVE\"\n".
    "  INTERCHANGE_FORMAT = \"ASCII\"\n".
    "  ROWS               = $rows\n".
    "  ROW_BYTES          = $row_bytes\n".
    "  COLUMNS            = $columns\n".
    "\n";

#====================
# Print OBJECT = COLUMN

for (my $col_num = 1;$col_num<=$columns;$col_num++)
{
    printv "Printing COLUMN $col_num statements\n";

    print CUMLBL
	"  OBJECT     = COLUMN\n".
	"    COLUMN_NUMBER    = $col_num\n";

    # print NAME
    print CUMLBL "    NAME             = ";
    if (length($name[$col_num]) < 54)
    {
	print CUMLBL "\"$name[$col_num]\"\n";
    }
    elsif (length($name[$col_num]) < 69)
    {
	print CUMLBL "\n        \"$name[$col_num]\"\n";
    }
    else
    {
	my $name = $name[$col_num];
	$name =~ /^(.{53})/;
	print CUMLBL "\"$1-\n";
	my $line = 0; # lines between first and end of NAME being parsed
	my $length = undef; # number of characters preceding active line
	while ( $line < ((length($name) - 53 - 69)/69) )
	{
	    $length = 53 + ($line * 69);
	    $name =~ /^(.{$length})(.{69})/;
	    print CUMLBL "        $2-\n";
	    $line++;
	}
	$length = 53 + ($line * 69);
	$name =~ /^(.{$length})(.+)$/;
	print CUMLBL "        $2\"\n";
    }
    # end printing NAME

    print CUMLBL 
	"    START_BYTE       = $start_byte[$col_num]\n".
	"    BYTES            = $bytes[$col_num]\n".
	"    DATA_TYPE        = \"$data_type[$col_num]\"\n";

    # print FORMAT
    print CUMLBL "    FORMAT           = \"";
    if ($data_type[$col_num] eq "INTEGER") { print CUMLBL "I"; }
    else { print CUMLBL "A"; }
    print CUMLBL "$bytes[$col_num]\"\n";
    # end print FORMAT

    # print DESCRIPTION
    print CUMLBL "    DESCRIPTION      = ";
    $description[$col_num] =~ /^(.+)\n/;
    if (length($description[$col_num]) < 54 || ($1 && length($1) < 55))
    {
	print CUMLBL "\"$description[$col_num]\"\n";
    }
    else
    {
	print CUMLBL "\n        \"$description[$col_num]\"\n";
    }
    # end print DESCRIPTION
    
    print CUMLBL "  END_OBJECT = COLUMN\n\n";
} # end print OBJECT = COLUMN for loop

#====================
# post OBJECT = COLUMN statements

print CUMLBL 
    "END_OBJECT = INDEX_TABLE\n".
    "\n".
    "END\n";

#====================
# Close temporary cumulative index label and fix the record length to 80

close(CUMLBL);
printv "Closed \"$tmpdir/cumindex.lbl.mkcumidx.tmp\" for writing\n";

printv "Fixing the record length of \"$tmpdir/cumindex.lbl.mkcumidx.tmp\"".
       " to \"$tmpdir/cumindex.lbl.mkcumidx.fixed.tmp\"\n".
       `fixlen -c $tmpdir/cumindex.lbl.mkcumidx.tmp $tmpdir/cumindex.lbl.mkcumidx.fixed.tmp 2>&1`;

printv "Finished creating temporary cumulative index label\n";


#==============================================================================
# Move temporary cumindx.* files to permanent location

# Specify output locations
if ($opt_o) { @out_loc = ($opt_o); }
else { foreach my $loc (@volume_list) { push(@out_loc,$loc."/index"); }}

#====================
# Copy temp files to new locations and rename

foreach my $loc (@out_loc)
{
    copy("$tmpdir/cumindex.lbl.mkcumidx.fixed.tmp",$loc."/cumindex.lbl");
    printv "Created \"$loc/cumindx.lbl\"\n";
    copy("$tmpdir/cumindex.tab.mkcumidx.tmp",$loc."/cumindex.tab");
    printv "Created \"$loc/cumindx.tab\"\n";
}

#====================
# Delete temp files
unlink("$tmpdir/cumindex.lbl.mkcumidx.tmp",
       "$tmpdir/cumindex.tab.mkcumidx.tmp",
       "$tmpdir/cumindex.lbl.mkcumidx.fixed.tmp");
printv "Deleted \"$tmpdir/cumindex.lbl.mkcumidx.tmp\"\n";
printv "Deleted \"$tmpdir/cumindex.lbl.mkcumidx.fixed.tmp\"\n";
printv "Deleted \"$tmpdir/cumindex.tab.mkcumidx.tmp\"\n";
printv "***mkcumidx process completed***\n";


#==============================================================================
# Functions

sub printv($)
# IF the -v option is used, the text will be printed to the screen
{
    my $line = $_[0];
    if ($opt_v) { print "mkcumidx: $line"; }
    return 1;
} # end &printv($)



sub convert_to_doy($)
# converts DATETYPE "DATE" and/or "TIME" from year-month-day to year-day
{
    my $date = shift(@_);
    if ($date =~ /^\d{4}-\d{3}/) { return $date; }
    elsif ($date =~ /^(\d{4})-(\d\d)-(\d\d)(.*)$/)
    {
	my ($year,$month,$day,$time) = ($1,$2,$3,$4);

	# days past for month:
	#       month =   0 1  2  3  4   5   6   7   8   9  10  11  12
	my @days_past = (-1,0,31,59,90,120,151,181,212,243,273,304,334);

	my $new_day = $days_past[$month] + $day;
	if ($month != 2) { $new_day += &is_leap_year($year)}
	if (length($new_day) == 1) { $new_day = "00$new_day"; }
	elsif (length($new_day) == 2) { $new_day = "0$new_day"; }

	return "$year"."-"."$new_day".$time;
    }
    else { die "ERROR: \"$date\" not valid format.\n"; }
} # end convert_to_doy()


sub convert_to_ymd($)
# converts DATETYPE "DATE" and/or "TIME" from year-day to year-month-day
{
    my $date = shift(@_);
    if ($date =~ /^\d{4}-\d\d-\d\d/) { return $date; }
    elsif ($date =~ /^(\d{4})-(\d{3})(.*)/) 
    {
	my ($year,$day,$time) = ($1,$2,$3);
	my $new_day = $day;
	my $month = 1;

	# calculate new day and month
	if ($new_day > 31) { $new_day -= 31; $month++;}
	if ($new_day > (28 + &is_leap_year($year)) && $month == 2)
	{ 
	    $new_day -= (28 + &is_leap_year($year));
	    $month++;
	}
	if ($new_day > 31 && $month == 3)  { $new_day -= 31; $month++;}
	if ($new_day > 30 && $month == 4)  { $new_day -= 30; $month++;}
	if ($new_day > 31 && $month == 5)  { $new_day -= 31; $month++;}
	if ($new_day > 30 && $month == 6)  { $new_day -= 30; $month++;}
	if ($new_day > 31 && $month == 7)  { $new_day -= 31; $month++;}
	if ($new_day > 31 && $month == 8)  { $new_day -= 31; $month++;}
	if ($new_day > 30 && $month == 9)  { $new_day -= 30; $month++;}
	if ($new_day > 31 && $month == 10) { $new_day -= 31; $month++;}
	if ($new_day > 30 && $month == 11) { $new_day -= 30; $month++;}

	# pad day to DD and month to MM
	if (length($new_day) == 1) { $new_day = "0$new_day"; }
	if (length($month)   == 1) { $month   = "0$month";   }

	return "$year"."-$month"."-$new_day$time";
    }
    else { die "ERROR: \"$date\" not valid format.\n"; }

} # end convert_toymd()

sub is_leap_year($)
# checks to see if a year is a leap year
{
    my $year = shift(@_);
    if (($year % 4 == 0) && (($year % 100) || ($year % 400 == 0))) { return 1;}
    else { return 0; }
} # end is_leap_year





1;
