#!/usr/bin/perl

use Time::localtime;
use Getopt::Std;
use lib "/n/sbnops01/lcltools/lib/perl5";
require "nextPDSODL.pl";

#
# mklcldd - Make a local Data Dictionary
#
#   This routine reads through all detached label and catalog files in a 
#   directory and compiles a list of all objects and keywords mentioned in
#   those files.  It then searches through the full data dictionary to 
#   extract the definitions of those entries and write out local version
#   of the dictionary (for, say, inclusion on a CDROM).  Any elements or
#   keywords not found in the full dictionary are retained as (blank)
#   templates in the local copy.  Standard values not found in the standard
#   value list are added to the local version.  If a listing file is 
#   requested, unresolved object names, elements and standard values are 
#   output to the named file.  Otherwise, they are merely flagged in the 
#   dictionary file.
#
# Format:  mklcldd [ -c ] [ -r ]
#                  [ -d <master dictionary> ] [ -l <listing file> ]
#                  [ -o <local dictionary> ] <directory>
#
#    options -    -c   Check non-.lbl and .cat files for PDS Labels
#                 -r   Recurse through subdirectories
#                 -l   File name for listing new objects, keywords and values
#                 -d   Alternate location for master PSDD and index
#                 -o   Output dictionary filename (otherwise >STDOUT)
#
# Data Structures:
#
#     psdd_offset{<keyword>}           Offset into psdd.full file
#     psdd_length{<keyword>}           Number of bytes in pdsdd.full entry
#     psdd_type{<keyword>}             'E' for element, 'G' for generic, 'S'
#                                          for specific object
#
#     elem{<keyword>}                  List of standard values, uppercase,
#                                          no underscores, in double quotes,
#                                          separated by commas
#     obj{<keyword>}                   Empty list to hold/sort object names
#     elmalias{<keyword>}              Element aliases
#     elmalcon{<keyword>}              Element alias context (an object)
#     objalias{<keyword>}              Object aliases
#
#     output_obal{<keyword>}           Output Object alias list
#     output_elal{<keyword>}           Output element alias list
#
#     Context for aliases is held in the @current_object array, where the 
#     first element is the current object
#
#  29 May 1997, acr: fixed a dumb bug which prevented finding generic objects
#                    when specific ones didn't match.
#  13 Jun 2001, acr: Trouble-shooting for changed Perl behavior regarding
#                    newlines in strings being matched; fixed bug with standard
#                    value strings containing commas being split into two 
#                    strings.
#  18 Jul 2001, acr: Minor modifications to use universal constants for PSDD.
#  26 Sep 2006, acr: Modified to use Getopt::Std, Time::localtime and "use lib"
#
#==============================================================================
$PSDD    = "/usr/local/doc/pds/pdsdd.full";
$PSDDIDX = "/usr/local/doc/pds/pdsdd.idx";

$current_object[0] = "FILE";

#
# Deal with input parameters; open the output file:
#

getopts('crd:l:o:');

if (@ARGV != 1)   # There must be a directory name left
  { print STDERR "Usage: mklcldd [-r] [-c] [-d <main dictionary>] ";
    print STDERR " [-l <new entries list>] [-o <output dictionary>] ";
    print STDERR "directory\n";
    exit(0);
  }

if ($opt_d)
  { $psdd = $opt_d; }
else
  { $psdd = $PSDD;
    $psdd =~ s/\.full$//;
  }

if ($opt_l)
  { open (LIST,"> $opt_l") || die "Could not open \"$opt_l\" for listing."; }

$outdd = ($opt_o)? $opt_o : "-";
open (OUT,"> $outdd") || die "Could not open output file \"$outdd\".";

#
#  Start by reading the psdd.idx file, because if you can't read that you can't
#  do any processing.
#

open(DDIDX,"$psdd.idx") || die "Could not open $psdd.idx for reading.";

$line=<DDIDX>;
while ($line !~ /^END/)
  { $line=<DDIDX>; }

while ($line=<DDIDX>)
  { chop $line;
    ($kw, $type, $offset, $length) = split(/\s+/,$line);
    $psdd_offset{$kw} = $offset;
    $psdd_length{$kw} = $length;
    $psdd_type{$kw}    = substr($type,0,1);
  }
close(DDIDX);

##open (T1,">i.lst");
##foreach $kw (sort keys(%psdd_offset))
##  { printf T1 "%-30.30s %s %6d %5d\n", $kw, $psdd_type{$kw},
##               $psdd_offset{$kw},$psdd_length{$kw}; }
##close(T1);

# Read the alias list from the data dictionary file:

open(PSDD,"$psdd.full") || die "Could not open $psdd.full for reading.";

$line="";
while ($line !~ /OBJECT_ALIAS_SEQUENCE/)
  { $line=<PSDD>; }

while (($line=<PSDD>) !~ /ELEMENT_ALIAS_SEQUENCE/)
  { chop $line;
    $line =~ s/[()']//g;
    $line =~ s/\s+//g;
    ($alias, $def) = split(/,/,$line);
    $objalias{$alias} = $def;
  }
while (($line=<PSDD>) !~ /END_OBJECT = ALIAS_LIST/)
  { chop $line;
    $line =~ s/[()']//g;
    $line =~ s/\s+//g;
    ($alias,$context,$def) = split(/,/,$line);
    $elmalias{$alias} = $def;
    $elmalcon{$alias} .= "'$context'";  # There may be more than one object
  }

##open (T2,">a.lst");
##foreach $alias (sort keys(%objalias))
##  { printf T2 "%-30.30s : %-30.30s\n",$alias, $objalias{$alias}; }
##print T2 "\n\n\n";
##foreach $alias (sort keys(%elmalias))
##  { printf T2 "%-30.30s : %-30.30s\n",$alias, $elmalias{$alias}; }
##close(T2);

#
#  Now loop through the label and catalog files to collect the lists of
#  keywords and elements.
#

&dodir($ARGV[0]);

##open (T1,">o.lst");
##foreach $name (sort keys(%obj))
##  { printf T1 "%-30.30s found %3d times.\n",$name,$obj{$name}; }
##close (T1);
##
##open (T2,">e.lst");
##foreach $name (sort keys(%elem))
##  { printf T2 "%-30.30s : %s\n",$name,$elem{$name}; }
##close(T2);

#
#  Ready to write the local dictionary.  Note that aliases have already been
#  compiled and translated.
#
#  Method:
#
#     Write header and alias lists
#
#     For each object:
#       If the object name appears in the %psdd_offset array
#           copy entry from pdsdd to output dictionary
#       else
#           write empty template
#           next object;
#
#     For each element:
#       If the element name appears in the $psdd_offset array
#           continue;
#       else
#           write empty template
#           next element;
#
#       retrieve element definition
#       if there is a standard value list
#           check for standard values, add missing ones
#       write the element definition
#

# Header and Alias lists:

$date = ctime(time);
$date =~ s/\s+$//;
printf OUT "/* Local Planetary Science Data Dictionary Listing */\n";
printf OUT "/* Generated: $date */\n";
printf OUT "/* Start of alias definitions */\n";
printf OUT "OBJECT = ALIAS_LIST\n";
printf OUT "  OBJECT_ALIAS_SEQUENCE = (";
foreach $alias (sort (@output_obal))
    { printf OUT "\n    ('$alias','$objalias{$alias}')"; }
printf OUT ")\n";

printf OUT "  ELEMENT_ALIAS_SEQUENCE = (";
foreach $alias (sort (@output_elal))
    { $line = $elmalcon{$alias};
      $line =~ s/^'//;
      $line =~ s/'$//;
      @clist = split(/''/,$line);

      foreach $con (@clist)
        { printf OUT "\n    ('$alias','$con','$elmalias{$alias}')"; }
    }
printf OUT ")\n";
printf OUT "END_OBJECT = ALIAS_LIST\n";
printf OUT "END\n";
printf OUT "/* End of alias definitions */\n";
printf OUT "/* Start of object definitions */\n";

# Object definitions:

foreach $name (sort keys(%obj))
  { if ($psdd_offset{$name})
      { &copy_object($name); }
    else
      { &write_blank_object($name); }
  }

printf OUT "/* End of object definitions */\n";
printf OUT "/* Start of element definitions */\n";

# Element Definitions:

foreach $name (sort keys(%elem))
  { if ($psdd_offset{$name})
      { &copy_element($name); }
    else
      { &write_blank_element($name); }
  }

printf OUT "/* End of element definitions */\n";
printf OUT "/* End of local Data Dictionary Listing */\n";
close(OUT);

# Done.

#=============================================================================
#=============================================================================

sub dodir

  # This routine runs the processing for a single directory.  It collects
  # a list of files with extensions ".lbl" and ".cat", and a second list
  # of subdirectories, if recursion is indicated ($opt_r not null).  It
  # calls &dolbl on all label files and itself on all subdirectory files.
  # If file-checking was requested ($opt_c not null), all non-directory
  # files without *.lbl or *.cat extensions are opened to see if they begin
  # with PDS labels.  A PDS label must have either an SFDU label or the
  # PDS_VERSION_ID keyword as the first line to be recognized.

{ local ($dir) = $_[0];
  local ($name);           # filename holder
  local (@subd);           # subdirectory list
  local (@file);           # file list
  local ($s,$f);           # list subscripts
  local ($line1);

  if (!(-d $dir))
    { printf STDERR "\"$dir\" is not a directory.\n";
      return;
    }

  opendir (DIR,$dir) || die "Error attempting to open directory \"$dir\".";

  $f = $s = 0;
  foreach $name (sort readdir(DIR))
    { next if ($name =~ /^\./);

      if ($name =~ /\.lbl$/  ||  $name =~ /\.cat$/  ||
          $name =~ /\.LBL$/  ||  $name =~ /\.CAT$/)
        { $file[$f] = "$dir/$name";
          $f++;
        }
      elsif (-d "$dir/$name")
        { $subd[$s] = "$dir/$name";
          $s++;
        }
      elsif ($opt_c)
        { $name = "$dir/$name";
          open (TMP,$name);
          read (TMP,$line1,80);

          if ($line1 =~ /SFDU/  ||  $line1 =~ /PDS_VERSION_ID\s+=/)
            { $file[$f] = $name;
              $f++;
            }
          close(TMP);
        }
          
    }
  closedir(DIR);

  # Process the lists:

  foreach $name (@file)
    { &dolbl($name); }

  if ($opt_r)
    {  foreach $name (@subd)
        { &dodir($name); }
    }

  # Done.

  return;
}

#----------------------------------------------------------------------------

sub dolbl

  #  This routine reads through a single label or catalog file and looks
  #  for elements and objects as yet not encountered, adding them to the
  #  global lists.  The program looks for the "END" statement which terminates
  #  label, to avoid trouble in the case of attached labels.  It will continue
  #  reading lines until it matches double quotes, which may produce an
  #  unexpected end of file in case of syntax errors.

{ local ($lbl) = $_[0];
  local ($line, $nextline);
  local ($elm,$sv);          # element name and standard value
  local ($qc);               # quote counter
  local ($mval);             # $val formatted for matching

  open (LBL,$lbl)  ||  die "Could not open ODL file \"$lbl\".";
  print STDERR "Processing $lbl...\n";

  #$line=&nextPDSODL(*LBL);   # Returns complete assignment with no blanks 
                              # leading, trailing, or around the "="
  while ($line=&nextPDSODL(*LBL))
    { # Stop at the end of the label: 

      last if ($line eq "END");

      # Skip past non-OBJECT and non-element lines:

      next if ($line !~ /=/    ||  $line =~ /END_OBJECT/  || 
               $line =~ /^\^/  ||  $line =~ /SFDU/          );

      # Element lines contain an element of up to 30 characters followed by
      # an '='.  Objects are the value half of an OBJECT= line.

      # OBJECT check:

      if ($line =~ /OBJECT=/)
        { $line =~ s/^.*=//;
          $line = &aliaschk($line,"OBJECT");
          $obj{$line}++;
          unshift (@current_object,$line);   # Update object context
        }

      elsif ($line =~ /END_OBJECT/)
        { shift(@current_object); }          # Update object context

      # Element check:

      elsif ($line =~ /=/)
        { # First, separate the element and value parts:

          $line =~ tr/a-z/A-Z/;
          $line =~ tr/\n/ /;
          $elm = $line;
          $elm =~ s/=.*$//;
          $sv = $line;
          $sv =~ s/^[^=]+=//;

          # Check for element aliases:

          $elm = &aliaschk($elm,"ELEMENT");

          # And prepare the value part for checking:

          $sv =~ s/^[({]//;    # Remove set and sequence delimiters
          $sv =~ s/[)}]$//;

          # If the element has been encountered before and the current 
          # value has not, add it to the list. (Don't save values for 
          # any DESCRIPTION fields, though.)  If the value is a list of
          # values, then we have to loop through the list:

          if ($elm =~ /NOTE/  ||  $elm =~ /DESC/)
            { $elem{$elm} = 1; }

          else
            { ($val,$sv) = &next_value($sv);
              $elem{$elm} = 1 if (!$val);     # Set flag even if there is no
              while ($val)                    #  value to check
                { $mval = quotemeta($val);
                  if (!$elem{$elm})           # No values yet
                    { $elem{$elm} = $val; }   # Save value
                  elsif ($elem{$elm} !~ /$mval/)  # This value is new
                    { $elem{$elm} = join(",",($elem{$elm},$val)); }
                  ($val,$sv) = &next_value($sv);
                }
            }
        }
    }

  # Done.

  close(LBL);
  return;
}

#----------------------------------------------------------------------------

sub aliaschk

  # Routine to check for the current object or element being in the PSDD
  # alias list.  If it is, the alias is translated to the actual object
  # name, and the alias encountered is added to an output list for later
  # printing.  Current object context is the first element of the 
  # @current_object array.

{ local ($thing,$chktype) = @_;
  local ($ob, $el);
  local (@fnd, $i);
  local ($context);
  local ($done);


  if ($chktype eq "OBJECT")
    { $ob = $thing;

      # OBJECT aliases.  First check for a real object by that name and return
      # it if found.  If not, try to find an alias and return that.  If neither
      # works, try removing prefixes (delimited by underscores) from the 
      # beginning of the object name and checking again for a more generic
      # object or alias.  If everything fails, return the original object name.

      $done = 0;
      do
        { if ($ob !~ /_/)       # Check for end condition
            { $done = 1; }

          if ($psdd_offset{$ob})
            { return $ob; }

          elsif ($objalias{$ob})
            { if (@output_obal == 0)
                { $output_obal[0] = $ob; }
              else
                { @fnd = grep($_ eq $ob,@output_obal);
                  if (!@fnd)
                    { push(@output_obal,$ob); }
                }
              return $objalias{$ob};
            }
          else
            { $ob =~ s/^[A-Za-z0-9]+_//; }
        } until ($done);

      return $thing;
    }

  else
    { $el = $thing;

      # ELEMENT aliases.  Check for an alias within the current context FIRST
      # and return it if found.  Otherwise, return the element itself.

      if ($elmalias{$el})
        { 
          # Found an alias, but before accepting it make sure it applies in
          # the current context.  Note, also, that there may be more than one
          # applicable context (maintained as a list in @elemalcon):

          $context = "'$current_object[0]'";
          if ($elmalcon{$el} =~ /N\/A/  ||  
              $elmalcon{$el} =~ /$context/)
            
            { # Return the alias: 

              if (@output_elal == 0)
                { $output_elal[0] = $el; }
              else
                { @fnd = grep($_ eq $el,@output_elal);
                  if (!@fnd)
                    { push(@output_elal,$el); }
                }
              return $elmalias{$el};
            }
        }

      return $el;

    }
}

#----------------------------------------------------------------------------

sub copy_object

  #  Routine to copy the object definition from the main PSDD into the local
  #  copy being written on OUT.  The PSDD is already open on that device.

{ local ($o) = $_[0];

  local ($buff);

  # Go to the start of the object definition:

  seek(PSDD,$psdd_offset{$o},0);
  read(PSDD,$buff,$psdd_length{$o});
  print OUT $buff;

  # Add the "END" statement, which is not included in the object length:

  print OUT "END\n";
}

#----------------------------------------------------------------------------

sub write_blank_object

  #  Routine to insert a blank object template for an object not found in the
  #  main data dictionary and log the new object to the listing file, if any.

{ local ($o) = $_[0];

  printf OUT "OBJECT = GENERIC_OBJECT_DEFINITION\n";
  printf OUT "  NAME = $o            /* NEW OBJECT */\n";
  printf OUT "  STATUS_TYPE =\n";
  printf OUT "  DESCRIPTION =\n";
  printf OUT "  REQUIRED_ELEMENT_SET = {}\n";
  printf OUT "  OPTIONAL_ELEMENT_SET = {}\n";
  printf OUT "  REQUIRED_OBJECT_SET = {}\n";
  printf OUT "  OPTIONAL_OBJECT_SET = {}\n";
  printf OUT "END_OBJECT = GENERIC_OBJECT_DEFINITION\n";
  printf OUT "END  \n";

  printf LIST "New OBJECT: $o\n" if ($opt_l);
}

#----------------------------------------------------------------------------

sub copy_element

  #  Routine to copy an element definition from the main PSDD to the output
  #  version.  A check is made to see if standard values are required, and if
  #  so the values found in the input data set are checked against the current
  #  values and added to the list if not found.

{ local ($e) = $_[0];
  local (@buff,$b);
  local ($i,$blen,$val,$sval);
  local ($svflag);
  local (@stdlist,$sl,@inplist,@fnd);

  # Go to the start of the element definition:

  seek (PSDD,$psdd_offset{$e},0);

  # Fill the definition buffer and look for a standard value set flag:

  $b = $svflag = 0;
  while (($buff[$b]=<PSDD>) !~ /^END\s*$/)
    { $svflag = $b if ($buff[$b] =~ /STANDARD_VALUE_SET/);
      $b++;
    }

  # If there is a standard value list, check the collected standard values
  # against it and adjust the definition buffer as needed:

  if ($i=$svflag)
    { $sl = 0;
      while ($buff[$i] !~ /}/)
        { $stdlist[$sl] = $buff[$i]; 
          $i++;
          $sl++;
        }
      $buff[$i] =~ s/}/,/;          # Remove closing brace (must be replaced!)
      $stdlist[$sl] = $buff[$i];    # catch the last one

      $blen = @buff;
      @inplist = split(/","/,$elem{$e});

      foreach $val (@inplist)
        { $val =~ s/("|)([^"]+)("|)/"$2"/;
          @fnd = grep($_ =~ /$val/,@stdlist);
          if (!@fnd)                           # Didn't match any standard value
            { $buff[$blen]   = $buff[$blen-1]; # Shift last two lines forward
              $buff[$blen-1] = $buff[$blen-2];
              $buff[$blen-2] = "     $val,   /* NEW VALUE */\n";       
                                               # Add new value to end of list
              $blen++;
              printf LIST "New Standard Value for $e: $val\n" if ($opt_l);
            }
        }

      $buff[$blen-3] =~ s/,([^,]*)$/}$1/;      # Place closing brace on list
    }

  # Now print the buffer to the output file:

  $blen = @buff;
  for ($i=0; $i<$blen; $i++)
    { print OUT $buff[$i]; }

  # Done.
}

#----------------------------------------------------------------------------

sub write_blank_element

  #  Routine to insert a blank template for unmatched PSDD elements.

{ local ($e) = $_[0];
  local ($svlist, $sval);      # Standard value list string; single value
  local (@buff,$b);            # Output standard values buffer
  local ($i);

  printf LIST "New ELEMENT: $e\n" if ($opt_l);

  printf OUT "OBJECT = ELEMENT_DEFINITION\n";
  printf OUT "  NAME = $e     /* NEW ELEMENT */\n";
  printf OUT "  STATUS_TYPE =\n";
  printf OUT "  GENERAL_DATA_TYPE =\n";
  printf OUT "  STANDARD_VALUE_TYPE =\n";
  printf OUT "  MAXIMUM_LENGTH = \n";
  printf OUT "  DESCRIPTION =\n";

  # If we've collected a list of standard values, print them here:

  if ($elem{$e}  &&  $elem{$e} ne "1")
    { $svlist = $elem{$e};
      $b = 0;
      ($buff[$b],$svlist) = &next_value($svlist);
      while ($buff[$b])
        { $b++;
          ($buff[$b],$svlist) = &next_value($svlist);
        }

      # Now print them:

      print OUT "  STANDARD_VALUE_SET = {\n";
      for ($i=0; $i<$b-1; $i++)
        { printf OUT  "     $buff[$i],   /* NEW VALUE */\n";
          printf LIST "    Standard Value for $e: $buff[$i]\n" 
              if ($opt_l);
        }
      printf OUT  "     $buff[$b-1]}   /* NEW VALUE */\n";
      printf LIST "    Standard Value for $e: $buff[$b-1]\n" if ($opt_l);
    }        
      
  printf OUT "END_OBJECT = ELEMENT_DEFINITION\n";
  printf OUT "END\n";
}

#----------------------------------------------------------------------------

sub next_value

  # Routine to take the input list and return the next value in standard
  # format: underscores replaced by blanks and in double quotes.  The 
  # returned value is removed from the string passed in.

{ local ($str) = @_[0];   # Copy of the $sv string
  local ($rv,$junk,$j);   # Return value and throw-away strings
  local ($rem);           # remainder of string


  # First, remove leading blanks:

  $str =~ s/^\s*//;

  # Items in the list must be separated by commas and may be quoted. To avoid
  # confusing commas inside quotes with those outside, we break on matching 
  # quotes first, then on a comma. Note that only double quotes are valid in
  # PDS labels:

  if ($str =~ /^"/)
    { $rv = $str;
      $rv =~ s/^"([^"]+)".*$/$1/;   # Take the first quoted string, sans quotes
      $str =~ s/^"[^"]+"//;         # Remove value from remaining string
      $str =~ s/^,//;               # Remove comma, if any
    }
  else
    { ($rv,$junk) = split(',',$str);  # Get first value
      $str =~ s/^[^,]+//;             # Remove first value
      $str =~ s/^,//;
    }

  # Prepare the return value:

  $rv =~ s/^\s+//;
  $rv =~ s/\s+$//;
  $rv = "\"$rv\"" if ($rv);
  $rv =~ s/_/ /g;

  # Clean up the value string:

  $str =~ s/^,//;
  $str =~ s/^\s*//;
  
  # Done:

  return ($rv,$str);
}
