#!/usr/bin/perl

use Getopt::Std;
use File::Temp;

# Routine to generate the required PDS volume index table file and detached
# label.
#
#   Format: % mkvolidx [-a] [-l|-u] [-p] [-v VOLUME_ID] [-r volume_root] \
#                [-t] [data_dir [...]]
#
#     where   -a             check all files for attached PDS labels
#             -p             print progress messages
#             -l             force file names to lower case in index
#             -u             force file names to upper case in index
#             -v VOLUME_ID   specifies the VOLUME_ID to be included as the 
#                            first column in the table
#             -r volume_root specifies the root of the PDS volume to be 
#                            indexed.  Default is "."
#             -t             keep the temporary file and display its name
#                            for debugging
#             data_dir       is the list of data subdirectories to be
#                            indexed. Default is "<volume_root>/data".
#
#     Default case for file specifications is what is found in the directory.
#
# 18 January 2006, A.C.Raugh.
#
# The files generated by this routine will meet the PDS requirements for the
# volume INDEX_TABLE file and label.  Additional fields that the SBN finds to
# be useful in the majority of their data sets (fields which are required to
# be in data file labels) are also included in the index.
#
# Note that the program cannot recognize figurative constants (like "N/A" or
# "NULL") that might appear as the value of keywords in labels.  The data 
# preparer should check for these where they might occur, and modify the 
# label definition for the affected columns to add the appropriate 
# *_CONSTANT.
#
#-----------------------------------------------------------------------------
#
# The following columns are included in the output.  All values come from the
# label files unless otherwise indicated:
#
#     VOLUME_ID                Read from the voldesc.cat file, if any,
#                              overridden by the -v option.  If no valid ID 
#                              is available, "NULL_0000" is used as a
#                              MISSING_CONSTANT. (Required)
#     FILE_SPECIFICATION_NAME  The fully qualified file name of the file
#                              containing the PDS label (Required)
#     DATA_SET_ID              Required
#     PRODUCT_ID               Required
#     PRODUCT_CREATION_TIME    Required
#     START_TIME               SBN
#     TARGET_NAME              SBN
#     INSTRUMENT_HOST_NAME     SBN
#     INSTRUMENT_NAME          SBN
#
# The index file and label are prepared in two passes.  The first pass reads
# through all the files in the data directory, collecting values into a 
# temporary file and noting maximum fields lengths.  The second pass goes
# back through the temporary file, clipping unnecessary trailing blanks,
# adding delimiters and creating the label.  The temporary file is deleted
# on completion.
#
# 16 Feb 2007, ACR: Added START_TIME and STOP_TIME to label (by checking
#                   START_TIME in product labels)
#
#=============================================================================
# Constants and initializations:

$TRUE  = 1;
$FALSE = 0;

$NULL_VOLUME_ID = "NULL_0000";

@colname    = ("VOLUME_ID",
               "FILE_SPECIFICATION_NAME",
               "DATA_SET_ID",
               "PRODUCT_ID",
               "PRODUCT_CREATION_TIME",
               "START_TIME",
               "TARGET_NAME",
               "INSTRUMENT_HOST_NAME",
               "INSTRUMENT_NAME");
%coldefsize = ("VOLUME_ID",                11,
               "FILE_SPECIFICATION_NAME", 255,
               "DATA_SET_ID",              40,
               "PRODUCT_ID",               40,
               "PRODUCT_CREATION_TIME",    23,
               "START_TIME",               23,
               "TARGET_NAME",             120,
               "INSTRUMENT_HOST_NAME",    120,
               "INSTRUMENT_NAME",          60);
%colmaxsize = ("VOLUME_ID",               0,
               "FILE_SPECIFICATION_NAME", 0,
               "DATA_SET_ID",             0,
               "PRODUCT_ID",              0,
               "PRODUCT_CREATION_TIME",   0,
               "START_TIME",              0,
               "TARGET_NAME",             0,
               "INSTRUMENT_HOST_NAME",    0,
               "INSTRUMENT_NAME",         0);
%coldesc    = ("VOLUME_ID",
                 " ",
               "FILE_SPECIFICATION_NAME",
                 "Path from volume root to label file",
               "DATA_SET_ID",
                 "From label",
               "PRODUCT_ID",
                 "From label",
               "PRODUCT_CREATION_TIME",
                 "From label",
               "START_TIME",
                 "From label",
               "TARGET_NAME",
                 "From label",
               "INSTRUMENT_HOST_NAME",
                 "From label",
               "INSTRUMENT_NAME",
                 "From label" );
%colmissing = ("VOLUME_ID",               "NULL_0000",
               "FILE_SPECIFICATION_NAME", "NULL",
               "DATA_SET_ID",             "MISSING_DSID",
               "PRODUCT_ID",              "MISSING_PRODUCT_ID",
               "PRODUCT_CREATION_TIME",   "9999-99-99",
               "START_TIME",              "9999-99-99",
               "TARGET_NAME",             "MISSING_TARGET_NAME",
               "INSTRUMENT_HOST_NAME",    "MISSING_HOST_NAME",
               "INSTRUMENT_NAME",         "MISSING_INST_NAME" );
%colquoted  = ("VOLUME_ID",               $TRUE,
               "FILE_SPECIFICATION_NAME", $TRUE,
               "DATA_SET_ID",             $TRUE,
               "PRODUCT_ID",              $TRUE,
               "PRODUCT_CREATION_TIME",   $FALSE,
               "START_TIME",              $FALSE,
               "TARGET_NAME",             $TRUE,
               "INSTRUMENT_HOST_NAME",    $TRUE,
               "INSTRUMENT_NAME",         $TRUE);
%coltype    = ("VOLUME_ID",               "CHARACTER",
               "FILE_SPECIFICATION_NAME", "CHARACTER",
               "DATA_SET_ID",             "CHARACTER",
               "PRODUCT_ID",              "CHARACTER",
               "PRODUCT_CREATION_TIME",   "TIME",
               "START_TIME",              "TIME",
               "TARGET_NAME",             "CHARACTER",
               "INSTRUMENT_HOST_NAME",    "CHARACTER",
               "INSTRUMENT_NAME",         "CHARACTER");

undef $vol_start_time;
undef $vol_stop_time;

undef %missing_data;   # TRUE if a MISSING_CONSTANT value is needed

# Collect arguments and options and make sanity checks before beginnning:

getopts('alupr:tv:');

$check_all = ($opt_a)? $TRUE : $FALSE;

# If both "-l" and "-u" were listed, ignore both and use native case (lower
# case for files this program creates):

if ($opt_l && $opt_u)
  { undef $opt_l;
    undef $opt_u;
  }

$volume_root = ($opt_r)? $opt_r : ".";
if (! -d $volume_root)
  { die "Invalid volume root, \"$volume_root\".\n"; }

# If no explicit data directory was given, check for a default:

if (@ARGV == 0)
  { if (-d "$volume_root/data")
      { $ARGV[0] = "data"; }
    elsif (-d "$volume_root/DATA")
      { $ARGV[0] = "DATA"; }
    else
      { die "No $volume_root/data (or DATA) directory found."; }
  }

# Locate the index directory and set the output file names with the
# corresponding case:

$index_dir = ($opt_u)? "$volume_root/INDEX" : "$volume_root/index";
if (-d "$volume_root/INDEX")
  { $index_dir = "$volume_root/INDEX";
    $index_table = "INDEX.TAB";
    $index_label = "INDEX.LBL";
  }
elsif (-d "$volume_root/index")
  { $index_dir = "$volume_root/index";
    $index_table = "index.tab";
    $index_label = "index.lbl";
  }
else
  { die "Index directory $volume_root/index (or INDEX) does not exist."; }

#*****
#printf "Volume root: $volume_root\n";
#printf "  Index dir: $index_dir\n"; 
#for ($i=0; $i<@ARGV; $i++)
#  { printf "    data %2d: $ARGV[$i]\n",$i+1; }
#*****

# Set up the temporary file to hold the first-pass index:

$unlink = ($opt_t)? 0 : 1;
($TMP, $tempfile) = File::Temp::tempfile(UNLINK => $unlink);

printf STDOUT "TEMPFILE: $tempfile\n" if ($opt_t);

#----------------------------------------------------------------------------
# First Pass - Collecting the index values
#
# First, we need to get the VOLUME_ID, which is constant in all lines of
# the index we're producing.  If the user passed one in, we use that.  If
# not, we look for a 'voldesc.cat' file in the volume root and take it from
# there.  Failing that, we'll use the default value of "NULL_0000".

if (-e "$volume_root/voldesc.cat")
  { $voldesc = "$volume_root/voldesc.cat"; }
elsif (-e "$volume_root/VOLDESC.CAT")
  { $voldesc = "$volume_root/VOLDESC.CAT"; }
else
  { undef $voldesc; }

if ($opt_v)
  { $volume_id = $opt_v; 
    $coldesc{"VOLUME_ID"} = "Value entered by data preparer";
  }
elsif ($voldesc)
  { open (VCAT, $voldesc) || die "Could not open $voldesc ($!)";
    $line = <VCAT>;
    while ($line !~ /^\s*VOLUME_ID\s*=/)
      { $line = <VCAT>; }
    $line =~ /^\s*VOLUME_ID\s*=\s*(\S+)\s*$/;
    $volume_id = $1;
    $volume_id =~ s/"//g;
    $volume_id =~ tr/a-z/A-Z/;
    close (VCAT);
    $voldesc =~ s/.*\///;
    $coldesc{"VOLUME_ID"} = "From $voldesc file";
  }
else
  { $volume_id = $NULL_VOLUME_ID;
    $coldesc{"VOLUME_ID"} = "Null placeholder";
  }

# Force the VOLUME_ID to uppercase and see if it meets PDS requirements:

$volume_id =~ tr/a-z/A-Z/;
if ($volume_id !~ /^[A-Z0-9]{1,6}_[0-9]{4}$/)
  { printf "WARNING: VOLUME_ID is not PDS-compliant\n"; }
$len = length($volume_id);
$coldefsize{"VOLUME_ID"} = $len;
$colmaxsize{"VOLUME_ID"} = $len;

# We'll loop through the input directories in the ARGV array.  We know there
# is at least one.  These directories are relative to the volume root.  We'll
# skip non-existent directories (with a warning message).

foreach $data_dir (sort @ARGV)
  { if (! -d "$volume_root/$data_dir")
      { printf STDERR "Directory \"$volume_root/$data_dir\" does not exist.";
        printf STDERR "  Ignored.\n";
        next;
      }

    do_dir($data_dir,$volume_root);
  }

close($TMP);

#----------------------------------------------------------------------------
# Second Pass - Cleaning up and writing the output

# We need to open the output files in the index/ directory and re-open the
# temp file for reading.  Note that any existing index.tab and index.lbl
# files will be overwritten without comment.

open (TAB,">$index_dir/$index_table") || 
  die "Could not open $index_dir/$index_table ($!)";
open (LBL,">$index_dir/$index_label") || 
  die "Could not open $index_dir/$index_label ($!)";
open (TMP,$tempfile)    || die "Could not re-open $tempfile ($!)";

printf "Writing index table to $index_dir/$index_table..." if ($opt_p);

$rows = 0;
while ($line=<TMP>)
  { $rows++;

    # First, we need to get the values stored in the temporary file record:

    $offset = 0;
    for ($i=0; $i<@colname; $i++)
      { $n = $colname[$i];
        $fld[$i] = substr($line,$offset,$colmaxsize{$n}); 
        $offset += $coldefsize{$n} + 1;
      }

    # Now write them out to the .tab file with the right delimiters:

    $n = $colname[0];
    if ($colquoted{$n})
      { printf TAB "\"%-*s\"",$colmaxsize{$n},$fld[0]; }
    else
      { printf TAB "%-*s", $colmaxsize{$n},$fld[0]; }

    for ($i=1; $i<@colname; $i++)
      { $n = $colname[$i];
        if ($colquoted{$n})
          { printf TAB ",\"%-*s\"",$colmaxsize{$n},$fld[$i]; }
        else
          { printf TAB ",%-*s", $colmaxsize{$n},$fld[$i]; }
      }
    printf TAB "\r\n";

    # Next line.
  }

close(TAB);
close(TMP);

printf "Done with index table file.\n" if ($opt_p);

# Now to write the label:

printf "Writing $index_dir/$index_label..." if ($opt_p);

$row_bytes = @colname + 1;  # For "," delimiters and the "\r\n" at the end
for ($i=0; $i<@colname; $i++)
  { $row_bytes += $colmaxsize{$colname[$i]};
    $row_bytes += 2 if ($colquoted{$colname[$i]}); 
  }

# Check for valid start and stop times:

$vol_start_time = "\"N/A\"" if (!$vol_start_time);
$vol_stop_time  = "\"N/A\"" if (!$vol_stop_time);

# Force the pointer value to the same case as the index file paths, if 
# requested:

if ($opt_u)
  { $index_table =~ tr/a-z/A-Z/; }
elsif ($opt_l)
  { $index_Table =~ tr/A-Z/a-z/; }

$fmt = "%-78.78s\r\n";
$columns = @colname;
printf LBL $fmt, "PDS_VERSION_ID = PDS3";
printf LBL $fmt, " ";
printf LBL $fmt, "RECORD_TYPE    = \"FIXED_LENGTH\"";
printf LBL $fmt, "FILE_RECORDS   = $rows";
printf LBL $fmt, "RECORD_BYTES   = $row_bytes";
printf LBL $fmt, " ";
printf LBL $fmt, "START_TIME     = $vol_start_time";
printf LBL $fmt, "STOP_TIME      = $vol_stop_time";
printf LBL $fmt, " ";
printf LBL $fmt, "^INDEX_TABLE   = \"$index_table\"";
printf LBL $fmt, " ";
printf LBL $fmt, "OBJECT     = INDEX_TABLE";
printf LBL $fmt, "  INDEX_TYPE         = \"SINGLE\"";
printf LBL $fmt, "  INTERCHANGE_FORMAT = \"ASCII\"";
printf LBL $fmt, "  ROWS               = $rows";
printf LBL $fmt, "  ROW_BYTES          = $row_bytes";
printf LBL $fmt, "  COLUMNS            = $columns";
printf LBL $fmt, " ";

# Looping through the columns:

$start_byte = 1;
for ($i=0; $i<@colname; $i++)
  { $n = $colname[$i];
    $start_byte++ if ($colquoted{$n});

    printf LBL $fmt, "  OBJECT     = COLUMN";

    if ($missing_data{$n})
      { $hold = ($colquoted{$n})? "\"$colmissing{$n}\"" : $colmissing{$n};

        printf LBL $fmt, "    NAME             = \"$n\"";
        printf LBL $fmt, "    START_BYTE       = $start_byte";
        printf LBL $fmt, "    BYTES            = $colmaxsize{$n}";
        printf LBL $fmt, "    DATA_TYPE        = \"$coltype{$n}\"";
        printf LBL $fmt, "    FORMAT           = \"A$colmaxsize{$n}\"";
        printf LBL $fmt, "    MISSING_CONSTANT = $hold";
        printf LBL $fmt, "    DESCRIPTION      = \"$coldesc{$n}\"";
      }
    else
      { printf LBL $fmt, "    NAME             = \"$n\"";
        printf LBL $fmt, "    START_BYTE       = $start_byte";
        printf LBL $fmt, "    BYTES            = $colmaxsize{$n}";
        printf LBL $fmt, "    DATA_TYPE        = \"$coltype{$n}\"";
        printf LBL $fmt, "    FORMAT           = \"A$colmaxsize{$n}\"";
        printf LBL $fmt, "    DESCRIPTION      = \"$coldesc{$n}\"";
      }


    printf LBL $fmt, "  END_OBJECT = COLUMN";
    printf LBL $fmt, " ";

    $start_byte++ if ($colquoted{$n});
    $start_byte += $colmaxsize{$n} + 1;
  }

# Wrap it up:

printf LBL $fmt, "END_OBJECT = INDEX_TABLE";
printf LBL $fmt, " ";
printf LBL $fmt, "END";

close (LBL);

printf "Done with index label file.\n" if ($opt_p);

# And we're done.  The temp file will be deleted on exit.

#____________________________________________________________________________
#____________________________________________________________________________
#____________________________________________________________________________

sub do_dir

  # Subroutine to loop through the contents of a directory, calling itself
  # recursively to handle subdirectories.

{ local ($dir,$vroot) = @_;      # Input directory to process and volume root

  my (@flist,$fl);
  my (@dlist,$dl);
  my ($file,$fullname);
  my ($i,$fullpath);

  $fl = $dl = 0;

  $fullpath = "$volume_root/$dir";
  opendir(DIR,$fullpath) || die "Could not open directory $fullpath ($!)";

  printf "In directory \"$dir\"...\n" if ($opt_p);

  foreach $file (readdir DIR)
    { next if ($file =~ /^\./);

      $fullname = "$fullpath/$file";

      # If this is a directory, add it to the subdirectory list:

      if (-d $fullname)
        { $dlist[$dl] = "$dir/$file";
          $dl++;
        }

      # If it's a label file, add it to the file list:

      elsif ($file =~ /\.lbl$/  ||  $file =~ /\.LBL$/)
        { $flist[$fl] = "$dir/$file";
          $fl++;
        }

      # And if it's not a label file but we're checking every file for
      # attached labels, look for the label and add the file if found:

      elsif ($opt_a && PDS_label($fullname))
        { $flist[$fl] = "$dir/$file";
          $fl++;
        }
    }

  # Now, loop through the labels, collecting lines for the index:

  foreach $file (sort @flist)
    { printf "  Parsing file $file\n" if ($opt_p);
      do_file($file,$vroot); 
    }

  # And loop through the subdirectories:

  foreach $file (sort @dlist)
    { do_dir($file,$vroot); }

  # And we're done.

  return;
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

sub do_file

  # Routine to read through a label, collect values, and write the 
  # preliminary index table record.

{ local ($label,$vroot) = @_;  # Label file and volume root

  my ($line);
  my ($name,$len);
  my (%val);
  my ($dsid,
      $product_id,
      $product_time,
      $start_time,
      $target_name,
      $host_name,
      $instrument_name);
  my ($filespec);

  undef %val;

  undef $dsid;
  undef $product_id;
  undef $product_time;
  undef $start_time;
  undef $target_name;
  undef $host_name;
  undef $instrument_name;

  open (LBL,"$vroot/$label")  || 
    die "Could not open $vroot/$label for reading ($!)";

  $line = <LBL>;
  while ($line !~ /^\s*END\s/)
    { foreach $name (@colname)
        { if ($line =~ /^\s*$name\s*=/)
            { $val{$name} = get_value($line,$label);
              $len = length($val{$name});
              $colmaxsize{$name} = ($len > $colmaxsize{$name})? 
                                    $len : $colmaxsize{$name};
            }
        }
      $line = <LBL>;
    }

  close (LBL);

  # Check that all values were found, supply missing values where not, and
  # check for maximum lengths:

  foreach $name (@colname)
    { # Skip this check for values that come from outside the label:

      next if ($name eq "VOLUME_ID"  ||  $name eq "FILE_SPECIFICATION_NAME");

      if (! $val{$name})
        { $val{$name} = $colmissing{$name}; 
          $len = length($val{$name});
          $colmaxsize{$name} = ($len > $colmaxsize{$name})? 
                                $len : $colmaxsize{$name};
          $missing_data{$name} = $TRUE;
        }
    }

  # If we have a START_TIME, we'll check it against the volume start time:

  if ($val{"START_TIME"}  &&  ($val{"START_TIME"} ne "N/A"))
    { $start_time = $val{"START_TIME"};
      if ($vol_start_time)
        { $vol_start_time = ($vol_start_time gt $start_time)? 
                            $start_time : $vol_start_time; 
          $vol_stop_time  = ($vol_stop_time  lt $start_time)?
                            $start_time : $vol_stop_time;
        }
      else
        { $vol_start_time = $start_time;
          $vol_stop_time  = $start_time;
        }
    }

  # We have to set the max for FILE_SPECIFICATION_NAME separately, since it's
  # not actually in the label:

  $name = "FILE_SPECIFICATION_NAME";
  $len  = length($label);
  $colmaxsize{$name} = ($len > $colmaxsize{$name})?
                        $len : $colmaxsize{$name};

  # Force case on the filespec if requested:

  $filespec = $label;
  if ($opt_u)
    { $filespec =~ tr/a-z/A-Z/; }
  elsif ($opt_l)
    { $filespec =~ tr/A-Z/a-z/; }

  # And print the index line to the temporary file open on $TMP:

  printf $TMP "%-*s", $coldefsize{"VOLUME_ID"},$volume_id;
  printf $TMP " %-*s",$coldefsize{"FILE_SPECIFICATION_NAME"},$filespec;
  for ($i=2; $i<@colname; $i++)
    { printf $TMP " %-*s",$coldefsize{$colname[$i]}, $val{$colname[$i]}; }
  printf $TMP "\r\n";

  # And we're done.

  return;
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

sub PDS_label

  # Returns TRUE if the file begins with a PDS label ("PDS_VERSION_ID" at
  # the beginning of the file).

{ local ($file) = $_[0];
  my ($check);

  open (CHK, $file)  ||  die "Could not open $file to check for label ($!)";
  read CHK,$check,60;
  close (CHK);

  if ($check =~ /PDS_VERSION_ID/)
    { return $TRUE; }
  else
    { return $FALSE; }
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

sub get_value

  # Extracts the keyword value from the label line.  Values are returned 
  # without quotes or blank padding.  If the value runs over more than 
  # one line it will be truncated (but it can be on the line after the 
  # keyword).  If there are multiple values for a keyword, an error is
  # signalled and the first value is returned.

{ local ($inline,$filename) = @_;    # Line to be parsed and label file name

  my ($key,$val,@v);
  my ($rem);
  my ($next,$i);

  # Parse the input line into keyword and remainder:

  $inline =~ /^\s*(\S+)\s*=/;
  $key = $1;
  $rem = substr($inline,index($inline,"=")+1);
  chomp $rem;
  $rem =~ s/^\s+//;
  $rem =~ s/\s+$//;

  # OK, now things get complicated.  First, we want to make sure we have
  # enough of the value to work with.  There are a number of cases:
  #
  #   1. Remainder is null (read the next line and continue)
  #   2. Remainder is just an opening quote 
  #   3. Remainder is a quote with some text after it but no closing quote
  #   4. Remainder contains matched quotes
  #   5. Remainder contains no quotes
  #
  # We'll try to deal with these in order of increasing complexity.  In no
  # case will we read more than two additional lines.  We are assuming, of
  # course, that all the labels are syntactically correct.

  if (length($rem)==0)
    { $rem = <LBL>;
      chomp $rem;
      $rem =~ s/^\s+//;
      $rem =~ s/\s+$//;
    }

  # If we got just an opening quote, read the next line:

  if ($rem eq "\"")
    { $next = <LBL>;
      chomp $next;
      $next =~ s/^\s+//;
      $next =~ s/\s+$//;
      $rem = "\"$next";

      # If we still don't have balanced quotes, signal an error but continue:

      if ($rem !~ /".*"/)
        { printf STDERR "ERROR: $filename - $key value too long.\n";
          $rem .= "\"";
        }
    }

  # If we have an opening quote without a closing quote, we read the next
  # line and concatenate it to what we've got:

  elsif ($rem =~ /"/  &&  $rem !~ /".+"/)
    { $next = <LBL>;
      chomp $next;
      $next =~ s/^\s+//;
      $next =~ s/\s+$//;
      $rem = "$rem $next";

      # If we still don't have balanced quotes, signal an error but continue:

      if ($rem !~ /".*"/)
        { printf STDERR "ERROR: $filename - $key value too long.\n";
          $rem .= "\"";
        }
    }

  # If we have no quotes, we drop everything after the first blank:

  elsif ($rem !~ /"/  &&  $rem =~ / /)
    { $i = index($rem," ");
      $rem = substr($rem,0,$i);
    }

  # Otherwise, we're set with a remainder string that contains at least two
  # quotes.

  # Next complication: sets and series.  If the first character of the remainder
  # is '(' or '{', we delete all [({] until we get to the start of the first
  # value, which we want to extract:

  if ($rem =~ /^[({]/)
    { $rem =~ s/^[({]+\s*//;

      # If what's left begins with a quote, we drop everything after the
      # second quote:

      if ($rem =~ /^("[^"]+")/)
        { $rem = $1; }

      # Otherwise, we drop everything after the identifier:

      else
        { $rem =~ s/[^A-Za-z0-9_].*$//; }

      # Now print the error message:

      printf STDERR "ERROR: $filename has multiple values for $key. ";
      printf STDERR "Only first value ($rem) included.\n";
    }

  # Now in the remainder we have either no quotes, or at least two quotes.
  # If there are no quotes, we're done.  If there are any quotes, we want
  # just what's in the first set:

  if ($rem =~ /^"([^"]+)"/)
    { $val = $1;}
  else
    { $val = $rem; }

  # We'll do a final trim just in case someon did something silly in quotes:

  $val =~ s/^\s+//;
  $val =~ s/\s+$//;

  # And we're done:

  return $val;
}
