#
#  Subroutine to read the next logical line of a PDS/ODL file.  The next 
#  logical line is the next line which is not blank or a comment, and if it
#  is an assignment statement, it includes the entire value part of the line,
#  even if this runs over more than one line (within quotes, of course).
#
#  NOTE: The full value of long text fields is NOT returned!
#
#  A null value is returned if the read fails.
#
#  06 June 1997, acr: Fixed problem detecting closing ")"
#  04 Jun 1999, acr: removed "die" statements and return 'undef' in case of 
#                    read error.
#  26 Aug 1999, acr: added code to deal with lines which have '=' characters
#                    within the value string
#  16 Jun 2000, acr: added 'splitODLline' to return key and value parts;
#                    modified 'nexPDSODL' to preserve carriage control and
#                    blanks in DESCRIPTION and NOTE fields
#  18 Sep 2000, acr: modified to retain line break after first part of NOTE
#                    and DESCRIPTION fields and to trim a terminal newline (\n) 
#                    before returning the line string; added 'trimquotes'
#  13 Dec 2000, acr: Fixed bug which caused problems with keywords having a 
#                    value of '0', without quotes.
#  29 Apr 2008, acr: Fixed faulty logic in "cleanline" subroutine that trimmed
#                    "comments" inside quoted strings of sets and series (found
#                    in mildly odd Rosetta label syntax)
#
#==============================================================================

sub nextPDSODL

  # Routine to return the next complete line from the input label file open
  # on <ODL>.

{ local(*ODL,$maxlen) = @_;   # Maximum return length; file handle to read from
  local($buff);       # Line buffer
  local($line);       # Line to be returned
  local($not_done);   # Loop flag
  local($val,$t);     # value part of line and remainder
  local($qt);         # Quote character
  local($br);         # Bracket character
  local($brmeta);     # Bracket character after quotemeta()
  local(@sv);         # Array to save line contents when 'split' finds
                      #  additional '=' characters
  local($in_quote);   # True if we're searching for the end of a quoted string
  local($in_desc);    # True if we're working on a DESCRIPTION or NOTE value

  $maxlen = 255 unless ($maxlen);   # Set default length
  $line = "";
  $in_quote = 0;

  # Get the next line, formatted for processing:

  return undef unless ($buff = &cleanline);

  # If there is no "=", return now:

  if ($buff !~ /=/)
    { $buf =~ s/\n$//;
      return $buff;
    }

  # We have at least the beginning of an assignment statement.  Save
  # it into $line:

  $line = "$buff\n";

  # Check for NOTE and DESCRIPTION fields:

  $in_desc = ($line =~ /^[^=]*DESCRIPTION\s*=/  ||
              $line =~ /^[^=]*DESC\s*=/         ||
              $line =~ /^[^=]*NOTE\s*=/);

  # Now, we have to make sure we have the entire value.  The cases are:
  #
  #  1. If nothing follows the "=", we must read the next line to find 
  #     the value.
  #  2. If a simple value follows the "=", return
  #  3. If a quote follows the "=", we must read lines until we find a
  #     matching end quote.
  #  4. If a bracket follows the "=", we must read lines until we find 
  #     a matching closing bracket.  In this case we ignore quotes and
  #     hope for the best.
  #

  ($t,$val,@sv) = split(/=/,$buff);

  $val = join("=",($val,@sv));

  # 1. Read next line

  if (! $val  &&  $val ne "0")
    { if ($buff = &cleanline)
        { $line = "$line $buff";
          $val = $buff;
        }
      else
        { return undef; }
    }

  # 2. Simple value (and check for a matching quote on this line):

  if ($val !~ /^("|'|{|\()/  ||  $val =~ /^".*"/   ||  $val =~ /^'.*'/  ||
                                 $val =~ /^\(.*\)/ ||  $val =~ /^\{.*\}/  )
    { $line =~ s/\n$//;
      return $line; 
    }

  # 3. Find matching quote.

  if ($val =~ /^("|')/)
    { $qt = substr($val,0,1);

      $not_done = 1;
      while ($not_done)
        { return undef unless $buff = &cleanline($qt,$in_desc); 

          # Look for closing quote.  If found, delete anything after it and set
          # loop flag:

          if ($buff =~ /$qt/)
            { $buff =~ s/$qt.*$/$qt/;
              $not_done = 0;
            }

          # If the line is more than $maxlen characters long, then assume we're 
          # in some kind of description field and ignore it.  Otherwise, 
          # concatenate it to the end of the current buffer.  The last line is
          # always added so that the quotes will balance.

          if (length($line) < $maxlen  ||  ! $not_done)
            { $line = "$line $buff"; }
        }
    }

  # 4. Find matching bracket.

  else
    { $br = ($val =~ /^\(/ ) ? ")" : "}" ;
      $brmeta = quotemeta($br);

      $not_done = 1;
      while ($not_done)
        { return undef unless $buff = &cleanline($in_quote,$in_desc); 

          # Look for closing bracket.  If found, delete anything after it and
          # set loop flag:

          if ($buff =~ /$brmeta/)
            { $buff =~ s/$brmeta.*$/$br/;
              $not_done = 0;
            }
          $line = "$line $buff";
        }
    }

  # If the very last character in the buffer is a newline, we delete it be
  # before returning:

  $line =~ s/\n$//;

  # Return the line:

  return $line;

}

#-----------------------------------------------------------------------------

sub cleanline

  # Reads the next line from <ODL> and cleans it up for processing. If we are
  # currently inside quotes, the $quote parameter is not null and contains the 
  # opening quote character we are looking to match. The second parameter will
  # be TRUE if we are also in a DESCRIPTION or NOTE field.

{ local($quote,$indesc) = @_; # quote character, desc. flag
  local($buff);               # input buffer
  local($ix,$qt,$qcnt);       # For local quote counting

  $buff = "";
  if (length($quote) > 1)    # Make sure we've got valid input
    { undef $quote;
      undef $indesc;         # If the first isn't valid, ignore the second
    }

  while ($buff eq "")
    { return undef unless ($buff = <ODL>);

      $buff =~ s/\s*$//;     # Remove all record delimiters

      # Remove comments.  This is tricky, since the comment delimiter may
      # appear within an open quote, or after a closing quote, and there may
      # be another '=' within the quotes. We may also already be inside 
      # quotes opened on a previous line.
      #
      # Note that we're coding some dangerous assumptions in here to save
      # some significant programming effort.  Specifically, we're assuming:
      #
      #  a. Any quoted value running over more than a line is descriptive
      #     text - i.e., a long scalar, not part of a sequence or set;
      #  b. Any sequence or set will have balanced quotes on any single 
      #     line in their definition - i.e., no single value will have an
      #     opening quote on one line and the closing quote on the next;
      #  c. All quotes on the line will be of the same type, double or single,
      #     as determined by the first one encountered;
      #  d. No quoted scalar value will ever have a unit specifier following
      #     it.
      #
      # These assumptions are generally valid, but are not required by PDS/ODL
      # syntax.

      # First, the easy case.  If we're looking for a closing quote and we find
      # one, we just discard anything after the closing quote:

      if ($quote)
        { $buff =~ s/${quote}.*$/$quote/; }

      # Next, if we have a value beginning with a quote, it's a quoted scalar
      # and we need to check for a closing quote.  If we have one, we discard
      # everything that follows it.  (We've explicitly coded an option for 
      # each type of quote):

      elsif ($buff =~ /^[^=]*=\s*("|')/)  
        { if ($buff =~ /^[^=]*=\s*"/)      # Double-quoted scalar
            { $buff =~ s/^([^=]*=\s*"[^"]*").*$/$1/; } 
          else                             # Single-quoted scalar
            { $buff =~ s/^([^=]*=\s*'[^']*').*$/$1/; }
        }

      # Next we check for the beginning of a set or sequence.  If we've got 
      # one, then we check for balanced quotes.  If we don't have balanced
      # quotes, we give up and signal a message that we hope is helpful.
      
      elsif ($buff =~ /^[^=]*=\s*(\(|{)/)   # Set or Sequence
        { # First, see if we have any quotes. If so, count the quotes and
          # check for an even number:

          if ($buff =~ /^[^"']*('|")/)
            { $qt = $1;
              $qcnt = 0;
              $ix = index($buff,$qt) + 1;
              while ($ix>0)
                { $qcnt++;
                  $ix = index($buff,$qt,$ix) + 1;
                }

              if ($qcnt % 2)     # True if $qcnt is NOT even
                { printf stderr "Could not parse syntax in '$buff'\n";
                  printf stderr "Giving up...\n";
                  return undef;
                }

              # OK, we've got balanced quotes, so we can trim comments on
              # the end of the line - after the last quote:

              $buff =~ s/\s*\/\*[^$qt]*$//;
            }

          # No quotes, so we can trim comments safely:

          else
            { $buff =~ s/\s*\/\*.*$//; }

          # Bracket matching is done elsewhere, so we're done for now:
        }

      # By process of elimination, we're left with a naked scalar (a number,
      # time or identifier).  In this case we can safely trim trailing blanks:

      else                                 # Plain scalar
        { $buff =~ s/\s*\/\*.*$//; }

      # If we're inside a NOTE or DESCRIPTION field we add a linefeed.
      # Otherwise we trim blank space and return.

      if ($indesc)
        { $buff .= "\n"; }
      else
        { $buff =~ s/^\s+//;
          $buff =~ s/\s*$//;
          $buff =~ s/\s*=\s*/=/;
        }
    }

  return $buff;
}

#.............................................................................

sub getval

  # mini routine to get the value field of an ODL line.  This routine
  # assumes the line has been prepared by 'nextPDSODL'.  Leading and
  # trailing blanks are trimmed.  If the value is a sequence or series,
  # blanks around commas are also trimmed.

  { local ($buff = $_[0]);
    local ($val);

    $val = $buff;
    $val =~ s/^[^=]*=//;  # Remove everything up to the FIRST '='

    # Clean up sequences and series:

    if ($val =~ /^[({]/)
      { $val =~ s/^([({])\s+/$1/;
        $val =~ s/\s+([)}])$/$1/;
        $val =~ s/\s*,\s*/,/g;
      }

    # Return the value:

    return $val;
  }

#.............................................................................

sub splitODLline

  # Routine to split a line prepared by 'nextPDSODL' into keyword and 
  # value parts.

  { local ($buff = $_[0]);
    local ($key,$val);

    # If the line contains a full DESCRIPTION or NOTE, there will be carriage
    # control in the string. So to get just the keyword value we first remove
    # all whitespace, to prevent the match from being stopped short at the 
    # linefeed characters. Then we delete the first '=' and everything after
    # it:

    $key = $buff;
    $key =~ s/\s*//g;
    $key =~ s/^([^=]*)=.*$/$1/;

    $val = $buff;
    $val =~ s/^[^=]*=\s*//;

    # Clean up sequences and series:

    if ($val =~ /^[({]/)
      { $val =~ s/^([({])\s+/$1/;
        $val =~ s/\s+([)}])$/$1/;
        $val =~ s/\s*,\s*/,/g;
      }

    # Return the values:

    return ($key,$val);
  }

#-----------------------------------------------------------------------------


sub isPDSlabel

# Function returning 1 if the given file starts with a PDS label, 0 otherwise.
# PDS labels must have the PDS_VERSION_ID keyword as the first in the label.
# or must have "OBJECT = REFERENCE" as the first line.

{ my ($file) = $_[0];
  my ($line);

  open (IN,$file) || die "Could not open '$file' for PDS-check.";

  while ($line = <IN>)
    { last if ($line !~ /^\s*$/  &&  $line !~ /^\s*\/\*/); }
  close(IN);

  if ($line =~ /^\s*PDS_VERSION_ID\s*=/)
    { return 1; }
  else
    { return 0; }
}

#-----------------------------------------------------------------------------

sub trimquotes

# Function to trim quotes around the input value (single or double). Quotes
# must be balanced to be removed.

{ my ($val) = $_[0];

  if ($val =~ /^"/  &&  $val =~ /"$/)
    { $val =~ s/^"//;
      $val =~ s/"$//;
    }

  elsif ($val =~ /^'/  &&  $val =~ /'$/)
    { $val =~ s/^'//;
      $val =~ s/'$//;
    }

  return $val;
}

#-----------------------------------------------------------------------------

1;
