/*

   nullval:  Routine to insert null value flags into blank fields in ASCII
             tables.

   Format:   % nullval [ -o <output file> ] <input file>
                       <start byte> <null value> [ ... ]

             where
                <input file> is an input ASCII file, preferably with fixed-
                             length records
                <start byte> is the first byte of the field to be checked
                             as it appears in the PDS label (i.e., the first
                             character in the record is byte 1)
                <null value> the character string to replace blank fields.
                             This mus be enclosed in quotes.

            if no <output file> is given, the results are directed to stdout.

  07 April 1995, acr.
*/

#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

#define MAXRECLENGTH   10000     /* Maximum record length */
#define MAXFIELDS        100     /* Maximum number of fields */

main(argc,argv)
  int   argc;
  char *argv[];

{ FILE  *ifp,*ofp;          /* input/output file pointers     */
  char  line[MAXRECLENGTH]; /* input line buffer              */
  char  file[250];          /* file name buffer               */
  int   start[MAXFIELDS];   /* start byte array               */
  char *nval[MAXFIELDS];    /* pointers to null values        */
  int   length[MAXFIELDS];  /* the field lengths              */
  int   linelength;         /* length of input line           */
  int   sv;                 /* subscript for MAXFIELDS arrays */
  char *rem;                /* remainder pointer              */
  int   blank;              /* blank field flag               */
  int   record;             /* record count                   */
  int   looplen;            /* string length for loop boundary*/
  int   i,j,k;              /* loop/subscript variables       */


  /* Collect command line arguments and open files: */

  /* ... check for minimum number of arguments ... */

  if (argc < 4  ||  (argc % 2 != 0))           
    { fprintf (stderr,"Usage: nullvall <input file> <start byte> ");
      fprintf (stderr,"<null value>  [ ... ]\n");
      exit(0);
    }

  if (strcmp(argv[1],"-o") == 0)
    { 
      /* ... re-check minimum arguments ... */

      if (argc < 6)
        { fprintf (stderr,"Usage: nullvall <input file> <start byte> ");
          fprintf (stderr,"<null value>  [ ... ]\n");
          exit(0);
        }

      /* ... open output file ... */

      if ((ofp = fopen(argv[2],"w")) == NULL)
        { fprintf (stderr,"nullval: Cannot open %s for writing (%s).\n",
                           argv[2],strerror(errno));
          exit(10);
        }

      /* ... set next argument pointer ... */

      i = 3;
    }
  else
    { ofp = stdout;
      i = 1;
    }

  /* ... Open input file ... */

  if ((ifp = fopen(argv[i],"r")) == NULL)
    { fprintf (stderr,"nullval: Cannot open %s for reading (%s).\n",
                       argv[i],strerror(errno));
      exit(11);
    }

  i++;

  /* Loop through remaining arguments: */

  sv = 0;
  for (j=i; j<argc; j=j+2)
    { start[sv] = strtol(argv[j],&rem,10) - 1;
      if ( (strcmp(rem,"") != 0)  ||  (start[sv] < 0) )
        { fprintf (stderr,"nullval: Invalid start byte (%s).\n",argv[j]);
          exit(20);
        }
       nval[sv] = argv[j+1];
       length[sv] = strlen(nval[sv]);
       sv++;
    }

/*--------------------------------------------------------------------------

  Now that the arguments are dealt with, process the file.

  As each line is read in, the successive fields are checked for blanks.
  When an entirely blank field is found, the null value supplied is used to
  replace it.
*/

  record = 0;
  while (fgets(line,MAXRECLENGTH,ifp) != NULL)
    { record++;

      /* First, make sure the last character in the array is a newline.  If
         it isn't you've exceeded the file record length:
      */

      linelength = strlen(line);
      if (line[linelength-1] != '\n')
        { fprintf (stderr, "nullval: Maximum record length (%d) exceeded ",
                            MAXRECLENGTH)   ;
          fprintf (stderr, "at record %f.\n", record);
        }

      /* Continue regardless: */

      for (i=0; i<sv; ++i)
        { 
          /* First, make sure the field lies within the record.  If not,
             skip this field:
          */

          if (start[i] > linelength)
            { fprintf (stderr, "nullval: Field %d starting point (byte %d) ",
                                i, start[i]);
              fprintf (stderr, "exceeds end of record %d (%d bytes).  ",
                                record, linelength);
              fprintf (stderr, "Ignored.\n");
              continue;
            }

          /* Second, check to see if the field extends beyond the end of 
             the record:
          */

          if (start[i]+length[i] >= linelength)
            { fprintf (stderr, "nullval: null value extends past ");
              fprintf (stderr, "end of record %d; truncated. \n",record);
              looplen = linelength - start[i];
            }
          else
            { looplen = length[i]; }

          /* Then, make sure the field is blank: */

          blank = 1;
          for (j=start[i]; j<start[i]+looplen; j++)
            { if (line[j] != ' ') blank = 0; }

          if (blank)
            {
              /* Transfer the value: */

              for (j=0; j<looplen; j++)
                { line[start[i]+j] = nval[i][j]; }
            }
        }
      fputs(line,ofp);
    }

  fclose(ifp);
  fclose(ofp);

  exit(0);
}
