#!/usr/bin/perl

#  Routine to generate an index file from PDS labels by extracting keywords
#  from the labels as specified by the input.
#
#  14 Nov 2007, amd: Adapted from version developed by acr to properly
#                    handle namespaces, to allow the extraction of
#                    individual values from sequences or sets, and to
#                    allow for the automatic expansion of sequences or 
#                    sets into multiple columns.
#
#  Input options:
#
#     -n            Do not use default keywords
#     -r            Descend recursively through subdirectories
#     -v            Include "VOLUME_ID" in default list.
#     -u            Remove units from numeric values.
#     -o <file>     Send output to <file>
#     -f <file>     Read keywords from <file>
#     -d <dir>      Start at directory <dir>
#
# Modification History
# ====================
#  15 Feb 2008, ACR: Added "use lib" statement for configuration

# Import necessary libraries

use lib "/n/sbnops01/lcltools/lib/perl5";
use ODLlib;

# Define the default keywords, one for when the option -v isn't supplied,
# and one for when it is.

@standardKeywords = ("^FILE_SPECIFICATION_NAME/-50",
		     "PRODUCT_ID/-30",
		     "DATA_SET_ID/-40",
		     "PRODUCT_CREATION_TIME/19");
@standardKeywordsV = ("^FILE_SPECIFICATION_NAME/-50",
		      "PRODUCT_ID/-30",
		      "DATA_SET_ID/-40",
		      "VOLUME_ID/-11/NULL_0001",
		      "PRODUCT_CREATION_TIME/19");

# Subroutine prototypes

sub main ( \@ );
sub parseArguments ( $ );
sub usage ();
sub getOutputStream ( $ );
sub getKeywordDefs ( $ );
sub ensureKeywordSyntax ( $ );
sub getMatchingFiles ( $ );
sub handleKeyword ( $$ );
sub fixOutput ( \@ );
sub dumpOutput ( $$$\@ );

# subroutines for try/catch/finally blocks

sub try (&$ );
sub catch (&;$ );
sub finally (&;$ );

# Run the program, via the 'main' subroutine

exit (main (@ARGV));

# Subroutine definitions

##############################################################################
# main
#
# Top-level program logic.
#
# Parameters: An array containing the program arguments
#
# Returns: 0 (signaling success)
#

sub main (\@)
{

    # Get parameters

    my (
	$argv # Command-line arguments passed to the program
	) = @_;

    # Declare local variables

    my (
	$arguments,    # Hash of arguments
	$outputStream, # Destination of output table
	$keywordDefs,  # Keyword definitions, default and user-given
	$files,        # List of files to parse
	$output,       # In-memory representation of index table
	$object,       # ODL object, representing individual labels
	$keyword       # Individual keyword definition
	);

    # Parse the arguments from the command line

    $arguments = parseArguments ($argv);

    # Get a handle to the output stream, creating a file if necessary

    $outputStream = getOutputStream ($arguments);

    # Get the keyword definitions, from the default lists, a file given by
    # the user, and keywords given on the command line

    $keywordDefs = getKeywordDefs ($arguments);

    # Get a list of files to index

    $files = getMatchingFiles ($arguments);

    # Initialize the output.

    @output = ();

    # Handle each file

    foreach $file (@$files)
    {

	# Create a representation of the label file as an ODL object,
	# PDS-based sanity checks instead of plain ODL.

	$object = ODLlib->create ($file, {base => "PDS"});

	# Initialize a row in the table for the file

	$row = [];

	# Handle each given keyword

	foreach $keyword (@$keywordDefs)
	{

	    # Put the value taken from the ODL object, for the given keyword,
	    # into the row.

	    push (@$row, handleKeyword ($object, $keyword));
	}

	# Add the row to the output table.

	push (@output, $row);
    }

    # Fix columns in the output table. This is needed when a value taken from
    # the ODL file is a sequence or set, since these values are expanded so
    # that each value in the sequence or set has it's own cell. When this
    # happens, the coresponding cells in other rows of the table must also 
    # be expanded into multiple cells, to ensure that later cells still
    # line up.

    fixOutput (@output);

    # Print the table to the output stream.

    dumpOutput ($arguments, $outputStream, $keywordDefs, @output);

    # Return 0 for success.

    return 0;
}

##############################################################################
# parseArguments
#
# Parse the command-line arguments into a more useful form
#
# Parameters: A reference to the command-line arguments
#
# Returns: A hash representing the command-line options
#

sub parseArguments ($)
{
    # Get parameters

    my (
	$argv # Command-line arguments
	) = @_;

    # Declare local variables

    my (
	$argument,        # An individual argument from the command-line
	$arguments,       # A hash of the parsed arguments
	$outputFile,      # The name of the output file
	$keywordListFile, # The name of the file containing keyword definitions
	$rootDirectory,   # The directory in which to search for files.
	@keywordDefs      # Array of command-line keyword definitions
	);

    # Initialize the arguments hash

    $arguments = {};

    # Continue while there are more arguments to parse

    while ($argument = shift (@$argv))
    {

	# If the argument begins with a hyphen...

	if ($argument =~ /^\-.*/)
	{

	    # remove the hyphen

	    $argument =~ s/^\-//;

	    # While there are more letters in the argument...

	    while (length ($argument) > 0)
	    {

		# Get the next letter

		$letter = substr ($argument, 0, 1, "");

		# If it's one of the switches (-n, -r, -v, -c -u)

		if ($letter =~ /^[nrvcu]$/)
		{

		    # Set that switches flag to 1 in the arguments hash

		    $arguments->{$letter} = 1;
		}

		# Otherwise, if it's an argument that is associated with
		# a value (-o <file>, -f <file>, -d <dir>:

		elsif ($letter =~ /^[ofd]$/)
		{

		    # If there are other letters after the argument, 
		    # it is malformed (e.g. -or <file> instead of -ro <file>)

		    if ($argument ne "")
		    {

			# Display usage notes for the program

			usage ();

			# Exit with an error

			exit (1);
		    }

		    # Otherwise, it's the last letter in the argument

		    else
		    {

			# Associate the NEXT argument with the letter, in the
			# arguments hash.
			# In this way, 
			#   -o output.tab
			# in the command line will result in
			#   $arguments->{o} == "output.tab"
			# in the program.

			$arguments->{$letter} = shift (@$argv);
		    }
		}

		# Otherwise, it's not a known argument

		else
		{

		    # Print the usage

		    usage ();

		    # It could be the -h option, which isn't really an
		    # error, so only exit with 1 (signaling an error) if
		    # the argument wasn't 'h'.

		    exit ($letter eq "h"?0:1);
		}
	    }
	}

	# It wasn't a command-line switch, so it must be the beginning
	# of the keyword definitions. Read the rest of the keyword
	# definitions from the command line into an array

	else
	{

	    # add the argument taken from $argv, followed by the rest
	    # of $argv.

	    push (@keywordDefs, $argument, splice(@$argv));
	}
    }

    # Store the keyword definitions in the arguments hash

    $arguments->{keywordDefs} = \@keywordDefs;

    # Return the arguments hash

    return $arguments;
}

##############################################################################
# usage
#
# Display usage notes.
# 
# Parameters: none
#

sub usage ()
{
    # Declare local variables

    my (
	$length # The length of $0 (the program name), used to make
	        # the output line up properly.
	);

    # Compute the length of $0, + 1 for a space after it

    $length = length ($0) + 1;

    # Print the usage

    print ("$0 [-n] [-r] [-v] [-u]\n".
	   " " x $length . "[-o <output file>]\n".
	   " " x $length . "[-f <keyword list file>]\n".
	   " " x $length . "[-d <root directory]\n".
	   " " x $length . "[ [ <keyword>/<width>[/<default>] ] ... ]\n"
	   );
}

##############################################################################
# Retrieve the appropriate output stream to print the resulting table to.
sub getOutputStream ($)
{
    # Get parameters

    my (
	$arguments # Arguments hash (reference)
	) = @_;

    # If the -o option was given, and the filename wasn't specified as '-'
    # (meaning STDOUT), then we are writing to a file

    if (exists ($arguments->{o}) && $arguments->{o} ne "-")
    {

	# Open the file for writing, or die trying

	open (outputFile, ">", $arguments->{o}) or
	    die ("Could not open the file ".$arguments->{o}." for writing.");

	# Return the output file stream

	return outputFile;
    }

    # Otherwise, we are writing to STDOUT

    else
    {

	# Return the standard output stream

	return STDOUT;
    }
}

##############################################################################
# getKeywordDefs
#
# Get the keyword definitions. These come from 3 places: The standard keyword
# list (with or without the VOLUME_ID keyword depending on the -n option),
# Any keyword specification file, and any command-line keywords
#
# Parameters: A hash representing the arguments passed to the program, as
#             returned by parseArguments.
#
# Returns: A reference to an array containing all the needed keyword
#          specifications.
#

sub getKeywordDefs ($)
{
    # Get parameters

    my (
	$arguments # Arguments hash (reference)
	) = @_;

    # Declare local variables

    my (
	@keywords, # Array of keywords
	$line      # A line from the keyword definition file
	);

    # unless the -n option was given, we're using default keywords

    unless (exists ($arguments->{n}))
    {
	# depending on whether or not the -v option was given, we initialize
	# the keyword definition to one of two arrays.

	@keywords = exists ($arguments->{v})?@standardKeywordsV:
	                                     @standardKeywords;
    }

    # If the user gave a file containing keyword definitions

    if (exists ($arguments->{f}))
    {

	# Open the file for reading, or die trying.

	open (keywordListFile, "<", $arguments->{f})
	    or die ("Could not open the file ".$arguments->{f}." for reading.");

	# For each line in the file

	while ($line = <keywordListFile>)
	{

	    # Remove the trailing new-line

	    chomp $line;

	    # Remove comments.

	    $line =~ s/\s*\#.*$//;

	    # Check if the line still has something to process

	    if ($line !~ /^\s*$/)
	    {

		# Remove any leading whitespace

		$line =~ s/^\s*//;

		# Add the keyword to the list

		push (@keywords, $line);
	    }
	}
    }

    # Add all of the command-line keyword definitions

    push (@keywords, @{$arguments->{keywordDefs}});

    # Ensure each keyword is a valid keyword definition

    ensureKeywordSyntax ($_) foreach (@keywords);

    # Return a reference to the keyword array

    return \@keywords;
}

##############################################################################
# ensureKeywordSyntax
#
# Make sure that the keyword is formatted correctly. Dies otherwise.
#
# Parameters: A keyword specification string
#
# Returns: nothing
#

sub ensureKeywordSyntax ($)
{
    # Get parameters

    my (
	$keyword # Keyword definition to check
	) = @_;

    # Split by the '/' delimiter

    @items = split (/\//, $keyword);

    # Check the format.

    if (scalar(@items) < 1 || # There must be 1, 2, or 3 items,
	scalar(@items) > 3 || # when splitting by /
	                      # Syntax check for the keyword:
	$items[0] !~ /^[\+\-]?                 # Leading + or -
	               \^?(\w|\:)+(\[-?\d+\])? # Identifier, and optional
	               (\.(\w|\:)+(\[\d+\])?)* # sub-identifiers, for objects
	               (\(\d+(?:,\d+)?\))?     # Optional indexes
	              $/x)
    {
	die ("Illegal keyword syntax: $keyword\n");
    }
}

##############################################################################
# getMatchingFiles
#
# Find all label files to be used when generating the index table
# 
# Parameters: A hash representing the arguments passed to the program, as
#             returned by parseArguments.
#
# Returns: A reference to an array containing all of the filenames of
#          label files to be included in the index.
#

sub getMatchingFiles ($)
{

    # Get parameters

    my (
	$arguments # Arguments hash (reference)
	) = @_;

    # Declare local variables

    my (
	$dir,    # The starting directory
	$output, # Output from the `find` program, locating label files
	@files   # Array of label filenames
	);

    # Check if the user gave a starting directory

    if (exists ($arguments->{d}))
    {

	# The user gave a starting directory, use it

	$dir = $arguments->{d};
    }
    else
    {

	# The user didn't give a starting directory, use the current
	# directory.

	$dir = ".";
    }

    # Check if we are searching recursively

    if ($arguments->{r})
    {

	# We are searching recursively, find all files

	$output = `find $dir -type f -iname \\*.lbl`;
    }
    else
    {

	# We're not searching recursively, only look 1 level deep (don't
	# search into subdirectories)

	$output = `find $dir -maxdepth 1 -type f -iname \\*.lbl`;
    }

    # Split the output by newlines

    @files = split (/\n/, $output);

    # Return a reference to the array of filenames

    return \@files;
}

##############################################################################
# handleKeyword
#
# Gets the value specified by a keyword specification string for the given
# ODL object. Special keywords are handled appropriately and sets or 1D
# sequence values are left as-is (they're not expanded into multiple cells)
#
# Parameters: An ODL object representing a label file
#             A keyword specification string
#
# Returns: An ODL value object representing the value associated with the
#          given keyword specification.
#

sub handleKeyword ($$)
{

    # Get parameters

    my (
	$object, # ODL object representing label file
	$keyword # Keyword definition to extract a value for
	) = @_;

    # Declare local variables

    my (
	$value, # The value extracted from the ODL file.
	$error, # An error string representing what went wrong.
	@pointers,
	$pointer,
	$recordBytes,
	$width,
	$default,
	$error
	);

    # Check if the keyword is a special keyword, starting with '^'

    if ($keyword =~ /^[\+\-]?\^([^\/]*)(?:\/([^\/]*)(?:\/(.*))?)?$/)
    {

	# Get the rest of the keyword

	$keyword = $1;

	# Handle each special keyword differently...

	# FILE_SPECIFICATION_NAME is the full path and filename

	if ($keyword eq "FILE_SPECIFICATION_NAME")
	{
	    $value = ODLlib::createTextValue($object->{FILENAME});
	}

	# PATH is the path to the file

	elsif ($keyword eq "PATH")
	{
	    $value = $object->{FILENAME};
	    $value =~ s/\/[^\/]*$//;
	    $value = ODLlib::createTextValue ($value);
	}

	# LABEL is just the filename, no path

	elsif ($keyword eq "LABEL")
	{
	    $value = $object->{FILENAME};
	    $value =~ s/^.*\/([^\/]*)\.[^\/\.]*$/\1/;
	    $value = ODLlib::createTextValue ($value);
	}

	# FILE is the name of the file that contains the data

	elsif ($keyword eq "FILE")
	{

	    # Get the pointers

	    @pointers = $object->getPointers();

	    # Only worry about the 1st pointer

	    $pointer = $pointers[0];

	    # Handle each pointer type separately:

	    if ($pointer->{value}->isaValue ("TEXT_STRING"))
	    {

		# For text strings, just get the string value

		$value = $pointer->{value}->getRawValue();
	    }
	    elsif ($pointer->{value}->isaNumber())
	    {

		# For numeric values, the filename is the same as the
		# label filename

		$value = $object->{FILENAME};
		$value =~ s/^.*\/([^\/]*)$/\1/;
	    }
	    elsif ($pointer->{value}->isaSequence("SEQUENCE_1D"))
	    {

		# For sequences, e.g. ("TABLE.TAB", 1024 <BYTES>),
		# extract the 1st value from the sequence.

		$value = $pointer->{value}->{value}->[0]->getRawValue();
	    }
	    else
	    {

		# Not a known pointer value type

		die ("Unknown type");
	    }

	    # Create an ODL text value from what we've computed

	    $value = ODLlib::createTextValue ($value);
	}

	# OFFSET is the offset (in bytes) of the data within the
	# file that contains the data.

	elsif ($keyword eq "OFFSET")
	{

	    # Get the pointers

	    @pointers = $object->getPointers();

	    # Only use the 1st poitner

	    $pointer = $pointers[0];

	    # Handle each pointer type separately
	    # If the pointer is just a string (a filename), then
	    # the offset is 0 (within that file)

	    if ($pointer->{value}->isaValue ("TEXT_STRING"))
	    {
		$value = 0;
	    }

	    # If the pointer is a number, then it's some sort of
	    # offset. When units aren't specified, they're assumed
	    # to be RECORD_BYTES

	    elsif ($pointer->{value}->isaNumber())
	    {

		# Get the pointer value as a perl-native number

		$value = $pointer->{value}->getPerlNumber();

		# Check the units of the value

		if ($pointer->{value}->{units} eq "")
		{

		    # If the units are blank, we assume that the value
		    # actually counts records, so we need to multiply by
		    # the number of bytes in a reccord.
		    # Get the value of the RECORD_BYTES keyword

		    $recordBytes = $object->getValueObject("RECORD_BYTES");

		    # Multiply it to the offset we got earlier

		    $value = $value * $recordBytes->getPerlNumber();
		}
	    }

	    # If the pointer is a sequence, then we take the 2nd item
	    # in the sequence as the offset. This should be a number.

	    elsif ($pointer->{value}->isaSequence("SEQUENCE_1D"))
	    {

		# Get the 2nd value in the seqence as a perl-native number

		$value = $pointer->{value}->{value}->[1]->getPerlNumber();

		# Check the units of the value

		if ($pointer->{value}->{value}->[1]->{units} eq "")
		{

		    # If the units are blank, multiply by the number of
		    # bytes in a record.
		    # Get the value of the RECORD_BYTES keyword

		    $recordBytes = $object->getValueObject("RECORD_BYTES");

		    # Multiply it to the offset we got earlier

		    $value = $value * $recordBytes->getPerlNumber();
		}
	    }

	    # Create an ODL text value from the computed byte offset

	    $value = ODLlib::createTextValue ($value);
	}

	# OBJECT refers to the name of the 1st pointer in the label.
	# So, if '^TABLE" is the 1st pointer, the OBJECT would be 'TABLE"

	elsif ($keyword eq "OBJECT")
	{

	    # Get the pointers

	    @pointers = $object->getPointers();

	    # Only use the 1st pointer

	    $pointer = $pointers[0];

	    # Get the object identifier from the pointer, and create a new
	    # ODL text value from it.

	    $value = ODLlib::createTextValue ($pointer->{keyword}->{object});
	}

	# The keyword isn't any known special keyword. Abort.

	else
	{
	    die ("Unknown keyword ^".$keyword);
	}
    }

    # It's not a special keyword... assume it's a normal LABEL keyword.

    else
    {

	# Parse the keyword specification, extracting the portions into
	# local variables.

	$keyword =~ /^[\+\-]?([^\/]*)(?:\/([^\/]*)(?:\/(.*))?)?$/;
	($keyword, $width, $default) = ($1, $2, $3);

	# Attempt to find the keyword's value. If this fails at any point,
	# we'll just use 'undef', and fill it with it's default value later.

	try
	{

	    # Get the value. The ODL library that created $object actually
	    # parses the keyword syntax used by pdsidx, so we don't need to
	    # worry about navigating object trees or finding indexed values.

	    $value = $object->getValueObject ($keyword);

	    # If it's a 2D sequence, then we don't handle it... abort!

	    if ($value->isaSequence("SEQUENCE_2D"))
	    {
		$error = ("Automatic expansion of 2D sequences".
			  " is not permitted.");
	    }

	    # If it's not a scalar, 1D sequence, or set, then we don't
	    # handle it... abort!

	    elsif (!($value->isaScalar() ||
		     $value->isaSequence("SEQUENCE_1D") ||
		     $value->isaValue("SET")))
	    {
		$error = ("The value type: ".$value->getType().
			  " is not supported");
	    }

	    # We're done... $value is the ODL value object for the keyword
	}
	catch
	{

	    # If there was any problem in the try block, use undef as the
	    # value.
	    #
	    # (note that the fatal errors just set $error, so they won't be
	    # handled here)

	    $value = undef;
	};

	# if $error was defined, abort the program with the error message

	die ($error) if ($error ne "");
    }

    # Return the value

    return $value;
}

##############################################################################
# fixOutput
#
# Handles expanding columns with set or 1D sequence values into multiple
# columns, making sure the rest of the columns line up by adding null values
# where necessary. The changes are made in-place in the matrix parameter.
#
# Parameters: A matrix of values, where each row represents a single label file
#             and each column represents a single keyword specification.
# 
# Returns: nothing
#

sub fixOutput (\@)
{
    # Get parameters

    my (
	$output
	) = @_;

    # Declare local variables

    my (
	@columns,
	$i,
	$row,
	$size,
	@newOutput,
	@newRow,
	$j,
	$need,
	$have,
	);

    # Create array to hold column-counts for the expanded columns

    @columns = ();

    # Initialize the columns array to all 1s

    for ($i = 0; $i < scalar (@{$output->[0]}); ++$i)
    {
	$columns[$i] = 1;
    }

    # For each row in the table, we're going to look for cells that are
    # set or sequence values, making note of that and of the maximum number
    # of values we find.

    foreach $row (@$output)
    {

	# Go through each cell in the row

	for ($i = 0; $i < scalar(@$row); ++$i)
	{

	    # If the cell has a value, and it's not a scalar, we need to check
	    # how many values it holds

	    if (defined ($row->[$i]) &&
		! $row->[$i]->isaScalar())
	    {

		# Get the number of values held in this cell

		$size = $row->[$i]->getSize();

		# Update the expanded-column size if it isn't already
		# big enough

		$columns[$i] = $size if ($columns[$i] < $size);
	    }
	}
    }

    # Create an array for the output table

    @newOutput = ();

    # Create a temporary reference to a row (for the table)

    $newRow = ();

    #
    # The 1st row in the table holds the keyword definitions, since there
    # may be more columns than keyword definitions, and this is a reasonably
    # easy way to tell what keyword created what column
    #

    # Go through each expanded-column specification.

    for ($i = 0; $i < scalar(@columns); ++$i)
    {

	# Push the index of the keyword specification as many times as
	# needed for the number of columns the expanded-column expands to.
	# To clairify, if column 5 had to represent 3 values, then
	#   $columns[4] == 3 (arrays are zero-indexed)
	# which would push '4' into the new row 3 times, making it looks
	# like this: (0,1,2,3,4,4,4)

	for ($j = 0; $j < $columns[$i]; ++$j)
	{
	    push (@$newRow, $i);
	}
    }

    # Save the column-keyword associations as the 1st row in the table.

    push (@newOutput, $newRow);

    # Now, go through the real output, one row at a time

    foreach $row (@$output)
    {

	# Create a temporary array ref to hold the row

	$newRow = [];

	# For each item in the row, we need to insert it's values, and possibly
	# some null values if the column is expanded into more columns than
	# the value provides

	for ($i = 0; $i < scalar(@$row); ++$i)
	{

	    # $need is the number of columns the column expands to.

	    $need = $columns[$i];

	    # If the value isn't defined, then we have one value,
	    # which is undef

	    if (!defined ($row->[$i]))
	    {

		# we have 1 value

		$have = 1;

		# add undef

		push (@$newRow, undef);
	    }

	    # If the value is a scalar, then we have one value, which is
	    # whatever the value is.

	    elsif ($row->[$i]->isaScalar())
	    {

		# We have 1 value

		$have = 1;

		# Add the value to the row

		push (@$newRow, $row->[$i]);
	    }

	    # Otherwise, it's a set or 1D sequence
	    else
	    {

		# We have as many values as the set or 1D sequence provides

		$have = $row->[$i]->getSize();

		# Push each item's value into the row

		for ($j = 0; $j < $have; ++$j)
		{
		    push (@$newRow, $row->[$i]->{value}->[$j]);
		}
	    }

	    # As long as we have fewer column values than we need to fill
	    # the expanded column, add undef to the end.

	    while ($have < $need)
	    {
		push (@$newRow, undef);
		++$have;
	    }
	}

	# Add the row to the table

	push (@newOutput, $newRow);
    }

    # Replace the contents of the output array in-place, so the callee sees the
    # modifications we made here.

    splice (@$output, 0, scalar(@$output), @newOutput);
}


##############################################################################
# dumpOutput
#
# Formats the values and prints them to the file
#
# Parameters: A reference to the arguments hash, as returned by parseArguments
#             The output stream, to write the index table to
#             A reference to the keyword definition array
#             The output matrix
#
# Returns: nothing.
#

sub dumpOutput ( $$$\@ )
{
    # Get parameters

    my (
	$arguments,
	$outputStream,
	$keywordDefs,
	$output
	) = @_;

    # Declare local variables

    my (
	$keywordIndexes,
	$row,
	$first,
	$i,
	$value,
	$keyword,
	$units,
	$width,
	$default,
	$chars,
	$text
	);

    # Get the column-keyword specification mappings, which is stored in the
    # first row in the table.

    $keywordIndexes = shift (@$output);

    # Process each (remaining) row in the output table

    foreach $row (@$output)
    {

	# $first is used to control the space-padding between columns

	$first = 1;

	# Process each cell in the row

	for ($i = 0; $i < scalar (@$row); ++$i)
	{

	    # Get the value

	    $value = $row->[$i];

	    # Get the associated keyword specification

	    $keyword = $keywordDefs->[$keywordIndexes->[$i]];

	    # Extract the meaningful portions from the keyword specification

	    $keyword =~ /^([\+\-]?)(\^?[^\/]*)(?:\/([^\/]*)(?:\/(.*))?)?$/;
	    ($units, $keyword, $width, $default) = ($1, $2, $3);

	    # If width isn't defined, default it to 10

	    if (!defined ($width))
	    {
		$width = 10;
	    }

	    # chars is the maximum number of characters we can have

	    $chars = $width;
	    $chars = -$chars if ($chars < 0);

	    # If this isn't the first column, print a space

	    printf $outputStream (" ") unless ($first);
	    $first = 0;
	    
	    # If the value is defined, we're going to print its value

	    if (defined ($value))
	    {

		# Check the units... If they say to suppress the units of the
		# value, or it was given on the command line and not overruled
		# by the keyword specification, then we strip the units

		if ($units eq "-" || 
		    ($units ne "+" && exists($arguments->{u})))
		{
		    $text = $value->getRawValueNoUnits();
		}

		# Otherwise, we get the raw value (with units in place)

		else
		{
		    $text = $value->getRawValue();
		}

		# Trim the text so that it's at most $chars bytes long

		if (length ($text) > $chars)
		{
		    $text = substr($text, 0, $chars);
		}

		# Print the text, formatted by printf (%5s for a 5 byte string)

		printf $outputStream ("%".$width."s", $text)
	    }

	    # Otherwise, we're going to need to use the default value

	    else
	    {

		# Print the default value

		printf $outputStream ("%".$width."s", $default);
	    }
	}

	# Start a new line for the next row.

	print $outputStream ("\n");
    }
}

##############################################################################
# try
#
# implemenation of try block for try/catch/finally
#

sub try (&$)
{
    # Get parameters
    my (
	$try,
	$catch
	) = @_;

    # Declare local variables
    my (
	$throw,
	$throwMessage
	);

    $throw = 0;

    # Run the try block

    eval { &$try () };

    # See if it threw any errors

    if ($@)
    {

	# It did, see if there's a catch block

	if ($catch->{catch})
	{

	    # There is, set $_ to the error message

	    local $error = $@;

	    # Run the catch block

	    eval { $_ = $error; &{$catch->{catch}} };

	    # See if the catch block threw any errors

	    if ($@)
	    {

		# It did, save the error for later use

		$throw = 1;
		$throwMessage = $@;
	    }
	}
	else
	{

	    # No catch block, save the error for later use

	    $throw = 1;
	    $throwMessage = $@;
	}
    }

    # See if there is a finally block

    if ($catch->{finally})
    {

	# There is, evaulate the finally block

	eval { &{$catch->{finally}} };

	# See if it threw any errors

	if ($@)
	{

	    # It did, save the error for later use
	    # This will override anything thrown above that hasn't been
	    # caught.

	    $throw = 1;
	    $throwMessage = $@;
	}
    }

    # See if we caught any unhandled errors above

    if ($throw)
    {

	# We did, re-throw the error.

	die ($throwMessage);
    }
}

##############################################################################
# catch
#
# Implementation of the catch block for try/catch/finally
#

sub catch (&;$)
{

    # If there's only one argument, it must be the catch block

    if (scalar (@_) == 1)
    {

	# return a hash, with the given catch block

	return {catch => $_[0]};
    }
    
    # Otherwise, there are two arguments, the catch block and 
    # any return values from later blocks.

    else
    {

	# merge the later blocks (probably just finally), with the
	# current catch block.

	return mergeHandlers ($_[1], {catch   => $_[0]});
    }
}

##############################################################################
# finally
#
# Implementation of the finally block for try/catch/finally
#

sub finally (&;$)
{

    # If there's only one argument, it must be the catch block.

    if (scalar (@_) == 1)
    {

	# return a hash, with the given finally block

	return {finally => $_[0]};
    }

    # Otherwise, there are two arguments, the finally block and any
    # return values from later blocks (this is rare, but it's possible to
    # place other blocks after the finally block.

    else
    {

	# Merge the later blocks with the current finally block.

	return mergeHandlers ($_[1], {finally => $_[0]});
    }
}

##############################################################################
# mergeHandlers
#
# Merges any catch/finally blocks for try/catch/finally
#

sub mergeHandlers ($$)
{
    # Get parameters:

    my (
	$handlers1,
	$handlers2
	) = @_;

    # Declare local variables

    my (
	$newHandler,
	$key
	);

    # Initialize a new handler as a copy of the 1st handler.

    $newHandler = {};
    $newHandler->{$_} = $handlers1->{$_} foreach keys (%$handlers1);

    # Attempt to add every item in the 2nd handler to the new handler,
    # aborting on any duplicates.

    foreach $key (keys (%$handlers2))
    {
	if (exists ($newHandler->{$key}))
	{
	    die ("Can't have more than one $key block for 'try'");
	}
	$newHandler->{$key} = $handlers2->{$key};
    }

    return $newHandler;
}
