#!/usr/bin/perl

use Getopt::Std;
#==============================================================================
#
#  Routine to check links in the given files to see if they meet PDS BROWSE
#  object restrictions.
#
#  Format:  % pdslinkchk [-r] [-v] [-o <log_file>] [ file ... ]
#
#   where  -r  indicates linkchk should recurse through subdirectories
#          -v  verbose listing - prints lines for files with no errors found
#          -o  redirects output to the named file
#
#  The input list may contain both file and directory names. File names
#  entered explicitly need not end in an HTML extension. However, only 
#  files with HTML extensions are checked in directory searches. 
#
#  Recognized HTML extensions are:
#
#    .html     .HTML         .shtml     .SHTML
#    .htm      .HTM          
#
#  Output is written to stdout if no output file is given. Links are flagged
#  as invalid if they are not strictly relative to the current directory.
#  File names in links are checked for existence.
#
#  15 June 2001, acr.
#  09 Aug 2001, acr: fixed silly bug in 'dofile' (found by Andrei Mesinger) 
#                    - omitted directory name in file open
#  15 Jan 2002, acr: Fixed stupid bug that failed to pass recursion option at
#                    top level.
#  26 Sep 2006, acr: Modified to use Getopt::Std
#             
#
#==============================================================================
# Begin by dealing with the input parameters:

getopts('o:rv');

# Open the output file, if any:

$outfile = ($opt_o)? $opt_o : "-";
open (LOG,"> $outfile") || die "Could not open '$outfile' for writing.";

# If there are no file arguments, check the current directory:

if (@ARGV == 0)
  { &dodir(".",$opt_r); }

# Otherwise loop through the arguments, separating files from directories:

else
  { $dl = 0;
    $fl = 0;
    while ($file = shift @ARGV)
      { if (-d $file)
          { $dlist[$dl] = $file;
            $dl++;
          }
        else
          { $flist[$fl] = $file;
            $fl++;
          }
      }

    # Now we'll call the file processing routine for each file:

    foreach $file (@flist)
      { &dofile($file,"."); }

    # And the directory routine for each directory:

    foreach $file (@dlist)
      { &dodir($file,$opt_r); }
  }

# Close the output file and we're done:

close(LOG);

#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------

sub dofile

# Subroutine to process the input file, looking for HTML anchors. We assume it
# makes sense to do so.

{ local ($file,$dir) = @_;     # Input file and path
  my    ($line,$ref,$link);
  my    ($i,$j,$k);
  my    ($linecount);
  my    ($filespec,$path);
  my    ($errcount);
  my    ($up,$spec);           # Parts of href reference


  # Begin by opening the file:

  open (HTML,"$dir/$file") || die "Could not open '$file' HTML file.";
  $linecount = 0;
  $errcount  = 0;

  # Now we're simply going to loop through the file, looking for the 'HREF'
  # tag. When we find it, we'll check the file reference:

  while ($line = <HTML>)
    { chop $line;
      $linecount++;

      $line =~ s/\s*=\s*/=/;
      if ($line =~ /href=/  ||  $line =~ /HREF=/)
        { $line =~ s/HREF=/href=/g;    # To make life easier

          # There might be more than one anchor in a line, so we have to
          # do this carefully:

          $i = index($line,"href=");

          while ($i>=0)
            { $ref = substr($line,$i+6);
              $ref =~ s/^([^"]*)".*$/$1/;

              # First we check for valid links. The links must be relative
              # to the current directory. So that means they can't start
              # with a protocal tag or a slash:

              if ($ref =~ /:\/\//)
                { $errcount++;
                  printf LOG "\n$dir/$file\n" if ($errcount==1);
                  printf LOG "  line %4d: External reference  - \"%s\"\n",
                              $linecount, $ref;
                }
              elsif ($ref =~ /^\//)
                { $errcount++;
                  printf LOG "\n$dir/$file\n" if ($errcount==1);
                  printf LOG "  line %4d: Absolute reference  - \"%s\"\n",
                              $linecount, $ref;
                }

              # If we have a valid reference form, we check the actual file.
              # First we have to construct the file name from the reference.
              # This is easy if there is no "../" notation - we just prepend
              # the current directory. If there are "../" prefixes, though,
              # we separate them out and crawl our way up the directory tree.

              else
                { $link = $ref;
                  if ($ref !~ /\.\.\//)
                    { $filespec = "$dir/$ref"; }
                  else
                    { $path = ($dir eq ".")? "" : $dir;

                      # Divide the reference into relative path info ("../")  
                      # and explicit file spec:

                      $up = $ref;
                      $up =~ s/^([\.\/]*).*$/$1/;
                      $spec = $ref;
                      $spec =~ s/^[\.\/]*//;

                      # Now loop until we're exhausted either the current path
                      # or the relative path:

                      while (length($up) > 0  &&  length($path) > 0)
                        { $up = substr($up,0,length($up)-3);
                          $path =~ s/(\/|)[^\/]+$//;
                        }

                      # If we have relative path left, trim the last '/':

                      chop $up;

                      # Now paste together the file specification:

                      $path = $up . $path;
                      $filespec = (length($path)>0)? "$path/$spec" : $spec;

                    }

                  # Now make sure the file exists:

                  if (!(-e $filespec))
                    { $errcount++;
                      printf LOG "\n$dir/$file\n" if ($errcount==1);
                      printf LOG "  line %4d: File does not exist - \"%s\"\n",
                              $linecount, $link;
                    }
                }

              # Next link in this line:

              $line = substr($line,$i+6);
              $i = index($line,"href=");
            }
        }
    }

  # If running in verbose mode, print a line for file with no errors found:

  if ($errcount == 0  &&  $opt_v)
    { printf LOG "\n$dir/$file - OK\n"; }

  # Close the file and we're done:

  close(HTML);
}

#------------------------------------------------------------------------------
#------------------------------------------------------------------------------

sub dodir

# Subroutine to run through a directory looking for HTML files to check.

{ local ($dir,$recurse) = @_;    # Directory name and recursion flag
  my    ($dlist,$dl);
  my    ($flist,$fl);
  my    ($file,$subdir);
 

  # Open the directory for reading:

  opendir (DIR,$dir) || die "Could not open directory '$dir'.";

  # Loop through the contents, making a list of HTML and subdirectory files:

  $dl = 0;
  $fl = 0;
  foreach $file (sort readdir(DIR))
    { next if ($file =~ /^\./);

      if (-d "$dir/$file")
        { $dlist[$dl] = "$dir/$file";
          $dl++;
        }

      elsif ($file =~ /\.html$/ || $file =~ /\.htm$/ || $file =~ /\.shtml$/ ||
             $file =~ /\.HTML$/ || $file =~ /\.HTM$/ || $file =~ /\.SHTML$/)
        { $flist[$fl] = $file;
          $fl++;
        }
    }
  closedir(DIR);

  # Now call the file routine for each file:

  foreach $file (@flist)
    { &dofile($file,$dir); }

  # If we're recursing through subdirectories, call the directory routine:

  if ($recurse  &&  $dl > 0)
    { foreach $file (@dlist)
        { &dodir($file,$recurse); }
    }

  # And we're done:
}

#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
