#!/usr/bin/perl

#  ppodl: Pretty Print ODL
#
#  This routine takes PDS ODL in nasty format and turns it into something I
#  like.  All input bytes are preserved except for white space leading, 
#  trailing, or surrounding the "=".  
#
#  Indentation is made uniform, a blank line is inserted before OBJECT 
#  definitions (where none exists), equal signs are aligned among adjacent 
#  keywords whenever possible, and lines are padded out to 80 bytes with \r\n
#  when the "-p" switch is used. Lines which exceed 80 bytes because of the
#  padding are left intact, and a list of them is written to STDERR. All 
#  keywords and symbolic values are forced to uppercase (including quoted
#  symbols).
#
#  Format: % ppodl [ -p ] [-nb] [ input [ output ] ]
#
#     where -p   means pad and add LF
#           -nb  means "no blanks" - do not insert blank padding lines
#
#  Standard I/O is used when no file names are supplied.
#
#  Modifications:
#
#    28 Nov 1995, acr: creation
#    25 Jun 1996, acr: corrected printing of first SFDU line;
#    21 Oct 1997, acr: adjusted output line counting (was missing inserted
#                      blank lines)
#    29 Jan 1998, acr: minor change to find_key_size to make checking for
#                      keyword lines more rigorous ('=' in description values
#                      was messing things up)
#    04 Feb 1999, acr: output format changes to preserve leading spaces in 
#                      continued text, and to align successive values of a 
#                      set when they occur on separate lines.
#    12 Mar 1999, acr: Fixed stupid bug that deleted prose lines that start
#                      in column 1
#    03 May 1999, acr: Added "-p" switch to indicate padding, otherwise 
#                      output is NOT padded and delimited with CRLF.
#    13 Dec 2000, acr: Add "inquote" check to avoid messing around with '='
#                      that appear inside long quoted text passages; fixed 
#                      a minor logical bug that then began manifesting itself.
#    14 Feb 2001, acr: added "forcesymbolsup" to force all non-text string 
#                      values to upper case. (No check is made for underscores
#                      vs blanks within quotes.)
#    16 May 2001, acr: Fixed error in 'forcesymbolsup' that was treating
#                      one-line NOTE and DESCRIPTION fields as symbols.
#    10 Oct 2001, acr: Fixed minor glitch  casued by forgetting that the 
#                      ODL comment delimiter is "/*", not '#'; added code
#                      to pad around comments.
#    04 Oct 2002, acr: Added "-nb" option
#    09 Apr 2003, acr: Fixed dumb bug that put padding around every line of
#                      comments
#    09 Jul 2003, acr: Changed handling of quotes around character values -
#                      now we quote everything, just in case.
#    01 Aug 2003, acr: Fixed handling of quotes around "PDS3" in labels; 
#                      fixed odd bug that put newline inside added quote.
#    09 Sep 2003, acr: Fixed bug that added quotes around things following
#                      equal signs in quotes text.
#    04 Mar 2005, acr: Fixed annoying bug which inserted quotes in the wrong
#                      place in files with trailing blanks.  PPODL now trims
#                      all trailing whitespace and carriage control before
#                      beginning processing.  Original carriage control is 
#                      preserved unless the -p option is used, in which case
#                      output is padded to 80-byte lines with "\r\n" carriage
#                      control.
#    18 Mar 2005, acr: Fixed the annoying, over-indenting problem that occurred
#                      when an object contained more that one NOTE or DESC
#                      keyword.
#
#-----------------------------------------------------------------------------
#
# Global variables: 
#
#     $sline    start line number - input
#     $fline    finish line number - output
#     @label    label lines - input
#     $lblsize  number of lines in the label - input
#     $maxkey   largest keyword size found - output
#     @long     list of line numbers of lines > 80 bytes
#     $nlong    subscript for @long array
#     $lnoffset line number offset - for added lines in output
#     $cl       current line number in label
#     $indent   indentation
#     $CR       carriage control characters

$TRUE  = 1;
$FALSE = 0;

# Parse the command line:

$PADCR    = 0;
$BLANKPAD = 1;
$infile   = "-";
$outfile  = "-";

if (@ARGV > 3)
  { die "Usage: ppodl [-p] [-nb] [ infile [ outfile ] ]\n"; }

for ($i=0; $i<@ARGV; $i++)
  { if ($ARGV[$i] eq "-p")
      { $PADCR = 1; }
    elsif ($ARGV[$i] eq "-nb")
      { $BLANKPAD = 0; }
    elsif ($infile eq "-")
      { $infile = $ARGV[$i]; }
    elsif ($outfile eq "-")
      { $outfile = $ARGV[$i]; }
  }

# Open file as needed:

open (ODL,$infile) || die "ppodl: Could not open input file ($infile)";
open (LBL,">$outfile") || die "ppodl: Could not open output file ($outfile)";

# Read entire label into array:

@label = <ODL>;
close(ODL);
$lblsize = @label;   # of lines in the label

# Find and save carriage-control and remove trailing blanks:

$CR = &fixlines;
if ($PADCR)
  { $CR = "\r\n"; }

# Read through the label, find all keywords and symbols and force them to
# (quoted) uppercase:

&forcekeysup;
&forcesymbolsup;

$sline  = 0;
$indent = 0;
$nlong  = 0;
$lnoffset = 1;     # First line is '1', not '0'.

# Call the recursive routine which handles the processing:

&do_object($sline);

# List any lines which exceeded 80 bytes on output:

if ($nlong > 0)
  { printf STDERR "The following lines exceeded 80 bytes on output:\n\n   ";

    foreach $n (sort @long)
      { printf STDERR " $n"; }
    printf STDERR "\n";
  }

# Done.

close(LBL);

#============================================================================

sub do_object

  # Routine to perform processing for one object or subobject.  This routine
  # is called recursively to deal with subobjects.

  # Global variables used: 
  #
  #     @label    label lines - input
  #     $lblsize  number of lines in the label - input
  #     @long     list of line numbers of lines > 80 bytes
  #     $nlong    subscript for @long array
  #     $indent   indentation        - input
  #
  # Arguments:
  #
  #     $sline    start line number  - input

{ local ($sline)=@_;

  local ($fline);             # Last line to be processed
  local ($status);            # function return status
  local ($maxkey);            # maximum keysize
  local ($fk,$fv);            # formatting widths for keyword and value 

  local ($line);              # line buffer
  local ($key,$val);          # line piece holding places
  local ($blanks);            # blank filler for indentation
  local ($tmpval);            # value holder
  local ($trline);            # line trimmed of leading blanks
  local ($i,$newindent);
  local ($inquotes);          # TRUE if we're in a text field
  local ($comment_line);      # TRUE if the current line is a comment
  local ($comment_next);      # TRUE if the next line is a comment
  local ($comment_last);      # TRUE is the last line was a comment

  $blanks   = "                                                           ";
  $inquotes = 0;
  $comment_next = 0;
  $comment_last = 0;


  # Start by finding maximum keysize and range of lines in this object:

  ($status,$maxkey,$fline) = &find_keysize($sline);

  # Output the lines in the range:

  for ($i=$sline; $i<$fline; $i++)
    { $line = $label[$i];
      $line =~ s/\s*$//;
      $trline = $line;
      $trline =~ s/^\s+//;
      $comment_line = ($line =~ /^\s*\/\*/);
      $comment_next = ($i+1 < $lblsize)? ($label[$i+1] =~ /^\s*\/\*/) : 1;

      # If the line before an OBJECT or comment line isn't blank or another
      # comment (and blank padding hasn't been disabled via the "nb" option), 
      # insert a blank line:

      if ($BLANKPAD  &&
          ($line =~ /^\s*OBJECT\s*=/  ||  ($comment_line  &&  !$comment_last))
          &&  ($i>0  &&  $label[$i-1] !~ /^\s*$/))
        { if ($PADCR)
            { printf LBL "%78.78s$CR",$blanks; }
          else
            { printf LBL "$CR"; }
          $lnoffset++;
        }


      # Print indentation, if any:

      printf LBL "%${indent}.${indent}s",$blanks if ($indent > 0);

      # Standard keyword=value lines: 

      if ($line =~ /^\s*[\^A-Z_0-9]+\s*=/  &&  $inquotes == 0)
        { $trline =~ s/\s*=\s*/=/;
          ($key,$val,@left) = split(/=/,$trline); # Allow for value fields
          $val = join("=",($val,@left));          #  which contain an "="

          # Check for unbalanced double quotes:

          $inquotes = ($val =~ /^\s*"/  &&  $val !~ /"\s*$/)? 1 : 0;

          # If we've opened an extended quote, figure out the minimum
          # leading white space so we can clip it from each subsequent 
          # line:

          $ldspace = checkmargin($i);
##printf "LDSPACE: $ldspace\n";
          # Treat OBJECT and END_OBJECT lines separately:

          if ($key eq "OBJECT"  ||  $key eq "END_OBJECT")
            { $fk = 10;
              $fv = 78 - ($fk + $indent + 3);
            }
          else
            { # If we're looking at the beginning of a set of values, we
              # prefer to print one on a line and indent to align the list
              # of values. To cover this we need to check for the braces,
              # calculate the additional indentation and add it as a blank
              # string to the beginning of the value:

              $INSET = ($val =~ /\{/  &&  $val !~ /\}/);

              ($fk,$fv) = find_formats($key,$val,$maxkey,$i);
            }

          printf LBL "%-${fk}.${fk}s = ",$key;
          if ($PADCR)
            { printf LBL "%-${fv}.${fv}s$CR",$val; }
          else
            { printf LBL "$val$CR"; }
        }

      # Lines that don't begin with a comment string ("/*") or END_OBKECT
      # are treated one of three ways:
      #  1. If we are in the middle of an INSET list, they are indented
      #     the width of the keyword plus ' = { ', to align with preceeding
      #     values. In this case we check for the end of the set.
      #  2. Otherwise, we're in the middle of a text field, so we preserve
      #     leading spacing and indent an additional 2 spaces.  We check for
      #     overflow, but we do not truncate.
      #  3. If this is the very first line of the file, we've encountered
      #     an SFDU label, which is never indented.

      elsif (!$comment_line  &&  $trline !~ /^END_OBJECT/)
        { if ($INSET)
            { # Check for the end of the list - it might be just the brace:

              if ($trline =~ /^\}/)
                { $fv = $maxkey + 3; 
                  $INSET = 0;
                }
              elsif ($trline =~ /\}/)
                { $fv = $maxkey + 5;
                  $INSET = 0;
                }
              else
                { $fv = $maxkey + 5; }

              # Attach the additional indent:

              $tmpval = sprintf "%${fv}.${fv}s$trline",$blanks;
              $line = $tmpval;

              # Check for overflow and print:

              if (length($line) > (78-$indent))
                { $long[$nlong] = $i + $lnoffset;
                  $nlong++;
                  printf LBL "$line$CR";
                }
              elsif ($PADCR)
                { $fv = 78 - $indent;
                  printf LBL "%-${fv}.${fv}s$CR",$line; 
                }
              else
                { printf LBL "$line$CR"; }
            }            

 
          elsif ($i == 0)
            { if ($PADCR)
                { printf LBL "%-78.78s$CR",$line; }
              else
                { printf LBL "$line$CR"; }
            }

          elsif ($line =~ /^\s*$/)   # Blank line
            { if ($PADCR)
                { $fv = 78 - $indent;
                  printf LBL "%-${fv}.${fv}s$CR",$line; 
                }
              else
                { printf LBL "$line$CR"; }
            }

          else
            { # OK, we have here a continuation of a prose description of some
              # sort. Ideally, we'd like the left edge of this to be indented
              # an additional 2 spaces, but we also want to preserve the 
              # relative indentation of tables and headings and such.  The
              # minimum lead space has been saved in the $ldspace variable.
              # So, we trim this minimum lead space from the $line buffer:

              $trline = substr($line,$ldspace);
              $line = $trline;              

              # Now we check for overflow and output the line:

              if (length($line) > (76-$indent))
                { $long[$nlong] = $i + $lnoffset;
                  $nlong++;
                  printf LBL "  $line$CR";
                }
              elsif ($PADCR)
                { $fv = 76 - $indent;
                  printf LBL "  %-${fv}.${fv}s$CR",$line;
                }
              else
                { printf LBL "  $line$CR"; }

              # Check for the closing quote:

              $inquotes = ($line =~ /"\s*$/)? 0 : 1;
            }
        }

      # Comment and END_OBJECT lines

      elsif (length($line) > (78-$indent))
        { $long[$nlong] = $i + $lnoffset;
          $nlong++;
          printf LBL "$line$CR";
        }
      elsif ($PADCR)
        { $fv = 78 - $indent;
          printf LBL "%-${fv}.${fv}s$CR",$trline;
        }
      else
        { printf LBL "$trline$CR"; }

      # Finally, if this is a comment line and the following line is not 
      # blank, another comment, or an OBJECT line (and $BLANKPAD is true), 
      # insert a line:

      if ($BLANKPAD  &&  $comment_line  &&  !$comment_next  &&  $i<$lblsize  
          &&  $label[$i+1] !~ /^\s*$/  &&  $label[$i+1] !~ /^\s*OBJECT\s*=/)
        { if ($PADCR)
            { printf LBL "%78.78s$CR",$blanks; }
          else
            { printf LBL "$CR"; }
          $lnoffset++;
        }

      $comment_last = $comment_line;

    }

   # Increment $sline:

   $sline = $fline;

   # Now, if we stopped at a subobject, call this routine to handle it with
   # an additional indentation:

   if ($status == 1)
     { $indent += 2;          # increase indentation
       &do_object($sline); 
       $indent -= 2;          # reset indentation
     }

   # If we stopped at an END_OBJECT, call this routine again but at the last
   # indent level, to handle any additional keywords:

   elsif ($status == 2)
     { $indent -= 2;
       &do_object($sline); 
     }

   # If we stopped at the END statement, finish processing the remaining
   # lines at the current indent level:

   elsif ($status == 3)
     { while ($sline < $lblsize)
         { $line = $label[$sline];
           $line =~ s/\s*$//;

           if (length($line) < 78)
             { if ($PADCR)
                 { printf LBL "%-78.78s$CR",$line; }
               else
                 { printf LBL "$line$CR"; }
             }
           else
             { $long[$nlong] = $sline + $lnoffset;
               $nlong++;
               printf LBL "$line$CR";
             }
           $sline++;
         }
     }

  # Otherwise, we just stopped at EOF, which is OK (or at least that's what
  # we'll assume), so exit:

  else
    { exit(0); }

}
#----------------------------------------------------------------------------

sub find_keysize

  # Routine to read until the next OBJECT, END_OBJECT or END statement and
  # determine the maximum size of keywords present.  

  # Global variables used: 
  #
  #     @label    label lines - input
  #     $lblsize  number of lines in the label - input
  #
  # Arguments:
  #
  #     $sline    start line number  - input
  #
  # Returns a completion status, the maximum keysize found and the number of
  # the last line in the range:
  #
  #     1   Stopped at OBJECT
  #     2   Stopped at END_OBJECT
  #     3   Stopped at END
  #     0   Stopped at EOF
  #     $maxkey   largest keyword size found - output
  #     $fline    finish line number - output

{ local ($sline)=$_[0]; # start line
  local ($fline);       # finish line
  local ($maxkey);      # largest key size
  local ($status);      # return status
  local ($done);        # loop flag
  local ($line);        # line buffer
  local ($keysize);     # current key size (characters)
  local ($buff,$len);   # temporary variables
  
  $status = -1;
  $fline  = $sline;
  $done   = 0;
  $maxkey = 0;

  while ($status < 0)
    { $line = $label[$fline];

      # First, check for end conditions:

      if ($line =~ /^\s*OBJECT\s+/)
        { $status = 1; }               # OBJECT is processed at same level

      elsif ($line =~ /^\s*END_OBJECT\s*/)  # Deal with END_OBJECT separately
        { if ($sline == $fline)             # ...first keyword encountered,
            { $maxkey = 10;                 #     continue
              $fline++;
              $status = 0 if ($fline >= $lblsize);
              next;
            }
          else                              # ...end of current range
            { $status = 2;
              next;
            }
        }

      elsif ($line =~ /^\s*END\s*/)
        { $status = 3;
          next;
        }

      elsif ($line !~ /^\s*[A-Z0-9_^-]+\s*=/)   # Skip over non-keyword lines
        { $fline++;
          $status = 0 if ($fline >= $lblsize);

          next;
        }

      # Next, isolate the keyword and find its size:

      $line =~ s/^\s+//;
      $line =~ s/\s*=.*\s*$//;

      $keysize = length($line);

      # Save the maximum size, provided this is actually a keyword line:

      $maxkey = ($maxkey > $keysize)? $maxkey : $keysize;

      # Next line:

      $fline++;
      if ($fline >= $lblsize)
        { $status = 0; }
    }

  # Done.  Return status, maximum keysize and ending line number:

  ($status,$maxkey,$fline,$lead);
}

#------------------------------------------------------------------------------

sub find_formats

  # Routine to determine the optimal output widths for the given input line.
  # Ideally, lines will not exceed 80 bytes total.  If necessary, padding bytes
  # will be removed (but not indent bytes) to fit the value in the 80 bytes.
  # Lines longer than 80 bytes, however, are printed where necessary.

  # Arguments:
  #
  #    $key     keyword, stripped of blanks
  #    $val     value, stripped of blanks
  #    $maxkey  maximum key size (for alignment of values)
  #    $cl      current line number in label
  #
  # Global variables:
  #
  #    @long    list of line numbers of lines > 80 bytes
  #    $nlong   subscript for @long array
  #    $indent  number of indentation spaces
  #
  # Returns:
  #
  #    $k       print width for keyword
  #    $v       print width for value

{ local ($key,$val,$maxkey,$cl) = @_;
  local ($total, $kw, $need, $k, $v);           # byte sizes

  # First, check to see if the usual case will work:

  $total = $indent + $maxkey + 3 + length($val);

  if ($total < 78)       # allow for \r\n
    { $k = $maxkey;
      $v = 78 - ($indent + $maxkey + 3);
    }

  else
    { # See if there is enough room in the blank padding to get in under 78:

      $need = $total - 78;
      $kw = $maxkey - length($key);

      if ($kw >= $need)
        { $k = $maxkey - $need;
          $v = length($val);
        }

      else                      # Give up
        { $long[$nlong] = $cl + $lnoffset;
          $nlong++;
          $k = length($key);
          $v = length($val);
        }
    }

  # Return the formatting values:

  ($k,$v);
}

#------------------------------------------------------------------------------

sub forcekeysup

  # Routine to read through the lines stored in the @label array, find the 
  # keywords, and force them to all uppercase (which is required for the 
  # rest of the logical tests to work).

  { my $keyword;
    my $i;
    my $inquotes;
    my $line;
    my $ind;

    # We have to keep track of when we're inside double quotes, to avoid
    # mistaking a non-keyword "=" for a keyword separator, and so we can 
    # recognize lines which consist of only keywords and no values (though
    # possibly followed by a comment).

    $inquotes = 0;

    for ($i=0; $i<@label; $i++)
      { $line = $label[$i];

        # If we're inside quotes, we just look for the end of the quotes:

        if ($inquotes)
          { $inquotes = ($line =~ /"/)? 0 : 1; 
            next;
          }

        # If we're not in quotes, we look for an '=':

        $ind = index($line,"=");

        # If we found the "=" we assume we've got a keyword and force 
        # everything before the "=" to uppercase and we're ready for the
        # next line:

        if ($ind > 0)
          { substr($line,0,$ind) =~ tr/a-z/A-Z/; }

        # Otherwise, we assume the first word (delimited by blanks)
        # is a keyword, and we force this to uppercase:

        elsif ($line !~ /^\s*$/  &&  $line !~ /^\s*\/\*/)
          { $line =~ s/^\s*//;    # Remove leading blanks
            $line =~ s/\s*$/ /;   # Make sure there is one trailing blank
            $ind = index($line," ");
            substr($line,0,$ind) =~ tr/a-z/A-Z/;
          }

        # Check to see if we've opened quotes without closing them:

        if ($line =~ /"/)
          { $ind = index ($line,"\"");
            if (index(substr($line,$ind+1),"\"") < 0)
              { $inquotes = 1; }
            else
              { $inquotes = 0; }
          }

        # Save modifications and go on to next line:

        $label[$i] = $line;
      }

    # And we're done:

    1;
  }

#------------------------------------------------------------------------------

sub forcesymbolsup

  # Routine to read through the lines stored in the @label array, find the
  # symbol-typed values, and force them to all uppercase. We add quotes 
  # around them while we're at it.

  { my ($key,$value);
    my $i;
    my $inquotes;
    my $line;
    my $ind,$valstart,$vallen;
    my $force;
    my $len;

    # We have to keep track of when we're inside double quotes to avoid
    # mistaking a non-keyword "=" for a keyword separator, and so we can
    # recognize lines which consist of only keywords and no values (though
    # possibly followed by a comment).

    $inquotes = 0;

    for ($i=0; $i<@label; $i++)
      { $line = $label[$i];

         # If we're inside quotes, we just look for the end of the quotes,
         # forcing to uppercase if this is a symbol string:

        if ($inquotes)
          { if ($line =~ /"/)
              { $inquotes = 0;
                if ($force)
                  { $vallen = index($line,"\"") - 1;
                    substr($line,0,$vallen) =~ tr/a-z/A-Z/;
                  }
              }
            next;
          }

        # If we're not in quotes, we look for an '=':

        $ind = index($line,"=");

        # If we've got no "=" then the line may consist of either a bare
        # keyword or a quoted string. If it's a bare keyword, we check the
        # type and set $force accordingly, then loop. 


        if ($ind < 0)
          { if ($line !~ /^\s*"/)        # Bare keyword
              { $key = $line;
                $key =~ s/\s+.*$//;
                $force = ($key !~ /NOTE$/  &&  $key !~ /DESCRIPTION$/ &&
                          $key !~ /_DESC$/);
                next; 
              }
            else
              { $valstart = 0; }
          }

        # If we did have an "=", we check the keyword to determine $force 
        # and set a pointer to the beginning of the value:

        else
          { $key = substr($line,0,$ind-1);
            $key =~ s/\s+$//;
            $force = ($key !~ /NOTE$/  &&  $key !~ /DESCRIPTION$/ &&
                      $key !~ /_DESC$/);
            $valstart = $ind+1;
            while (substr($line,$valstart,1) eq " ") { $valstart++; }
          }

        # If we've made it this far we know we have a value and where it
        # starts (including quotes). We next determine whether we have
        # balanced quotes on this line. Note that $vallen includes 
        # enclosing quotes, if any:

        $ind = index($line,"\"");
        if ($ind >= 0)
          { $valstart = $ind + 1;
            $vallen   = index(substr($line,$valstart),"\"");

            if ($vallen >= 0)
              { $inquotes = 0; }
            else
              { $inquotes = 1; 
                $vallen   = length($line) - $valstart + 1;
              }
          }
        else
          { $vallen = length($line) - $valstart - 1; }

        # SHORT CUT: At this point we've set all our flags, so if we're not
        # forcing values we can loop now:

        next if (!$force);

        # Finally, if we're not in an extended quote and the value is not a
        # quoted string, we set the end of value to be the first blank space
        # to avoid forcing in-line comments to uppercase:

        if (!$inquotes  &&  $line !~ /"/)
          { $ind = index(substr($line,$valstart)," ");
            if ($ind >= 0)
              { $vallen = $ind - 1; }
            else
              { $vallen = length($line) - $valstart; }
          }

        # Now to force the symbols. If the quotes are balanced, we force the
        # value in between; otherwise we force the rest of the line:

        if ($inquotes)      # UNBALANCED quotes
          { substr($line,$valstart) =~ tr/a-z/A-Z/; }
        else                # balanced quotes
          { substr($line,$valstart,$vallen) =~ tr/a-z/A-Z/; }

        # Finally, if we're not in quotes and the value is not currently
        # enclosed in quotes but should be, add them:

        if (!$inquotes  &&  
            $line !~ /^\s*(END_|)OBJECT/   &&
            $line !~ /^\s*(END_|)GROUP/    &&
            $line !~ /^\s*PDS_VERSION_ID/  &&
            substr($line,$valstart-1) !~ /^"/)

          { $value = substr($line,$valstart,$vallen);
            if ($value =~ /^[A-Z][0-9A-Z_]*$/)
              { substr($line,$valstart) =~ s/$value/"$value"/; }
          }


        # Save modifications and go on to next line:

        $label[$i] = $line;
      }

    # And we're done:

    1;
  }

#------------------------------------------------------------------------------

sub fixlines

  # Routine to record the line delimiter in $CR and trim trailing blanks and
  # carriage returns (if any) from the input lines in @label.

  { my ($i,$j,$len);
    my ($cr);

    # We begin by examining the first line to figure out what the carriage
    # control characters are:

    $len = length($label[0]);
    if ($len > 1  &&  $label[$len-2] == '\r')
      { $cr = "\r\n"; }
    else
      { $cr = "\n"; }

    # Now loop through the label lines, removing all trailing white space,
    # including carriage control:

    for ($i=0; $i<@label; $i++)
      { $label[$i] =~ s/\s*$//; }

    # And we're done.  We return the carriage control string:

    return $cr;
  }

#------------------------------------------------------------------------------

sub checkmargin

  # Routine to check lines in extended text fields to determine the 
  # amount of leading whitespace common to all lines.  That number of
  # blanks is returned and later used to adjust indenting.

  { my ($startline) = $_[0];     # The current line pointer passed
    my ($a,$b,$c);
    my ($line,$len);
    my ($ldspace);
    my ($not_done);

    # The pointer passed in points to the line with the opening
    # quote.  We loop through subsequent lines until we find the
    # closing quote, checking for leading whitespace and saving 
    # the minimum size:

    $a        = $startline + 1;
    $ldspace  = 1000;
    $not_done = 1;

    while ($not_done)
      { return $ldspace if ($a >= $lblsize);

        $line = $label[$a];
        $line =~ /^( *)[^ ]/;
        $len = length($1);
        $ldspace = ($len < $ldspace)? $len : $ldspace;

        if ($line =~ /"/)
          { $not_done = 0; }
        else
          { $a++; }
      }

    # Return the minimum common leading space and we're done:

    return $ldspace;
  }
