#!/usr/bin/perl

use Getopt::Std;

# Routine to remove a PDS COLUMN definition from a label file, adjust all 
# related keywords accordingly, and also to produce a new copy of the label
# file with the data COLUMN deleted. 
#
# Format:  % rmcolumn [-o] [-b <byte_count>] [-d] <COLUMN_NUMBER> <input_label>
#
#   where   -o   indicates the new version of the file(s) should overwrite
#                existing versions.  The default is to use the same file name
#                with an additional extension of ".RC"
#           -d   indicates that the datafile mentioned in the TABLE pointer
#                should also be modified, by having the corresponding bytes
#                removed
#           -b   allows the user to specify a number of bytes to extract, 
#                rather than allowing the value to default to the COLUMN width
#                plus any (presumably blank) padding (i.e., the difference in
#                START_BYTEs determines the bytes deleted).
#
# 26 Oct 1998, acr.
# 26 Sep 2006, acr: Modified to use Getopt::Std
#
#============================================================================

# Check for input parameters:

getopts('odb:');

if (@ARGV != 2)
  { die "Usage: rmcolumn [-o] [-d] [-b <bytes>] <column_number> <label>\n"; }

$colnum = $ARGV[0];
$inlbl  = $ARGV[1];
$tmplbl = "$inlbl.RC";

# Make sure any input byte count is an integer: 

if ($opt_b  &&  $opt_b !~/^[0-9]+$/)
  { die "Invalid byte count ($opt_b).\n"; }

#----------------------------------------------------------------------------
# We begin by opening the input label and scanning through it for some basic
# information.  We're assuming everything is verified and well behaved here,
# so the basic assumption is that the label is correct and accurate.  We do
# have to collect some parameters and check for valid input, though.  We must
# check:
#
#   1. That this is an ASCII table object
#   2. Whether it has a COLUMN_LABEL header
#   3. If there is no COLUMN_LABEL header, that the table is alone in the file
#   4. That the column to be deleted exists
#   5. The dimensions and location of the column to be deleted
#
#----------------------------------------------------------------------------

open(LBL,$inlbl) ||  die "Unable to open input label '$inlbl'.";

# Loop through the input label one record at a time:

$header_present = 0;
$not_done       = 1;

while (($line=<LBL>) && $not_done)
  { chomp $line;
    if ($line =~ /RECORD_BYTES/)
      { $record_bytes = &getval($line); }
    elsif ($line =~ /COLUMNS/)
      { $columns = &getval($line); }
    elsif ($line =~ /INTERCHANGE_FORMAT/)
      { $intformat = &getval($line); }
    elsif ($line =~ /\^HEADER/)
      { $header_present = 1; }
    elsif ($line =~ /HEADER_TYPE/)
      { $header_type = &getval($line); }
    elsif ($line =~ /\^TABLE/)
      { $datafile = &getval($line); }
    elsif ($line =~ /^\s*OBJECT\s*=\s*COLUMN/)
      { $not_done = 0; }
  }

# We have now encountered the first COLUMN definition.  If the column we seek
# exists, then we loop through all the column definitions, saving parmaters as
# we encounter them:

if ($colnum > $columns)
  { die "Requested column ($colnum) does not exist.\n"; }

$thiscol = 1;
while ($thiscol <= $columns)
  { if ($line =~ /^\s*END_OBJECT\s*=\s*COLUMN/)
      { $thiscol++; }
    elsif ($line =~ /^\s*START_BYTE\s*=/)
      { $start_byte[$thiscol] = &getval($line); }
    elsif ($line =~ /^\s*BYTES\s*=/)
      { $bytes[$thiscol] = &getval($line); }
    $line=<LBL>;
  }

close(LBL);

# OK, let's see what we got:

if ($header_present  &&  $header_type ne "COLUMN_LABEL")
  { die "Cannot handle table files with non-label HEADERs.\n"; }

if ($intformat ne "ASCII")
  { die "Cannot handle non-ASCII tables.\n"; }

if (!$datafile)
  { die "Data file pointer not found.\n"; }
elsif ($opt_d)
  { $datafile =~ s/[()]//g;  # Remove everything but the file name
    $datafile =~ s/,[0-9]+//;
    $datafile =~ s/["']//g;

    # Try to locate the data file:

    $df = $datafile;
    if (! -e $df)
      { $df = $datafile;
        $df =~ tr/a-z/A-Z/;
        if (! -e $df)
          { $df =~ tr/A-Z/a-z/;
            if (! -e $df)
              { die "Unable to find datafile '$datafile'.\n"; }
          }
      }
    $datafile = $df;
    $tmptab   = "$datafile.RC";
  }

if ($opt_b  &&  ($start_byte+$opt_b > $record_bytes-$opt_b))
  { die "Given field width ($opt_b) extends beyond end of record.\n"; }

# Determine the number of bytes to be deleted.  If this was input as an option,
# we use that value.  Otherwise we use either the differenc in START_BYTES 
# between the column being deleted and its successor, or the BYTES value from 
# the deleted column, if it is the last in the label:

if ($opt_b)
  { $delbytes = $opt_b; }
elsif ($colnum == $columns)
  { $delbytes = $bytes[$colnum]; }
else
  { $delbytes = $start_byte[$colnum+1] - $start_byte[$colnum]; }

#----------------------------------------------------------------------------
# Now we've got all the information we need, we can begin the modifications of
# the label file:

open(LBL, "rmcr $inlbl | varlen |");
open(NEW, "| fixlen -r 79 | addcr >$tmplbl");

$record_bytes = $record_bytes - $delbytes;
$row_bytes    = $record_bytes;
$columns--;

# Loop through the top-level keywords in the label:

$line = <LBL>;
while ($line !~ /^\s*OBJECT\s*=/)
  { if ($line =~ /RECORD_BYTES/)
      { $line =~ s/=\s*[0-9]+/= $record_bytes/; }
    print NEW $line;
    $line = <LBL>;
  }

# If we have a COLUMN_LABEL header, we need to treat it as well:

if ($line =~ /^\s*OBJECT\s*=.*HEADER/)
  { print NEW $line;
    $line = <LBL>;
    while ($line !~ /END_OBJECT/)
      { if ($line =~ /BYTES\s*=/)
          { $line =~ s/=\s*[0-9]+/= $record_bytes/; }
        print NEW $line;
        $line = <LBL>;
      }

    while ($line !~ /OBJECT\s*=\s*TABLE/)
      { print NEW $line;
        $line = <LBL>;
      }
  }

# Otherwise, we take care of the TABLE keywords:

print NEW $line;
$line = <LBL>;
while ($line !~ /OBJECT\s*=\s*COLUMN/)
  { if ($line =~ /ROW_BYTES/)
      { $line =~ s/=\s*[0-9]+/= $row_bytes/; }
    elsif ($line =~ /COLUMNS/)
      { $line =~ s/=\s*[0-9]+/= $columns/; }
    print NEW $line;

    $line = <LBL>;
  }

$thiscol = 0;
while ($line !~ /^\s*END\s*$/)
  { if ($line =~ /^\s*OBJECT\s*=\s*COLUMN/)
      { $thiscol++;

        # Columns preceeding the column to be deleted are passed through
        # unchanged:

        if ($thiscol < $colnum)
          { while ($line !~ /END_OBJECT/)
              { print NEW $line;
                $line = <LBL>;
              }
            print NEW $line;
          }

        # The target column is read but not written:

        elsif ($thiscol == $colnum)
          { while ($line !~ /END_OBJECT/)
              { $line = <LBL>; }
          }

        # Columns after the target column must be modified:

        else
          { while ($line !~ /END_OBJECT/)
              { if ($line =~ /COLUMN_NUMBER/)
                  { $num = $thiscol-1;
                    $line =~ s/=\s*[0-9]+/= $num/;
                  }
                elsif ($line =~ /START_BYTE/)
                  { $sb = $start_byte[$thiscol] - $delbytes;
                    $line =~ s/=\s*[0-9]+/= $sb/;
                  }
                print NEW $line;

                $line = <LBL>;
              }
            print NEW $line;
          }
      }
    else
      { print NEW $line; }
    $line = <LBL>;
  }

print NEW $line;
#----------------------------------------------------------------------------
# At this point we're done with the input label and the output label.  We close
# both:

close(LBL);
close(NEW);

# If we're overwriting the old label, we do that now:

if ($opt_o)
  { rename $tmplbl, $inlbl; }



# And we're done.

#----------------------------------------------------------------------------
# Now, if we're modifying the data file as well, we'll open that and write a
# new one:

if ($opt_d)
  { open (TAB,$datafile);
    open (NEW,">$tmptab") || 
      die "Could not open temporary datafile '$tmptab'";

    # Find the record sections to save:

    $firstpart = $start_byte[$colnum] - 1;
    $lastpart  = $firstpart + $delbytes;


    while ($line=<TAB>)
      { $outline  = substr($line,0,$firstpart);
        $outline .= substr($line,$lastpart);
        print NEW $outline;
      }

    close(TAB);
    close(NEW);

    # Overwrite the input file as indicated:

    if ($opt_o)
      { rename $tmptab, $datafile; }
  }

# And we're done. 

#============================================================================
#============================================================================

sub getval

  #  Extracts the simple value portion of the input line and trims blanks:

  { my ($buff) = $_[0];
    my ($val);

    $val = $buff;
    $val =~ s/^.*=\s*//;
    $val =~ s/\s+$//;
    $val =~ s/^["']//;
    $val =~ s/["']$//;
    $val;
  }
