#!/usr/bin/perl

#  sbnlc- PDS Detached Label checker.
#
#  This routine makes sure that all required fields are present in the
#  input label.
#
#  Format: % sbnlc[ -e | -s | -a ] [ <label> ]
#
#  Modification History:
#
#    06 Mar 1996, acr: fixed pattern match for object pointer
#    27 Mar 1996, acr: added check for valid DATA_TYPEs
#    03 May 1996, acr: modified to ignore RECORD_BYTES and FILE_RECORDS when
#                      RECORD_TYPE is STREAM.
#    07 Jun 1996, acr: added DATE and TIME to list of valid DATA_TYPEs; added
#                      blank line before file name print
#
#=============================================================================


# Check for and gather options:

if (@ARGV > 0)
  { $arg = shift @ARGV;

    if ($arg eq "-e"  ||  $arg eq "-s"  ||  $arg eq "-a")
      { $label_type = $arg; }
    else
      { unshift (@ARGV,$arg); 
        $label_type = "-s";
      }
  }

if (@ARGV > 1)
  { die "Usage: sbnlc[ -e | -s | -a ] [<label>].\n"; }
elsif (@ARGV == 1)
  { $filename = $ARGV[0]; }

#--------------------------------------------------------------------------
# Required Keywords:
#
# ... all files:

  @rkw = ("PDS_VERSION_ID",              #  0
          "RECORD_TYPE",                 #  1
          "RECORD_BYTES",                #  2
          "FILE_RECORDS",                #  3
          "DATA_SET_ID",                 #  4
          "PRODUCT_ID",                  #  5
          "PRODUCT_NAME",                #  6
          "PRODUCT_CREATION_TIME",       #  7
          "OBJECT ",                     #  8
          "END ",                        #  9
          "\\^",                         # 10

          "TARGET_NAME",                 # 11
          "INSTRUMENT_HOST_NAME",        # 12
          "INSTRUMENT_NAME",             # 13
          "START_TIME",                  # 14
          "STOP_TIME",                   # 15

          "SPACECRAFT_CLOCK_START_COUNT",# 16
          "SPACECRAFT_CLOCK_STOP_COUNT", # 17
          "SPACECRAFT_NAME");            # 18 alternate for INSTRUMENT_HOST_NAME

  $all   = 11;
  $data  = 16;
  $space = 18;

$interchange = "BINARY";       # Assume BINARY data

# Initialize output:

$indent = "   ";
if ($filename)
  { printf "\n$filename\n"; }


# Loop through the input label records looking for the INTERCHANGE_FORMAT
# value and the presence of the required keywords.  This stops when the
# OBJECT definition line is encountered:

$lc = 0;        # Line counter
while ($line=<>)
  { chop $line;
    $lc++;
    $line =~ s/^\s+//;

    if ($line =~ /INTERCHANGE_FORMAT/)
      { $interchange = $line;
        $interchange =~ s/^=\s*("|)//;
        $interchange =~ s/("|)\s*$//;
        $interchange =~ tr/a-z/A-Z/;
        if ($interchange ne "ASCII"  &&  $interchange ne "BINARY")
          { printf "$indent Invalid INTERCHANGE FORMAT ($interchange)";
            printf "at line $lc.\n";
          }
      }
    else
      { for ($i=0; $i<@rkw; $i++)
          { if ($line =~ /^$rkw[$i]/)
              { $found[$i] = 1; 

                # Keep track of the RECORD_TYPE - it matters later:

                if ($i == 1)   # RECORD_TYPE
                  { $record_type = $line;
                    $record_type =~ s/^.*=\s*("|)//;
                    $record_type =~ s/("|)\s*$//;
                  }
              }
          }
      }

    last if ($line =~ /^OBJECT /);
  }
$not_done = 1 if ($line =~ /^OBJECT /);

# If there are more lines in the file (there should be), pass through them
# checking the DATA_TYPEs and looking for the "END" statement:

if ($not_done)
  { while ($line=<>)
      { chop $line;
        $lc++;
        $line =~ s/^\s+//;

        if ($line =~ /INTERCHANGE_FORMAT/)    # It might occur in the object
          { $interchange = $line;
            $interchange =~ s/^.*=\s*("|)//;
            $interchange =~ tr/a-z/A-Z/;
            $interchange =~ s/([A-Z0-9\-_]+)("|)\s*$/$1/;
            if ($interchange ne "ASCII"  &&  $interchange ne "BINARY")
              { printf "$indent Invalid INTERCHANGE FORMAT ($interchange)";
                printf "at line $lc.\n";
              }
          }
        elsif ($line =~ /DATA_TYPE/)
          { 
            # If this is a DATA_TYPE line, check the type against the
            # valid types in use at the SBN.  The return value is true 
            # if the type was valid:
 
            $error++ if !(&check_datatype($line,$interchange,$lc));
          }
        else
          { $found[9] = 1 if ($line =~ /^END /);
          }
      }
  }


# Now check for all the required fields:

printf "\n";

#  If this was a spacecraft file, SPACECRAFT_NAME may replace 
#  INSTRUMENT_HOST_NAME:

if ($label_type eq "-s")
  { $found[12] = ($found[12] == 1)? $found[12] : $found[18]; }

$errors = 0;

# ... for all files:

for ($i=0; $i<10; $i++)
  { if ($found[$i] != 1)
      {
        # RECORD_BYTES and FILE_RECORDS are optional when RECORD_TYPE=STREAM,
        # so check for this case:

        next if (($i==2 || $i==3) &&  $record_type eq "STREAM");

        # Otherwise, log the error:
      
        $errors++;
        printf "$indent Missing $rkw[$i]\n";
      }
  }

# ... special check for object pointer:

if ($found[10] != 1)
  { $errors++;
    printf "$indent Missing ^<OBJECT> pointer\n";
  }

# ... for data files:

if ($label_type eq "-e"  ||  $label_type eq "-s")
  { for ($i=11; $i<16; $i++)
      { if ($found[$i] != 1)
          { $errors++;
            printf "$indent Missing $rkw[$i]\n";
          }
      }
  }

# ... for spacecraft files:

if ($label_type eq "-s")
  { for ($i=16; $i<18; $i++)
      { if ($found[$i] != 1)
          { $errors++;
            printf "$indent Missing $rkw[$i]\n";
          }
      }
  }

#-----------------------------
# Done.
#==============================================================================
#==============================================================================

sub check_datatype

  # This routine checks for valid DATA_TYPE values with the following
  # assumptions:
  #
  #   1. $interchange contains the correct INTERCHANGE_FORMAT.
  #   2. The data set was formatted by and/or for the SBN, so that BINARY
  #         DATA_TYPE values are also well-determined.
  #   3. The current PDS_VERSION_ID is 3.2
  #
  #  Valid ASCII types are:         Valid BINARY types are:
  #
  #      CHARACTER                       CHARACTER
  #      DATE                            DATE            (always in ASCII)
  #      TIME                            TIME            (always in ASCII)
  #      ASCII_COMPLEX                   IEEE_COMPLEX
  #      ASCII_INTEGER                   IEEE_REAL
  #      ASCII_REAL                      MSB_INTEGER
  #                                      MSB_UNSIGNED_INTEGER
  #                                      MSB_BIT_STRING
  #
  #  Any DATA_TYPE not in this list is reported as invalid.  A value if 1
  #  (true) is returned for valid types, 0 (false) otherwise.
  #

  { local ($buff,$icf,$count) = @_; # Input line, INTERCHANGE_FORMAT & line 
                                    # count
    local ($type);                  # DATA_TYPE values
    local ($invalid);               # valid type flag
    local ($i);                     # loop variable

    local (@ascii)  = ("CHARACTER", "DATE", "TIME", 
                       "ASCII_COMPLEX", "ASCII_INTEGER", "ASCII_REAL");
    local (@binary) = ("CHARACTER", "DATE", "TIME",
                       "IEEE_COMPLEX", "IEEE_REAL",
                       "MSB_INTEGER", "MSB_UNSIGNED_INTEGER", 
                       "MSB_BIT_STRING");


    # First, retrieve the DATA_TYPE value from the input buffer:

    $type = $buff;

    $type =~ s/^.*DATA_TYPE\s*=\s*//;
    if ($type =~ /^"/)
      { $type =~ s/\s+"/"/g;
        $type =~ s/^(".+").*$/$1/;
        $type =~ s/"//g;
        $type =~ s/ /_/g;
        $type =~ tr/a-z/A-Z/;
      }
    else
      { $type =~ s/^(\S+)\s+.*$/$1/; }

    # Now, break on INTERCHANGE_FORMAT to check for valid type:

    $invalid = 1;

    if ($icf eq "ASCII")
      { for ($i=0; $i<@ascii; $i++)
          { if ($type eq $ascii[$i])
              { $invalid = 0;
                last;
              }
          }
      }
    elsif ($icf eq "BINARY")
      { for ($i=0; $i<@binary; $i++)
          { if ($type eq $binary[$i])
              { $invalid = 0;
                last;
              }
          }
      }
    else
      { return 1; }

    # Check result:

    if ($invalid)
      { printf "$indent Invalid DATA_TYPE ($type) at line $lc.\n";
      }

    return !($invalid);
  }       
