#!/usr/bin/perl

use Getopt::Std;

#==============================================================================
#
#  SBNLV - SBN Label Verifier Tool
#
#  This utility checks syntax of the standard SBN-produced PDS data labels 
#  against the specifications found in the supplied data dictionary.  
#  
#  For the list of capabilities, see the accompanying man page.
#  
#  Format:  % sbnlv [-d <data dictionary>] [-o <output file>] <ODL file>
#  
#    where  -d <dictionary>   is the name of the data dictionary to be used.
#                             This should be the full name of the dictionary
#                             file.  The index file is assumed to have the 
#                             same file name and an extension of ".idx".  
#                             The default file is set at the top of the program.
#  
#           -o <output file>  is the name of the report file.  Default is to 
#                             write the report to stdout.
#  
#    The input ODL file name must be stated.
#
#  Modification History:
#
#   27 Nov 1996, acr: added code to handle sets and sequences; separated
#                     errors and warnings in the output
#   16 Dec 1996, acr: stamped a number of bugs to handle catalog files 
#                     properly (stop looking for object pointers there) and
#                     deal properly with lines with missing values.  Also
#                     removed chop commands after line input since the string
#                     processing seems to handle it OK.
#
#   03 Apr 1997, acr: Fixed bug that prevented sbnlv from allowing "UNK" and
#                     "N/A" values in time fields.
#
#   22 Apr 1997, acr: Fixed bug that messed up line splitting process when
#                     string contained '='; corrected spelling of PDS_VERSION_ID
#                     in error messages
#
#   02 Aug 1997, acr: Fixed dumb bug that required /^SFDU$/ pattern match 
#                     rather than just /^SFDU/
#
#   29 Jan 1998, acr: Fixed a couple of dumb typos that caused optional object
#                     lists to be ignored
#
#   07 Oct 1998, acr: Changed 'require' to use default system libraries
#
#   15 Oct 1998, acr: Changed pointer file name to allow upper and lower case;
#                     Added specific check for underscores in DATA_SET_ID
#
#   29 Sep 2000, acr: Expanded file name in pointer to 27.3 format
#
#   02 May 2001, acr: Added checks for 'z' in time field, unit specifiers in
#                     element values
#
#   26 Sep 2006, acr: Modified to use Getopt::Std
#  
#  **************************************************************************
#  
#  Global Variables
#  
#  A number of global variables are used to hold information used by different
#  subroutines.  These are:
#  
#  Program Accounting:
#    $lcount        Current line in input label/catalog file
#    $errcount      Total errors reported
#    $wrncount      Total warnings reported
#    $wrnarray      List of warning messages (saved to be printed separately)
#    $ATTACHED      True if this is an attached data label
#  
#  Index arrays:
#    %ddoff         Byte offset into the DD file for the object (from the index)
#    %ddtyp         Type of definition.  Used for finding matching objects/elem
#  
#  Alias lists:
#    %objalias      Object aliases
#    %elmalias      Elements aliases; context is included in subscript
#    
#  Object Parameters:
#    $ob_name       NAME
#    $ob_type       OBJECT = type
#    $ob_reqelm     Required elements list (concatenated into a string)
#    $ob_optelm     Optional elements
#    $ob_reqobj     Required objects
#    $ob_optobj     Optional objects
#  
#  Element Parameters:
#    $ed_name       NAME
#    $ed_type       OBJECT = type
#    $ed_maxlen     MAXIMUM_LENGTH (for character data)
#    $ed_maxval     MAXIMUM (for numeric data)
#    $ed_minval     MINIMUM (for numeric data)
#    $ed_svset      Standard value set, concatenated into a string
#  
#  File Handles:
#    DD             Data dictionary (stays open)
#    LBL            Input label/catalog file
#    RPT            Output report file (defaults to STDOUT)
#
#  
#  **************************************************************************
#  
#  Subroutines
#  
#  All subroutines except for the library function Getopt are included in this
#  file.  Subroutines are listed alphabetically after the main routine.  They
#  are:
#  
#    aliases             Reads the alias lists from the DD into %obj/%elmalias
#    checkdatetime       Called by check_element to check DATE and TIME values
#    check_element       Checks element value type, range and syntax
#    check_pointer       Checks syntax of pointer element; sets $ATTACHED flag
#    check_sequence      Checks sequence notation and member syntax
#    do_object           Main (recursive) driver routine for checking an OBJECT
#                         definition as it is read from the label
#    get_dd_set          Retrieves a set of values in DD format and concatenates
#                         them into a single string with blanks between
#    get_element_def     Reads the element definition from the DD into $ed_*
#    get_object_def      Reads the object definition from the DD into $ob_*
#    leapyear            Returns 1 if the given year is a leap year, 0 otherwise
#    next_statement      Reads the next ODL statement from the input file,
#                         ignoring blank lines and comments and concatenating
#                         text fields which span multiple records into a single
#                         string, then returns keyword and value pair
#    pdsddidx            Reads the DD index file into %ddtyp and %ddoff
#
#==============================================================================

#--------------------------------------
#  Global scalars and default values:
#--------------------------------------

$PSDD     = "/n/sbnops01/lcltools/doc/pdsdd.full";# Standard PSDD and index files
$PSDDIDX  = "/n/sbnops01/lcltools/doc/pdsdd.idx";
$lcount   = 0;                                 # Label line counter
$errcount = 0;
$wrncount = 0;
$ATTACHED = 0;
$OBJECT   = "(IMPLICIT) FILE";                 # Current object type
$report   = "-";                               # Default report file name
$label_type = "";                              # FILE for label, else CATALOG 
$underline = "___________________________________________";
$underline .= $underline;                      # Report file dingbat


#--------------------------------------
#  PRELIMINARIES
#--------------------------------------


# Start by dealing with the input parameters:

getopts('d:o:');

if (@ARGV != 1)    # There must be an input file name
  { die "Usage: sbnlv [-d <DD>] [-o <output>] <ODL file>.\n"; }
else
  { $label = $ARGV[0]; }

if ($opt_d)        # There was a dictionary file specified
  { $PSDD = $opt_d;
    $PSDDIDX = $PSDD;
    $PSDDIDX =~ s/\.[^\.]*$/.idx/    # Replace current extension with ".idx"
  }

if ($opt_o)        # Output file name specified
  { $report = $opt_o; }


# Open the label file and confirm that it is a valid file by checking the first
# non-blank line for either "SFDU", "PDS_VERSION_ID" or "OBJECT =":

open (LBL,$label)  ||  die "sbnlv: Unable to open label file $label.";

$K = "ERR";

while ($K eq "ERR") 
  { ($K,$V) = &next_statement; }   # Read first data line

die "$K" if ($K eq "EOF");

# Now, all files should begin with PDS_VERSION_ID or an SFDU label followed
# by the version id.  If the very next line begins an OBJECT definition, then
# this is a catalog file.  Otherwise, it' a label:

if ($K =~ /^PDS_VERSION_ID$/  ||  $V =~ /^SFDU/)
  { while ($K =~ /^PDS_VERSION_ID$/  ||  $V =~ /^SFDU/)
      { ($K,$V) = &next_statement; }
    if ($K =~ /^OBJECT$/)
      { $label_type = "CATALOG"; }
    else
      { $label_type = "(IMPLICIT) FILE"; }
  }
elsif ($K =~ /^OBJECT$/)
  { $label_type = "CATALOG"; 
    $wbuff1 = sprintf ("line %4d: ODL File does not begin with ", $lcount);
    $wbuff2 = "either PDS_VERSION_ID or an SFDU label.";
    $wrnarray[$wrncount] = "$wbuff1 $wbuff2";
    $wrncount++;
  }
else
  { die "$label is not recognizable as a PDS label or catalog file.\n"; }


seek(LBL,0,0);    # Rewind to beginning of file and reset line counter
$lcount = 0;


# Open the report file:

open (RPT,">$report")  ||  die "sbnlv: Could not open report file %s.";


# Open the data dictionary:

open (DD,$PSDD) || die "sbnlv: Could not open data dictionary file $PSDD.";


# Read in the index file:

&pdsddidx;


# Read in the alias tables:

&aliases;

# Write the report file heading:

printf RPT "                   SBN Label Verifier Error Report\n";
printf RPT "                   ===============================\n";
printf RPT "\n\n";
printf RPT     "     Input File: $label\n";
if ($label_type eq "(IMPLICIT) FILE")
  { printf RPT "      File Type: Label\n"; }
else
  { printf RPT "      File Type: Catalog\n"; }
printf RPT     "Data Dictionary: $PSDD\n";
printf RPT     "\n%79.79s\n", $underline;
printf RPT     "\nERRORS:\n\n";


#--------------------------------------
#  MAIN PROCESSING
#--------------------------------------


# Now to start the loop through the file.  If this is a label file, there is
# a default file object which must be ended by the "END" statement.  The whole
# thing can be handled by a setup and call to the object routine:

if ($label_type eq "(IMPLICIT) FILE")
  { 
    # The implicit FILE object of a label has different requirements than
    # the standard FILE object, so instead of calling the get_object_def
    # routine we set the global variables here by hand:

    $ob_name = "(IMPLICIT) FILE";
    $ob_type = "GENERIC_OBJECT_DEFINITION";
    $ob_reqelm  = " PDS_VERSION_ID  RECORD_TYPE  RECORD_BYTES";
    $ob_reqelm .= "  FILE_RECORDS  LABEL_RECORDS";
    $ob_reqelm .= "  DATA_SET_ID  PRODUCT_ID  TARGET_NAME";
    $ob_reqelm .= "  START_TIME  STOP_TIME  PRODUCT_CREATION_TIME ";
    $ob_optelm  = " PSDD ";
    $ob_reqobj  = "";
    $ob_optobj  = " PSDD ";

    
    $status = &do_object("(IMPLICIT) FILE");

  }

else
  { # For catalog files, we need to run a loop at this level to step through
    # the objects (there may be one or more).  There should be no keywords
    # found outside of objects apart from the initial PDS_VERSION_ID and
    # optional SFDU label, and there should be an "END" statement after 
    # the last END_OBJECT:

    ($key,$obj) = &next_statement;

    while ($key ne "END"  &&  $key ne "EOF")
      { if ($key eq "OBJECT")
          { &get_object_def($obj);
            unless (&do_object($obj) == 1)
              { printf "line %4d: Missing END_OBJECT for $obj.\n",$lcount; }
          }
        elsif ($key ne "ERR"  &&  $key ne "PDS_VERSION_ID" &&  $obj ne "SFDU")
          { printf RPT "line %4d: Keyword \"$key\" does not appear ", $lcount;
            printf RPT "within an OBJECT definition.\n";
            $errcount++;
          }

        ($key,$obj) = &next_statement;
      }

    if ($key eq "EOF")
      { printf RPT "line %4d: Missing END statement.\n", $lcount; 
        $errcount++;
      }
  }


# Finish up the report, printing any warnings encountered:

if ($wrncount > 0)
  { printf RPT "\n\nWARNINGS:\n\n";

   for ($i=0; $i<$wrncount; $i++)
     { printf RPT "$wrnarray[$i]\n"; }
  }

# Report the final totals:

printf RPT "\n%79.79s\n", $underline;

printf RPT "\n\nProcessing finished: ";
if ($errcount == 1)
  { printf RPT "1 error"; }
else
  { printf RPT "$errcount errors"; }
if ($wrncount == 0)
  { printf RPT ".\n"; }
elsif ($wrncount == 1)
  { printf RPT "; 1 warning.\n"; }
else
  { printf RPT "; $wrncount warnings.\n"; }

# Done.  Close files:

close(DD);
close(LBL);
close(RPT);

#--------------------------------------
#  EXIT
#--------------------------------------

#==============================================================================
#==============================================================================

sub aliases

  # This routine reads in the alias lists from the top of the full data 
  # dictionary file, already open on the DD handle, into the global
  # associative arrays, %objalias and %elmalias.

  { local ($lbuff,@fld);    # Input line buffer and sratch array
    local ($sub);           # for generating the element array subscripts

    # First, read past the stuff at the top, looking for the OBJECT_ALIAS_
    # SEQUENCE keyword:

    $lbuff = "";

    while ($lbuff !~ /OBJECT_ALIAS_SEQUENCE/)
      { $lbuff = <DD>;
        chop $lbuff;
      }

    # Object aliases are easy, as they are unique.  Loop through the lines,
    # stopping at the ELEMENT_ALIAS_SEQUENCE keyword.  Each line has all
    # quotes and parentheses stripped and is the split on the comma into 
    # alias and translation:

    chop ($lbuff = <DD>);
    while ($lbuff !~ /ELEMENT_ALIAS_SEQUENCE/)
      { $lbuff =~ s/[()']+//g;
        @fld = split(/,/,$lbuff);
        $objalias{$fld[0]} = $fld[1];
        chop ($lbuff = <DD>);
      }

    # Now we're at the element aliases.  These are trickier, since and alias
    # can mean different things in different contexts.  If it is the same in
    # all contexts, the context is noted as "N/A".  To get around this, the
    # associative array subscript is formed by joining the alias and context
    # together with a colon and using the resulting string for the subscript.
    # The end of the alias list is signalled by the END_OBJECT statement.

    chop ($lbuff = <DD>);
    while ($lbuff !~ /END_OBJECT/)
      { $lbuff =~ s/[()']//g;
        @fld = split(/,/,$lbuff);
        $sub = "$fld[0]:$fld[1]";
        $elmalias{$sub} = $fld[2];
        chop ($lbuff = <DD>);
      }

    # All done.  Reset the DD handle to point to the beginning of the file:

    seek(DD,0,0);
  }

#==============================================================================

sub checkdatetime

  # Subroutine to check the format of the input date/time field and make
  # sure it is a valid format.  Also performs a sanity check on subfield
  # values.  Note that at the moment PDS treats both DATE and TIME fields
  # identically.

  { local ($dt,$type) = @_;   # Input date/time and type of string
    local (@f,@d,@t);         # String subfields
    local ($date,$time);      # Separate substrings for checking
    local ($ok);              # logical flag
    local (@mdays);           # Number of days in each month

    @mdays = (0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);

    $ok = 1;

    # A "Z" at the end of the string is optional, but if it appears it must
    # be upper case. We check for it and delete it before further processing:

    $dt =~ s/Z$//;
    if ($dt =~ /z$/)
      { printf RPT "line %4d: 'z' at end of TIME value should be 'Z'.\n";
        $dt =~ s/z$//;
      }

    # Checks on the string generally: 
    #    o  The values "N/A" and "UNK" should be ignored
    #    o  Date/time values should not be enclosed in quotes
    #    o  Subfields must be zero-padded (i.e., no blanks)

    if ($dt eq "N/A"  ||  $dt eq "UNK")
      { return 1; }

    if ($dt =~ /"/)
      { printf RPT "line %4d: DATE/TIME values should not be enclosed ",$lcount;
        printf RPT "in quotes.\n";
        $errcount++;
        $ok = 0;
      }

    if ($dt =~ /\s/)
      { printf RPT "line %4d: DATE/TIME values must be zero-filled.\n",$lcount;
        $ok = 0;
        $errcount++;
      }

    # There are two recognized formats for PDS dates and times :
    # yyyy-mn-ddThh:mm:ss[.fff] and yyyy-dddThh:mm:ss[.fff]  
    # Either format may be truncated after the year, month, or day fields.

    { if ($dt =~ /T/)
        { @f = split(/T/,$dt);      # Separate date and time
          @d = split(/-/,$f[0]);    # split up date fields
          @t = split(/:/,$f[1]);    # split up time fields
          $time = 1;
        }
      else
        { @d = split(/-/,$dt);
          $time = 0;
        }

      # Checks on the date field components:

      if (@d < 1  || @d > 3)
        { $ok = 0; }

      elsif (@d == 3)
        { $ok = 0 if (length($d[0]) != 4);
          $ok = 0 if (length($d[1]) != 2);
          $ok = 0 if (length($d[2]) != 2);
          $mdays[2] += &leapyear($d[0]); 
          $ok = 0 if ($d[2] < 1  ||  $d[2] > $mdays[$d[1]]);
          $ok = 0 if ($d[1] < 1  ||  $d[1] > 12);
        }

      elsif (@d == 2)    # Could be day number or just a month
        { $ok = 0 if (length($d[0]) != 4);
          $ok = 0 if (length($d[1]) < 2);
          $ok = 0 if ($d[1] < 1  ||  $d[1] > (365 + &leapyear($d[0])));
        }
      else               # Just a year
        { $ok = 0 if (length($d[0]) != 4); }


      # Check on the time field components:

      if ($time  &&  @t == 3)
        { $ok = 0 if (length($t[0]) != 2);
          $ok = 0 if (length($t[1]) != 2);
          $ok = 0 if (length($t[2]) <  2  ||  length($t[2]) > 6);
          $ok = 0 if ($t[0] < 0  ||  23 < $t[0]);
          $ok = 0 if ($t[1] < 0  ||  59 < $t[1]);
          $ok = 0 if ($t[2] < 0  ||  60 < $t[2]);
          if ($t[2] == 60)
            { $wbuff1 = sprintf ("line %4d: Time field has '60' in", $lcount);
              $wbuff2 = "seconds field.  Is this a leap second?";
              $wrnarray[$wrncount] = "$wbuff1 $wbuff2";
              $wrncount++;
            }
        }
      elsif ($time  &&  @t != 3)
        { $ok = 0; }

      }

    # All done checking.  Issue an error message if warranted and return the 
    # final status:

    unless ($ok)
      { printf RPT "line %4d: Invalid DATE/TIME field \"%s\".\n", 
                   $lcount,$dt;
        $errcount++;
      }
    return $ok;
  }

#==============================================================================

sub check_element

  # This routine takes the input element and value, retrieves the matching
  # dictionary element, and checks the value for syntax, range, standard value,
  # etc.
  #
  # Various global elements are referenced here.  These are:
  #
  #    $lcount           Label line count
  #    %objalias         Object aliases
  #    %elmalias         Element aliases
  #    %ddoff            PSDD offset to definition (in bytes)
  #    %ddtyp            PSDD object type
  #    $OBJECT           current object type
  #
  #    Element definition variables
  #
  #    $ed_name          NAME
  #    $ed_type          GENERAL_DATA_TYPE
  #    $ed_maxlen        MAXIMUM_LENGTH
  #    $ed_maxval        MAXIMUM
  #    $ed_minval        MINIMUM
  #    $ed_svset         STANDARD_VALUE_SET (concatenated list of all values)

  { local ($ky,$vl) = @_;   # Input keyword and value
    local ($offset);        # Offset into PSDD of definition
    local ($tmpkey);        # Temporary key 
    local ($alias);         # Holding place for matching alias
    local ($sv);            # Holding place to format value for standard check
    local ($ok);            # error flag

    # First, make sure the keyword is in all caps:

    if ($ky =~ /a-z/)
      { printf RPT "line %4d: Keyword contains lower-case letters (%s).\n",
                       $lcount, $ky;
        $ky =~ tr/a-z/A-Z/;
        $errcount++;
      }

    # Now make sure it has a valid format:

    if ($ky !~ /^[A-Z0-9_]+$/)
      { printf RPT "line %4d: Invalid element name (%s).\n",$lcount,$ky;
        $errcount++;
        return -1;
      }

    # If there's no index entry for the keyword itself, try looking for an
    # alias:

    $offset = 0;

    if (!($ddtyp{$ky}))
      { 
        # First, try looking for an element alias with a specific context.
        # The current object type is concatenated to the element name and the
        # look-up is retried.  If this fails, "N/A" (indicating all contexts)
        # is appended and we try again.  If an alias is found, the keyword is
        # set to the equivalent name (in this routine).  Finally, if no match
        # is found an error is returned.

        $tmpkey = join(":",($ky,$OBJECT));
        if ($elmalias{$tmpkey})
          { $alias = $elmalias{$tmpkey}; }
        else
          { $tmpkey = join(":",($ky,"N/A"));
            if ($elmalias{$tmpkey})
              { $alias = $elmalias{$tmpkey}; }
            else
              { printf RPT "line %4d: \"%s\" not found in ",$lcount,$ky;
                printf RPT "the data dictionary.\n";
                $errcount++;
                return -2;
              }
          }
        printf RPT "line %4d: *INFO* Substituted \"%s\" for alias \"%s\".\n",
                      $lcount, $alias, $ky;
        $ky = $alias;
      }

  # Now there should DEFINITELY be an index entry, so signal a major error if
  # there still isn't:

  if (!($ddtyp{$ky}))
    { printf RPT "line %4d: *PROGRAM ERROR*\"%s\" not found ",$lcount,$ky;
      printf RPT "in the data dictionary, and it should have been.\n";
      $errcount++;
      return -100;
    }

  # Make sure this is an element and not an object:

  if ($ddtyp{$ky} !~ /ELEMENT/)
    { printf RPT "line %4d: \"%s\" is an OBJECT not an ELEMENT.\n",
                    $lcount, $ky;
      $errcount++;
      return -3;
    }


  # Load up the element definition in the global variables:

  &get_element_def($ky);

  # Look for a unit identifier on the end of the value. If it is there,
  # we remove it before further processing (no attempt to verify contents):

  $vl =~ s/\s*<.*$//;


  # Now, check the value against the dictionary parameters:

  $ok = 1;

  # ...checking improper use of quotes...

  if ($ed_type eq "REAL"     ||  $ed_type eq "EXPONENTIAL"  ||  
      $ed_type eq "INTEGER"  ||  $ed_type eq "NONDECIMAL"   ||
      $ed_type eq "DATE")
    { if ($vl =~ /["']/)
        { printf RPT "line %4d: $ed_type values should not be ",$lcount;
          printf RPT "enclosed in quotes.\n";
          $errcount++;
          $ok = 0;
        }
    }

  # Convert the input value to standard format (all upper-case, no blanks) 
  # for the remaining comparisons:

  $sv = $vl;
  $sv =~ s/^"([^"]*)"$/$1/;  # Remove enclosing quotes
  $sv =~ s/^'([^']*)'$/$1/;
  $sv =~ tr/a-z/A-Z/;
  $sv =~ s/\s+/_/g;

  # ...checking data type...  (Note: The type DECIMAL is not used anywhere
  #                                  in the PSDD)

  if ($ed_type eq "ALPHABET")
    { $ok = 0 if ($sv !~ /^[A-Za-z ]+$/); }
  elsif ($ed_type eq "ALPHANUMERIC")
    { $ok = 0 if ($sv !~ /^[A-Za-z0-9 ]+$/); }
  elsif ($ed_type eq "REAL")
    { $ok = 0 if ($sv !~ /^(\+|-|)[0-9]*(\.[0-9]*|)(E|e|)(\+|-|)[0-9]*$/); }
  elsif ($ed_type eq "EXPONENTIAL")
    { $ok = 0 if ($sv !~ /^(\+|-|)[0-9]*(\.[0-9]*|)(E|e)(\+|-|)[0-9]+$/); }
  elsif ($ed_type eq "INTEGER")
    { $ok = 0 if ($sv !~ /^(\+|-|)[0-9]+$/); }
  elsif ($ed_type eq "NONDECIMAL")       # (e.g., a bit-mask)
    { $ok = 0 if ($sv !~ /^[0-9]+#(\+|-|)[A-Fa-f0-9]+#$/); }
  elsif ($ed_type eq "DATE"  ||  $ed_type eq "TIME")
    { $sv = $vl;    # Check this one as is
      $ok = &checkdatetime($sv,$ed_type);
    }

  if (!($ok))
    { printf RPT "line %4d: \"%s\" is not a valid %s value.\n",
                  $lcount, $vl, $ed_type;
      $errcount++;
    }

  # ...checking value constraints...  (length or range)

  if ($ed_maxlen  &&  length($sv) > $ed_maxlen)
    { printf RPT "line %4d: Maximum length (%d) exceeded ", $lcount, $ed_maxlen;
      printf RPT "for $ed_type element (%d bytes).\n", length($sv);
      $errcount++;
      if (length($sv) > (1.5 * $ed_maxlen))
        { printf RPT "           ...Could this be a runaway string?\n"; }
      $ok = 0;
    }

  if ($ed_maxval  &&  $sv > $ed_maxval)
    { printf RPT "line %4d: Maximum value ($ed_maxval) exceeded.\n", $lcount; 
      $errcount++;
      $ok = 0;
    }

  if ($ed_minval  &&  $sv < $ed_minval)
    { printf RPT "line %4d: Minimum value ($ed_minval) not reached.\n", $lcount;
      $errcount++;
      $ok = 0;
    }

  # If there is a standard value set, see if the input value occurs.  To do 
  # this, convert the input value to all upper-case, remove quotes and replace
  # all blanks with underscores:

  if ($ed_svset  &&  $ed_svset !~ / $sv /)
    { $wbuff1 = sprintf ("line %4d: Non-standard value for $ky: $vl",$lcount);
      $wrnarray[$wrncount] = $wbuff1;
      $wrncount++;
    }

  # Before we return, we make a special check for any particular problem
  # keywords that have requirements that are not automatically gleanable from
  # the PSDD definition:

  if ($ky eq "DATA_SET_ID"  &&  $vl =~ /_/)
    { printf RPT "line %4d: Invalid DATA_SET_ID: %s ",$lcount, $vl;
      printf RPT "(may not contain underscores)\n";
      $errcount++;
      $ok = 0;
    }
    
  # All done.  Return the status:

  return $ok;
}

#==============================================================================

sub check_pointer

  # Routine to determine the type of the input pointer and check syntax for
  # data pointers.  It returns the dictionary entry matched if a data pointer,
  # 0 otherwise.  It also sets the $ATTACHED flag to true if this is an
  # attached data label.

  { local ($ptr,$val) = @_;    # Input pointer and value
    local ($fname,$offset);    # Patterns for matching parts of pointer values

    $fname  = "\"[A-Za-z0-9_-]{1,27}\\.[A-Za-z0-9_-]{1,3}\"";
    $offset = "[0-9]+\\s*(<BYTES>|)";

    # Begin by looking at the pointer itself and checking to see if it
    # represents a data pointer.  Data pointers will contain as the rightmost
    # part of its name the name of a data_dictionary object.  Note that there
    # may be prefixes on this:

    $ptr =~ s/^\^//;

    while ($ptr ne "") 
      { if ($ddtyp{$ptr} =~ /OBJECT/)
          {
            # Found an object pointer.  Check syntax. Note that parentheses
            # are required if both filename and offset are present, but are 
            # optional otherwise.  Dividing the check on the presence of
            # parentheses makes it easier to make sure there is a matching
            # set of them:

            if ($val =~ /\(/ )
              { if (  $val !~ /^\(\s*$fname\s*,\s*$offset\s*\)$/  
                  &&  $val !~ /^\(\s*($fname|$offset)\s*\)$/    )
                  { printf RPT "line %4d: Invalid object pointer syntax",
                               $lcount;
                    printf RPT "($val).\n";
                    $errcount++;
                  }
              }
            elsif ($val !~ /^$fname$/  &&  $val !~ /^$offset$/)
              { printf RPT "line %4d: Invalid object pointer syntax",$lcount;
                printf RPT "($val).\n";
                $errcount++;
              }

            # If there is no filename - just an offset - in the value, then
            # set $ATTACHED to true:

            $ATTACHED = 1 if ($val !~ /$fname/);

            return $ptr;
          }

        else
          {  # Couldn't match current value.  Try removing the prefix and 
             # check again:

             if ($ptr =~ /_/)     # There is a prefix - clip it
               { $ptr =~ s/^[^_]+_//; }
             else                 # No prefix - give up
               { $ptr = ""; }
          }
      }

    # If we got this far we never matched the pointer to an OBJECT.  All other
    # types of pointers must match the filename-only syntax, so check this 
    # before returning:

    if ($val !~ /^$fname$/)
      { printf RPT "line %4d: Invalid file pointer value ($val).\n", $lcount; 
        $errcount++;
      }

    return 0;
  }

#==============================================================================

sub check_sequence

  # Recursive routine to check the syntax of a sequence of values.

  { local ($key,$seq);          # Keyword and sequence list
    local (@mem,$m);            # Set members
    local ($SYM, $INT);         # Patterns for syntax-checking

    # Start by stripping off the outermost parentheses and preparing for 
    # splitting:

    $seq =~ s/^\(\s*//;
    $seq =~ s/\s*\)$//;
    @mem = split(/,/,$seq);

    # Now, for each member, determine if it is a scalar or another sequence,
    # and call the appropriate checking routine:

    foreach $m (@mem)
      { $m =~ s/^\s+//;
        $m =~ s/\s+$//;

        if ($m =~ /^\(/)   # sub-sequence
          { &check_sequence($k,$m); }

        else               # scalar
          { &check_element($k,$m); }
      }

    return 1;
  }

#==============================================================================

sub do_object

  # Main routine for descending through objects and checking each element and
  # sub-object.  This routine is called recursively, so provision is made to 
  # store the current object definition in local variables before calling the
  # routine on a sub-object, which will reset the global object variables.

  { local ($object) = $_[0];         # Current object name

    local ($save_name, $save_type,   # Holding places for object parameters
           $save_re, $save_oe,
           $save_ro, $save_oo);
    local ($reqelm_check);           # Check list for required elements
    local ($reqobj_check);           # Check list for required objects
    local ($k,$v);                   # Keyword and value from input line
    local (@fld);                    # Temporary output array
    local ($pointers,$ptr);          # Pointer list for label FILE, temp pointer
    local ($stat);                   # Function return status
    local ($itsalabel);              # True if the object is an implicit FILE
    local ($mem);                    # Single member of a set or sequence

    # Copy the required element and object lists into the check list
    # variables.  As each is encountered it is removed from the list, so 
    # that eventually the list should be empty.

    $reqelm_check = $ob_reqelm;
    $reqobj_check = $ob_reqobj;

    # If this is the implicit file object for a label file, we also want to
    # keep a list of object pointers encountered, so we can make sure all
    # pointers have a corresponding OBJECT definition at the top level.  This
    # list will be compiled as the label is read and the pointers encountered.

    if ($object eq "(IMPLICIT) FILE")
      { $itsalabel = 1;
        $pointers = "";
      }
    else
      { $itsalabel = 0;
        undef $pointers;
      }

    # The label file handle can be assumed to be pointing at either the
    # top of file or the "OBJECT =" line.  If this is a catalog file, the 
    # first line might be an SFDU label.  If so, skip it and go on:

    ($k,$v) = &next_statement;

    if ($itsalabel  &&  $v eq "SFDU")
      { ($k,$v) = &next_statement; }

    # From this point, loop through the lines of the object definition, 
    # checking elements and calling this routine recursively for sub-objects.

    while ($k ne "END_OBJECT"  &&  $k ne "END"  &&  $k ne "EOF")
      { if ($k eq "OBJECT")
          { 
            # Several procedures are required for sub-objects.  First do
            # do the accounting to make sure the object is allowed and to
            # indicate it was found if required:

            if ($ob_reqobj =~ / $v /)
              { $reqobj_check =~ s/ $v /  /; }
            elsif ($ob_optobj !~ / $v /  &&  $ob_optobj !~ / PSDD /)
              { printf RPT "line %4d: $v is not an allowed ",$lcount;
                printf RPT "sub-object of $ob_name.\n";
                $errcount++;
              }

            # If this the implicit file object in a label, also check to
            # see if this object has an entry on the pointers list.  If so,
            # remove it.  If not, complain about it:

            if ($object eq "(IMPLICIT) FILE")
              { if ($pointers =~ / $v /)
                  { $pointers =~ s/ $v / /; }
                else
                  { printf RPT "line %4d: No data pointer found for ",$lcount;
                    printf RPT "$v object.\n";
                    $errcount++;
                  }
              }

            # Now save the current global values, lookup the new object, 
            # call this routine again, and restore the values on return:

            $save_name = $ob_name;
            $save_type = $ob_type;
            $save_re   = $ob_reqelm;
            $save_oe   = $ob_optelm;
            $save_ro   = $ob_reqobj;
            $save_oo   = $ob_optobj;

            &get_object_def($v);
            last if (&do_object($v) == 0);   # i.e., if the routine returns an
                                             # error or EOF
            $ob_name   = $save_name;
            $ob_type   = $save_type;
            $ob_reqelm = $save_re;
            $ob_optelm = $save_oe;
            $ob_reqobj = $save_ro;
            $ob_optobj = $save_oo;
          }

        elsif ($k =~ /^\^/)
          {
            # We've got a pointer.  If we're currently working the implicit
            # file object of a label, check the syntax and determine whether it
            # is a data pointer (data pointers return the object name from the 
            # checking routine as it appears in the data dictionary, others
            # return 0).  Data pointers are added to the list of pointers 
            # which must have corresponding OBJECT definitions at this level.
            # Other pointers are ignored.  Note that the pointer syntax is
            # always checked:

            if (&check_pointer($k,$v)  &&  $itsalabel)
              { $k =~ s/^\^//;
                $pointers .= " $k ";
              }
          }

        elsif ($k ne "ERR")
          { # For elements, first make sure the element is allowed, check it 
            # off the list if it's required, and call the routine to verify
            # the value:

            if ($ob_reqelm =~ / $k /)
              { $reqelm_check =~ s/ $k /  /; }
            elsif ($ob_optelm !~ / $k /  &&  $ob_optelm !~ / PSDD /)
              { printf RPT "line %4d: $k is not an allowed keyword ",$lcount;
                printf RPT "in $ob_name.\n";
                $errcount++;
              }

            # If this is a set of values, each one must be verified, so split
            # them out and loop through them:

            if ($v =~ /^{/)
              { $v =~ s/[{}]//g;
                @fld = split(/,/,$v);
                foreach $mem (@fld)
                  { $mem =~ s/^\s+//;
                    $mem =~ s/\s+$//;
                    &check_element($k,$mem);
                  }
              }

            # If this is a sequence, it might be composed of other sequences.
            # Each member and sub-member must be verified.  To do this, call
            # a recursive routine to handle the potential loop:

            elsif ($v =~ /^\(/)
              { &check_sequence($k,$v); }

            else
              { &check_element($k,$v); }
          }

        # Next line:

        ($k,$v) = &next_statement;

      }

    # Now, see how the loop ended.  If this is an implicit FILE object, it 
    # should have ended with an END statement.  If it's some other object,
    # it must end in an END_OBJECT, and if that END_OBJECT has a value it
    # must match the OBJECt line value:

    if ($itsalabel)
      { if ($k ne "END")
          { printf RPT "line %4d: Missing END statement.\n", $lcount;
            $errcount++;
          }
      }
    else
      { if ($k eq "END_OBJECT"  &&  $v ne ""  &&  $v ne $object)
          { printf RPT "line %4d: END_OBJECT value ($v) does not ",$lcount;
            printf RPT "match OBJECT value ($object).\n";
            $errcount++;
          }
      }

    # All done.  Make sure the required elements and objects were all found. 
    # The required elements list has to be adjusted for detached data labels, 
    # though, so do that first:

    if ($object eq "(IMPLICIT) FILE")
      { if (!$ATTACHED)
          { $reqelm_check =~ s/ LABEL_RECORDS //; }

        # Also, if there are any object names left in the pointers list,
        # these dangling references should be indicated:

        if ($pointers !~ /^\s*$/)
          { $pointers =~ s/^\s*//;
            $pointers =~ s/\s*$//;
            @fld = split(/\s+/,$pointers);

            foreach $ptr (sort (@fld))
              { printf RPT "line %4d: No OBJECT definition found ",$lcount;
                printf RPT "for $ptr pointer.\n";
                $errcount++;
              }
          }
              
      }

    if ($reqelm_check)
      { if ($reqelm_check !~ /^\s*$/)    # i.e., if it isn't blank
          { $reqelm_check =~ s/^\s+//;
            $reqelm_check =~ s/\s+$//;
            @fld = split(/\s+/,$reqelm_check);
            if (@fld == 1)
              { printf RPT "line %4d: The required element $fld[0] was ",
                             $lcount;
                printf RPT "not found for object $ob_name.\n";
                $errcount++;
              }
            else
              { printf RPT "line %4d: The following required ",$lcount;
                printf RPT "elements were not found for $ob_name:\n";

                for ($i=0; $i<@fld; $i++)
                  { printf RPT "     $fld[$i]\n"; 
                    $errcount++;
                  }
              }
          }
      }

    if ($reqobj_check)
      { if ($reqobj_check !~ /^\s*$/)    # i.e., if it isn't blank
          { $reqobj_check =~ s/^\s+//;
            $reqobj_check =~ s/\s+$//;
            @fld = split(/\s+/,$reqobj_check);
            if (@fld == 1)
              { printf RPT "line %4d: The required sub-object $fld[0] was ",
                             $lcount;
                printf RPT "was not found for object $ob_name.\n";
                $errcount++;
              }
            else
              { printf RPT "line %4d: The following required ",$lcount;
                printf RPT "sub-objects were not found for $ob_name:\n";

                for ($i=0; $i<@fld; $i++)
                  { printf RPT "     $fld[$i]\n"; 
                    $errcounr++;
                  }
              }
          }
      }

    # And return.  Return value is true if "END_OBJECT" was encountered, 0 if
    # error or EOF, -1 if "END" was:

    if ($k eq "END_OBJECT")
      { $stat = 1; }
    elsif ($k eq "END")
      { $stat = -1; }
    else
      { $stat = 0; }

    return $stat;
  }

#==============================================================================

sub get_dd_set

  # Routine to retrieve the current set from the DD file and return it as a
  # single long string concatenation of all values.

  { local ($string,$line);   # output and input buffers
    local ($value);          # next value retrieved

    $string = "";
    $line = "";

    while ($line !~ /\}/)
      { chop ($line=<DD>);
        $line =~ s/^\s+//;
        $line =~ s/\s+$//;
        $word = $line;
        $word =~ s/[",}]+//g;
        $word =~ s/ /_/g;

        $string .= " $word ";
      }

    return $string;
  }

#==============================================================================

sub get_element_def

  # This routine retrieves the definition of the input element from the data
  # dictionary (making use of the index arrays) and saves the various 
  # parameters into global variables for access elsewhere in the program.
  # The data dictionary is assumed to be open on the DD handle.
  #
  # Element definition variables:
  #
  #    $ed_name          NAME
  #    $ed_type          GENERAL_DATA_TYPE
  #    $ed_maxlen        MAXIMUM_LENGTH
  #    $ed_maxval        MAXIMUM
  #    $ed_minval        MINIMUM
  #    $ed_svset         STANDARD_VALUE_SET (concatenated list of all values)

  { local ($element) = $_[0];   # Element to be retrieved
    local ($line);              # input buffer
    local ($not_done);          # loop flag
    local (@fld); 

    # Clear the global variables:

    undef $ed_name;
    undef $ed_type;
    undef $ed_maxlen;
    undef $ed_maxval;
    undef $ed_minval;
    undef $ed_svset;

    # First, make sure there is a match in the index arrays:

    if (!($ddtyp{$element})  ||  $ddtyp{$element} !~ /ELEMENT_DEFINITION/)
      { printf RPT "line %4d: No element \'$element\' found in the data ";
        printf RPT "dictionary.\n";
        $errcount++;
        return 0;
      }

    # Position the DD handle to the start of the definition:

    seek(DD,$ddoff{$element},0);

    # Read through the definition until the END_OBJECT is found, saving the
    # desired fields.  We can assume proper formatting in the data dictionary:

    $line = "";

    while ($line !~ /END_OBJECT/)
      { chop($line=<DD>);
        $line =~ s/^\s+//;
        $line =~ s/\s+$//;
        @fld = split(/ = /,$line);

        if ($fld[0] eq "NAME")
          { $ed_name = $fld[1];
            if ($ed_name ne $element)
              { printf RPT "MAJOR PROBLEM: Index entry for $element ";
                printf RPT "pointed to $ed_name.\n";
                $errcount++;
                return 0;
              }
          }

        elsif ($fld[0] eq "GENERAL_DATA_TYPE")
          { $ed_type = $fld[1]; }
 
        elsif ($fld[0] eq "MAXIMUM_LENGTH")
          { $ed_maxlen = $fld[1]; }

        elsif ($fld[0] eq "MAXIMUM")
          { $ed_maxval = $fld[1]; }

        elsif ($fld[0] eq "MINIMUM")
          { $ed_minval = $fld[1]; }

        elsif ($fld[0] eq "STANDARD_VALUE_SET")
          {
            # The standard value set is concatenated into a single list, with
            # each value surrounded by blanks so that a simple match can later
            # be performed against the whole list.  Each value is forced to
            # upper case, has outside quotes removed and all imbedded blanks 
            # replaced by underscores.

            $not_done = 1;

            while ($not_done)
              { chop ($line=<DD>);
                $not_done = 0 if ($line =~ /}/);

                $line =~ s/^\s*"//;
                $line =~ s/".*$//;
                $line =~ tr/a-z/A-Z/;
                $line =~ s/ /_/g;

                $ed_svset = $ed_svset . " $line ";
              }
          }
      }

    # Done.  Return successful status:

    return 1;
  }
    
#==============================================================================

sub get_object_def

  # This routine retrieves the object definition from the PSDD and saves it 
  # into the following global variables.  Note that these values must be 
  # saved locally for recursive calls to the object handling routine!
  #
  #    $ob_name           NAME
  #    $ob_type           GENERIC vs SPECIFIC_OBJECT_DEFINITION
  #    $ob_reqelm         Required element list
  #    $ob_optelm         Optional element list
  #    $ob_reqobj         Required sub-objects
  #    $ob_optobj         Optional sub-objects
  #
  # Also, users may have added a descriptive prefix to a generic object name,
  # so an iterative search is used to determine if the object is in the data
  # dictionary.

  { local ($object) = $_[0];
    local ($buff,$word);     # string buffers
    local ($n);              # throw-away string space
    local ($not_done);       # loop flag

    # Begin by clearing the current global values:

    undef $ob_name;
    undef $ob_type;
    undef $ob_reqelm;
    undef $ob_optelm;
    undef $ob_reqobj;
    undef $ob_optobj;

    # Now, look for the given object name in the index list.  If not found,
    # try translating it in the alias list.  If that doesn't work either,
    # peel off the leftmost prefix and try again:

    $not_done = 1;
    while ($not_done)
      { if ($ddtyp{$object} =~ /OBJECT/)
          { $ob_name = $object;
            $not_done = 0; }
        elsif ($ddtyp{$objalias{$object}} =~ /OBJECT/)
          { $ob_name = $objalias{$object};
            printf RPT "line %4d: *INFO* Substituted $ob_name for ", $lcount;
            printf RPT "alias $object.\n";
            $not_done = 0;
          }
        elsif ($object =~ /_/)
          { $object =~ s/^[^_]+_//; }
        else
          { $not_done = 0; }
      }

    # If $ob_name is still not defined, we never found a match.  Otherwise,
    # set the DD handle to point to the start of the object definition:

    if ($ob_name)
      { seek (DD, $ddoff{$ob_name}, 0); }
    else
      { printf RPT "line %4d: Could not find OBJECT \"%s\" ",$lcount,$object;
        printf RPT "in the data dictionary.\n";
        $errcount++;

        # For undefined objects the type is set to UNDEF as a flag, the 
        # required element/object lists are set to null and the optional
        # lists are set to "PSDD".  This way processing can continue on the
        # element which might appear in the dictionary even if the object
        # cannot be verified.

        $ob_type = "UNDEF";
        $ob_reqelm = "";
        $ob_optelm = "  PSDD  ";
        $ob_reqobj = "";
        $ob_reqelm = "  PSDD  ";

        return 0;
      }

    # Let the user know if we've matched a beheaded form of the object name 
    # found instead of the whole thing:

    if ($ob_name ne $_[0])
      { printf RPT "line %4d: *INFO* Could not find $_[0] in the data ", 
                     $lcount;
        printf RPT "dictionary. Will use the more general $ob_name instead.\n";
      }

    # Now read through lines until the end of the definition, saving the 
    # values encountered:

    $buff = "";
    while (($buff=<DD>) !~ /END_OBJECT/)
      { chop $buff;
        $buff =~ s/^\s+//;
        $buff =~ s/\s+$//;
        if ($buff =~ /OBJECT_DEFINITION/)
          { ($n, $ob_type) = split(/ = /,$buff); }
        elsif ($buff =~ /^NAME = /)
          { ($n, $ob_name) = split(/ = /,$buff); }
        elsif ($buff =~ /REQUIRED_ELEMENT_SET/)
          { $ob_reqelm = &get_dd_set; }
        elsif ($buff =~ /OPTIONAL_ELEMENT_SET/)
          { $ob_optelm = &get_dd_set; }
        elsif ($buff =~ /REQUIRED_OBJECT_SET/)
          { $ob_reqobj = &get_dd_set; }
        elsif ($buff =~ /OPTIONAL_OBJECT_SET/)
          { $ob_optobj = &get_dd_set; }
      }

    # Done.

    return 1;
  }

#==============================================================================

sub leapyear

  # Returns 1 if the given year (two or four digits) is a leap year, 0 
  # otherwise.

  { local ($year) = $_[0];

    return 0 if (($year % 4) != 0);

    if (($year % 100) == 0  &&  ($year % 400) == 0)
      { return 1; }
    else
      { return 0; }
  }

#==============================================================================

sub next_statement

  # This routine reads the next statement from the label file open on LBL, 
  # parses it, and returns the keyword and value in separate scalars.  Note
  # that only values in quotes are allowed to span lines; and that this 
  # routine concatenates those values into a single string (so it may read 
  # more than one line per call).  Blank lines and lines consisting entirely
  # of comments are ignored.  In-line comments following values are ignored.
  # End of file should NOT be encountered in this routine, as the calling 
  # routine should be keeping an eye out for the END statement, so when EOF 
  # is encountered in this routine it is an error.

  # The global line count variable, $lcount, is updated in here.  Note that
  # label lines should end with carriage return AND linefeed, so 2 characters
  # are chopped from the end of each input line.

  { local ($line,$key,$val);  # Work spaces
    local ($still_looking);   # Loop flag
    local ($term);            # Value terminator
    local ($start_line);      # starting point for multi-line value string
    local ($qi);              # '=' index (pointer for splitting)

    # Start by skipping over blank and comment lines to the next statment line:

    $still_looking = 1;

    while ($still_looking)
      { if (!($line = <LBL>))
          { printf RPT "line %4d: Unexpected end of input file.\n", $lcount;
            $errcount++;
            return ("EOF",0);
          }

        $lcount++;
##        chop $line; chop $line;  

        next if ($line =~ /^\s*$/);       # Blank lines
        next if ($line =~ /^\s*\/\*/);    # Comment-only lines

        # If we made it this far we've got one:

        $still_looking = 0;
      }

    # This must be a statement line.  I.e., either it consists of a single,
    # unadorned keyword (possibly followed by a comment) or it has an "=":

    if ($line !~ /=/)
      {
        # Before checking for a lone keyword, remove trailing comments:

        $line =~ s/\/\*.*$//;

        if ($line =~ /^\s*[A-Za-z0-9_]+\s*$/)
          { $key = $line;
            $key =~ s/\s+//g;
            $val = "";
          }
        else
          { printf RPT "line %4d: Assignment expected but not found.\n",
                           $lcount;
            $errcount++;
            return ("ERR",0);
          }
      }

    else
      {
        # OK, this looks like an assignment.  Split it into keyword and 
        # value and do some syntax checking on the keyword:

        $line =~ s/([^'"]*)\/\*.*$/$1/;     # Remove comments
        $line =~ s/("[^"]*")\s*\/\*.*$/$1/;
        $line =~ s/('[^']*')\s*\/\*.*$/$1/;
        $line =~ s/\s*=\s*/=/;    # Remove space around the first '='
        $line =~ s/^\s+//;        # Remove leading and trailing blanks
        $line =~ s/\s+$//;

        # There may be an '=' in the value part of the string, so we have to 
        # do this carefully:

        $qi = index($line,"=");
        $key = substr($line,0,$qi);
        $val = substr($line,$qi+1);

        # If the value field is null, read in the next line and interpret it
        # as the value field.  But check for another "=", just in case the
        # Value was left entirely blank.

        while ($val eq "")
          { if (!($line = <LBL>))
              { printf RPT "line %4d: Unexpected end of input file.\n", $lcount;
                $errcount++;
                return ("EOF",0);
              }
            $lcount++;
##            chop $line; chop $line;

            # Clean up the input line:

            $line =~ s/^([^'"]*)\/\*.*$/$1/;
            $line =~ s/("[^"]*")\s*\/\*.*$/$1/;
            $line =~ s/('[^']*')\s*\/\*.*$/$1/;
            $line =~ s/^\s+//;
            $line =~ s/\s*$//;

            # Now check for another "=" and deal:

            if ($line =~ /=/)
              { printf RPT "line %4d: No value given for $key\n",$lcount-1;
                $errcount++;

                # Recover by dropping the old keyword, splitting the new
                # line and letting the loop continue:

                $line =~ s/\s*=\s*/=/;
                ($key,$val) = split(/=/,$line);
              }
            else
              { $val = $line; }
          }

        if ($key !~ /^(\^|)[A-Za-z0-9_]+$/)
          { printf RPT "line %4d: Invalid keyword ($key)\n",$lcount;
            $errcount++;
            return ("ERR",0);
          }

        # Values in quotes, sets and sequences may run over one line.  If any 
        # of these are true, make sure the entire value is collected before
        # returning:

        # ...QUOTES

        if ($val =~ /^["']/)
          { $term = substr($val,0,1);
            if ($term eq "'")
              { $wbuff1 = sprintf ("line %4d: PDS recommends double", $lcount);
                $wbuff2 = "quotes only.";
                $wrnarray[$wrncount] = "$wbuff1 $wbuff2";
                $wrncount++;
              }

            # There is an opening quote.  If there's a matching quote on this
            # line, we're done.  Remove trailing blanks and comments and check
            # for garbage after the closing quote:

            if ($val =~ /^'[^']*'/)
              { $val =~ s/\s+$//;
                $val =~ s/\s*\/\*.*$//;
                
                if ($val !~ /'$/)
                  { $wbuff1 = sprintf ("line %4d: Extra characters", $lcount);
                    $wbuff2 = "following end of value.";
                    $wrnarray[$wrncount] = "$wbuff1 $wbuff2";
                    $wrncount++;
                  }
              }
            elsif ($val =~ /^"[^"]*"/)
              { $val =~ s/\s+$//;
                $val =~ s/\s*\/\*.*$//;

                if ($val !~ /"$/)
                  { $wbuff1 = sprintf ("line %4d: Extra characters",$lcount);
                    $wbuff2 = "following end of value.";
                    $wrnarray[$wrncount] = "$wbuff1 $wbuff2";
                    $wrncount++;
                  }
              }

            else
              { # No closing quote.  We'll keep reading in lines until a 
                # matching terminating quote (in $term) is found.  Leading
                # and trailing blanks are ignored, as nothing is actually
                # done to the value - we just keep concatenating until we
                # find the end or run out of file.

                $start_line = $lcount;
                $still_looking = 1;

                while ($still_looking)
                  { if (!($line=<LBL>))
                      { printf RPT "line %4d: Unexpected end of file while", 
                                    $start_line;
                        printf RPT " looking for end of value string.\n";
                        $errcount++;
                        return ("EOF",0);
                      }
                    $lcount++;
##                    chop $line; chop $line;

                    if ($line =~ /$term/)
                      { 
                        # Check for garbage after closing quote:

                        $line =~ s/\s+$//;        # Remove trailing blanks
                        $line =~ s/\s*\/\*.*$//;  # Remove trailing comments

                        if ($line !~ /$term$/ )
                          { $wbuff1 = sprintf ("line %4d: Extra", $lcount);
                            $wbuff1 = "$wbuff1 characters following end of"; 
                            $wbuff2 = "value\"$line\".";
                            $wrnarray[$wrncount] = "$wbuff1 $wbuff2";
                            $wrncount++;
                          }

                        # Add line to value string and set ending flag:
 
                        $line =~ s/$term.*$/$term/;
                        $val = $val . $line;
                        $still_looking = 0;
                      }

                    else    # Just concatenate the line and continue
                      { $val = $val . $line; }
                  } # WHILE looking for closing quote 
              } # ELSE no closing quote
          } # ELSE in quotes


        # ...SETS AND SEQUENCES

        elsif ($val =~ /^[{\(]/)
          { $line = $val;
            $val = "";

            # Some local variables for looping through characters to match
            # sequence/set markers:

            local (@c, $i,$paren_count);
            local ($in_quote, $in_sequence,$not_done);

            @c = split(//,$line);
            $i = 1;
            $in_quote = 0;
            $not_done = 1;
            if ($c[0] eq "{")
              { $in_sequence = 0; }
            else
              { $in_sequence = 1; 
                $paren_count = 1;
              }

            while ($not_done)
              { if ($in_quote)
                  { if ($c[$i] eq $term)
                      { $in_quote = 0; }
                  }

                elsif ($c[$i] eq "\""  ||  $c[$i] eq "'")
                  { $in_quote = 1;
                    $term = $c[$i];
                  }

                # Remove comments outside of quotes:

                elsif ($c[$i] eq "/"  &&  $c[$i+1] eq "*")
                  { $line =~ s/\/\*.*$//; }


                # Check for end of sequence or set markers:

                elsif ($in_sequence)
                  { if ($c[$i] eq "(")
                      { $paren_count++; }

                    elsif ($c[$i] eq ")")
                      { $paren_count--;
                        if ($paren_count == 0)
                          { $not_done = 0; }
                      }
                  }

                else   # in a set
                  { if ($c[$i] eq "{")
                      { printf RPT "line %4d: Sets may not be nested.\n",
                                     $lcount; 
                        $errcount++;
                      }

                    elsif ($c[$i] eq "}")
                      { $not_done = 0; }
                  }

                $i++;

                # If we've reached the end of this line, add it to the current `
                # value string and fetch the next line:

                if ($i >= length($line)  &&  $not_done)
                  { $val = $val . $line;
                    if (!($line = <LBL>))
                      { printf RPT "line %4d: Unexpected end of input file.\n",
                                     $lcount;
                        $errcount++;
                        return ("EOF",0);
                      }
                    $lcount++;
##                    chop $line;
                    $i = 0;
                    @c = split(//,$line);
                  }
              }

            # Found end. Remove everything after the closing paren/brace:

            if ($in_sequence)
              { $line =~ s/\)[^\)]*$/)/; }
            else
              { $line =~ s/}.*$/}/; }
 
            $val = $val . $line;

          }

        # If the value is not enclosed in quotes or brackets, then it better
        # not contain any space (once trailing comments are removed, i.e.),
        # although it may be followed by a unit specifier:

        else
          { # No quotes:

            $val =~ s/\s*\/\*.*$//;

            if ($val =~ /\s/  &&  $val !~ /^\S+\s+<.*>/)
              { printf RPT "line %4d: Invalid value format ($val).\n",
                                $lcount;
                $errcount++;
                return ("ERR",0);
              }
          }

      } # ELSE assignment line

    # Return the keyword and value:

    return ($key,$val);

  }

#==============================================================================

sub pdsddidx

  # Routine to read the PSDD index file into the (global) associative arrays
  # %ddoff and %ddtyp for later reference.

  { local ($lbuff,@fld);   # Line and field buffers
    local ($notfound);     # Loop flag

    # First, open the index file:

    open (IDX,$PSDDIDX) ||  die "sbnlv: Could not open $PSDDIDX for reading.";

    # Read past the attached label file:

    $notfound = 1;
    while ($notfound)
      { $lbuff = <IDX>;
        chop $lbuff;
        if ($lbuff =~ /^\s*END\s*$/)
          { $notfound = 0; }
      }

    # We're now ready to start reading the index lines.  The first field in 
    # each line is the element/object name, the second is the type of 
    # definition, the third is the byte offset of the beginning of the 
    # definition from the start of the full PSDD file.

    while ($lbuff=<IDX>)
      { chop $lbuff;
        @fld = split(/\s+/,$lbuff);
        $ddoff{$fld[0]} = $fld[2];
        $ddtyp{$fld[0]} = $fld[1];
      }

    # We're done with the index file.  Close it up and return:

    close(IDX);
  }

#==============================================================================
#==============================================================================
#==============================================================================
