#!/usr/bin/perl

use Getopt::Std;

# Routine to list the REFERENCE_KEY_ID fields found in a list of files
# Each ID is shown only once and the output list is alphabetized.
#
# Format:  
#
#   % showrefkeys [-r] [-a] [-c] [-l] [-x <filename>] [ file/directory_list ]
#
#  where   -r  indicates directories should be searched recursively
#          -a  check all files for PDS labels (must begin with PDS_VERSION_ID)
#          -c  indicates a count of the number of times each ID appears should
#              be included in the output listing
#          -l  list the files in which each key occurs
#          -x  exclude files called <filename> (no wildcards)
#
# NOTE: This routine assumes there will only be one value for each 
#       REFERENCE_KEY_ID. This is NOT required byt he standards.
#
# Global Variables:
# -----------------
#  $opt_a     true for "-a"
#  $opt_c     true for "-c"
#  $opt_l     true for "-l"
#  $opt_r     true for "-r"
#  $opt_x     contains "<filename>"     
#  %refkey    master collection of all unique REFERENCE_KEY_IDs
#  %keyfiles  list of files containing each REFERENCE_KEY_ID
#
#
#  16 May 2000, acr.
#  26 Sep 2006, acr: Modified to use Getopt::Std
#
#=============================================================================
# Get the input options and arguments:

getopts('aclrx:');

if (@ARGV == 0)
  { $ARGV[0] = "."; }

# Loop through the command line arguments, separating files from directories:

undef $dlist;
undef $flist;
$dl = 0;
$fl = 0;

foreach $file (@ARGV)
  { if (-d $file)
      { $dlist[$dl] = $file;
        $dl++;
      }
    elsif ($file =~ /\.cat$/  ||  $file =~ /\.CAT$/  ||
           $file =~ /\.lbl$/  ||  $file =~ /\.LBL$/  ||
          ($opt_a  &&  &isPDSlabel($file)) )
      { if ($opt_x  &&  $file ne $opt_x)
          { $flist[$fl] = $file;
            $fl++;
          }
      }
  }

# Now call the routines to process each file and directory:

foreach $file (@flist)
  { &dofile($file); }

foreach $dir  (@dlist)
  { &dodir($dir); }

# Now print the final list (format determined by command-line options):

if ($opt_l)     # Long format, with counts and file names   
  { foreach $id (sort (keys %refkey))
      { @list = sort (split(/,/,$keyfiles{$id}));
        $file = shift @list;
        printf "%-20.20s %5d  %s\n",$id, $refkey{$id}, $file;
        while ($file = shift(@list))
          { printf "%26.26s  %s\n"," ",$file; }
      }
  }
elsif ($opt_c)  # Display of counts requested
  { foreach $id (sort (keys %refkey))
      { printf "%-20.20s %5d\n", $id, $refkey{$id}; }
  }

else            # Simple display
  { foreach $id (sort (keys %refkey))
      { printf "$id\n"; }
  }

# Done.

#______________________________________________________________________________
#______________________________________________________________________________
#

sub isPDSlabel

# Function returning 1 if the given file starts with a PDS label, 0 otherwise.
# PDS labels must have the PDS_VERSION_ID keyword as the first in the label,
# or must have "OBJECT = REFERENCE" as the first line.

{ my ($file) = $_[0];
  my ($line);

  open (IN,$file) || die "Could not open '$file' for PDS-check.";

  while ($line = <IN>)
    { last if ($line !~ /^\s*$/  &&  $line !~ /^\s*\/\*/); }
  close(IN);

  if ($line =~ /^\s*PDS_VERSION_ID\s*=/  ||
      $line =~ /^\s*OBJECT\s*=\s*REFERENCE/)
    { return 1; }
  else
    { return 0; }
}

#______________________________________________________________________________
#

sub dofile

# Subroutine to open and read the ODL (or ODL-labelled) file given and look
# for all REFERENCE_KEY_ID values. These are added to the collection in 
# %refkey.

{ my ($file) = $_[0];
  my ($line);
  my ($not_done);
  my ($id);

  # Open the file:

  open (IN,$file) || die "Could not open '$file' for reading.";

  # Loop through the file until end of file or end of label is found:

  $line=<IN>;
  $not_done = 1;
  while ($not_done)
    { last if ($line =~ /^\s*END\s*$/ ||  $line =~ /^\s*END\s*\/\*/);

      if ($line =~ /^\s*REFERENCE_KEY_ID\s*=/)
        { chomp $line;
          $line =~ s/\/\*.*$//;
          $line =~ s/^.*=\s*//;
          $line =~ s/\s*$//;
          $line =~ s/"//g;
          $refkey{$line}++;
          if ($opt_l)
            { $keyfiles{$line} .= "$file,"; }
        }

      $not_done = ($line=<IN>);
    }

  # Close the file and we're done:

  close(IN);
  return 1;
}

#______________________________________________________________________________
#

sub dodir

# Subroutine to open a directory, find the PDS labelled and catalog files, 
# and call the 'dofile' subroutine on each. Subdirectories are examined
# recursively if the user included the '-r' option on the command line.
# Hidden files are ignored.

{ my ($dir) = $_[0];
  my ($file);
  my (@flist,@dlist);
  my ($fl,$dl);

  # Open the directory and initialize lists:

  opendir (DIR,$dir)  ||  die "Unable to open directory '$dir'.";

  undef @flist;
  undef @dlist;
  $fl = 0;
  $dl = 0;

  # Loop through the directory contents:

  foreach $file (sort readdir(DIR))
    { next if ($file =~ /^\./);
      next if ($opt_x  &&  $file eq $opt_x);

      $file = "$dir/$file";
      next if ($opt_x  &&  $file eq $opt_x);
      if ($opt_r  &&  -d $file)
        { $dlist[$dl] = $file;
          $dl++;
        }
      elsif ($file =~ /\.cat$/  ||  $file =~ /\.CAT$/  ||
             $file =~ /\.lbl$/  ||  $file =~ /\.LBL$/  ||
            ($opt_a  &&  &isPDSlabel($file)) )
        { $flist[$fl] = $file;
          $fl++;
        }
    }

  # Now call the appropriate subroutine for each list:

  foreach $file (@flist)
    { &dofile($file); }

  foreach $file (@dlist)
    { &dodir($file); }

  # And we're done.

}
