#!/usr/bin/perl

# Routine to split a reference catalog file into separate files for each
# reference.
#
# Format: split_ref [-P] [-d output_dir] [ reference_file ]
#
#    where  -P             Add the "PDS_VERSION_ID" and "END" lines to each 
#                          output file
#           -d output_dir  Directory to hold the output files (default is
#                          current working directory)
#
# Output files are given ISO-compliant names (but lowercase) based on the
# REFERENCE_KEY_ID value.  Non-compliant characters are replaced by 
# underscores; consecutive underscores are collapsed to a single character;
# leasing and trailing underscores are removed.  The extension ".cat" is
# used on the output files.  No attempt is made to adjust padding or 
# carriage control in the output files - it will be the same as in the 
# input file.
#
#==============================================================================

use Getopt::Std;

$TRUE  = 1;
$FALSE = 0;

# Deal with the input parameters and arguments:

getopts ('Pd:');

if ($opt_d)
  { if (-d $opt_d)
      { $outdir = $opt_d; }
    else
      { die "Output directory '$opt_d' does not exist"; }
  }
else
  { $outdir = "."; }

if (@ARGV == 0)
  { $reffile = "-"; }
elsif (@ARGV == 1)
  { $reffile = $ARGV[0]; }
else
  { die "Usage: split_ref [-P] [-d output_dir] [ reference_file ]\n"; }

# All righty.  Now we'll open the input file and prepare to loop through it:

open (REF,$reffile) || die "Could not open $reffile ($!)";

# Read the first link and determine carriage control:

$line = <REF>;
$CRLF = ($line =~ /\r/)? "\r\n" : "\n";

$not_done = $TRUE;
$count = 0;

while ($not_done)
  { # Read to the next reference object:

    while (($line=<REF>)  &&  ($line !~ /REFERENCE_KEY_ID/))
      { ; }

    if (eof REF)
      { $not_done = $FALSE;
        next;
      }

    # Get the REFERENCE_KEY_ID value: 

    if ($line =~ /"/)
      { $line =~ /"(.+)"/; 
        $id = $1;
      }
    else
      { $line =~ /=\s*(\S+)\s/; 
        $id = $1;
      }
    
    # Convert it to an ISO-compliant file name:

    $name = $id;
    $name =~ tr/A-Z/a-z/;
    $name =~ s/[^a-z0-9_]+/_/g;
    $name =~ s/__+/_/g;
    $name =~ s/^_//;
    $name =~ s/_$//;
    $name = substr($name,0,28) if (length($name) > 28);

    # Open the file in the output directory and write the opening
    # line, if requested ("-P" option):

    open (OUT, "> $outdir/$name.cat")  ||
      die "Could not open $outdir/$name.cat for writing ($!)";

    if ($opt_P)
      { printf OUT "%-78.78s$CRLF", "PDS_VERSION_ID = PDS3"; 
        printf OUT "%-78.78s$CRLF", " ";
      }

    # Print the OBJECT = REFERENCE line, the REFERENCE_KEY_ID line,
    # and then read lines to the output file until the end of the
    # object:

    printf OUT "%-78.78s$CRLF", "OBJECT     = REFERENCE";
    printf OUT $line;

    while ($line !~ /END_OBJECT/)
      { $line=<REF>;
        printf OUT $line;
      }

    # Write the "END" line, if requested:

    if ($opt_P)
      { printf OUT "%-78.78s$CRLF", " "; 
        printf OUT "%-78.78s$CRLF", "END";
      }

    # Close the file and look for the next reference.

    close (OUT);
    $count++;

  }

# Close up the input file and we're done.

close(REF);

printf stderr "$count files written.\n";
