#!/usr/bin/perl

use Getopt::Std;

# Routine to search the 'allrefs.cat' file for references matching the
# input string.
#
# Format:  srchref [ -t ] [ -f <ref_file> ] [ -o <output> ] <string>
#
#  where  -t   attempts to match the string anywhere in the reference
#              text (case-sensitive)
#         -f   provides a file other than "allrefs.cat" to be searched
#         -o   directs the output to the named file
#
#  Default behaviour is to force the input string to uppercase and search
#  only the REFERENCE_KEY_ID field.  All matching REFERENCE objects are
#  returned, in their entirety
#
# 30 Apr 2003, A.C.Raugh
#
# Modification History
# ====================
#
# 17 Jul 2003  A.C.Raugh   Added message for no match found
#
#============================================================================

getopts('tf:o:');

$TRUE  = 1;
$FALSE = 0;
$no_match = $TRUE;

# Sort out the various options:

$REFFILE = ($opt_f)? $opt_f : "/n/sbnops01/lcltools/doc/allrefs.cat";
$OUTFILE = ($opt_o)? $opt_o : "-";

if (@ARGV != 1)
  { die "Usage: srchref [ -t ] [ -f <ref_file> ] [ -o <output> ] <string>\n"; }

# Open files:

open (REF,$REFFILE) || die "Could not open reference file '$REFFILE' ($!) ";
open (OUT,"> $OUTFILE") ||
  die "Could not open output file '$OUTFILE' ($!) ";

# Determine whether we're searching IDs or text.  If IDs, we need to get an
# uppercase version of the argument string to work with.  Otherwise, we'll
# preserve case:

$srch_string = $ARGV[0];
if (!$opt_t)
  { $srch_string =~ tr/a-z/A-Z/; }

# We're letting the user pass anything here, so he may have inserted meta-
# characters for wildcard searches, for example.  If he can get it past the
# shell, we'll try to run with it.  Forcing things to uppercase would mess
# that up, but searches on ID fields are going to be fairly simple anyway,
# so we'll let it ride.

# We'll save the reference lines in an array, so we can display them without
# having to re-read them.  The first line will always be the OBJECT line, so
# we can initialize that and forget about it:

$ref[0] = "OBJECT     = REFERENCE\n";

# Now we'll step through the reference file, one object at a time, looking
# for matches:

$not_done = nextref($line);
$r = 1;

while ($not_done)
  { # We split here based on whether we're doing an ID search or a text 
    # search:

    if ($opt_t)
      { $matched_it = $FALSE;

        while ( !$matched_it  &&  $line !~ /^\s*END_OBJECT/ )
          { $matched_it = ($line =~ /$srch_string/i);
            $ref[$r] = $line;
            $r++;

            $line = <REF>;   # Next line
          }

        # We check for a match. If we found one, we read in the rest of 
        # the reference and print the whole thing out:

        if ($matched_it)
          { while ($line !~ /^\s*END_OBJECT/)
              { $ref[$r] = $line;
                $r++;
                $line = <REF>;
              }
            $ref[$r] = $line;
            $r++;

            for ($i=0; $i<$r; $i++)
              { printf OUT $ref[$i]; }
            printf OUT "\n";

            $no_match = $FALSE;
          }

        # Non-matches are ignored.  Either way we need to re-initialize for
        # the next reference:

        $r = 1;
      }

    # In the case of the ID match, we only need check the line returned by
    # 'nextref' to know whether we've got a match.

    elsif ($line =~ /$srch_string/)
     { # We have a match, so we read and print the rest of the reference:

        printf OUT $ref[0];
        printf OUT $line;

        while ($line !~ /^\s*END_OBJECT/)
          { $line = <REF>;
            printf OUT $line;
          }
        printf "\n";

        $no_match = $FALSE;
      }

    # Advance to the next REFERENCE object:

    $not_done = nextref($line);
  }

# If we found no matches, tell the user:

printf OUT "No match found for '$ARGV[0]'.\n" if ($no_match);

# Close up files and we're done:

close(REF);
close(OUT);


#=============================================================================
sub nextref

  # Routine to find the beginning of the next REFERENCE object.  The 
  # REFERENCE_KEY_ID line is returned in the argument; the function returns
  # TRUE if a REFERENCE was found or FALSE if we encountered end of file.
  # This routine is a notational convenience, and uses the global REF file
  # handle and values of $TRUE and $FALSE.

{ my ($line);

  while (!eof(REF)  &&  $line !~ /^\s*REFERENCE_KEY_ID\s*=/)
    { $line = <REF>; }

  if (eof(REF))
    { return $FALSE; }
  else
    { @_[0] = $line; 
      return $TRUE;
    }
}

 
