#!/usr/bin/perl

use Getopt::Std;

# TBLTMPL - create a label template for ASCII table files
#
#  This program reads through the given input file and generate a template
#  TABLE description complete with COLUMN definitions.  The template contains
#  a best-guest physical description of the data based on what the program 
#  finds as it scans through the file.  Logical fields (like NAME, DESCRIPTION,
#  UNIT, etc.) must, of course, be filled in later.
#
#  17 February 1998, acr.
#
#  Format:
#
$USAGE = "tbltmpl [ -t ] [ ASCII_file  [ output_file ] ]";
#
#  where  -t   indicate TABLE object only.  Otherwise a complete label 
#              template (SBN preferred form) is written.
#
#  16 Sep 1998, acr: Fixed a couple of bugs that omitted output of the final
#                    COLUMN and the END_OBJECT = TABLE statement.
#  07 Oct 1998, acr: Changed 'require' to use default system libraries
#
#===============================================================================
# Begin by getting the command-line options and opening the files:

getopts('t');

die "Usage: $USAGE\n" if (@ARGV > 2);
$outfile= (@ARGV > 1)? $ARGV[1] : "-";
$infile = (@ARGV > 0)? $ARGV[0] : "-";

open (DAT,$infile) || die "Unable to open $infile for reading";
open (TMP,">$outfile") || die "Unable to open $outfile for writing";

#______________________________________________________________________________
#
# FILE READING
#______________________________________________________________________________
#
#  We read through the input file once to compile the information needed.
#  We count character classes, total lines, line length and check for the
#  presence of carriage-control:
#

$linecount = 0;
$maxlength = 0;
$minlength = 1.e30;
$CR = 0;
undef @nonblank;      # Non-blank characters
undef @special;       # Non-printing, non-blank characters
undef @blank;         # Blank               [ ]
undef @alpha;         # Letters             [A-Za-z]
undef @digits;        # Digits              [0-9]
undef @number;        # Number characters   [0-9.+-eEdD]
undef @point;         # Decimal points      [.]
undef @expind;        # Exponent indicators [eEdD]
undef @sign;          # Sign characters     [-+]

while ($line=<DAT>)
  { $linecount++;
    $len = length($line);
    $minlength = ($len < $minlength)? $len : $minlength;
    $maxlength = ($len > $maxlength)? $len : $maxlength;

    @l = split(//,$line);

    # Check carriage-control:

    if ($l[$len-2] eq "\r")
      { $CR++;
        $len--;
      }

    # Loop through bytes and catalog the types of characters found:

    for ($i=0; $i<$len; $i++)
      { $c = $l[$i];           # Notational convenience
 
        if ($c eq " ")
          { $blank[$i]++; }

        else
          { $nonblank[$i]++;

            if ($c =~ /\s/)
              { $special[$i]++; }

            elsif ($c =~ /[a-zA-Z]/)
              { $alph[$i]++; 
                if ($c =~ /[eEdD]/)
                  { $expind[$i]++; }
              }

            elsif ($c =~ /[0-9]/)
              { $digit[$i]++; }

            elsif ($c eq ".")
              { $point[$i]++; }

            elsif ($c =~ /[-+]/)
              { $sign[$i]++; }

            if ($c =~ /[-0-9.+eEdD]/)
              { $number[$i]++; }
          }
      }
  }

close(DAT);

#______________________________________________________________________________
#
# ANALYSIS
#______________________________________________________________________________

# Now, let's see what we found.  First, we check for errors and inconsistencies:

if (!$CR)
  { printf STDOUT "$infile does not contain carriage-return characters.\n"; }

elsif ($CR != $linecount)
  { printf STDOUT "Inconsistent use of carriage-control: \n";
    printf STDOUT "     %5d records had CR/LF delimiters, \n",$CR;
    printf STDOUT "     %5d records had LF delimiters only.\n",$linecount;
  }

if ($maxlength != $minlength)
  { printf STDOUT "$infile does not contain fixed-length records: \n";
    printf STDOUT "     Minimum length: %5d bytes\n",$minlength;
    printf STDOUT "     Maximum length: %5d bytes\n",$maxlength;
  }

# Next, we want to determine the number of distinct columns in the file.  
# For this, we assumed columns are separated by a gutter of at least one blank
# byte.  So, we step through the 'blank' array previously generated and look
# for non-zero values:

$i=0; 
$colcount = -1;
undef @colstart;
undef @colstop;

while ($i<$maxlength)
  {
    # Begin by skipping blank columns:

    while (($i < $maxlength)  &&  ($nonblank[$i] == 0))
      { $i++; }

    if ($i < $maxlength)
      { $colcount++;
        $colstart[$colcount] = $i;
      }

    # Now look for next blank column:

    while ($i < $maxlength  &&  $nonblank[$i] != 0)
      { $i++; }

    $colstop[$colcount] = $i-1;

    # And repeat the loop:
  }

$colcount++;
# printf STDERR "Found $colcount columns in $infile.\n";


# Now we want to loop through the columns found, and determine the following
# attributes: START_BYTE, BYTES, DATA_TYPE, and FORMAT.

undef @start_byte;
undef @bytes;
undef @data_type;
undef @format;

for ($col=0; $col<$colcount; $col++)
  { $start_byte[$col] = $colstart[$col]+1;    # PDS START_BYTE is 1-based

    $bytes[$col] = $colstop[$col] - $colstart[$col] + 1;

    # To determine DATA_TYPE and FORMAT, we must examine the types of
    # characters encountered in each byte of the field.  We start with 
    # logical flags set to TRUE for each DATA_TYPE, then test each 
    # column for violation of the content constraints for each type:

    $integer = 1;
    $float   = 1;
    $scifloat= 1;
    undef $decimal;
    undef $expchar;

    # Everything that isn't numeric is a string.  Blanks are always allowed.

    for ($i=$colstart[$col]; $i<=$colstop[$col]; $i++)
      { if ($blank[$i]+$digit[$i]+$sign[$i] != $linecount)
          { $integer = 0; }

        if ($blank[$i]+$digit[$i]+$point[$i]+$sign[$i] != $linecount)
          { $float = 0; }

        if ($blank[$i]+$number[$i] != $linecount)
          { $scifloat = 0; }

        # Save location of special characters that always occur in the same 
        # column:

        if ($point[$i]+$blank[$i] == $linecount)
          { $decimal = $i; }

        if ($expind[$i]+$blank[$i] == $linecount)
          { $expchar = $i; }
      }

    # OK, set DATA_TYPE and FORMAT according to what we found:

    if ($integer)
      { $data_type[$col] = "ASCII_INTEGER";
        $format[$col] = "I$bytes[$col]";
      }

    elsif ($float)
      { $data_type[$col] = "ASCII_REAL";
        $width = $bytes[$col];
        $precision = ($decimal)? $bytes[$col]-($decimal-$colstart[$col]+1) : 0;
        $format[$col] = "\"F$width.$precision\"";
      }

    elsif ($scifloat)
      { $data_type[$col] = "ASCII_REAL";
        $width = $bytes[$col];
        if ($decimal && $expchar)
          { $precision = $bytes[$col] - ($decimal-$colstart[$col]+1) 
                                      - ($colstop[$col]-$expchar+1);
          }
        elsif ($decimal)
          { $precision = $bytes[$col] - ($decimal-$colstart[$col]+1); }
        else
          { $precision = 0; }
        $format[$col] = "\"E$width.$precision\"";
      }

    else
      { $data_type[$col] = "CHARACTER";
        $format[$col] = "A$bytes[$col]";
      }
  }

#______________________________________________________________________________
#
# OUTPUT
#______________________________________________________________________________
#
# If we're outputting a complete label template, write the general file info
# section:

if (!$opt_t)
  { $record_type = ($maxlength == $minlength)? "FIXED_LENGTH" : "STREAM";
    $INFILE = $infile;
    $INFILE =~ tr/a-z/A-Z/;
    printf TMP "PDS_VERSION_ID = PDS3\n";
    printf TMP "\n";
    printf TMP "RECORD_TYPE  = $record_type\n";
    printf TMP "RECORD_BYTES = $maxlength\n";
    printf TMP "FILE_RECORDS = $linecount\n";
    printf TMP "\n";
    printf TMP "^TABLE       = \"$INFILE\"\n";
    printf TMP "\n";
    printf TMP "DATA_SET_ID  = \" \"\n";
    printf TMP "PRODUCT_NAME = \" \"\n";;
    printf TMP "PRODUCT_ID   = \"$INFILE\"\n";
    printf TMP "\n";
    printf TMP "INSTRUMENT_HOST_NAME  = \n";
    printf TMP "INSTRUMENT_NAME       = \n";
    printf TMP "TARGET_NAME           = \n";
    printf TMP "START_TIME            = \n";
    printf TMP "STOP_TIME             = \n";
    printf TMP "PRODUCT_CREATION_TIME = \n";
    printf TMP "\n";
  }

# Now we're ready to print the TABLE object template.  It's tedious, but here
# we go:

printf TMP "OBJECT     = TABLE\n";
printf TMP "  INTERCHANGE_FORMAT = ASCII\n";
printf TMP "  NAME               = \n";
printf TMP "  ROWS               = $linecount\n";
printf TMP "  ROW_BYTES          = $maxlength\n";
printf TMP "  COLUMNS            = $colcount\n";
printf TMP "  DESCRIPTION        = \" \"\n";

# Looping through the columns we found:

for ($i=0; $i<$colcount; $i++)
  { printf TMP "\n";
    printf TMP "  OBJECT     = COLUMN\n";
    printf TMP "    COLUMN_NUMBER = %3d\n",$i+1;
    printf TMP "    NAME          = \n";
    printf TMP "    START_BYTE    = %3d\n",$start_byte[$i];
    printf TMP "    BYTES         = %3d\n",$bytes[$i];
    printf TMP "    DATA_TYPE     = $data_type[$i]\n";
    printf TMP "    FORMAT        = $format[$i]\n";
    printf TMP "    UNIT          = \n";
    printf TMP "    DESCRIPTION   = \" \"\n";
    printf TMP "  END_OBJECT = COLUMN\n";
  }

# Add the END statement for the TABLE object:

printf TMP "\nEND_OBJECT = TABLE\n";

# Add the END statement if this is a label template:

printf TMP "\nEND\n" if (!$opt_t);

# And we're done:

close TMP;

