#!/usr/bin/perl

use Getopt::Long;
use File::Find;

# Routine to check for required volume structure and compliant file names in
# a PDS volume delivery.
#
# Format: % vc_files [-u] [-o output] [--sbn-ignore ] \
#                    [--exclude pattern] root_dir
#
#  where  -u  indicates all file and directory names will be in upper case
#             (lowercase is the default)
#         -o output
#             directs the output to the named file (stdout is the default)
#         --sbn-ignore
#             indicates "root_dir" is an SBN archive root, and that archive
#             support directories and files should be ignored.
#         --exclude pattern or -x pattern
#             ignore all files and directories matching "pattern" at any
#             point in the fully qualified file name (may be repeated)
#
# Note that the file name testing code is copied from the 'chkisonames'
# routine using the tighter PDS constraints.
#
# 25 February 2008, A.C.Raugh
#
# Modification History:
# =====================
#
# 03 Mar 2008, ACR: Added --sbn-ignore and --exclude options
# 
#
#============================================================================

# Some constants:

$TRUE  = 1;
$FALSE = 0;

# Check for clean run:

$error_count = 0;

# Parse the input arguments, do sanity-checking and open the output stream:

GetOptions("u"   => \$opt_u,
           "o=s" => \$opt_o,
           "sbn-ignore"   => \$opt_sbnignore,
           "x|exclude=s"  => \@exclude);

$output = ($opt_o)? $opt_o : "-";
$UPPER = $opt_u;

if (@ARGV != 1  ||  !(-d $ARGV[0]))
  { printf "Usage: vc_files [-u] [-o output] [--sbn-ignore] ";
    printf "[--exclude pattern] root_dir";
    die "\n";
  }
$ROOT = $ARGV[0];
$ROOT =~ s/\/$//;

# If we're ignoring SBN archive support files/directories, we'll add them
# to whatever we might already have in the @exclude array:

if ($opt_sbnignore)
  { $i = @exclude;
    $exclude[$i++] = "dataset\.html";
    $exclude[$i++] = "DOWNLOAD";
    $exclude[$i++] = "CODE";
    $exclude[$i++] = "WORK";
    $exclude[$i++] = "NOTES";
    $exclude[$i++] = "VOLUME";
  }

open(OUT,">$output") ||
  die "Could not open output file '$output' ($!).";

# Print a header with the root directory name:

printf OUT "VC_FILES File and Directory Name Check Report\n";
printf OUT "=============================================\n\n";
printf OUT "For the volume with root \"$ROOT\"\n\n";

#
# We're ready to go.  We'll use the File::Find package to do a pre-order
# traversal of the root directory passed in to check the file and directory
# identifiers for PDS compliance:
#

find({wanted   => \&check_filename, 
      no_chdir => $FALSE},
      $ROOT);

# Now we need to check to make sure all the required files are present.
# We'll do this by directory querying for the presence of these files.
# For neatness, we'll do this in a subroutine.

check_reqfiles();

# Report positive result if we haven't printed any errors:

printf OUT "  - OK\n" if ($error_count == 0);
printf OUT "\n";

# Close the output file and we're done:

close(OUT);

#----------------------------------------------------------------------------

sub check_filename

  # Routine used by File::Find to check PDS-compliance in file names

  { my ($name,$ext);          # Separated file name and extension
    my ($nlen,$xlen);         # Name and extension length
    my ($fullname);           # file name without path
    my ($fqname);             # Fully-qualified name (name + path from root)
    my (@parts);              # parsing array
    my ($dir);                # local copy of directory name, sans root
    my ($i, $j);

    # If we've got any exlusion patterns, we ignore matching files:

    foreach $pattern (@exclude)
      { return if ($File::Find::name =~ /$pattern/); }


    # See if we've got a directory.  Note that we chdir into the
    # current directory (no_chdir => $FALSE in the call to "find").

    if (-d)
      { # If it's the root directory, we ignore it and return:

        return if ($File::Find::name =~ /^$ROOT$/);

        # Otherwise we parse it and check it:

        $dir = $File::Find::name;
        $dir =~ s/$ROOT\///;      # Use this for display
        $dir =~ s/\/$//;          # Remove final '/', if any

        @parts = split(/\//,$File::Find::name);
        $i = @parts - 1;
        $name = $parts[$i];

        # Check length and character set:

        if (length($name) > 30)
          { printf OUT "$name - directory name too long.\n"; 
            $error_count++;
          }


        if ($UPPER)
          { if ($name =~ /[a-z]/)
              { printf OUT "$dir/ - lower case characters in ";
                printf OUT "directory name\n";
                $error_count++;
              }
            elsif ($name !~ /^[A-Z0-9_]+$/)
              { printf OUT "$dir/ - invalid characters in ";
                printf OUT "directory name.\n";
                $error_count++;
              }
          }
        else
          { if ($name =~ /[A-Z]/)
              { printf OUT "$dir/ - upper case characters in ";
                printf OUT "directory name\n";
                $error_count++;
              }
            elsif ($name !~ /^[a-z0-9_]+$/)  
              { printf OUT "$dir/ - invalid characters in ";
                printf OUT "directory name.\n";
                $error_count++;
              }
          }

        # That's all we need to do for directory names:

        return;
      }

    # PDS file names, and filenames + path info, must meet a strict
    # subset of the ISO 9660 requirements...

    # ISO 9660: fully qualified name may not exceed 255 characters:

    $fqname = $File::Find::name;
    $fqname =~ s/^$ROOT\///;     # Remove root directory from path
    if (length($fqname) > 255)
      { printf OUT "$fqname - total path + filename length ";
        printf OUT "exceeds 255 characters (ISO 9660 violation).\n";
        $error_count++;
      }

    # Check for invalid characters and mixed/incorrect case:

    $fullname = $_;
    if ($UPPER)
      { if ($fullname =~ /[a-z]/)
          { printf OUT "$fqname - lower case characters in ";
            printf OUT "file identifier\n";
            $error_count++;
          }
        if ($fullname !~ /^[A-Z0-9_.]+$/  &&  $fullname !~ /[a-z]/)
          { printf OUT "$fqname - invalid characters in ";
            printf OUT "file identifier\n";
            $error_count++;
          }
      }
    else
      { if ($fullname =~ /[A-Z]/)
          { printf OUT "$fqname - upper case characters in ";
            printf OUT "file identifier\n";
            $error_count++;
          }
        if ($fullname !~ /^[a-z0-9_.]+$/  &&  $fullname !~ /[A-Z]/)
          { printf OUT "$fqname - invalid characters in ";
            printf OUT "file identifier\n";
            $error_count++;
          }
      }

    # Now we'll split the identifier into name and extension and check
    # each separately:

    @parts = split(/\./,$fullname);

    if (@parts > 2)
      { printf OUT "$fqname - invalid file identifier, ";
        printf OUT "too many '.'\n";
        $error_count++;
      }

    elsif (@parts == 1)
      { # PDS files should generally have both name and extension, except 
        # in a software directory.  We'll assume anything under a directory
        # named "software", "code" or "sw" is OK without an extension:

        if ($fullname !~ /\./)
          { if ($UPPER  &&  
                 ($fqname !~ /SOFTWARE\//  &&
                  $fqname !~ /\/SW\//      &&
                  $fqname !~ /\/CODE\// )
               )
              { printf OUT "$fqname - file identifier has no extension\n"; 
                $error_count++;
              }
            elsif (!$UPPER  &&  
                   ($fqname !~ /software\//  &&
                    $fqname !~ /\/sw\//      &&
                    $fqname !~ /\/code\//)
                  )
              { printf OUT "$fqname - file identifier has no extension\n";  
                $error_count++;
              }
          }
        else
          { printf OUT "$fqname - file identifier has null extension\n";  
            $error_count++;
          }
      }

    elsif (@parts == 2)
      { $name = $parts[0];
        $ext  = $parts[1];

        $nlen = length($name);
        $xlen = length($ext);

        # Check various length constraints:

        if (($nlen + $xlen) > 30)
          { printf OUT "$fqname - total identifier length > 31\n";   
            $error_count++;
          }

        if (($nlen + $xlen) == 0)
          { printf OUT "$fqname - invalid null file identifier\n";   
            $error_count++;
          }

        if ($nlen > 27)
          { printf OUT "$fqname - file name too long (> 27 characters)\n";   
            $error_count++;
          }

        if ($xlen > 3)
          { printf OUT "$fqname - file extension too long ";
            printf OUT "(> 3 characters)\n";
            $error_count++;
          }

        if ($nlen == 0)
          { printf OUT "$fqname - null file name (hidden files ";
            printf OUT "are not allowed)\n";
            $error_count++;
          }
      }

    # And that wraps it up for identifier checking:

    return;

  }

#----------------------------------------------------------------------------

sub check_reqfiles

  # Routine to check for the existence of required files and directories
  # under the input root directory.

  { my (@req_list);        # Required directory array
    my ($file,$FILE);

    @req_list = ("aareadme.txt"         ,
                 "aareadme.htm"         ,
                 "aareadme.lbl"         ,
                 "voldesc.cat"          ,
                 "catalog"              ,
                 "catalog/catinfo.txt"  ,
                 "index"                ,
                 "index/indxinfo.txt"   ,
                 "index/index.tab"      ,
                 "index/index.lbl"      ,
                 "data"                 ,
                 "calib"                ,
                 "calib/calinfo.txt"    ,
                 "document"             ,
                 "document/docinfo.txt" ,
                 "extras"               ,
                 "extras/extrinfo.txt"  ,
                 "gazetter"             ,
                 "gazetter/gazinfo.txt" ,
                 "gazetter/gazetter.txt",
                 "gazetter/gazetter.tab",
                 "gazetter/gazetter.lbl",
                 "geometry"             ,
                 "geometry/geominfo.txt",
                 "label"                ,
                 "label/labinfo.txt"    ,
                 "software"             ,
                 "software/softinfo.txt");

    foreach $file (@req_list)
      { if ($UPPER)
          { $FILE = $file;
            $FILE =~ tr/a-z/A-Z/;
            $req{$file} = "$ROOT/$FILE";
          }
        else
          { $req{$file} = "$ROOT/$file"; }
      }

    # These file and directories are always required:

    if (!(-f $req{"aareadme.txt"}))
      { if (-f $req{"aareadme.htm"}  &&  -f $req{"aareadme.lbl"})
          { printf OUT "WARNING: No '%s', but ", $req{"aareadme.txt"};
            printf OUT "found '%s' and label\n", $req{"aareadme.htm"};
          }
        else
          { printf OUT "Missing required '%s' file\n",$req{"aareadme.txt"}; 
            $error_count++;
          }
      }

    if (!(-f $req{"voldesc.cat"}))
      { printf OUT "Missing required '%s' file\n", $req{"voldesc.cat"}; 
            $error_count++;
      }

    if (-d $req{"catalog"})
      { if (!(-f $req{"catalog/catinfo.txt"}))
          { printf OUT "Missing required '%s' file\n",
                       $req{"catalog/catinfo"}; 
            $error_count++;
          }
      }
    else
      { printf OUT "Missing required '%s' directory\n", $req{"catalog"}; }

    if (-d $req{"index"})
      { if (!(-f $req{"index/indxinfo.txt"}))
          { printf OUT "Missing required '%s' file\n",
                        $req{"index/indxinfo.txt"}; 
            $error_count++;
          }
        if (!(-f $req{"index/index.tab"}))
          { printf OUT "Missing required '%s' file\n",
                        $req{"index/indexinfo.txt"}; 
            $error_count++;
          }
        if (!(-f $req{"index/index.lbl"}))
          { printf OUT "Missing required '%s' file\n",
                        $req{"index/index.lbl"}; 
            $error_count++;
          }
      }
    else
      { printf OUT "Missing required '%s' directory\n", $req{"index"}; 
        $error_count++;
      }

    # These are optional directories, but if they exist their required
    # files must be present:

    if (-d $req{"calib"})
      { if (!(-f $req{"calib/calinfo.txt"}))
          { printf OUT "Missing required '%s' file\n",
                        $req{"calib/calinfo.txt"}; 
            $error_count++;
          }
      }

    if (-d $req{"document"})
      { if (!(-f $req{"document/docinfo.txt"}))
          { printf OUT "Missing required '%s' file\n",
                        $req{"document/docinfo.txt"};  
            $error_count++;
          }
      }

    if (-d $req{"extras"})
      { if (!(-f $req{"extras/extrinfo.txt"}))
          { printf OUT "Missing required '%s' file\n",
                        $req{"extras/extrinfo.txt"};  
            $error_count++;
          }
      }

    if (-d $req{"gazetter"})
      { if (!(-f $req{"gazetter/gazinfo.txt"}))
          { printf OUT "Missing required '%s' file\n",
                        $req{"gazetter/gazinfo.txt"}; } 
            $error_count++;
          
        if (!(-f $req{"gazetter/gazetter.txt"}))
          { printf OUT "Missing required '%s' file\n",
                        $req{"gazetter/gazetter.txt"};  
            $error_count++;
          }
        if (!(-f $req{"gazetter/gazetter.tab"}))
          { printf OUT "Missing required '%s' file\n",
                        $req{"gazetter/gazetter.tab"};  
            $error_count++;
          }
        if (!(-f $req{"gazetter/gazetter.lbl"}))
          { printf OUT "Missing required '%s' file\n",
                        $req{"gazetter/gazetter.lbl"};  
            $error_count++;
          }
      }

    if (-d $req{"geometry"})
      { if (!(-f $req{"geometry/geominfo.txt"}))
          { printf OUT "Missing required '%s' file\n",
                        $req{"geometry/geominfo.txt"};  
            $error_count++;
          }
      }

    if (-d $req{"label"})
      { if (!(-f $req{"label/labinfo.txt"}))
          { printf OUT "Missing required '%s' file\n",
                        $req{"label/labinfo.txt"};  
            $error_count++;
          }
      }

    if (-d $req{"software"})
      { if (!(-f $req{"software/softinfo.txt"}))
          { printf OUT "Missing required '%s' file\n",
                        $req{"software/softinfo.txt"};  
            $error_count++;
          }
      }

    # Finally, check for a data/ directory.  This isn't required or
    # reserved, but nearly every volume uses it for SBN:

    if (!(-d $req{"data"}))
      { printf OUT "WARNING: No '%s' directory found\n",$req{"data"}; }

    # Done.

    return;
  }
