#!/usr/bin/perl

use lib "/n/sbnops01/lcltools/lib/perl5";

use Getopt::Long;
use File::Find;
require "nextPDSODL.pl";

#============================================================================
#
# VC_INDEX
#
# Routine to check the index file of a PDS volume.  It checks only the 
# index.tab file and its label, NOT the cumindex.tab file, if any.  First
# it locates the required index file and checks the label to make sure the
# required keywords/colums are present.  Then it cross-checks the files
# listed in the index table against the actual files in the volume.
#
# Format: % vc_index [-o output] [--exclude pattern] [--sbn-ignore] root_dir
#
#  where -o output     directs output to the named file
#        --exclude     ignores any files or directories matching the 
#                      given pattern (may be repeated)
#        --sbn-ignore  ignores SBN on-line archive support files
#        root_dir      is the volume root
#
# 14 March 2008, A.C.Raugh
#
# 30 Apr 2008, ACR: added "calib/" to the not-data directory list; tweaked
#                   PRODUCT_ID missing message; fixed bug checking column
#                   names for PRODUCT_CREATION_TIME
# 05 May 2008, ACR: added code to skip over one optional HEADER object 
#                   if found prior to the INDEX_TABLE object
# 01 Jun 2009, ACR: fixed inadvertent typo that caused the program to 
#                   fail when path and filename were in separate columns.
#
#============================================================================
# Constants:

$TRUE  = 1;
$FALSE = 0;
$MAXLEN = 10000;   # Maximum input ODL line length

# We'll use hashes for each file-related keyword to track existence, start
# byte and length.  We'll initialize these here so they will have global
# scope:

$filespec{"found"} = $FALSE;
$path{"found"}     = $FALSE;
$filename{"found"} = $FALSE;

# We just need to track existence for these:

$pid_found         = $FALSE;
$volid_found       = $FALSE;
$dsid_found        = $FALSE;
$pctime_found      = $FALSE;

# Here's a hash of reserved directory names in upper and lower case, so 
# we can identify data directories (those that AREN'T on this list):

%resdir = ( "BROWSE"   => 1, "browse"   => 1,
            "CALIB"    => 1, "calib"    => 1,
            "CATALOG"  => 1, "catalog"  => 1,
            "DOCUMENT" => 1, "document" => 1,
            "EXTRAS"   => 1, "extras"   => 1,
            "GEOMETRY" => 1, "geometry" => 1,
            "INDEX"    => 1, "index"    => 1,
            "LABEL"    => 1, "label"    => 1,
            "SOFTWARE" => 1, "software" => 1);

# Parse the input arguments and open the output file:

GetOptions("o=s"        => \$opt_o,
           "sbn-ignore" => \$opt_sbnignore,
           "x|exclude=s"  => \@exclude);

$output = ($opt_o)? $opt_o : "-";

if (@ARGV != 1  ||  !(-d $ARGV[0]))
  { printf "Usage: vc_index [-o output] [--exclude pattern] ";
    printf "[--sbn-ignore] root_dir\n"; 
    die;
  }

$ROOT = $ARGV[0];
$ROOT =~ s/\/$//;

# If we're ignoring SBN archive support files/directories, we'll add them
# to whatever we already have in @exclude:

if ($opt_sbnignore)
  { $i = @exclude;
    $exclude[$i++] = "dataset\.html";
    $exclude[$i++] = "\/DOWNLOAD\/";
    $exclude[$i++] = "\/CODE\/";
    $exclude[$i++] = "\/WORK\/";
    $exclude[$i++] = "\/NOTES\/";
    $exclude[$i++] = "\/VOLUME\/";
  }

open (OUT,">$output") ||
  die "Could not open output file '$output' ($!).";

# Print a header with the root directory name:

printf OUT "VC_INDEX Required Index File Check Report\n";
printf OUT "=========================================\n\n";
printf OUT "For the volume with root \"$ROOT\"\n\n";

# Let's find our index directory and the index.tab and index.lbl files:

if (-d "$ROOT/index")
  { $INDEX = "$ROOT/index"; }
elsif (-d "$ROOT/INDEX")
  { $INDEX = "$ROOT/INDEX"; }
else
  { die "Could not find an index directory under $ROOT."; }

if (-f "$INDEX/index.tab"  &&  -f "$INDEX/index.lbl")
  { $xtab = "$INDEX/index.tab";
    $xlbl = "$INDEX/index.lbl";
  }
elsif (-f "$INDEX/INDEX.TAB"  &&  -f "$INDEX/INDEX.LBL")
  { $xtab = "$INDEX/INDEX.TAB";
    $xlbl = "$INDEX/INDEX.LBL";
  }
else
  { die "Could not find volume index table and label files in $INDEX"; }

# First Stage -  we'll call a subroutine to parse the index label, flag
# errors in it and collect the information we'll need to read the index
# table file:

printf OUT "Checking $xlbl file...\n";

$error_count = 0;
$offset      = 0;     # Offset to beginning of index
$rows = parse_xlbl($xlbl);
if ($error_count == 0)
  { printf OUT "  - OK\n"; }
printf OUT "\n";

## Check the parse results:
##
#if ($filespec{"found"})
#  { printf "Found FILE_SPECIFICATION_NAME ";
#    printf "(offset = %d, ",$filespec{"offset"};
#    printf "bytes = %d)\n",$filespec{"bytes"};
#  }
#if ($path{"found"}  &&  $filename{"found"})
#  { printf "Found PATH_NAME (offset = %d, ",$path{"offset"};
#    printf "bytes = %d) and ", $path{"bytes"};
#    printf "FILE_NAME (offset = %d, ",$filename{"offset"};
#    printf "bytes = %d)\n",$filename{"bytes"};
#  }
#printf "Table rows = $rows\n";

# Second Stage - read the table file, collect the label file names and 
# check that they all exist.  We'll also saves their names in a hash for 
# reference in stage three:

printf OUT "Checking files listed in $xtab exist...\n";

$error_count = 0;
$missing_files = read_xtab($xtab);
if ($missing_files == 0)
  { printf OUT "  - OK\n"; }
elsif ($missing_files == 1)
  { printf OUT "\n  $missing_files missing file total\n"; }
else
  { printf OUT "\n  $missing_files missing files total\n"; }
printf OUT "\n";

# Third Stage - find the label files in the data directories and
# make sure every label is mentioned in the index.tab file.  We'll
# use the File::Find package to do this:

printf OUT "Checking data directories for unindexed labels...\n"; 

# When we're done, error_count will have the number of unmatched labels
# found.

$error_count = 0;
find({wanted  => \&check_data,
      nochdir => $FALSE},        # We'll chdir to each subdirectory
     $ROOT);
if ($error_count == 0)
  { printf OUT "  - OK\n\n"; }
elsif ($error_count == 1)
  { printf OUT "\n  $error_count unindexed label total\n\n"; }
else
  { printf OUT "\n  $error_count unindexed labels total\n\n"; }

# And we're done.

#----------------------------------------------------------------------------

sub parse_xlbl

  # Routine to read through the index.lbl file passed in, check for the
  # required keywords and columns, and save the data needed to read the
  # file names from the index.tab file.  We return the number of rows
  # the table thinks it has.

  { local ($lbl) = $_[0];     # Input index.lbl file (with path)
    my ($line);               # ODL line buffer
    my ($key,$val);           # for parsing ODL
    my ($rec_byts);           # For calculating offset
    my ($table_type);         # main index object type
    my (%col);                # local hash to simplify column checking
    my ($rows);               # ROWS (returned to the calling routine)

    # Open the label file:

    open(LBL,$lbl)  ||  die "Could not open $lbl for reading ($!)";

    # Now we'll loop through it until we get to the first object, keeping
    # track of any required parameters that might appear as keywords 
    # rather than columns:

    $line = nextPDSODL(LBL,$MAXLEN);
    while ($line !~ /^\s*OBJECT\s*=\s*/)
      { ($key,$val) = splitODLline($line);
        $val =~ s/"//g;

        if ($key eq "VOLUME_ID")
          { $volid_found = $TRUE; }
        elsif ($key eq "DATA_SET_ID")
          { $dsid_found = $TRUE; }
        elsif ($key eq "PRODUCT_CREATION_TIME")
          { $pctime_found = $TRUE; }
        elsif ($key eq "RECORD_BYTES")
          { $rec_bytes = $val; }
        elsif ($key eq "^INDEX_TABLE")
          { # Save the offset, if any, just in case:

            if ($val =~ /\(/)
              { $val =~ /^.*,\s*([0-9]+)/;
                $offset = $1 - 1;
                $offset = ($val =~ /<BY/)? $offset : $offset * $rec_bytes;
              }
            else
              { $offset = 0; }
          }

        # Next line...

        $line = nextPDSODL(LBL,$MAXLEN);
      }

    ($key,$val) = splitODLline($line);

    # We found the first object definition.  It might be a header, which we'll
    # skip:

    if ($val =~ /HEADER$/)
      { $line = nextPDSODL(LBL,$MAXLEN);
        while ($line !~ /^\s*OBJECT\s*=\s*/)
          { $line = nextPDSODL(LBL,$MAXLEN); }
       ($key,$val) = splitODLline($line);
      }
    
    # OK, this object better be the INDEX_TABLE.  Because this is so often 
    # missed, we'll check for that:

    if ($val ne "INDEX_TABLE")
      { printf OUT "  $lbl ERROR - Does not use INDEX_TABLE object.\n"; 
        $error_count++;
      }

    if ($val !~ /TABLE$/)
      { printf OUT "   $lbl FATAL ERROR - Invalid OBJECT definition found.\n";
        close(OUT);
        die "  $lbl FATAL ERROR - Invalid OBJECT definition found.\n"; 
      }

    # Save the object type for comparison later:

    $table_type = $val;

    # Now we loop through the columns of the index table, looking for NAMEs
    # that match the key fields we're interested in:

    $not_done = $TRUE;
    while ($not_done)
      { if (!($line = nextPDSODL(LBL,$MAXLEN)))
          { printf OUT "  $lbl FATAL ERROR - Unexpected end of file.\n";
            close(OUT);
            die "  $lbl FATAL ERROR - Unexpected end of file.\n"; 
          }
        
        # Check for end of label:

        if ($line =~ /^\s*END\s/)
          { printf OUT "  $lbl ERROR - No END_OBJECT statement for ",
            printf OUT "$table_type\n";
            $error_count++;
            $not_done = $FALSE;
            next;
          }

        # Check for end of table:

        if ($line =~ /^\s*END_OBJECT/)
          { ($key,$val) = splitODLline($line);
            if ($val  &&  ($val ne $table_type))
              { printf OUT "  $lbl ERROR - Mismatched table OBJECT types\n";
                $error_count++;
              }
            $not_done = $FALSE;
            next;
          }

        # Check for ROWS keyword:

        if ($line =~ /^\s*ROWS\s*=/)
          { ($key,$val) = splitODLline($line);
            $val =~ s/"//g;
            $rows = $val;
          }

        # Check for column definition start:

        next if ($line !~ /^\s*OBJECT\s*=\s*COLUMN/);

        # Got a column - start a subloop

        undef %col;   # Reset
        $line = nextPDSODL(LBL,$MAXLEN);
        while ($line !~ /^\s*END_OBJECT/)
          { ($key,$val) = splitODLline($line);
            $key =~ s/"//g;
            $val =~ s/"//g;
            $col{$key} = $val;
            $line = nextPDSODL(LBL,$MAXLEN);
            if (!$line)
              { printf OUT "  $lbl FATAL ERROR - Could not find END_OBJECT ";
                printf OUT "for column\n";
                die;
              }
          }

        # Now check to see if we found one of the fields we needed:

        if (!$col{"NAME"}  ||  !$col{"START_BYTE"}  ||  !$col{"BYTES"})
          { printf OUT "  $lbl FATAL ERROR - Column without ";
            printf OUT "required fields\n";
            die;
          }

        if ($col{"NAME"} eq "PRODUCT_ID")
          { $pid_found = $TRUE; }
        elsif ($col{"NAME"} eq "VOLUME_ID")
          { $volid_found = $TRUE; }
        elsif ($col{"NAME"} eq "DATA_SET_ID")
          { $dsid_found = $TRUE; }
        elsif ($col{"NAME"} eq "PRODUCT_CREATION_TIME")
          { $pctime_found = $TRUE; }
        elsif ($col{"NAME"} eq "FILE_SPECIFICATION_NAME")
          { $filespec{"found"}  = $TRUE;
            $filespec{"offset"} = $col{"START_BYTE"} - 1;
            $filespec{"bytes"}   = $col{"BYTES"};
          }
        elsif ($col{"NAME"} eq "PATH_NAME")
          { $path{"found"}  = $TRUE;
            $path{"offset"} = $col{"START_BYTE"} - 1;
            $path{"bytes"}   = $col{"BYTES"};
          }
        elsif ($col{"NAME"} eq "FILE_NAME")
          { $filename{"found"}  = $TRUE;
            $filename{"offset"} = $col{"START_BYTE"} - 1;
            $filename{"bytes"}  = $col{"BYTES"};
          }

        # Now look for the next COLUMN (or end of TABLE)...
      }

    # Done with the label:

    close(LBL);

    # Check that all required fields were found:

    if (!$filespec{"found"}  &&  !($path{"found"} && $filename{"found"}))
      { printf OUT "  $lbl FATAL ERROR - Incomplete file locations\n";
        die;
      }

    if (!$pid_found)
      { printf OUT "  $lbl WARNING - missing PRODUCT_ID.  ";
        printf OUT "PRODUCT_ID is required unless it is always equal ";
        printf OUT "FILE_NAME or FILE_SPECIFICATION_NAME.\n";
      }

    if (!$volid_found)
      { printf OUT "  $lbl ERROR - missing VOLUME_ID\n";
        $error_count++;
      }

    if (!$dsid_found)
      { printf OUT "  $lbl ERROR - missing DATA_SET_ID\n";
        $error_count++;
      }

    if (!$pctime_found)
      { printf OUT "  $lbl ERROR - missing PRODUCT_CREATION_TIME\n";
        $error_count++;
      }

    # Done:

    return $rows;
  }

#----------------------------------------------------------------------------

sub read_xtab

  # Routine to read the index.tab file and check for the existence of 
  # each file listed in it.  

  { local ($tab) = $_[0];      # Name of the table file
    my ($line);                # input buffer
    my ($fullname);            # Full file spec
    my ($tmpname);             # holding place for case variations
    my ($name,$path);          # parts of the file spec
    my ($fs_offset,$fs_bytes); # Time-savers for resolving hash references
    my ($p_offset,$p_bytes);
    my ($n_offset,$n_bytes);
    my ($full_spec);           # Flag for file spec format
    my ($linecount) = 0;       # Place in file
    my ($bad_fs_quotes) = $FALSE;   # Specific error flags
    my ($bad_n_quotes)  = $FALSE;
    my ($bad_p_quotes)  = $FALSE;
    my ($missing_files) = 0;

    our (%xfile);              # Hash of file names found in this index
 
    # First, open the file for reading and position the read pointer at
    # the previously calculated offset:

    open (TAB,$tab)  ||  die "Could not open $tab for reading ($!)\n";
    seek (TAB,$offset,SEEK_SET);

    # Set up some flags for faster access:

    if ($filespec{"found"})  # File location is in a single field
      { $full_spec = $TRUE;
        $fs_offset = $filespec{"offset"};
        $fs_bytes  = $filespec{"bytes"};
      }
    else
      { $full_spec = $FALSE;
        $p_offset  = $path{"offset"};
        $p_bytes   = $path{"bytes"};
        $n_offset  = $filename{"offset"};
        $n_bytes   = $filename{"bytes"};
      }

    # Now check each line:

    while ($line=<TAB>)
      { $linecount++;

        # Get the file name we want to check:

        if ($full_spec)     # Full in a single field
          { $fullname = substr($line,$fs_offset,$fs_bytes); 
            if (!$bad_fs_quotes  &&  $fullname =~ /"/)
              { printf OUT "  ERROR - quotes inside the ";
                printf OUT "FILE_SPECIFICATION_NAME field\n";
                $bad_fs_quotes = $TRUE;
                $fullname =~ s/"//g;
              }
            $fullname =~ s/^\s+//;
            $fullname =~ s/\s+$//;
          }
        else
          { $path = substr($line,$p_offset,$p_bytes);
            if (!$bad_p_quotes  &&  $path =~ /"/)
              { printf OUT "  ERROR - quotes inside the PATH_NAME field\n";
                $bad_p_quotes = $TRUE;
                $path =~ s/"//g;
              }
            $path =~ s/^\s+//;
            $path =~ s/\s+$//;
            $path =~ s/\/$//;

            $name = substr($line,$n_offset,$n_bytes);
            if (!$bad_n_quotes  &&  $name =~ /"/)
              { printf OUT "  ERROR - quotes inside the FILE_NAME field\n";
                $bad_n_quotes = $TRUE;
                $name =~ s/"//g;
              }
            $name =~ s/^\s+//;
            $name =~ s/\s+$//;

            $fullname = "$path/$name";
          }

        # Save the full file name and path for later reference (forced
        # case is for faster checking later):

        $tmpname = $fullname;
        $tmpname =~ tr/A-Z/a-z/;
        $xfile{"$ROOT/$tmpname"} = $TRUE;

        # We'll check up to three times to find the file name: in the
        # case exactly as listed in the index.tab file; forced to lower-
        # case; and forced to uppercase. As soon as we determine the file
        # exists, we're done. 

        next if (-f "$ROOT/$fullname");
        
        # Try lowercase:

        next if (-f "$ROOT/$tmpname");

        # Finally, try uppercase:

        $tmpname =~ tr/a-z/A-Z/;
        next if (-f "$ROOT/$tmpname");

        # If we're here, we failed to find the file:

        printf OUT "  $fullname not found.\n";
        $missing_files++;

        # Next line...
      }

    # We're done.  We'll set the error flag before exitting:

    $error_count++ if ($bad_fs_quotes || $bad_n_quotes || $bad_p_quotes);
    $error_count++ if ($missing_files);

    return $missing_files;
  }

#----------------------------------------------------------------------------

sub check_data

  # Routine called by File::Find::find to search for PDS label files that
  # should be in the index table, but aren't.

  { my ($tmpname);
    my ($file,$dir);
    my ($line);
    my ($has_label,$is_indexed);

    # If the file we're currently looking at is a directory, do nothing:

    return if (-d $_);

    # If the file we're looking at is in the volume root, do nothing:

    return if ($File::Find::dir eq $ROOT);

    # If the file is in a not-data directory, do nothing:

    foreach $dir (keys %resdir)
      { return if ($File::Find::name =~ /\/$dir\//); }

    # If the file matches anything in the exclude list, do nothing:

    foreach $file (@exclude)
      { return if ($File::Find::name =~ /$file/); }

    # Open the file and see if it contains a PDS label, which we'll assume
    # is true if we find the "PDS_VERSION_ID" keyword in the first few
    # hundred bytes:

    open (LBL,$_)  ||  
      die "Could not open $File::Find::name to check for label ($!)";

    read(LBL,$line,300);
    if ($line =~ /PDS_VERSION_ID\s*=/)
      { $has_label = $TRUE; }
    close(LBL);

    # Now see if the file name is in the index.  We force the case to 
    # lower for this check:

    $tmpname = $File::Find::name;
    $tmpname =~ s/$ROOT\///;
    $tmpname =~ tr/A-Z/a-z/;
    $tmpname = "$ROOT/$tmpname";

    $is_indexed = $xfile{$tmpname};

    # And check for the mismatches in existence and indexing:

    if ($has_label  &&  !$is_indexed)
      { printf OUT "  $File::Find::name has a PDS label but ";
        printf OUT "is not listed in the index table.\n";
        $error_count++;
      }

    if ($is_indexed  &&  !$has_label)
      { printf OUT "  $File::Find::name appears in the index table but ";
        printf OUT "does not appear to contain a PDS label.\n";
        $error_count++;
      }

    # Other combinations are OK, so we move on:

  }
