#!/usr/bin/perl

use lib "/n/sbnops01/lcltools/lib/perl5";

use Getopt::Long;
use File::Find;
require "nextPDSODL.pl";

# Routine to ensure that every file on a PDS volume has a label, either 
# attached or detached.  The nextPDSODL.pl file is a rather old library
# from SBN that will read an ODL file one logical line at a time.
#
# Format: % vc_labels [ -o output] [--sbn-ignore] [--exlude pattern] root_dir
#
# 26 Feb 2008, A.C.Raugh
#
# 11 Mar 2008, ACR: Bug fix for pointers with braces; added check for 
#                   .asc file extension in document directory text files.
# 05 May 2008, ACR: Bug fix for bad exclude array name in ignoring files
#
#============================================================================
# Constants:

$TRUE  = 1;
$FALSE = 0;
$MAXLEN = 10000;   # Maximum input ODL line length

# Check for clean run:

$error_count = 0;

# Parse the input arguments and open the output file:

GetOptions("o=s"         => \$opt_o,
           "sbn-ignore"  => \$opt_sbnignore,
           "x|exclude=s" => \@exclude);  

$output = ($opt_o)? $opt_o : "-";

if (@ARGV != 1  ||  !(-d $ARGV[0]))
  { printf "Usage: vc_labels [-o output] [--sbn-ignore]";
    printf " [--exclude pattern] root_dir\n";
    die "\n";
  }

$ROOT = $ARGV[0];
$ROOT =~ s/\/$//;

# If we're ignoring SBN archive support files/directories, we'll add them
# to whatever we already have in @exclude:

if ($opt_sbnignore)
  { $i = @exclude;
    $exclude[$i++] = "DOWNLOAD";
    $exclude[$i++] = "NOTES";
    $exclude[$i++] = "VOLUME";
    $exclude[$i++] = "CODE";
    $exclude[$i++] = "WORK";
    $exclude[$i++] = "dataset\.html";
  }

open(OUT,">$output") ||
  die "Could not open output file '$output' ($!).";

# Print a header with the root directory name:

printf OUT "VC_LABELS Missing Label/File Report\n";
printf OUT "===================================\n\n";
printf OUT "For volume with root \"$ROOT\"\n\n";

# We're going to use the File::Find package to locate and chdir to the 
# directories in the volume:

find({wanted   => \&check_directory,
      no_chdir => $FALSE},
      $ROOT);

# Report a positive result if we haven't found any errors: 

printf OUT "  - OK\n" if ($error_count == 0);
printf OUT "\n";

# Close the output file and we're done:

close(OUT);

#----------------------------------------------------------------------------

sub check_directory

  # Routine used by Find::File to navigate the directory tree

  { my ($dir);
    my (%lab,%data);       # hashes for tracking files
    my ($file);            # looping variable
    my (%last_pointer);    # for tracking backward references
    my (%multiple_ref);    # data files with multiple labels
    my (%no_data);         # referenced data files that don't exist
    my ($line,$key,$val);  # for reading and parsing label lines
    my ($label);
    my (@flist,$fl);       # List of files from pointer and list length
    my ($not_done);
    my ($ignore);          # Ignore-file flag

    # Make sure there's nothing left over from last time:

    undef %lab;
    undef %data;
    undef %last_pointer;
    undef %no_data;
    undef %multiple_ref;
    undef @flist;

    # If the current entry is not a directory, we do nothing:
   
    return if (!(-d));

    # If the current entry matches something in @exclude, ignore it:

    foreach $pattern (@exclude)
      { return if ($File::Find::name =~ /$pattern/); }

    # But if it is a directory, we are going to open it and gather the
    # names of files in it:

    opendir(DIR,$_) || 
      die "Could not open directory '$_' for reading ($!)";

    foreach $file (readdir(DIR))
      { next if ($file =~ /^\./);
        next if (-d "$_/$file");

        # Save labels and data files; ignore format and catalog files:

        next if ($file =~ /\.cat$/  ||  $file =~ /\.CAT$/);
        next if ($file =~ /\.fmt$/  ||  $file =~ /\.FMT$/);

        # Ignore files matching an exclusion pattern:

        $ignore = $FALSE;
        foreach $pattern (@exclude)
          { $ignore = $TRUE if ("$File::Find::dir/$file" =~ /$pattern/); }
        next if ($ignore);

        if ($file =~ /\.lbl$/ || $file =~ /\.LBL$/)
          { $lab{$file} = 0; }

        else
          { $file =~ tr/A-Z/a-z/;      # Force case for convenience
            $data{$file} = 1;
          }
      }
    closedir(DIR);

    # The general procedure from now on is this - for each label file:
    #
    #  Open the label
    #  Find an object pointer
    #  Get the file name(s) from the pointer
    #  Check to see if the file named is present
    #  If it is, increment its %data entry count
    #    -if the count > 1, compare the label name to the %last_pointer
    #  Next pointer
    #
    # After all labels have been read, we'll look for dangling files and
    # references.
    
    foreach $label (keys %lab)
      { open(LBL,"$_/$label") || 
          die "Could not open '$File::Find::name/$label' for reading ($!)";

        # Loop through, looking for pointers:

        while ($line = nextPDSODL(LBL,$MAXLEN))
          { ($key,$val) = splitODLline($line);
            last if ($key eq "END");

            # If this isn't a pointer statement, keep going:

            next if ($key !~ /^\^/);

            # If it is a pointer, make sure it's not a structure or
            # descriptive pointer:

            next if ($key =~ /STRUCTURE/    ||  $key =~ /CATALOG/);
            next if ($key =~ /DESCRIPTION/  ||  $key =~ /TEXT/);

            # This is a pointer.  Now we're going to make an assumption
            # which is not actually encoded in the PDS Standards.  We're
            # going to assume that if pointer contains a sequence it will
            # contain only one sequence, and thus only one file name; and 
            # if it contains a set, it is a set of file names without 
            # offsets.

            $val =~ s/\(//;    # Remove parentheses, if any
            $val =~ s/\)//;
            $val =~ s/\{//;    # Remove braces, if any
            $val =~ s/\}//;
            $val =~ s/"//g;    # "

            @flist = split(/\,/,$val);
            $fl = @flist;

            # If the second element in the list is an offset, delete it: 

            if (@flist == 2  &&  $flist[1] =~ /^[0-9]\s*(<|)/)
              { $fl = 1; }

            # Now process the list:

            for ($i=0; $i<$fl; $i++)
              { $flist[$i] =~ tr/A-Z/a-z/; }

            # Loop through the array of file references, setting the
            # various trackers:

            for ($i=0; $i<$fl; $i++)
              { $file = $flist[$i];

                if ($data{$file})
                  { $data{$file}++; }
                else
                  { $data{$file} = 1;
                    $no_data{$file} = "$File::Find::dir/$file";
                  }

                if ($data{$file} > 2  &&  $last_pointer{$file} ne $label)
                  { $multiple_ref{$file}++; }
                $last_pointer{$file} = $label;

                # While we're here, we'll also make sure that 
                # DOCUMENT objects are not pointing to .TXT files:

                if ($key =~ /DOCUMENT$/  &&  $file =~ /\.txt$/)
                  { printf OUT "  $File::Find::name/$label: ";
                    printf OUT "DOCUMENT objects may not use ";
                    printf OUT ".txt/.TXT extensions.  Use ";
                    printf OUT ".asc/.ASC instead.\n";
                    $error_count++;
                  }
              }

            # Next line:
          }

        if ($key ne "END")
          { printf OUT "  $File::Find::name/$label has ";
            printf OUT "no END statement\n";
            $error_count++;
          }

        close(LBL);

        # Next label:
  
      }

    # Now if we have any data files that have not been reference by a label,
    # we check to see if they have an attached label.  
    #
    # NOTE: We're about to make another assumption that isn't coded into
    # the PDS Standards.  Specifically, we're going to assume that if a 
    # file has an attached label, it will not have pointers in that label
    # to other data files as well.  Might actually have to do something 
    # about that later.
    #
    # We're also assuming that the ODL at the beginning of an otherwise
    # unreferenced data file is actually an attached label.  This should
    # be a reasonably safe assumption.

    foreach $file (sort (keys %data))
      { if ($data{$file} == 1)
          { $FILE = $file;
            $FILE =~ tr/a-z/A-Z/;

            open(LBL,"$_/$file") || open(LBL,"$_/$FILE");

            # The 'nextPDSODL' routine returns the next complete logical
            # ODL statement.  It ignores blank lines and comment lines.
            # Any PDS label less than a decade old should start with the
            # PDS_VERSION_ID line...

            $line = nextPDSODL(LBL,$MAXLEN);
            if ($line =~ /PDS_VERSION_ID/)
              { $data{$file}++; }
          }

        if ($data{$file} == 1)
          { printf OUT "  $File::Find::name/$file has no label\n";
            $error_count++;
          }

        if ($multiple_ref{$file})
          { printf OUT "  WARNING: $File::Find::name/$file is referenced ";
            printf OUT "by %d labels\n", $multiple_ref{$file}+1;
            $error_count++;
          }
      }

    if (%no_data)
      { foreach $file (sort (keys %no_data))
          { printf OUT "  $File::Find::name/%s references ",$no_data{$file};
            printf OUT "non-existent data file '$file'\n";
            $error_count++;
          }
      }

    # And that should do it for this directory:

  }
