#!/usr/bin/perl

use lib "/n/sbnops01/lcltools/lib/perl5";

use Getopt::Long;
use File::Find;
require "nextPDSODL.pl";

# Default locations for externally referenced files:

$ALLREFS = "/n/sbnops01/lcltools/doc/allrefs.out";
$PDSDD   = "/n/sbnops01/lcltools/doc/pdsdd.full";

#============================================================================
# VC_REFINT
# ---------
# Routine to check referential integrity in a PDS volume structure.  The
# nextPDSODL.pl file is a rather old library from SBN that will read an ODL
# file one logical line at a time.  Not pretty, but reliable.
#
# Format: % vc_refint [-u] [-s] [-o output] [-h]
#                     [--sbn-ignore] \
#                     [--ignore-defined] [--allrefs allrefs_file] \
#                     [--pdsdd pdsdd_file] \
#                     [--exclude pattern] root_dir
#
#  where -u   requires all file/directory names to be uppercase (lowercase
#             is the default)
#        -s  or  --summary-only
#             reports only the summary
#        -o output
#             directs output to the named file rather than stdout
#        -h or --help or -?
#             print option summary
#        --sbn-ignore
#             ignores files and directories that are part of the SBN archive 
#             RAID support files
#        --ignore-defined
#             ignores any otherwise missing database key value that already
#             appears in either the pdsdd.full file or the allrefs.out file
#        --allrefs allrefs_file
#             Alternate location for the allrefs.out file
#        --pdsdd pdsdd_file
#             Alternate location for the pdsdd.full file
#        --exclude pattern
#             ignores files/directories matching the given regular expression
#
# NOTE: 
#
#   This routine assumes valid syntax, but does not attempt to check or 
#   enforce PDS constraints on the number or type of catalog object in 
#   each catalog file,  Every catalog file in the catalog directory is 
#   read through.
#
#
# 03 March 2008, A.C.Raugh
# 07 Mar 2008, ACR: Added "--ignore-defined" and related options
#
#============================================================================
# Constants:

$TRUE   = 1;
$FALSE  = 0;
$MAXLEN = 10000;   # Maximum returned ODL line length

# Check for clean run:

$error_count = 0;
$syntax_err  = 0;

# Initialize the hashes we'll use to generalize the search for key field
# values:

%obj_key = ("MISSION"                     => "MISSION_NAME",
            "INSTRUMENT_HOST"             => "INSTRUMENT_HOST_ID",
            "INSTRUMENT_HOST_INFORMATION" => "INSTRUMENT_HOST_NAME",
            "INSTRUMENT"                  => "INSTRUMENT_ID",
            "INSTRUMENT_INFORMATION"      => "INSTRUMENT_NAME",
            "DATA_SET_COLLECTION"         => "DATA_SET_COLLECTION_ID",
            "DATA_SET"                    => "DATA_SET_ID",
            "TARGET"                      => "TARGET_NAME",
            "REFERENCE"                   => "REFERENCE_KEY_ID",
            "SOFTWARE"                    => "SOFTWARE_ID",
            "PERSONNEL"                   => "PDS_USER_ID");

%val_hash= ("MISSION"                     => \%mission_name,
            "INSTRUMENT_HOST"             => \%host_id,
            "INSTRUMENT_HOST_INFORMATION" => \%host_name,
            "INSTRUMENT"                  => \%inst_id,
            "INSTRUMENT_INFORMATION"      => \%inst_name,
            "DATA_SET_COLLECTION"         => \%dscollid,
            "DATA_SET"                    => \%dsid,
            "TARGET"                      => \%target_name,
            "REFERENCE"                   => \%refkeyid,
            "SOFTWARE"                    => \%software_id,
            "PERSONNEL"                   => \%pds_user_id);

%def_hash=   ("MISSION_NAME"              => \%mission_name,
              "INSTRUMENT_HOST_ID"        => \%host_id,
              "INSTRUMENT_HOST_NAME"      => \%host_name,
              "INSTRUMENT_ID"             => \%inst_id,
              "INSTRUMENT_NAME"           => \%inst_name,
              "DATA_SET_COLLECTION_ID"    => \%dscollid,
              "DATA_SET_ID"               => \%dsid,
              "TARGET_NAME"               => \%target_name,
              "REFERENCE_KEY_ID"          => \%refkeyid,
              "SOFTWARE_ID"               => \%software_id,
              "PDS_USER_ID"               => \%pds_user_id);

%nodef_hash= ("MISSION_NAME"              => \%no_mission_name,
              "INSTRUMENT_HOST_ID"        => \%no_host_id,
              "INSTRUMENT_HOST_NAME"      => \%no_host_name,
              "INSTRUMENT_ID"             => \%no_inst_id,
              "INSTRUMENT_NAME"           => \%no_inst_name,
              "DATA_SET_COLLECTION_ID"    => \%no_dscollid,
              "DATA_SET_ID"               => \%no_dsid,
              "TARGET_NAME"               => \%no_target_name,
              "REFERENCE_KEY_ID"          => \%no_refkeyid,
              "SOFTWARE_ID"               => \%no_software_id,
              "PDS_USER_ID"               => \%no_pds_user_id);

%old_hash  = ("MISSION_NAME"              => \%old_mission_name,
              "INSTRUMENT_HOST_ID"        => \%old_host_id,
              "INSTRUMENT_HOST_NAME"      => \%old_host_name,
              "INSTRUMENT_ID"             => \%old_inst_id,
              "INSTRUMENT_NAME"           => \%old_inst_name,
              "DATA_SET_COLLECTION_ID"    => \%old_dscollid,
              "DATA_SET_ID"               => \%old_dsid,
              "TARGET_NAME"               => \%old_target_name,
              "REFERENCE_KEY_ID"          => \%old_refkeyid);

# Parse the input arguments and open the output file:

GetOptions("u"              => \$opt_u,
           "s|summary-only" => \$opt_s,
           "h|help|?"       => \$opt_help,
           "ignore-defined" => \$opt_defignore,
           "allrefs=s"      => \$opt_allrefs,
           "pdsdd=s"        => \$opt_pdsdd,
           "o=s"            => \$opt_o,
           "sbn-ignore"     => \$opt_sbnignore,
           "x|exclude=s"    => \@exclude);

if ($opt_help)
  { printf "Format: % vc_refint [-u] [-s] [-o output] [-h]\n";
    printf "                    [--sbn-ignore] \ \n";
    printf "                    [--ignore-defined] \ \n";
    printf "                    [--allrefs allrefs_file] \ \n";
    printf "                    [--pdsdd pdsdd_file] \ \n";
    printf "                    [--exclude pattern] root_dir\n";
    printf "\n";
    printf "where -u   requires all file/directory names to be uppercase\n";
    printf "           (lowercase is the default)\n";
    printf "      -s  or  --summary-only\n";
    printf "           reports only the summary\n";
    printf "      -o output\n";
    printf "           directs output to the named file rather than stdout\n";
    printf "      -h or --help or -?\n";
    printf "           print option summary\n";
    printf "      --sbn-ignore\n";
    printf "           ignores files and directories that are part of the\n";
    printf "           SBN archive RAID support files\n";
    printf "      --ignore-defined\n";
    printf "           ignores any otherwise missing database key value\n"; 
    printf "           that already appears in either the pdsdd.full file\n";
    printf "           or the allrefs.out file\n";
    printf "      --allrefs allrefs_file\n";
    printf "           Alternate location for the allrefs.out file\n";
    printf "      --pdsdd pdsdd_file\n";
    printf "           Alternate location for the pdsdd.full file\n";
    printf "      -x pattern or --exclude pattern\n";
    printf "           ignores files/directories matching the given regular\n";
    printf "           expression\n";
    printf "\n";
    exit(0);
  }

$output  = ($opt_o)? $opt_o : "-";
$allrefs = ($opt_allrefs)? $opt_allrefs : $ALLREFS;
$pdsdd   = ($opt_pdsdd)? $opt_pdsdd : $PDSDD;

# If we're ignoring defined key values, we need to be sure we have the
# required reference files handy, then call a routine to fill some more
# reference hashes:

if ($opt_defignore)
  { if (!(-f $allrefs))
      { die "Could not find reference file '$allrefs'."; }
    if (!(-f $pdsdd))
      { die "Could not find data dictionary file '$pdsdd'."; }
    read_defined();
  }

if (@ARGV != 1  ||  !(-d $ARGV[0]))
  { printf "Usage: vc_refint [-u] [-s] [-o output] [--sbn-ignore] ";
    printf "[--exclude regexp] root_dir\n";
    printf "Try 'vc_refint --help' for more information\n";
    die "\n";
  }

$ROOT = $ARGV[0]; 
$ROOT =~ s/\/$//;

# If we're ignoring SBN archive support files/directories, we'll add them
# to whatever we already have in @exclude:

if ($opt_sbnignore)
  { $i = @exclude;
    $exclude[$i++] = ($opt_u)? "DATASET\.HTML" : "dataset\.html";
    $exclude[$i++] = "DOWNLOAD";
    $exclude[$i++] = "CODE";
    $exclude[$i++] = "WORK";
    $exclude[$i++] = "NOTES";
    $exclude[$i++] = "VOLUME";
  }

open(OUT,">$output") ||
  die "Could not open output file '$output' ($!).";

# Print a header with the root directory name:

printf OUT "VC_REFINT Referential Integrity Report\n";
printf OUT "======================================\n\n";
printf OUT "For the volume with root \"$ROOT\"\n\n";

#----------------------------------------------------------------------------
#
# OK, We're ready to get to work.  For this task, we'll crawl the directories
# manually.  First task is to generate the list of catalog keys that are 
# actually defined in the catalog directory.  So we'll read through the .cat
# files in the catalog directory collecting the key fields we'll need to 
# reference in a series of hashes:
#
#  %mission_name     MISSION_NAME
#  %host_name        INSTRUMENT_HOST_NAME
#  %host_id          INSTRUMENT_HOST_ID
#  %inst_name        INSTRUMENT_NAME
#  %inst_id          INSTRUMENT_ID
#  %dsid             DATA_SET_ID
#  %target_name      TARGET_NAME
#  %refkeyid         REFERENCE_KEY_ID
#  %software_id      SOFTWARE_ID
#  %pds_user_id      PDS_USER_ID
#  %dscollid         DATA_SET_COLLECTION_ID
#
# Each of the above keywords, when used outside the catalog object that 
# defines them, must correspond exactly to the value list in the defining
# catalog file.  So we begin by finding and reading the contents of the 
# catalog directory.

$cat_dir = ($opt_u)? "$ROOT/CATALOG" : "$ROOT/catalog";

if (-d $cat_dir)
  { read_catdir($cat_dir); }
else
  { die "Could not find catalog directory '$cat_dir'\n"; }

## Checking what we managed to store in each hash:
#
# print_hashes();

# OK, we've initialized our key hashes.  Now we're ready to loop through
# all the files and directories in the given root directory.  We can use
# the File::Find package for this, since we actually will want to check
# the catalog directory files again for foreign keys.

find({wanted   => \&check_refint,
      no_chdir => $FALSE},      # We'll chdir to each directory searched
     $ROOT);

# Print a summary if we reported errors, or a positive results if we haven't:

if ($error_count > 0)
  { printf OUT "\n" if (!$opt_s);
    printf OUT "Summary of Undefined Key Values\n";
    printf OUT "===============================\n";

    # Checking each of the no-definition hashes individually:

    if (keys %no_mission_name)
      { printf OUT "\n%-40.40s Occurrences\n","Undefined MISSION_NAME values:";
        foreach $name (sort (keys %no_mission_name))
          { $pname = (length($name)>37)? substr($name,0,37) . "..." : $name;
            printf OUT "  %-40.40s   %2d\n",$pname, $no_mission_name{$name}; 
          }
      }

    if (keys %no_host_name)
      { printf OUT "\n%-40.40s Occurrences\n",
               "Undefined INSTRUMENT_HOST_NAME values:";
        foreach $name (sort (keys %no_host_name))
          { $pname = (length($name)>37)? substr($name,0,37) . "..." : $name;
            printf OUT "  %-40.40s   %2d\n",$pname, $no_host_name{$name}; 
          }
      }

    if (keys %no_host_id)
      { printf OUT "\n%-40.40s Occurrences\n",
               "Undefined INSTRUMENT_HOST_ID values:";
        foreach $name (sort (keys %no_host_id))
          { $pname = (length($name)>37)? substr($name,0,37) . "..." : $name;
            printf OUT "  %-40.40s   %2d\n",$pname, $no_host_id{$name};  
          }
      }

    if (keys %no_inst_name)
      { printf OUT "\n%-40.40s Occurrences\n",
                   "Undefined INSTRUMENT_NAME values:";
        foreach $name (sort (keys %no_inst_name))
          { $pname = (length($name)>37)? substr($name,0,37) . "..." : $name;
            printf OUT "  %-40.40s   %2d\n",$pname, $no_inst_name{$name};  
          }
      }

    if (keys %no_inst_id)
      { printf OUT "\n%-40.40s Occurrences\n",
                   "Undefined INSTRUMENT_ID values:";
        foreach $name (sort (keys %no_inst_id))
          { $pname = (length($name)>37)? substr($name,0,37) . "..." : $name;
            printf OUT "  %-40.40s   %2d\n",$pname, $no_inst_id{$name};  
          }
      }

    if (keys %no_dscollid)
      { printf OUT "\n%-40.40s Occurrences\n",
               "Undefined DATA_SET_COLLECTION_ID values:";
        foreach $name (sort (keys %no_dscollid))
          { $pname = (length($name)>37)? substr($name,0,37) . "..." : $name;
            printf OUT "  %-40.40s   %2d\n",$pname, $no_dscollid{$name};  
          }
      }

    if (keys %no_dsid)
      { printf OUT "\n%-40.40s Occurrences\n","Undefined DATA_SET_ID values:";
        foreach $name (sort (keys %no_dsid))
          { $pname = (length($name)>37)? substr($name,0,37) . "..." : $name;
            printf OUT "  %-40.40s   %2d\n",$pname, $no_dsid{$name};  
          }
      }

    if (keys %no_target_name)
      { printf OUT "\n%-40.40s Occurrences\n","Undefined TARGET_NAME values:";
        foreach $name (sort (keys %no_target_name))
          { $pname = (length($name)>37)? substr($name,0,37) . "..." : $name;
            printf OUT "  %-40.40s   %2d\n",$pname, $no_target_name{$name};  
          }
      }

    if (keys %no_refkeyid)
      { printf OUT "\n%-40.40s Occurrences\n",
                   "Undefined REFERENCE_KEY_ID values:";
        foreach $name (sort (keys %no_refkeyid))
          { $pname = (length($name)>37)? substr($name,0,37) . "..." : $name;
            printf OUT "  %-40.40s   %2d\n",$pname, $no_refkeyid{$name};  
          }
      }

    if (keys %no_software_id)
      { printf OUT "\n%-40.40s Occurrences\n","Undefined SOFTWARE_ID values:";
        foreach $name (sort (keys %no_software_id))
          { $pname = (length($name)>37)? substr($name,0,37) . "..." : $name;
            printf OUT "  %-40.40s   %2d\n",$pname, $no_software_id{$name};  
          }
      }

    if (keys %no_pds_user_id)
      { printf OUT "\n%-40.40s Occurrences\n","Undefined PDS_USER_ID values:";
        foreach $name (sort (keys %no_pds_user_id))
          { $pname = (length($name)>37)? substr($name,0,37) . "..." : $name;
            printf OUT "  %-40.40s   %2d\n",$pname, $no_pds_user_id{$name};  
          }
      }

    if ($syntax_err == 1)
      { printf "\n1 syntax error found\n"; 
        printf OUT "\n1 syntax error found\n"; 
      }
    elsif ($syntax_err > 1)
      { printf "\n $syntax_err syntax errors found\n"; 
        printf OUT "\n $syntax_err syntax errors found\n"; 
      }
  }
else
  { printf OUT "  - OK\n" if ($error_count == 0); }

printf OUT "\n";

# Close the output file and we're done:

close(OUT);

#============================================================================

sub read_catdir

  # Routine to read the files in the catalog directory and fill the 
  # name/id hashes with the defined names/ids.

  { local ($dir) = $_[0];   # Catalog directory, which we know exists

    my (@flist,$fl);        # List of catalog files and subscript
    my (@dlist,$dl);        # List of subdirectories and subscript
    my ($file);

    $fl = 0;
    $dl = 0;

    # Open the catalog directory and loop through entries:

    opendir(DIR,$dir)  ||  die "Could not open $dir for reading ($!)";

    foreach $file (readdir(DIR))
      { next if ($file =~ /^\./);

        # There might be subdirectories, so we'll need to deal with them:

        if (-d "$dir/$file")
          { $dlist[$dl] = "$dir/$file";
            $dl++;
            next;
          }

        # And we'll ignore anything that doesn't actually have a catalog
        # extension:

        next if (!$opt_u  &&  $file !~ /\.cat$/);
        next if ($opt_u  &&  $file !~ /\.CAT$/);

        # Add the catalog file to the list for processing:

        $flist[$fl] = "$dir/$file";
        $fl++;
      }

    closedir(DIR);

    # We'll call a routine to read each catalog file in succession:

    foreach $file (@flist)
      { read_catfile($file); }

    # And now we'll call this routine recursively for each subdirectory we
    # found:

    foreach $file (@dlist)
      { read_catdir($file); }

    # And we're done with the catalog directory...

    return;
  }

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

sub read_catfile

  # Routine to read through a single catalog file, recognize the defined key
  # fields, and add them to the hashes.

  { local ($file) = $_[0];    # Catalog file to read

    my ($line);               # next ODL line from the label
    my ($key,$val);           # for parsing the ODL line
    my ($looking);            # flag 
    my ($object);             # current object

    # Open the file and loop through it to the end:

    open(CAT,$file) || die "Could not open $file for reading ($!)";

    $not_done = $TRUE;
    while ($not_done)
      { # Read until we find a catalog object:

        $looking = $TRUE;
        while ($looking)
          { $line = nextPDSODL(CAT,$MAXLEN);

            # Check for premature end of file:

            if (!$line)
              { printf OUT "$file: SYNTAX ERROR -- no 'END' statement\n";
                $error_count++;
                $syntax_err++;
                close(CAT);
                return;
              }

            ($key,$val) = splitODLline($line);
            if ($key ne "END")
              { if (($key eq "OBJECT")  &&  ($obj_key{$val}))
                  { $looking = $FALSE; }
              }
            else
              { $looking = $FALSE;
                $not_done = $FALSE;
                next;
              }
          }

        next if (!$not_done);

        # Set the name of the key field we're looking for and the associated
        # hash reference according to the value of the OBJECT keyword found:

        $object   = $val;
        $seek_key = $obj_key{$val};
        $hash     = $val_hash{$val};

        # Now continue reading lines until we find the key field (NOTE that
        # we're assuming valid syntax and content - so key fields will always
        # be present and single-valued):

        $looking = $TRUE;
        while ($looking)
          { $line = nextPDSODL(CAT,$MAXLEN);

            if (!$line)
              { printf OUT "$file: SYNTAX ERROR -- no 'END' statement\n";
                $error_count++;
                $syntax_err++;
                close(CAT);
                return;
              }

            ($key,$val) = splitODLline($line);
            if ($key eq "END")
              { printf OUT "$file: SYNTAX ERROR -- $object has no ";
                printf OUT "$seek_key element\n";
                $error_count++;
                $syntax_err++;
                close(CAT);
                return;
              }

            if ($key eq $seek_key)
              { $val =~ s/^"\s*//;   #" Remove quotes and white space
                $val =~ s/\s*"$//;   #"
                $val =~ s/\s+/ /g;   # Replace multiple blanks with 1
                $hash->{$val} = 1;
                $looking = $FALSE;
              }
          }

        # Done with this object.  Go look for the next object:
      }

    # Done with this label.  Close it and return:

    close(CAT);

    return;
  }

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

sub check_refint

  # Routine called on each file to check for referential integrity in the
  # database key fields.  And field value that is not defined is added to 
  # a "not_found" hash corresponding to the key field type.  Every file is
  # checked for ODL content at the top of the file.

  { my ($line,$key,$val,@vallist);  # For parsing ODL lines
    my ($not_done);                 # Flag
    my ($hash,$nodef,$olddef);      # Indirect hash references

    # Note that we are chdired to the directory containing the file when
    # this routine is called.

    # Nothing to do for directories:

    return if (-d $_);

    # Ignore anything matching an excludion pattern:

    foreach $pattern (@exclude)
      { return if ($File::Find::name =~ /$pattern/); }

    # For all other files, we open the file and check for ODL.  We are 
    # making a BIG assumption that any files we encounter will be in 
    # reasonable, modern PDS format - specifically, that somewhere in 
    # the first 300 bytes will be the string "PDS_VERSION_ID =", and 
    # that this won't occur at the top of non-ODL files.  We have to 
    # check all files because of the possibility of attached labels.
    # We check the first 300 bytes just in case some data preparer put
    # comments or an SFDU label before the PDS_VERSION_ID, despite the
    # standards.

    open (ODL,$_) || die "Could not open $File::Find::name ($!)";
    read(ODL,$line,300);
    if ($line !~ /PDS_VERSION_ID\s*=/)
      { close(ODL);
        return;
      }

    # All righty.  We've got a nice, open file that contains ODL.  We
    # rewind back to the beginning and start looking for database keys:

    seek(ODL,0,SEEK_SET);
    $not_done = $TRUE;
    while ($not_done)
      { $line = nextPDSODL(ODL,$MAXLEN);
        if (!$line)
          { $not_done = $FALSE; 
            next;
          }
        ($key,$val) = splitODLline($line);
        if ($key eq "END")
          { $not_done = $FALSE;
            next;
          }

        # Check for a database key.  Database keys will have a value 
        # in %ndef_hash.

        next if (!($def_hash{$key}));

        # Also allow the value "N/A" as the only value in the list:

        next if ($val eq "\"N/A\"");

        # It is a database key.  Now we need to check the current value
        # against the appropriate list of defined values.  Note, though,
        # that the current value may be a list or sequence.  So we're 
        # going to parse the value string into the @vallist array:

        undef @vallist;
        if ($val =~ /^\(/)     # Parse the sequence
          { $val =~ s/^\(//;
            $val =~ s/\)$//;
            @vallist = splitlist($val,$File::Find::name);
          }
        elsif ($val =~ /^{/)  # Parse the set
          { $val =~ s/^{//;
            $val =~ s/}$//;
            @vallist = splitlist($val,$File::Find::name);
          }
        else
          { $val =~ s/^"\s*//; #" Remove extraneous leading/trailing space
            $val =~ s/\s*"$//; #"
            $val =~ s/\s+/ /g; #  Collapse multiple blanks to one
            $vallist[0] = $val;
          }

        # Now we check each element of @vallist to see if it's defined. If
        # not, we flag it:

        $hash   = $def_hash{$key};
        $nodef  = $nodef_hash{$key};
        if ($old_hash{$key})
          { $olddef = $old_hash{$key}; }
        else
          { undef $olddef; }

        foreach $val (@vallist)
          { if (!($hash->{$val}))
              { next if ($opt_defignore && $olddef && ($olddef->{$val}));  
                printf OUT "$File::Find::name: $key \"$val\" not defined\n"
                   if (!$opt_s);
                $nodef->{$val}++;
                $error_count++;
              }

          }

        # Next line:
      }

    # Done with this file:

    close(ODL);
    
    return;
  }

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

sub splitlist

  # Routine to split a list of values into an array of values.

  { local ($list,$file) = @_;  # Input value string & current file name
    my (@temp);                # Working array
    my (@res,$r);              # Results array
    my ($i,$elm); 

    # The biggest problem is that some - but not necessarily all - of the
    # values may be enclosed in double quotes. (We ignore the possibility of
    # single quotes.)  The value separator, a comma, may be part of a quoted
    # string.  So we have to be rather careful about how we peel off each
    # value.

    # If there are no quotes, this is easy - we split on commas:

    if ($list !~ /"/) #"
      { @res = split(/\s*,\s*/,$list); }

    # If there are quotes, we split on the commas anyway, but we'll have to 
    # check the results:

    else
      { @temp = split(/\s*,\s*/,$list);
        $r = 0;
        $i = 0;
        while ($i < @temp)
          { if ($temp[$i] =~ /^"/) #"
              { if ($temp[$i] =~ /"$/)   #" Balanced quotes - remove them
                  { $temp[$i] =~ s/^"\s*//; #"
                    $temp[$i] =~ s/\s*"$//; #"
                    $temp[$i] =~ s/\s+/ /g; # Collapse white space
                    $res[$r++] = $temp[$i];
                  }
                else     # Unbalanced quotes - add next value and loop:
                  { # There is potential for disaster here if there is
                    # a synatx error resulting in unbalance quotes, so 
                    # we'll go ahead and check for that:

                    if (($i+1) >= @temp)
                      { printf "DEADLY SYNTAX ERROR in $file - ";
                        printf "unbalanced quotes\n"; 
                        die;
                      }

                    $elm = "$temp[$i] $temp[$i+1]";
                    $elm =~ s/\s+/ /g;   # Collapse white space
                    $temp[$i+1] = $elm;
                  }
              }
            else
              { $res[$r++] = $temp[$i]; }
            $i++;
          }
      }

    # And we're done:

    return @res;
  }

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

sub read_defined

  # Routine to read the lists of known database key values from the 
  # pdsdd.full and allrefs.out files, initializing some reference
  # hashes.

  { my ($line);
    my ($key,$val);
    my (@temp);
    my (@db_key,@db_hash);
    my ($i);

    # We'll start with the pdsdd.full file.  Open it for reading:

    open(DD,$pdsdd) || die "Could not open $pdsdd for reading ($!)";

    # Initialize @db_key:

    @db_key = ("DATA_SET_COLLECTION_ID",
               "DATA_SET_ID",
               "INSTRUMENT_HOST_ID",
               "INSTRUMENT_HOST_NAME",
               "INSTRUMENT_ID",
               "INSTRUMENT_NAME",
               "MISSION_NAME",
               "TARGET_NAME");

    # And set references to the hashes we're initializing:

    @db_hash= ( \%old_dscollid,
                \%old_dsid,
                \%old_host_id,
                \%old_host_name,
                \%old_inst_id,
                \%old_inst_name,
                \%old_mission_name,
                \%old_target_name);

    # We take advantage of the fact the the PDSDD file is traditionally
    # sorted alphabetically, and look for the elements that have standard
    # values in sequence:

    $line = <DD>;
    for ($i=0; $i<@db_key; $i++)
      { while ($line !~ /^\s*NAME = $db_key[$i]\s*$/)
          { $line = <DD>; }

        while ($line !~ /^\s*STANDARD_VALUE_SET =/)
          { $line = <DD>; }

        # Parsing out the standard values and assigning them to their
        # respective hashes:

        while ($line !~ /\}\s*$/)   # Last element in std value list
          { $line = <DD>;
            $line =~ /"(.*)"/;
            $db_hash[$i]->{$1} = 1;
          }

        # Done with this database key
      }

    # Done with the pdsdd.full file:

    close(DD);

    # Next is the reference key values from allrefs.out:

    open(REF,$allrefs) || die "Could not open $allrefs for reading ($!)";

    # The first field, delimited by "|", ineach record is the 
    # REFERENCE_KEY_ID value we want to track.  There may be more than
    # one record for a key, but it is fast to assign than do a string
    # comparison to check, so we don't bother:

    while ($line=<REF>)
      { $line =~ /^([^|]+)|/;
        $old_refkeyid{$1} = 1;
      }
    close(REF);

    # And we're done.

    return;
  }

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

sub print_hashes

  # Routine to dump hashes for debugging

  { # Checking what we managed to store in each hash:

    printf "Mission name:\n";
    foreach $key (keys %mission_name)
      { printf "  $key   $mission_name{$key}\n"; }
    printf "Host name:\n";
    foreach $key (keys %host_name)
      { printf "  $key   $host_name{$key}\n"; }
    printf "Host ID:\n";
    foreach $key (keys %host_id)
      { printf "  $key   $host_id{$key}\n"; }
    printf "Instrument name:\n";
    foreach $key (keys %inst_name)
      { printf "  $key   $inst_name{$key}\n"; }
    printf "Instrument ID:\n";
    foreach $key (keys %inst_id)
      { printf "  $key   $inst_id{$key}\n"; }
    printf "DSID:\n";
    foreach $key (keys %dsid)
      { printf "  $key   $dsid{$key}\n"; }
    printf "Target Name:\n";
    foreach $key (keys %target_name)
      { printf "  $key   $target_name{$key}\n"; }
    printf "Reference:\n";
    foreach $key (keys %refkeyid)
      { printf "  $key   $refkeyid{$key}\n"; }
    printf "PDS user ID:\n";
    foreach $key (keys %pds_user_id)
      { printf "  $key   $pds_user_id{$key}\n"; }
    printf "Software ID:\n";
    foreach $key (keys %software_id)
      { printf "  $key   $software_id{$key}\n"; }
    printf "DS Collection ID:\n";
    foreach $key (keys %dscollid)
      { printf "  $key   $dscollid{$key}\n"; }

    return;
  }
