#!/usr/bin/perl

use Getopt::Long;
use File::Find;

# These are the locations of the external, single-file checking routines
# that we'll be invoking:

$CHKASCII = "/n/sbnops01/lcltools/bin/chkascii";
$SCANLEN  = "/n/sbnops01/lcltools/bin/scanlen";
#============================================================================
# VC_TEXT
# -------
# Routine to check text files in a PDS volume to make sure they meet the 
# minimal criteria (7-bit printing ASCII, CR/LF carriage control).
#
# Format: % vc_text [-o output] [-x pattern] volume_root
#
#  where  -o output    Directs output to the named file
#         -x pattern   Ignores files/paths matching the given pattern
#         --sbn-ignore Is shorthand for excluding all the SBN support stuff
#
#  Note: At some point we'll probably want to check attached PDS labels for
#        some of this stuff as well, which is why we're using the long
#        form for GetOpt (allows mutiple occurences of the same option
#        for listing mutiple extensions to be checked, e.g.).  But this 
#        is time-consuming to implement and not immediately necessary,
#        so I'm deferring it.
#
#        In the case of attached labels, probably the most efficient thing
#        to do is create a temporary file to hold just the label, and then
#        treat it as an attached label, but without the 80-byte line length
#        limit.
#
# 22 February 2008, A.C.Raugh
# 05 May 2008, ACR: Added exclude options because support files were annoying
#                   me.
# 10 Mar 2009, ACR: Fixed stupid bug that ignored the first error returned
#                   by CHKASCII
#
#============================================================================

# We'll use a hash as a quick method for checking if file extensions
# correspond to files that should contain compliant text:

%ODL = ("LBL" => 1,  "lbl" => 1,
        "FMT" => 1,  "fmt" => 1,
        "CAT" => 1,  "cat" => 1);
%TXT = ("TXT" => 1,  "txt" => 1,
        "TAB" => 1,  "tab" => 1,
        "HTM" => 1,  "htm" => 1,
        "CSV" => 1,  "csv" => 1,
        "ASC" => 1,  "asc" => 1);

# And here are some constants we might need:

$TRUE  = 1;
$FALSE = 0;

# We'll use this to note a completely clean run:

$error_flag = $FALSE;

# Get the input parameters, do some sanity checking and set up the 
# output stream:

GetOptions("o=s" => \$opt_o,
           "x|exclude=s" => \@exclude,
           "sbn-ignore"  => \$opt_sbnignore);

if (@ARGV == 1 && -d $ARGV[0])
  { $ROOT = $ARGV[0]; }
else
  { die "Usage: vc_text [-o out] volume_root\n"; }

$output_file = ($opt_o)? $opt_o : "-";

open(OUT,">$output_file") ||
  die "Could not open output stream \"$output_file\" ($!).\n";

# If we're ignoring SBN archive support files/directories, we'll add them
# to whatever we might already have in the @exclude array:

if ($opt_sbnignore)
  { $i = @exclude;
    $exclude[$i++] = "dataset\.html";
    $exclude[$i++] = "DOWNLOAD";
    $exclude[$i++] = "CODE";
    $exclude[$i++] = "WORK";
    $exclude[$i++] = "NOTES";
    $exclude[$i++] = "VOLUME";
  }

# Print a header for the output"

printf OUT "VC_TEXT 7-bit, Printable ASCII Compliance Report\n";
printf OUT "================================================\n\n";
printf OUT "For volume with root \"$ROOT\"\n\n";

# OK, we're ready to begin the recursive process of finding and checking
# text files.  We'll use the "find" function to do a pre-order traversal
# of the volume root.
#
# NOTE: We're avoiding the "proprocess" option in case we want to be able
# to use the "follow" or "follow-fast" options to follow symbolic links.

find({wanted => \&id_text, 
      no_chdir => $FALSE},    # chdir our way down the tree
      $ROOT);

# If we saw no errors, print a success message:

printf OUT "  - OK\n" if (!$error_flag);
printf OUT "\n";

# Close the output file and we're done:

close(OUT);

#----------------------------------------------------------------------------

sub id_text

# Routine called by the "find" function to determine if the current file
# is a text file that should be checked.  If so, a routine is called to 
# check it for text file compliance.

{ my ($file) = $_;  # The name of the current file
  my ($ext);        # File extension
  my ($apttern);    # looping variable for exclusion check
  my (@parts);      # For parsing lines
  my ($result);     # Output line from pipe
  my ($errcount);   # Error count for this file
  my ($reccount,$minlen,$maxlen);  # Values from 'scanlen' output
  
  return if (-d $file  ||  $file =~ /^\./);

  # If the full file spec matches anything in an exclusion option we
  # ignore it:

  foreach $pattern (@exclude)
    { return if ($File::Find::name =~ /$pattern/); }

  # Extract the extension, if any:

  @parts = split(/\./,$file);

  # If there's no extension, we're done:

  return if (@parts != 2);

  $ext = $parts[1];

  # We check the extension - either it's an ODL file or a flat text
  # file, but we need to keep track of which.  (Logically only one
  # can be true unless there's an overlap in the hash initialization,
  # which will never happen, RIGHT?).  We check for the option of 
  # attached labels, which are ODL files, and note if text files
  # occur in the document/ directory:

  $is_odl = ($ODL{$ext}  ||  $attached{$ext});
  $is_txt = ($TXT{$ext});
  $is_in_doc = ($txt  &&  
                ($File::Find::dir =~ /document/  ||  
                 $File::Find::dir =~ /DOCUMENT/)   &&
                $txt !~ /info/  &&  $txt !~ /INFO/);

  # We're done if this isn't an ODL or text file:

  return if (!$is_odl  &&  !$is_txt);

  # Now we're ready to start checking the file.  First test is to run
  # 'scanlen', parse the results and check for appropriate values for
  # the file type.  When the "-t" option is used, the first line of 
  # output from scanlen is the line delimiter type.  The second line
  # contains the total number of lines and the maximum and minimum 
  # line length:

  $errcount = 0;

  open(SL,"$SCANLEN -t $file |") || 
    die "Could not open $SCANLEN pipe ($!)";

  $result = <SL>;
  if ($result !~ /CR\/LF/)
    { printf OUT "$File::Find::name:\n";
      printf OUT "  Does not have CR/LF carriage control.\n";
      $errcount++;
    }

  $result = <SL>;
  $result =~ /^[^0-9]+([0-9]+)[^0-9]+([0-9]+)[^0-9]+([0-9]+)/;
  $reccount = $1;
  $maxlen   = $2;
  $minlen   = $3;
  close(SL);

  # If this is a detached label file or catalog file, it should have 
  # fixed length records not more than 80 bytes wide:

  if ($ext eq "lbl"  ||  $ext eq "cat"  ||  $ext eq "LBL"  ||  $ext eq "CAT")
    { if ($maxlen != $minlen)
        { printf OUT "$File::Find::name:\n" if ($errcount == 0);
          $errcount++;
          printf OUT "  WARNING -- Variable length records\n";
        }

      if ($maxlen > 80)
        { printf OUT "$File::Find::name:\n" if ($errcount == 0);
          $errcount++;
          printf OUT "  WARNING -- Maximum record length > 80\n";
        }
    }

  # Now to check for non-ASCII characters.  If this is a text file in a
  # document directory, then formfeed characters are allowed; otherwise
  # we do the stricter check:

  if ($is_txt && $is_in_doc)
    { open(CA,"$CHKASCII -ff $file |")  ||
        die "Could not open chkascii -ff pipe ($!) \n";
    }
  else
    { open(CA,"$CHKASCII $file |")  ||
        die "Could not open chkascii pipe ($!) \n";
    }

  # Check the first line for successful completion:

  $result = <CA>;
  if ($result =~ /^No invalid/)
    { close(CA); }

  # Otherwise, just transfer the error lines:

  else
    { printf OUT "$File::Find::name: \n" if ($errcount == 0);
      printf OUT "  $result";
      $errcount++;

      while ($result =<CA>)
        { printf OUT "  $result";
          $errcount++;
        }
      close(CA);
    }

  # If we found any errors, set the global error flag:

  $error_flag = $TRUE if ($errcount > 0);

  return;
}
